
#include "lightmap.h"


typedef struct fxeLDrawVertex
{
  // x,y,z of vertex
  float x;    // 0

  float y;    // 4

  float z;    // 8

  float q;    // 12

  // Texture Coordinates for tmu 0
  float u0,v0;  // 16

  // Texture Coordinates for tmu 1
  float u1,v1;  // 24
} fxeLDrawVertex;

typedef struct
{

  fxeLDrawVertex v0;
  fxeTextureCoordinate e0;
  fxeLDrawVertex v1;
  fxeTextureCoordinate e1;

} fxeLLine;





typedef struct fxeLandscape
{
  unsigned int *Landscapes;
  char NumLandscapesX;
  fxeVertex VertexWorld;
  fxeVertex VertexCamera;
  float Radius;
  fxeLDrawVertex VerticesWorld[4];
  fxeLDrawVertex VerticesCamera[4];
  fxeTextureHandle TH;
  char TextureMode;
  fxeLDrawVertex ClippedVertices1[8];
  fxeLDrawVertex ClippedVertices2[8];
  char Clipped;
  char PartlyClipped;
  char NumAddedVertices1;
  char NumAddedVertices2;
  fxeLightmap Lightmap;
	fxeVector Normal1,Normal2;
	char Displayed;
} fxeLandscape, *fxeLandscapePtr;

typedef struct fxeLandscapesParent
{
	float Width, Depth;
	int NumPolysX;
	float *HeightBuffer;

	// A pointer to a buffer with pointers to landscapes.
	// Pointers to landscapes that will be rendered
	// is stored in this list.
	fxeLandscape **ppLandscape;
	int NumLandscapesInList;

	fxeLandscape *pLandscape;
}	fxeLandscapesParent, *fxeLandscapesParentHandle;

typedef struct fxeHeightmapInfo
{
  int Width;
  float *MapData;
} fxeHeightmapInfo, *fxeHeightmapInfoPtr;

typedef struct fxeNormalBuffer
{
  fxeVector *Normals;
  int Width;
} fxeNormalBuffer, *fxeNormalBufferPtr;

typedef fxeLandscapePtr fxeLandscapeHandle;

extern fxeLandscapeHandle CurrLandscape;
extern fxeLandscapesParentHandle CurrLandscapeParent;

float fxeLandscapeGetHeightXZ(fxeLandscapesParentHandle LspH, float x, float z);

// Returns FXETRUE if object collided with landscape and
// FXEFALSE otherwise.
int fxeObjectDetectCollisionLandscape(fxeObjectHandle OH,
	fxeLandscapesParentHandle LspH);


fxeLandscapesParentHandle fxeLandscapeLoadFromFile(char *Geometry, char *Color, fxeTextureHandle TH,
	float Width, float Depth, float Height, fxeVector *LightNormal, 
	float LightIntensity);

fxeLandscapesParentHandle fxeLandscapeCreate64x64(float Width1, float Depth1,
  fxeTextureHandle TerrainTexture, int RandomSeed);

fxeLandscapeHandle
  fxeLandscapeCreateRandom(float Width, float Depth, fxeHeightmapInfoPtr
  pHeightmap, fxeNormalBufferPtr pNormalBuffer,
  fxeTextureHandle TerrainTexture, int Seed);

fxeLandscapePtr SetupLandscape(int LandscapeLevel,
  float Width, float Depth, int x, int z, fxeHeightmapInfoPtr pHeightmap,
  fxeTextureHandle TerrainTexture, int NumChildsX, fxeNormalBufferPtr
  pNormalBuffer);

fxeHeightmapInfoPtr fxeCreateHeightMap(
  int NumRecs, int RecWidth, int RecHeight, int MapWidth);

void fxeFreeHeightMap(fxeHeightmapInfoPtr pHeightmap);


void fxeLandscapeAddToWorld(fxeLandscapesParentHandle pLandscape);

void fxeLandscapeRemoveFromWorld(void);

