

// This is the source for the particle system
// for FxEengine. Many thanks goes to Daniel
// M Jones. I didn't actually copy his code 
// but I got a lot of ideas on how to do it from
// his code. 

#define _FXPARTICLE_

#ifndef _FXENGINE_
#include "fxenginedrv.h"
#endif



typedef struct
{
  // world position:
  float x,y,z;

  // The direction the Particle is moving to 
  fxeVector DirectionVector;

  // Radius of particle:
  float Radius;

  // Current r,g,b color:
  fxeColor CurrColor;

  // Current Alpha value:
  float a;

  // Current r,b,g decrease/increase variables:
  fxeColor IncColor;

  // Current alpha decrese/increase variable:
  float ia;

  // The number of frames left this particle will last:
  int NumFramesLeft;

	fxeVector ForceVector;

  void *Prev;
  void *Next;

} fxeParticle;

typedef struct
{
	float x, y, z;
	float cz;
  fxeColor InitColor;
  fxeColor FinishColor;
  fxeTextureHandle TexHandle;
  int TotalParticles;
  int NumParticles;
  // This variable is always pointing to the first particle in the
  // linked list of particles in this emitter.
  fxeParticle *UsedParticles;
  // This is a pointer to the first unused space for a particle in the
  // linked list
  fxeParticle *UnusedParticles;
  fxeVector ForceVector;
  // The ForceVector is multiplied by this constant each frame:
  // If ForceAcceleration is 2 the force is doubling it's strenght each frame
  float ForceAcceleration;
	float ParticleRadius;
	float Magnitude;
	void *Prev;
	void *Next;
} fxeEmitter, *fxeEmitterHandle;




fxeEmitterHandle fxeEmitterCreate(float x, float y, float z, 
	int NumParticles, fxeColor *InitColor,
  fxeColor *FinishColor, fxeTextureHandle TexHandle,
  fxeVector *ForceVector, float ForceAcceleration, float
	ParticleRadius, float Magnitude);

void fxeEmitterAddParticles(fxeEmitterHandle pEmitter, int NumParticles);

void fxeEmitterAddParticle(fxeEmitterHandle pEmitter);

void fxeEmitterTranslateFrame(fxeEmitterHandle pEmitter);

void fxeEmitterClipAndDraw(fxeEmitterHandle pEmitter);

void fxeEmitterUpdateFrame(fxeEmitterHandle pEmitter);

void fxeEmitterAddToWorld(fxeEmitterHandle pEmitter);

void fxeEmitterRemoveFromWorld(fxeEmitterHandle pEmitter);

void fxeParticleSystemInit(void);

void fxeEmitterChangeProperties(fxeEmitter *pEmitter, 
	float x, float y, float z, fxeVector *ForceVector);

fxeEmitter *pEmitters;


