

// Very simple FxEngine application - just loads a object
// and shows it on the screen. Press ESC to exit.

// As this is a windows application we need to include 
// windows.h
#include <windows.h>

// Include the header files for FxEngine.
#include "fx.h"

// Include function to set up our main window and maximize
// it.
#include "..\common files\wininit.h"


// The window handle of our main window is stored here.
HWND WindowHandle;

// The usual window callback procedure that handles all
// messages send from windows to this application.
LRESULT CALLBACK WndProc (HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam);

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    char * szCmdLine, int iCmdShow)
{
	MSG msg;
	fxeObjectHandle OH;
	
	// Setup and display main window:
	WindowHandle = SetupMainWindow(hInstance,WndProc);

	// Initialize FxEngine:
	// We need to pass a valid window handle and a resolution
	// to use to the FxEngine initialization function.
	if (fxeInitialize(WindowHandle,FXE_RES_640x480) == FXEFALSE)
	{
		// Couldn't initialize FxEngine, exit back to windows
		// with a error message:
		MessageBox("Couldn't initialize FxEngine!","Error",MB_OK,0);
		return 1;
	}

	// Load a 3DS object from disk, scale it to 100 units

	OH = fxeObjectLoad("tank.3ds",OBJECT_3DS,100.0f,FXETRUE);

	// Add the loaded object to the world so it will be rendered.

	fxeObjectAdd(OH);

	// Add a white lightsource

	fxeLightsourceAdd(1,1,1,-5000,-5000,5000,0,NULL);

	// Move the object to world position 0,0,150

	fxeObjectMove(OH,0.0f,0.0f,150.0f);

	// Rotate the object 130 degrees around the y axis
	// and 45 degrees around the x axis.

	fxeObjectRotate(OH,130,45,0);

	// Make the object flat shaded.

	fxeObjectSetShadingType(OH,FLAT_SHADING);

	// Clear the background

	fxeBackgroundClear();

	// Render the world, in this case the tank object

	fxeWorldRender(FXETRUE);

	// Here we enters a eternal loop that only stops when
	// the user presses the ESC key.

	while (1)
	{
		// Perform standard window message handling
		if	(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE) == TRUE)
    {
			// A message was send to the application:
			// Load the message sent to the application
			if	(!GetMessage(&msg,NULL,0,0))
			{
				// Terminate message was sent, 
				// exit application.

				// Before we exit we need to shut down FxEngine.
				fxeShutdown();

				return msg.wParam;
			}
				DispatchMessage(&msg);
    }
	}
}

LRESULT CALLBACK WndProc (HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{

	switch (iMsg)
  {

		case WM_KEYDOWN:

			// Check if the escape key was pressed:
			if (wParam == VK_ESCAPE)
			{
				PostQuitMessage(0);
				return 0;
			}
			return 0;

  }

	// Perform default handling of unhandled messages sent to the
	// application:
	return DefWindowProc(hwnd, iMsg, wParam, lParam);
}


