

// The smoke system is very close to the particle system
// Every smoke particle has a starting x,y,z position
// Then a random number tranlates it in x,y,z every frame
// The difference is that it is multiplied by the 
// background scene. A smoke emitter kan be created
// A smoke emitter has a radius and a opacity and a life. Larger
// radius means the starting position of every smoke particle
// can vary in radius. Opacity means how many smoke particles
// actually is generated. Example: 10 in opacity means
// that 10 smoke particles is added each frame. 
// Life is how many frames the smoke 
// particles can exist.



#include "FxSmoke.h"
#include "FxGlobals.h"
#include <glide.h>
#include "FxMalloc.h"
#include "FxTexture.h"


void fxeSmokeEmitterAddToWorld(fxeSmokeEmitterHandle pSmokeEmitter)
{
	if (pSmokeEmitters == NULL)
	{
		pSmokeEmitters = pSmokeEmitter;
		pSmokeEmitter->Next = NULL;
		pSmokeEmitter->Prev = NULL;
	}
	else
	{
		pSmokeEmitters->Next = pSmokeEmitter;
		pSmokeEmitter->Prev = pSmokeEmitters;
		pSmokeEmitter->Next = NULL;
		pSmokeEmitters = pSmokeEmitter;
	}
}

void fxeSmokeEmitterRemoveFromWorld
	(fxeSmokeEmitterHandle pSmokeEmitter)
{
	if (pSmokeEmitter->Next == NULL && pSmokeEmitter->Prev == NULL)
	{
		pSmokeEmitters == NULL;
	}
	else if (pSmokeEmitter->Next == NULL)
	{
		((fxeSmokeEmitter*)(pSmokeEmitter->Prev))->Next = NULL;
		if (pSmokeEmitters == pSmokeEmitter)
			pSmokeEmitters = pSmokeEmitter->Prev;
	}
	else if (pSmokeEmitter->Prev == NULL)
	{
		((fxeSmokeEmitter*)(pSmokeEmitter->Next))->Prev = NULL;
	}
	else
	{
		((fxeSmokeEmitter*)(pSmokeEmitter->Prev))->Next = pSmokeEmitter->Next;
		((fxeSmokeEmitter*)(pSmokeEmitter->Next))->Prev = pSmokeEmitter->Prev;
	}
}

void fxeSmokeSystemInit(void)
{
	pSmokeEmitters = NULL;
}


// Returns a handle to a emitter if successful and NULL
// if failure.
fxeSmokeEmitterHandle fxeSmokeEmitterCreate(float x, float y, float z, 
	int NumParticles, fxeTextureHandle TexHandle, float Radius,
  fxeVector *ForceVector, float ForceAcceleration, float
	ParticleRadius, float Magnitude, int Life, int Opacity)
{
  int Loop;
  fxeSmokeEmitter *pEmitter;
	
  pEmitter = (fxeSmokeEmitter*)FxMalloc(sizeof(fxeSmokeEmitter));
	pEmitter->ParticleCount = 0;
	pEmitter->x = x;
	pEmitter->y = y;
	pEmitter->z = z;
	pEmitter->TexHandle = TexHandle;
	pEmitter->Radius = Radius;
  pEmitter->TotalParticles = NumParticles;
  pEmitter->NumParticles = 0;
  pEmitter->UsedParticles = (fxeSmokeParticle*)FxMalloc(sizeof(fxeSmokeParticle)*NumParticles);
	if (pEmitter->UsedParticles == NULL)
		return NULL;
  pEmitter->UnusedParticles = pEmitter->UsedParticles;
  pEmitter->ForceVector.x = ForceVector->x;
  pEmitter->ForceVector.y = ForceVector->y;
  pEmitter->ForceVector.z = ForceVector->z;
  pEmitter->ForceAcceleration = ForceAcceleration;
	pEmitter->ParticleRadius = ParticleRadius;
	pEmitter->Magnitude = Magnitude;
	pEmitter->Life = Life;
	pEmitter->Opacity = Opacity;

  for (Loop=0;Loop<(NumParticles-1);Loop++)
  {
    pEmitter->UsedParticles[Loop].Next =
      (void*)&pEmitter->UsedParticles[Loop+1];
  }
  pEmitter->UsedParticles[NumParticles-1].Next = NULL;
  return pEmitter;
}


void fxeSmokeEmitterAddParticles(fxeSmokeEmitterHandle pEmitter)
{
	int Loop;

	if (pEmitter->Opacity == 0)
		return;

	for (Loop=0; Loop<pEmitter->Opacity; Loop++)
	{
		if (pEmitter->ParticleCount = 10)
		{
			pEmitter->ParticleCount = 0;
			fxeSmokeEmitterAddParticle(pEmitter);
		}
		else
			pEmitter->ParticleCount++;
	}
}


// Adds a new particle with a special direction vector
// and a special life length to a emitter specified:

void fxeSmokeEmitterAddParticle(fxeSmokeEmitterHandle pEmitter)
{
	fxeSmokeParticle *pCurrParticle;

	// only add particle if it's room for it:
	if (pEmitter->NumParticles == pEmitter->TotalParticles)
		return;

	pEmitter->UnusedParticles->x = pEmitter->x+
		rand()%((int)pEmitter->Radius*2)-pEmitter->Radius;
	pEmitter->UnusedParticles->y = pEmitter->y;
	pEmitter->UnusedParticles->z = pEmitter->z+
		rand()%((int)pEmitter->Radius*2)-pEmitter->Radius;
	pEmitter->UnusedParticles->Alpha = 128.0f;
	pEmitter->UnusedParticles->AlphaInc = (float)(128.0f/((float)pEmitter->Life));
	pEmitter->UnusedParticles->NumFramesLeft = (pEmitter->Life);
	
	
	pEmitter->UnusedParticles->Radius = pEmitter->ParticleRadius;

	pEmitter->UnusedParticles->DirectionVector.x = 0.0f;
	pEmitter->UnusedParticles->DirectionVector.y = 1.0f*pEmitter->Magnitude;
	pEmitter->UnusedParticles->DirectionVector.z = 0.0f;

	pEmitter->UnusedParticles->ForceVector.x = ((float)(rand()%1000))/1000.0f-0.5f;
	pEmitter->UnusedParticles->ForceVector.y = ((float)(rand()%1000))/1000.0f-0.5f;
	pEmitter->UnusedParticles->ForceVector.z = ((float)(rand()%1000))/1000.0f-0.5f;

	pCurrParticle = pEmitter->UnusedParticles;

	pEmitter->UnusedParticles = (fxeSmokeParticle*)pEmitter->UnusedParticles->Next;
	pEmitter->NumParticles++;

	pCurrParticle->Next = pEmitter->UsedParticles;
	pEmitter->UsedParticles->Prev = pCurrParticle;
	pEmitter->UsedParticles = pCurrParticle;
}


void fxeSmokeEmitterUpdateFrame(fxeSmokeEmitterHandle pEmitter)
{
	fxeSmokeEmitterTranslateFrame(pEmitter);
	fxeSmokeEmitterClipAndDraw(pEmitter);
}

void fxeSmokeEmitterTranslateFrame(fxeSmokeEmitterHandle pEmitter)
{
	int Loop;
	fxeSmokeParticle *pParticle;
	int NumParticles;
	fxeSmokeParticle *SavedParticle;

	if (pEmitter->NumParticles == 0)
		return;

	pParticle = pEmitter->UsedParticles;

	NumParticles = pEmitter->NumParticles;

	for (Loop=0;Loop<NumParticles;Loop++)
	{
		if (pParticle->NumFramesLeft == 0)
		{
			if (pParticle == pEmitter->UsedParticles)
			{
				// This particle is the first particle in the
				// emitter. We need to change the first particle to
				// this particle.
				pEmitter->UsedParticles = pParticle->Next;
				SavedParticle = pParticle->Next;
				pEmitter->NumParticles--;
				pParticle->Next = pEmitter->UnusedParticles;
				pEmitter->UnusedParticles = pParticle;
				pParticle = SavedParticle;
			}
			else if (Loop==(NumParticles-1))
			{
				// This is the last used particle in the emitter,
				// this means we just have to decrement the particle
				// counter.

				pEmitter->NumParticles--;
				pParticle->Next = pEmitter->UnusedParticles;
				pEmitter->UnusedParticles = pParticle;
			}
			else
			{
				// This is not the first particle and not the last
				// so we need to take into account previous and next
				// particle
				SavedParticle = pParticle->Next;
				pEmitter->NumParticles--;
				((fxeSmokeParticle*)(pParticle->Prev))->Next = 
					pParticle->Next;
				((fxeSmokeParticle*)(pParticle->Next))->Prev =
					pParticle->Prev;
				pParticle->Next = pEmitter->UnusedParticles;
				pEmitter->UnusedParticles = pParticle;
				pParticle = SavedParticle;
				
			}
		}

		else
		{
			// The particle have frames left to live, so translate
			// it and change its rgb color and decrement its life
			// time.

			pParticle->NumFramesLeft--;
			
			pParticle->x += pParticle->DirectionVector.x 
				+ pParticle->ForceVector.x;
			pParticle->y += pParticle->DirectionVector.y 
				+ pParticle->ForceVector.y;
			pParticle->z += pParticle->DirectionVector.z 
				+ pParticle->ForceVector.z;

			pParticle->Alpha = pParticle->Alpha -
				pParticle->AlphaInc;

			pParticle->ForceVector.x *= pEmitter->ForceAcceleration;
			pParticle->ForceVector.y *= pEmitter->ForceAcceleration;
			pParticle->ForceVector.z *= pEmitter->ForceAcceleration;
			pParticle = pParticle->Next;
		}

	}
}

void fxeSmokeEmitterClipAndDraw(fxeSmokeEmitterHandle pEmitter)
{
	fxeDrawVertex v1,v2,v3,v4;
  void *VertexPointers[4];
	int Loop;
	fxeSmokeParticle *pParticle;
	fxeTextureHandle TexHandle;
	fxeRectangle2d Rec2d;
	fxeRectCoordinate *RC;
	fxeDrawVertex WorldVertex, CameraVertex;
	float CurrQ;
	


	
	if (pEmitter->NumParticles == 0)
		return;

	WorldVertex.x = pEmitter->x;
	WorldVertex.y = pEmitter->y;
	WorldVertex.z = pEmitter->z;
	MatXDrawVerToDrawVer(CameraMatrix,&WorldVertex,&CameraVertex);
	pEmitter->cz = CameraVertex.z;

	pParticle = pEmitter->UsedParticles;

	if (GlobalFogMode != NO_FOGGING)
	{
    grFogMode(GR_FOG_DISABLE);
	}
	
	grColorCombine( GR_COMBINE_FUNCTION_SCALE_OTHER,
		GR_COMBINE_FACTOR_LOCAL,GR_COMBINE_LOCAL_CONSTANT,
		GR_COMBINE_OTHER_TEXTURE,FXFALSE);


	// always pass
	grDepthBufferFunction(GR_CMP_ALWAYS);


	grAlphaCombine( GR_COMBINE_FUNCTION_LOCAL_ALPHA,
    GR_COMBINE_FACTOR_ONE,GR_COMBINE_LOCAL_CONSTANT,
    GR_COMBINE_OTHER_TEXTURE,FXFALSE );
	

	// setup alpha blending to blend
	// source pixel with
	// destination pixel.
	grAlphaBlendFunction( GR_BLEND_SRC_ALPHA, GR_BLEND_ONE,
    GR_BLEND_ZERO, GR_BLEND_ZERO );

	TexHandle = pEmitter->TexHandle;

	SetupTexture(TexHandle);


	// Setup texture unit with particle texture:

	VertexPointers[0] = (void*)&v1;
	VertexPointers[1] = (void*)&v2;
	VertexPointers[2] = (void*)&v3;
	VertexPointers[3] = (void*)&v4;
  
	for (Loop=0; Loop<pEmitter->NumParticles; Loop++)
	{
		
		// Transform to camera space:
		WorldVertex.x = pParticle->x;
		WorldVertex.y = pParticle->y;
		WorldVertex.z = pParticle->z;

		MatXDrawVerToDrawVer(CameraMatrix,&WorldVertex,&CameraVertex);

		// Check if the particle need to be clipped:
		if ((CameraVertex.x + pParticle->Radius) < (-CameraVertex.z) ||
			(CameraVertex.x - pParticle->Radius) > CameraVertex.z ||
			(CameraVertex.y + pParticle->Radius) < (-CameraVertex.z) ||
			(CameraVertex.y - pParticle->Radius) > CameraVertex.z ||
			CameraVertex.z < NearClippingPlane || 
			CameraVertex.z > FarClippingPlane)
		{
			pParticle = pParticle->Next;
		}
		else
		{
			if ((CameraVertex.x + pParticle->Radius) < CameraVertex.z &&
				(CameraVertex.x - pParticle->Radius) > (-CameraVertex.z) &&
				(CameraVertex.y + pParticle->Radius) < CameraVertex.z &&
				(CameraVertex.y - pParticle->Radius) > (-CameraVertex.z))
			{

				v1.q = 1.0f/CameraVertex.z;      //pEmitter->cz;
				CurrQ = 1.0f/CameraVertex.z;
				v1.x = ((CameraVertex.x-pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				Snap(&v1.x);
				v1.y = ((CameraVertex.y+pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;
				Snap(&v1.y);
				v1.u0 = 0.0f;
				v1.v0 = 0.0f;
				v1.u1 = 0.0f;
				v1.v1 = 0.0f;

				v2.q = v1.q;
				v2.x = ((CameraVertex.x-pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				Snap(&v2.x);
				v2.y = ((CameraVertex.y-pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;
				Snap(&v2.y);
				v2.u0 = 0.0f;
				v2.v0 = 255.0f*v1.q;
				v2.u1 = 0.0f;
				v2.v1 = v2.v0;

				v3.q = v1.q;
				v3.x = ((CameraVertex.x+pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				Snap(&v3.x);
				v3.y = ((CameraVertex.y-pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;
				Snap(&v3.y);
				v3.u0 = 255.0f*v1.q;
				v3.v0 = v3.u0;
				v3.u1 = v3.u0;
				v3.v1 = v3.u0;

				v4.q = v1.q;
				v4.x = ((CameraVertex.x+pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				Snap(&v4.x);
				v4.y = ((CameraVertex.y+pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;
				Snap(&v4.y);
				v4.u0 = 255.0f*v1.q;
				v4.v0 = 0.0f;
				v4.u1 = v4.u0;
				v4.v1 = 0.0f;
			}
			else
			{
				// Project all vertices:

				v1.q = 1.0f/CameraVertex.z; //pEmitter->cz;
				CurrQ = 1.0f/CameraVertex.z;
				v1.x = ((CameraVertex.x-pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				v1.y = ((CameraVertex.y+pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;

				v2.q = v1.q;
				v2.x = ((CameraVertex.x-pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				v2.y = ((CameraVertex.y-pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;

				v3.q = v1.q;
				v3.x = ((CameraVertex.x+pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				v3.y = ((CameraVertex.y-pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;

				v4.q = v1.q;
				v4.x = ((CameraVertex.x+pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				v4.y = ((CameraVertex.y+pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;

				RC = Rec2d.RectCoordinates;
				RC[0].x = v1.x;
				RC[0].y = v1.y;
				RC[1].x = v2.x;
				RC[1].y = v2.y;
				RC[2].x = v3.x;
				RC[2].y = v3.y;
				RC[3].x = v4.x;
				RC[3].y = v4.y;

				ClipTextureRectangle2d(&Rec2d);

				v1.x = RC[0].x;
				v1.y = RC[0].y;
				Snap(&v1.x);
				Snap(&v1.y);
				v1.u0 = RC[0].u*255.0f*v1.q;
				v1.v0 = RC[0].v*255.0f*v1.q;
				v1.u1 = RC[0].u*255.0f*v1.q;
				v1.v1 = RC[0].v*255.0f*v1.q;

				v2.x = RC[1].x;
				v2.y = RC[1].y;
				Snap(&v2.x);
				Snap(&v2.y);
				v2.u0 = RC[1].u*255.0f*v1.q;
				v2.v0 = RC[1].v*255.0f*v1.q;
				v2.u1 = RC[1].u*255.0f*v1.q;
				v2.v1 = RC[1].v*255.0f*v1.q;

				v3.x = RC[2].x;
				v3.y = RC[2].y;
				Snap(&v3.x);
				Snap(&v3.y);
				v3.u0 = RC[2].u*255.0f*v1.q;
				v3.v0 = RC[2].v*255.0f*v1.q;
				v3.u1 = RC[2].u*255.0f*v1.q;
				v3.v1 = RC[2].v*255.0f*v1.q;

				v4.x = RC[3].x;
				v4.y = RC[3].y;
				Snap(&v4.x);
				Snap(&v4.y);
				v4.u0 = RC[3].u*255.0f*v1.q;
				v4.v0 = RC[3].v*255.0f*v1.q;
				v4.u1 = RC[3].u*255.0f*v1.q;
				v4.v1 = RC[3].v*255.0f*v1.q;


			}

			Performance.NumPolys++;

			grConstantColorValue((((unsigned int)pParticle->Alpha)<<24)+
				(64<<16)+(64<<8)+64);
	
			grDrawVertexArray(GR_TRIANGLE_FAN,4,VertexPointers);
			
			pParticle = pParticle->Next;
		}
		

	}

	grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
		GR_COMBINE_FACTOR_LOCAL,GR_COMBINE_LOCAL_ITERATED,
		GR_COMBINE_OTHER_TEXTURE,FXFALSE);
	
  grAlphaCombine(GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_ONE,
    GR_COMBINE_LOCAL_NONE,GR_COMBINE_OTHER_CONSTANT,FXFALSE);

	grConstantColorValue(255<<24);

  grAlphaBlendFunction(GR_BLEND_SRC_ALPHA,GR_BLEND_ONE_MINUS_SRC_ALPHA,
      GR_BLEND_ZERO,GR_BLEND_ZERO);

	grDepthBufferFunction(GR_CMP_LEQUAL);

	if (GlobalFogMode != NO_FOGGING)
	{
    grFogMode(GR_FOG_WITH_TABLE_ON_Q);
	}
}




void fxeSmokeEmitterChangeProperties(fxeSmokeEmitter *pEmitter, 
	float x, float y, float z, fxeVector *ForceVector)
{
	pEmitter->x = x;
	pEmitter->y = y;
	pEmitter->z = z;
	pEmitter->ForceVector.x = ForceVector->x;
	pEmitter->ForceVector.y = ForceVector->y;
	pEmitter->ForceVector.z = ForceVector->z;
}

void fxeSmokeEmitterChangeOpacity(fxeSmokeEmitter *pSmokeEmitter,
	int Opacity)
{
	pSmokeEmitter->Opacity = Opacity;
}










