
#include "fxglobals.h"
#include "FxMalloc.h"
#include <string.h>
#include "FxLog.h"

#define MEM_2MB 2097152

fxeTexture *pFirstTexture = NULL;

FxU32 StartAddressTmu0;
FxU32 StartAddressTmu1;
FxU32 DownloadAddressTmu0;
FxU32 DownloadAddressTmu1;
FxU32 EndAddressTmu0;
FxU32 EndAddressTmu1;



fxeTextureHandle CurrTexTmu0, CurrTexTmu1;

// Returns NULL if texture doesn't exist and else
// a handle to a texture already loaded.
fxeTextureHandle fxeTextureExist(char *Filename)
{
	fxeTexture *pTexture;

	if (pFirstTexture == NULL)
		return NULL;
	
	pTexture = pFirstTexture;

	while (pTexture != NULL)
	{
		if (pTexture->Filename != NULL)
		{
			if (strcmp(Filename,pTexture->Filename) == 0)
				return pTexture;
		}
		pTexture = pTexture->Next;
	}
	return NULL;
}

void fxeTextureCacheInit(void)
{
	fxeTexture *pTexture;
	pFirstTexture = NULL;

	StartAddressTmu0 = grTexMinAddress(GR_TMU0);
	EndAddressTmu0 = grTexMaxAddress(GR_TMU0);
	DownloadAddressTmu0 = StartAddressTmu0;

	if (NumTexelFx > 1)
	{
		StartAddressTmu1 = grTexMinAddress(GR_TMU1);
		EndAddressTmu1 = grTexMaxAddress(GR_TMU1);
		DownloadAddressTmu1 = StartAddressTmu1;
	}

	// allocate some memory in tmu0 for largest texture
	// possible. This space is needed if texture memory
	// is full and we need to use a texture that isn't
	// downloaded to tmu.

	pTexture = (fxeTexture*)FxMalloc(sizeof(fxeTexture));
	
	pTexture->TexInfo.format = GR_TEXFMT_RGB_565;
	pTexture->TexInfo.smallLodLog2 = GR_LOD_LOG2_256;
	pTexture->TexInfo.largeLodLog2 = GR_LOD_LOG2_1;
	pTexture->TexInfo.aspectRatioLog2 = GR_ASPECT_LOG2_1x1;

	pTexture->NumBytes = grTexTextureMemRequired(GR_MIPMAPLEVELMASK_BOTH,
		&pTexture->TexInfo);
	pTexture->NumBytesInTmu = pTexture->NumBytes;
	pTexture->Used = FXEFALSE;
	pTexture->DownloadedToTRAM = FXETRUE;
	pTexture->Changed = FXEFALSE;

	pTexture->TexType = FXE_TEXTURE_MAP;
	pTexture->VideomemoryAddress = fxeTMUAllocateMemory(GR_TMU0,
		pTexture->NumBytes);

	pTexture->Next = pFirstTexture;
	pFirstTexture = pTexture;

	// We must do the same thing for tmu1 if there is more tmu's than
	// one.
	if (NumTexelFx > 1)
	{
		// allocate some memory in tmu0 for largest texture
		// possible. This space is needed if texture memory
		// is full and we need to use a texture that isn't
		// downloaded to tmu.

		pTexture = (fxeTexture*)FxMalloc(sizeof(fxeTexture));
	
		pTexture->TexInfo.format = GR_TEXFMT_RGB_565;
		pTexture->TexInfo.smallLodLog2 = GR_LOD_LOG2_256;
		pTexture->TexInfo.largeLodLog2 = GR_LOD_LOG2_1;
		pTexture->TexInfo.aspectRatioLog2 = GR_ASPECT_LOG2_1x1;

		pTexture->NumBytes = grTexTextureMemRequired(GR_MIPMAPLEVELMASK_BOTH,
			&pTexture->TexInfo);
		pTexture->NumBytesInTmu = pTexture->NumBytes;
		pTexture->Used = FXEFALSE;
		pTexture->DownloadedToTRAM = FXETRUE;
		pTexture->Changed = FXEFALSE;

		pTexture->TexType = FXE_TEXTURE_SPECIAL;
		pTexture->VideomemoryAddress = fxeTMUAllocateMemory(GR_TMU1,
			pTexture->NumBytes);

		pTexture->Next = pFirstTexture;
		pFirstTexture = pTexture;

	}
	else
	{
		// allocate some memory in tmu1 for largest texture
		// possible. This space is needed if texture memory
		// is full and we need to use a texture that isn't
		// downloaded to tmu.

		pTexture = (fxeTexture*)FxMalloc(sizeof(fxeTexture));
	
		pTexture->TexInfo.format = GR_TEXFMT_RGB_565;
		pTexture->TexInfo.smallLodLog2 = GR_LOD_LOG2_256;
		pTexture->TexInfo.largeLodLog2 = GR_LOD_LOG2_1;
		pTexture->TexInfo.aspectRatioLog2 = GR_ASPECT_LOG2_1x1;

		pTexture->NumBytes = grTexTextureMemRequired(GR_MIPMAPLEVELMASK_BOTH,
			&pTexture->TexInfo);
		pTexture->NumBytesInTmu = pTexture->NumBytes;
		pTexture->Used = FXEFALSE;
		pTexture->Changed = FXEFALSE;
		pTexture->DownloadedToTRAM = FXETRUE;

		pTexture->TexType = FXE_TEXTURE_SPECIAL;
		pTexture->VideomemoryAddress = fxeTMUAllocateMemory(GR_TMU0,
			pTexture->NumBytes);

		pTexture->Next = pFirstTexture;
		pFirstTexture = pTexture;
	}
}


void PrepareTexturesForFrame(void)
{
	fxeTexture *pTexture;

	if (pFirstTexture == NULL)
		return;

	pTexture = pFirstTexture;

	while(pTexture != NULL)
	{
		pTexture->NumPolygons = 0;
		pTexture = pTexture->Next;
	}
}

void MakeUnusedTexturesUnused(void)
{
	fxeTexture *pTexture;

	if (pFirstTexture == NULL)
		return;

	pTexture = pFirstTexture;

	while(pTexture != NULL)
	{
		if (pTexture->NumPolygons == 0)
			pTexture->Used = FXEFALSE;
		pTexture = pTexture->Next;
	}
}	


void SetupTexture(fxeTextureHandle TexHandle)
{
	fxeTexture *pTexture;

	TexHandle->NumPolygons++;
	TexHandle->Used = FXETRUE;

	if (TexHandle->TexType == FXE_TEXTURE_MAP ||
		NumTexelFx == 1)
	{
		if (TexHandle->DownloadedToTRAM == FXETRUE)
		{
			// Setup texture unit 0's texture source
			grTexSource(GR_TMU0,TexHandle->VideomemoryAddress,
				GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);

			if (TexHandle->Changed == FXETRUE)
			{
				// The data of the texture have changed, we must download
				// it. 
				
				grTexDownloadMipMap(GR_TMU0,TexHandle->VideomemoryAddress,
					GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
				TexHandle->Changed = FXEFALSE;

			}

			return;
		}
		else
		{
		
			pTexture = pFirstTexture;

		
			// Search for a texture that is downloaded to vram
			// and that isn't used and has the same size as the
			// texture, if found then download our texture in the
			// same address of the tmu as the found texture. 
			while (pTexture != NULL)
			{
				if (pTexture->DownloadedToTRAM == FXETRUE &&
					pTexture->Used == FXEFALSE && pTexture->
						NumBytesInTmu >= TexHandle->NumBytes &&
					(pTexture->TexType == FXE_TEXTURE_MAP ||
					NumTexelFx == 1))
				{
					TexHandle->VideomemoryAddress = 
						pTexture->VideomemoryAddress;

					pTexture->DownloadedToTRAM = FXEFALSE;
					TexHandle->DownloadedToTRAM = FXETRUE;

					TexHandle->NumBytesInTmu = pTexture->NumBytesInTmu;
					TexHandle->Changed = FXEFALSE;

					grTexDownloadMipMap(GR_TMU0,
						TexHandle->VideomemoryAddress,
							GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);

					grTexSource(GR_TMU0,TexHandle->VideomemoryAddress,
						GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
				
					return;

				}
				pTexture = pTexture->Next;
			}


			pTexture = pFirstTexture;


			// Search for a texture that is downloaded to vram
			// and has the same size as the
			// texture
			while (pTexture != NULL)
			{
				if (pTexture->DownloadedToTRAM == FXETRUE &&
					pTexture->NumBytesInTmu >= TexHandle->NumBytes &&
					(pTexture->TexType == FXE_TEXTURE_MAP ||
					NumTexelFx == 1))
				{
					TexHandle->VideomemoryAddress = 
						pTexture->VideomemoryAddress;

					pTexture->DownloadedToTRAM = FXEFALSE;
					TexHandle->DownloadedToTRAM = FXETRUE;

					TexHandle->NumBytesInTmu = pTexture->NumBytesInTmu;
					TexHandle->Changed = FXEFALSE;

					grTexDownloadMipMap(GR_TMU0,
						TexHandle->VideomemoryAddress,
							GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);

					grTexSource(GR_TMU0,TexHandle->VideomemoryAddress,
						GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
					return;
				}
				pTexture = pTexture->Next;
			}
		}
	}
	else
	{
		if (TexHandle->DownloadedToTRAM == FXETRUE)
		{
				// Setup texture unit 1's texture source
				grTexSource(GR_TMU1,TexHandle->VideomemoryAddress,
					GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);

				if (TexHandle->Changed == FXETRUE)
				{
					// The data of the texture have changed, we must download
					// it. 
				
					grTexDownloadMipMap(GR_TMU1,TexHandle->VideomemoryAddress,
						GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
					TexHandle->Changed = FXEFALSE;
				}



			return;
		}
		else
		{
		
			pTexture = pFirstTexture;

		
			// Search for a texture that is downloaded to tram
			// and that isn't used and has the same size as the
			// texture, if found then download our texture in the
			// same address of the tmu as the found texture. 

				
			
			while (pTexture != NULL)
			{
				if (pTexture->DownloadedToTRAM == FXETRUE &&
					pTexture->Used == FXEFALSE && pTexture->
						NumBytesInTmu >= TexHandle->NumBytes &&
					pTexture->TexType == FXE_TEXTURE_SPECIAL)
				{
					TexHandle->VideomemoryAddress = 
						pTexture->VideomemoryAddress;

					pTexture->DownloadedToTRAM = FXEFALSE;
					TexHandle->DownloadedToTRAM = FXETRUE;

					TexHandle->NumBytesInTmu = pTexture->NumBytesInTmu;
					TexHandle->Changed = FXEFALSE;

					if (NumTexelFx > 1)
					{
						grTexDownloadMipMap(GR_TMU1,
							TexHandle->VideomemoryAddress,
								GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);

						grTexSource(GR_TMU1,TexHandle->VideomemoryAddress,
							GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
					}
					else
					{
						grTexDownloadMipMap(GR_TMU0,
							TexHandle->VideomemoryAddress,
								GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);

						grTexSource(GR_TMU0,TexHandle->VideomemoryAddress,
							GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
					}
				
					return;

				}
				pTexture = pTexture->Next;
			}


			pTexture = pFirstTexture;


			// Search for a texture that is downloaded to tram
			// and has the same size as the
			// texture
			while (pTexture != NULL)
			{
				if (pTexture->DownloadedToTRAM == FXETRUE &&
					pTexture->NumBytesInTmu >= TexHandle->NumBytes &&
					pTexture->TexType == FXE_TEXTURE_SPECIAL)
				{
					TexHandle->VideomemoryAddress = 
						pTexture->VideomemoryAddress;

					pTexture->DownloadedToTRAM = FXEFALSE;
					TexHandle->DownloadedToTRAM = FXETRUE;

					TexHandle->NumBytesInTmu = pTexture->NumBytesInTmu;
					TexHandle->Changed = FXEFALSE;

					grTexDownloadMipMap(GR_TMU1,
						TexHandle->VideomemoryAddress,
						GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);

					grTexSource(GR_TMU1,TexHandle->VideomemoryAddress,
						GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);

					return;
				}
				pTexture = pTexture->Next;
			}
		}

	}
}


// Returns 0xFFFFFFFF if failure and address in tmu if success. 

FxU32 fxeTMUAllocateMemory(GrChipID_t tmu, FxU32 Bytes)
{
	if (tmu == GR_TMU0)
	{
		if ((DownloadAddressTmu0+Bytes) <=
			EndAddressTmu0)
		{
			if (DownloadAddressTmu0 < (MEM_2MB) &&
				(DownloadAddressTmu0+Bytes) > (MEM_2MB))
				DownloadAddressTmu0 = MEM_2MB;
			DownloadAddressTmu0 += Bytes;
			return DownloadAddressTmu0 - Bytes;
		}
		else
			return 0xffffffff;
	}
	else if (tmu == GR_TMU1 && NumTexelFx > 1)
	{
		if ((DownloadAddressTmu1+Bytes) <=
			EndAddressTmu1)
		{
			if (DownloadAddressTmu1 < (MEM_2MB) &&
				(DownloadAddressTmu1+Bytes) > (MEM_2MB))
				DownloadAddressTmu1 = MEM_2MB;

			DownloadAddressTmu1 += Bytes;
			return DownloadAddressTmu1 - Bytes;
		}
		else
			return 0xffffffff;
	}
	else
		return 0xffffffff;
}


// Loads a 3df texture file from disk, returns
// NULL if failure and handle to texture if
// successful.

fxeTextureHandle fxeTextureLoad(char *filename, char TexType)
{
	Gu3dfInfo TexInfo;
  fxeTexture *Texture;
	char TexelFxNo;

	FxU32 DownloadAddr;

	if ((Texture = fxeTextureExist(filename)) != NULL)
		return Texture;



	

	// get the texture info:
  if (gu3dfGetInfo(filename,&TexInfo) == FXFALSE)
    return NULL;

	fxeLogDeviceWriteString("Loading texture: ");
	fxeLogDeviceWriteString(filename);
	fxeLogDeviceWriteString("\n\n");

  // allocate memory for the mipmap:
  TexInfo.data = (void*)FxMalloc(TexInfo.mem_required);
  if (TexInfo.data == NULL)
    return NULL;

  // Load the mipmap from disk:
  gu3dfLoad(filename,&TexInfo);


	Texture = (fxeTexture*)FxMalloc(sizeof(fxeTexture));
	if (Texture == NULL)
	{
    free(TexInfo.data);
		return NULL;
	}

	Texture->Filename = (char*)FxMalloc(strlen(filename)+1);
	strcpy(Texture->Filename,filename);

  Texture->TexInfo.data = TexInfo.data;
	Texture->TexInfo.format = TexInfo.header.format;
  Texture->TexInfo.aspectRatioLog2 = TexInfo.header.aspect_ratio;
  Texture->TexInfo.smallLodLog2 = TexInfo.header.small_lod;
  Texture->TexInfo.largeLodLog2 = TexInfo.header.large_lod;

	Texture->NumBytes = TexInfo.mem_required;
	Texture->NumBytesInTmu = TexInfo.mem_required;
	Texture->TexType = TexType;
	Texture->Used = FXETRUE;

	if (TexType == FXE_TEXTURE_MAP)
	{
		DownloadAddr = fxeTMUAllocateMemory(GR_TMU0,TexInfo.mem_required);
		TexelFxNo = GR_TMU0;
	}
	else
	{
		if (NumTexelFx > 1)
		{
			DownloadAddr = fxeTMUAllocateMemory(GR_TMU1,TexInfo.mem_required);
			TexelFxNo = GR_TMU1;
		}
		else
		{
			DownloadAddr = fxeTMUAllocateMemory(GR_TMU0,TexInfo.mem_required);
			TexelFxNo = GR_TMU0;
		}
	}

	if (DownloadAddr != 0xffffffff) 
	{
		// Download texture to tmu0:
		grTexDownloadMipMap(TexelFxNo,DownloadAddr,
			GR_MIPMAPLEVELMASK_BOTH,&Texture->TexInfo);
		Texture->DownloadedToTRAM = FXETRUE;
		Texture->VideomemoryAddress = DownloadAddr;
	}
	else
	{
		Texture->DownloadedToTRAM = FXEFALSE;
	}

	Texture->Next = pFirstTexture;
	pFirstTexture = Texture;
	return Texture;
}
