

#include "dynlight.h"
#include "fxenginedrv.h"
#include "FxMalloc.h"

fxeDynLightsource *DynamicLightsources = NULL;
fxeDynLightsource *FirstDynLightsource = NULL;


void fxeDynLightsourceChangeRadius(fxeDynLightsourceHandle DLSH, float NewRadius)
{
	if (DLSH->Radius != NewRadius)
	{
		DLSH->Radius = NewRadius;
		DLSH->Oor = 1.0f/NewRadius;
		DLSH->Changed = FXETRUE;
	}
}

void fxeDynLightsourceChangeRGB(fxeDynLightsourceHandle DLSH, float NewR, float NewG, float NewB)
{
	if (NewR == DLSH->r && NewG == DLSH->g && NewB == DLSH->b)
		return;
	else
	{
		DLSH->Changed = FXETRUE;
		DLSH->r = NewR*31.0f;
		DLSH->g = NewG*31.0f;
		DLSH->b = NewB*31.0f;
		DLSH->RDiv31 = NewR;
		DLSH->GDiv31 = NewG;
		DLSH->BDiv31 = NewB;
	}
}

void fxeDynLightsourceChangeXYZ(fxeDynLightsourceHandle DLSH, float NewX, float NewY, float NewZ)
{
	if (NewX == DLSH->x && NewY == DLSH->y && NewZ == DLSH->z)
		return;
	else
	{
		DLSH->Changed = FXETRUE;
		DLSH->HaveMoved = FXETRUE;
		DLSH->x = NewX;
		DLSH->y = NewY;
		DLSH->z = NewZ;
	}
}

void fxeDynLightsourceActivate(fxeDynLightsourceHandle DLSH)
{
	DLSH->Active = FXETRUE;
	DLSH->Changed = FXETRUE;
}

void fxeDynLightsourceDisable(fxeDynLightsourceHandle DLSH)
{
	DLSH->Active = FXEFALSE;
	DLSH->Changed = FXETRUE;
}



fxeDynLightsourceHandle fxeDynLightsourceAdd(float Radius, float x, float y, float z, 
	float r, float g, float b)
{
	fxeDynLightsourceHandle DLSH;

	DLSH = (fxeDynLightsourceHandle)FxMalloc(sizeof(fxeDynLightsource));
	DLSH->Radius = Radius;
	DLSH->x = x;
	DLSH->y = y;
	DLSH->z = z;
	DLSH->Oor = 1.0f/Radius;
	DLSH->r = r*31.0f;
	DLSH->g = g*31.0f;
	DLSH->b = b*31.0f;
	DLSH->RDiv31 = r;
	DLSH->GDiv31 = g;
	DLSH->BDiv31 = b;
	DLSH->Active = FXETRUE;
	DLSH->Changed = FXETRUE;
	DLSH->HaveMoved = FXETRUE;

	if (DynamicLightsources == NULL)
	{
		DynamicLightsources = DLSH;
		DynamicLightsources->Next = NULL;
		DynamicLightsources->Prev = NULL;
		FirstDynLightsource = DLSH;
	}
	else
	{
		DLSH->Prev = NULL;
		DLSH->Next = FirstDynLightsource;
		FirstDynLightsource = DLSH;
	}
	return FirstDynLightsource;
}

void fxeDynLightsourcesMakeUnchanged(void)
{
	if (FirstDynLightsource == NULL)
		return;

	DynamicLightsources = FirstDynLightsource;
	while (DynamicLightsources != NULL)
	{
		DynamicLightsources->Changed = FXEFALSE;
		DynamicLightsources = DynamicLightsources->Next;
	}
}

void fxeDynLightsourcesMakeUnMoved(void)
{
	if (FirstDynLightsource == NULL)
		return;

	DynamicLightsources = FirstDynLightsource;
	while (DynamicLightsources != NULL)
	{
		DynamicLightsources->HaveMoved = FXEFALSE;
		DynamicLightsources = DynamicLightsources->Next;
	}
}