#include "D_Engine.h"
//light.cpp

// All lightsources in world
D_Light *Lightsources;
// Ambient light level.
D_Color Ambient;
// Ambient light related functions:
D_Light::D_Light()
{
	NumFlares = 0;
}
void D_AmbientLight(float r, float g, float b)
{
  Ambient.r = r;
  Ambient.g = g;
  Ambient.b = b;
}

// Lightsource related functions:
void D_DrawLights()
{
	if(Lightsources == NULL) return;
	D_Light *Lights;
	Lights = Lightsources;
	grDepthMask(FXFALSE);
	//CalculateCameraTransformationMatrix();
	while(Lights != NULL)
	{

		if(Lights->Visible == D_TRUE)
		{
			//Lights->D_LightsourceSetUpDraw();  //do glide preparations
			Lights->D_LightsourceDraw();
			if(Lights->NumFlares)
			{
				if((Lights->Light.ScreenPos.x >= 0.0f && Lights->Light.ScreenPos.x <= HorRes) &&
					(Lights->Light.ScreenPos.y >= 0.0f && Lights->Light.ScreenPos.y <= VerRes))
					Lights->D_DrawLensFlare(Lights->Light.ScreenPos);
			}
		}
		Lights = Lights->Next;
	}
	grDepthMask(FXTRUE);
}

// inserts it into the
// linked list of lightsources
void D_Light::D_LightsourceAdd()
{
	D_Light *Lightsource;

	Lightsource = Lightsources;

	Next = NULL;
	// go to last lightsource in linked list of lightsources
	if (Lightsources == NULL)
	{
		// This is the first lightsource
		Prev = NULL;
	    Lightsources = this;
	}
	else
	{
		while (Lightsource->Next != NULL)
		{
			Lightsource = Lightsource->Next;
		}
	    Prev = Lightsource;
        Lightsource->Next = this;
	    //Lightsources = Lightsource;
	}
	//return Lightsources;
}

void D_Light::D_LightsourceInit(int mode, int shadetype)
{
	Lit = D_FALSE;
	Visible = D_FALSE;

	if(mode&LIT)  //we render the light's influence
		Lit = D_TRUE;    
	if(mode&VISIBLE)  //we draw the light
		Visible = D_TRUE;


	if(shadetype&NORMAL_SHADING)
		ShadingType = NORMAL_SHADING;
	else if(shadetype&DISTANCE_SHADING)
		ShadingType = DISTANCE_SHADING;

//	Pos = (D_Vector *)malloc(sizeof(D_Vector));
}


void D_Light::D_LightsourceMove(float x, float y, float z)
{
	Pos.x = x;
	Pos.y = y;
	Pos.z = z;
	Pos.w = 1.0f;
	Light.WorldPos.x = x;
	Light.WorldPos.y = y;
	Light.WorldPos.z = z;
	Light.WorldPos.w = 1.0f;
}

void D_Light::D_LightsourceNormal(float x, float y, float z)
{
	Normal.x = x;
	Normal.y = y;
	Normal.z = z;
}

void D_Light::D_LightsourceLight(float r, float g, float b, float dist)
{
	Color.r = r;
	Color.g = g;
	Color.b = b;
	Light.Color.r = r;
	Light.Color.g = g;
	Light.Color.b = b;
	Distance = dist;
}


void D_Light::D_LightsourceScale(float scale)
{
	Light.Scale = scale;
}

void D_Light::D_LightsourceTexture(char *filename)
{
	Light.Texture = D_TextureLoad(filename);
}


// Note that if LightsourceHandle isn't a valid lightsource undefined
// results may occure such as a halting system crash!

void D_Light::D_LightsourceDelete()
{
	if ((Prev == NULL) && (Next == NULL))
	{
		Lightsources = NULL;
	}
	else if (Prev == NULL)
	{
		Lightsources = Next;
		Lightsources->Prev = NULL;
	}
	else if(Next == NULL)
	{
		Prev->Next = NULL;
	}
	else
	{
		Prev->Next = Next;
		Next->Prev = Prev;
	}
	Next = NULL;
	Prev = NULL;

}

void D_AllLightsourceDelete()
{
	if(Lightsources == NULL) return;

	D_Light *Nextl;
	D_Light *Light;
	Nextl = Lightsources;
	
	while(Nextl != NULL)
	{
		Light = Nextl;
		Nextl = Light->Next;
		Light->D_LightsourceDelete();
	}
} 



void D_Light::D_LightsourceDraw()
{
	Light.SetUpDraw();
	Light.Draw();
}
 
void D_Light::D_InitLensFlare(int num)
{
//allocate memory
	NumFlares = num;
	LensFlare = (D_LensFlare **)malloc(sizeof(D_LensFlare *)*(NumFlares));
	for(int n = 0; n < num; n++)
	{
		//allocate and set up each one
		LensFlare[n] = (D_LensFlare *)malloc(sizeof(D_LensFlare));
		LensFlare[n]->size = 40.0f + RandomNum()*30.0f;
		LensFlare[n]->distance = RandomNum()*2.0f - 1.0f;  //percentage 
		LensFlare[n]->Texture = D_TextureLoad("flare2.3df");
	}


}
void D_Light::D_EditFlare(char *texturename, float pixsize, float dist)
{


}


void fxeLightsourceDrawLensflare(fxeLightsource *pLightsource)
{
	fxeVertex VertexCamera;
	fxeVertex VertexWorld;
	fxeDrawVertex ProjectedVertex;
	float VecLen;
	fxeVector LightVector;
	fxeVector CameraVector;
	float Dp3d;
	int Loop;
	float WindowCenterX, WindowCenterY;
	float FlareRelX, FlareRelY;
	float FlareX, FlareY;
	fxeDrawVertex Lensflare[4];

	if (pLightsource->NumLensflares == 0) 
		return;

	// We must transform lightsource to camera coordinates:
	VertexWorld.x = pLightsource->x;
	VertexWorld.y = pLightsource->y;
	VertexWorld.z = pLightsource->z;

	MatXVer(CameraMatrix,&VertexWorld,&VertexCamera);


	// We don't want divide by zero...
	if (VertexCamera.z <= NearClippingPlane)
		return;

	// Project the camera vertex so we can clip the lightsource in
	// 2D.

	ProjectedVertex.q = 1.0f/VertexCamera.z;

	ProjectedVertex.x = (VertexCamera.x*ProjectedVertex.q+1.0f)*
		ProjectWidthDiv2+ProjectStartX;
	Snap(&ProjectedVertex.x);

	ProjectedVertex.y = (VertexCamera.y*ProjectedVertex.q+1.0f)*
		ProjectHeightDiv2+ProjectStartY;
	Snap(&ProjectedVertex.y);


	// Don't use this flare if its outside the render window:
	if (ProjectedVertex.x < ProjectStartX || ProjectedVertex.x >
		ProjectEndX || ProjectedVertex.y < ProjectStartY ||
		ProjectedVertex.y > ProjectEndY)
		return;

	// As this is only made once per frame for a limited amount of
	// lightsources we can perform some advanced code here.
	
	// We need to create a normalized vector from the origin of 
	// camera space to the lightsource position in camera space.
	// Aahhh, this is simple because we already have the vector,
	// we just need to make its lenght 1.0.


	LightVector.x = VertexCamera.x;
	LightVector.y = VertexCamera.y;
	LightVector.z = VertexCamera.z;

	VecLen = VectorLenght(&LightVector);

	LightVector.x /= VecLen;
	LightVector.y /= VecLen;
	LightVector.z /= VecLen;

	CameraVector.x = 0.0f;
	CameraVector.y = 0.0f;
	CameraVector.z = 1.0f;
	
	Dp3d = DotProduct3d(&CameraVector,&LightVector);

	// Now Dp3d should be a value from 0..1. 1 means the lightsource
	// is just in front of the camera and should be drawn with maximum
	// intensity. This works perfect for lightsources with very long
	// distance such as the sun but not very well lightsources with
	// a limited distance.

	Alpha = 255.0f*Dp3d;

	WindowCenterX = ProjectWidthDiv2+ProjectStartX;
	WindowCenterY = ProjectHeightDiv2+ProjectStartY;
	FlareRelX = ProjectedVertex.x-WindowCenterX;
	FlareRelY = ProjectedVertex.y-WindowCenterY;

	
	grDepthBufferMode(GR_DEPTHBUFFER_DISABLE);
	grTexFilterMode(GR_TMU0,GR_TEXTUREFILTER_POINTSAMPLED,
		GR_TEXTUREFILTER_POINTSAMPLED);
	if (NumTexelFx > 1) 
		grTexFilterMode(GR_TMU1,GR_TEXTUREFILTER_POINTSAMPLED,
			GR_TEXTUREFILTER_POINTSAMPLED);

	//alpha blending stuff...
  grColorCombine( GR_COMBINE_FUNCTION_SCALE_OTHER,
                       GR_COMBINE_FACTOR_LOCAL_ALPHA,
                       GR_COMBINE_LOCAL_NONE,
                       GR_COMBINE_OTHER_TEXTURE,
                       FXFALSE );
  grAlphaCombine( GR_COMBINE_FUNCTION_LOCAL_ALPHA,
                        GR_COMBINE_FACTOR_NONE,
                        GR_COMBINE_LOCAL_CONSTANT,
                        GR_COMBINE_OTHER_NONE,
                        FXFALSE );
	grConstantColorValue(((int)Alpha)<<24);
	grAlphaBlendFunction( GR_BLEND_ONE, GR_BLEND_ONE,GR_BLEND_ZERO,GR_BLEND_ZERO );
  grAlphaTestFunction( GR_CMP_ALWAYS );

	//this is where the magic happens...
	

	for (Loop = 0; Loop<pLightsource->NumLensflares; Loop++)
	{
		FlareX = pLightsource->Lensflares[Loop].MulFactor *
			FlareRelX + WindowCenterX;
		FlareY = pLightsource->Lensflares[Loop].MulFactor *
			FlareRelY + WindowCenterY;

		// Upper left corner:

		Lensflare[0].q = 1.0f;
		Lensflare[0].x = FlareX - pLightsource->Lensflares[Loop].Radius;
		Lensflare[0].y = FlareY + pLightsource->Lensflares[Loop].Radius;
		Lensflare[0].a = Alpha;
		Lensflare[0].u0 = 0.0f;
		Lensflare[0].u1 = 0.0f;
		Lensflare[0].v0 = 0.0f;
		Lensflare[0].v1 = 0.0f;

		// Down left corner:

		Lensflare[1].q = 1.0f;
		Lensflare[1].x = FlareX - pLightsource->Lensflares[Loop].Radius;
		Lensflare[1].y = FlareY - pLightsource->Lensflares[Loop].Radius;
		Lensflare[1].a = Alpha;
		Lensflare[1].u0 = 0.0f;
		Lensflare[1].u1 = 0.0f;
		Lensflare[1].v0 = 255.0f;
		Lensflare[1].v1 = 255.0f;

		// Down Right corner:
	
		Lensflare[2].q = 1.0f;
		Lensflare[2].x = FlareX + pLightsource->Lensflares[Loop].Radius;
		Lensflare[2].y = FlareY - pLightsource->Lensflares[Loop].Radius;
		Lensflare[2].a = Alpha;
		Lensflare[2].u0 = 255.0f;
		Lensflare[2].u1 = 255.0f;
		Lensflare[2].v0 = 255.0f;
		Lensflare[2].v1 = 255.0f;

		// Upper right corner:

		Lensflare[3].q = 1.0f;
		Lensflare[3].x = FlareX + pLightsource->Lensflares[Loop].Radius;
		Lensflare[3].y = FlareY + pLightsource->Lensflares[Loop].Radius;
		Lensflare[3].a = Alpha;
		Lensflare[3].u0 = 255.0f;
		Lensflare[3].u1 = 255.0f;
		Lensflare[3].v0 = 0.0f;
		Lensflare[3].v1 = 0.0f;

		// Setup tmu[s] with texture of lensflare:
		SetupTmusWithTexture
			(pLightsource->Lensflares[Loop].LensTexture);

		grDrawTriangle(&Lensflare[0],&Lensflare[1],&Lensflare[2]);
		grDrawTriangle(&Lensflare[2],&Lensflare[3],&Lensflare[0]);
	}

	// Restore default glide state for FxEngine:

	grTexFilterMode(GR_TMU0,GR_TEXTUREFILTER_BILINEAR,
		GR_TEXTUREFILTER_BILINEAR);
	if (NumTexelFx > 1) 
		grTexFilterMode(GR_TMU1,GR_TEXTUREFILTER_BILINEAR,
			GR_TEXTUREFILTER_BILINEAR);

  grColorCombine( GR_COMBINE_FUNCTION_SCALE_OTHER,
                       GR_COMBINE_FACTOR_LOCAL,
                       GR_COMBINE_LOCAL_ITERATED,
                       GR_COMBINE_OTHER_TEXTURE,
                       FXFALSE );
  grAlphaCombine( GR_COMBINE_FUNCTION_LOCAL_ALPHA,
                        GR_COMBINE_FACTOR_NONE,
                        GR_COMBINE_LOCAL_CONSTANT,
                        GR_COMBINE_OTHER_NONE,
                        FXFALSE );
	grConstantColorValue(255<<24);
	grAlphaBlendFunction( GR_BLEND_ONE, GR_BLEND_ZERO,GR_BLEND_ZERO,GR_BLEND_ZERO );
	grDepthBufferMode(GR_DEPTHBUFFER_WBUFFER);
}

	

void D_Light::D_DrawLensFlare(D_Point2d p) //screen position of the light source
{


	int distx = (HalfHorRes) - p.x;
	int disty = (HalfVerRes) - p.y;
	float alpha = p.oow * 255.0f;   //the farther away, the less the lens flare
	if(alpha < 5.0f) return;   // far clip
	
	GrVertex vtx[4];  //vertexes of the polygons we are going to calculate and draw
	GrState oldState;
	grGlideGetState(&oldState);   //save the state of the alpha stuff so we can restore 
	
	//alpha blending stuff...
    grTexCombineFunction(GR_TMU0, GR_TEXTURECOMBINE_DECAL); //needed
    grColorCombine( GR_COMBINE_FUNCTION_SCALE_OTHER,
                        GR_COMBINE_FACTOR_LOCAL_ALPHA,
                        GR_COMBINE_LOCAL_NONE,
                        GR_COMBINE_OTHER_TEXTURE,
                        FXFALSE );
    grAlphaCombine( GR_COMBINE_FUNCTION_SCALE_OTHER,
                        GR_COMBINE_FACTOR_ONE,
                        GR_COMBINE_LOCAL_NONE,
                        GR_COMBINE_OTHER_TEXTURE,
                        FXFALSE );
    guAlphaSource( GR_ALPHASOURCE_ITERATED_ALPHA );
	grAlphaBlendFunction( GR_BLEND_ONE, GR_BLEND_ONE,GR_BLEND_ZERO,GR_BLEND_ZERO );
    grAlphaTestFunction( GR_CMP_ALWAYS );

	//this is where the magic happens...
	
	for(int n = 0; n < NumFlares; n++)
	{
		if (GlobalDetailLevel == LOW_DETAIL)
		{
			 grTexSource(GR_TMU0,LensFlare[n]->Texture->VideoMemoryAddressTMU0,
				  GR_MIPMAPLEVELMASK_BOTH,&LensFlare[n]->Texture->TexInfo);

		}

		else if (GlobalDetailLevel == HIGH_DETAIL)
		{
            // Setup texture unit 0's texture source
			grTexSource(GR_TMU0,LensFlare[n]->Texture->VideoMemoryAddressTMU0,
				GR_MIPMAPLEVELMASK_EVEN,&LensFlare[n]->Texture->TexInfo);

            // Setup texture unit 1's texture source
			grTexSource(GR_TMU1,LensFlare[n]->Texture->VideoMemoryAddressTMU1,
				GR_MIPMAPLEVELMASK_ODD,&LensFlare[n]->Texture->TexInfo);
		}

		float flare_x_pos = (HalfHorRes) - float(distx*LensFlare[n]->distance);
		float flare_y_pos = (HalfVerRes) - float(disty*LensFlare[n]->distance);
		

		vtx[0].x = (float)flare_x_pos - (LensFlare[n]->size);  //upper left
		vtx[0].y = (float)flare_y_pos - (LensFlare[n]->size);  
		vtx[0].oow = 1.0f;
		vtx[0].a = alpha;
		vtx[0].tmuvtx[0].sow = 0.0f; //texture S coord
		vtx[0].tmuvtx[0].tow = 0.0f; //texture T coord

		vtx[1].x = (float)flare_x_pos + (LensFlare[n]->size); //upper right
		vtx[1].y = (float)flare_y_pos - (LensFlare[n]->size); 
		vtx[1].oow = 1.0f;
		vtx[1].a = alpha;
		vtx[1].tmuvtx[0].sow = 255.0f;
		vtx[1].tmuvtx[0].tow = 0.0f;


		vtx[2].x = (float)flare_x_pos - (LensFlare[n]->size);  //lower left
		vtx[2].y = (float)flare_y_pos + (LensFlare[n]->size);  
		vtx[2].oow = 1.0f;
		vtx[2].a = alpha;
		vtx[2].tmuvtx[0].sow = 0.0f;
		vtx[2].tmuvtx[0].tow = 255.0f;


		vtx[3].x = (float)flare_x_pos + (LensFlare[n]->size);  //lower right
		vtx[3].y = (float)flare_y_pos + (LensFlare[n]->size);  
		vtx[3].oow = 1.0f;
		vtx[3].a = alpha;
		vtx[3].tmuvtx[0].sow = 255.0f;
		vtx[3].tmuvtx[0].tow = 255.0f;
		
		grDrawTriangle( &vtx[0], &vtx[3], &vtx[2]);   //bottom left most triangle
		grDrawTriangle( &vtx[0], &vtx[1], &vtx[3]);   //top rightmost triangle
	
	
	}
	grGlideSetState(&oldState);   //restore the state of the alpha stuff


}






