

#include "fxenginedrv.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "fxlandscapedrv.h"
#include "fxglobals.h"
#include "FxMalloc.h"


// S hr skapas en height map, frst skapas en heightmap i hgupplsning
// (65x65), sen skapas en i lgre upplsning (17*17), den i lgre upplsning
// tar hjdinfo frn den hgupplsta heightmapen, sen skapas en heightmap
// med nnu lgre upplsning (5*5) den tar hjdinfo frn den nst hgst
// upplsta heightmappen.
// Den hgst upplsta heightmapen skapas frn en formel som placerar
// ut en bestmt antal rektanglar med hjdinformation. Varje rectangels
// position bestms av slumpade tal medan storleken p
// rectanglarna bestms av en parameter. Hjdinformationen fr varje
// rectangel ges av fljande formel: Nuvarandevrde * 2
// Nr en rektangel skrivs
// i heightmapen multiplicaras dess hjdvrde till nuvarande hjdvrde i heightmapen
// Innan denna rutin anropas mste heightmappens alla element sttas till 1.

int TotLs = 0;
fxeVector *pLightNormal;
float GlobalLightIntensity;

// This function returns the height of the landscape relative to the origin of the world
// at the specified x,z value in world space. X and Z must be in the area
// of the landscape of the world. If successful function returns the height and -1
// if failure.
float fxeLandscapeGetHeightXZ(fxeLandscapesParentHandle LspH, float x, float z)
{
	int ix,iz;
	float PolygonWidth;
	float PolygonDepth;
	
	x+=(LspH->Width/2.0f);
	z=-z;
	z+=(LspH->Depth/2.0f);

	PolygonWidth = LspH->Width/LspH->NumPolysX;
	PolygonDepth = LspH->Depth/LspH->NumPolysX;
	
	ix = (int)(x/PolygonWidth);
	iz = (int)(z/PolygonDepth);
	
	if (ix >= 0 && ix < LspH->NumPolysX && iz >= 0 && iz <= LspH->NumPolysX)
		return LspH->HeightBuffer[iz*LspH->NumPolysX+ix];
	else
		return -1.0f;
}

void fxeFreeHeightMap(fxeHeightmapInfoPtr pHeightmap)
{
  free(pHeightmap->MapData);
}

fxeNormalBufferPtr fxeCreateNormalBuffer(int BufferWidth)
{
  int Loop;
  fxeNormalBuffer *pNormalBuffer;

  pNormalBuffer = (fxeNormalBuffer*)FxMalloc(sizeof(fxeNormalBuffer));
  if (pNormalBuffer == NULL)
    return NULL;

  pNormalBuffer->Normals = (fxeVector*)FxMalloc(sizeof(fxeVector)*
    (BufferWidth*BufferWidth));

  if (pNormalBuffer->Normals == NULL)
    return NULL;
    
  pNormalBuffer->Width = BufferWidth;

  for (Loop=0; Loop<(BufferWidth*BufferWidth); Loop++)
  {
    pNormalBuffer->Normals[Loop].x = 0.0;
    pNormalBuffer->Normals[Loop].y = 0.0;
    pNormalBuffer->Normals[Loop].z = 0.0;
  }
  return pNormalBuffer;
}

void fxeFreeNormalBuffer(fxeNormalBufferPtr pNormalBuffer)
{
  free(pNormalBuffer->Normals);
}

fxeHeightmapInfoPtr fxeCreateHeightMap2(int MapWidth, void *Data,
	float Height)
{
  int Loop, Loop2;
  fxeHeightmapInfoPtr pHeightmap;
	float HeightValue;
	unsigned char *CharData = (unsigned char*)Data;

  pHeightmap = (fxeHeightmapInfoPtr)FxMalloc(sizeof(fxeHeightmapInfo));
  if (pHeightmap == NULL)
    return NULL;

  // Create high resolution heightmap:
  pHeightmap->Width = MapWidth;

  // allocate memory for MapData:
  pHeightmap->MapData = (float *)FxMalloc(sizeof(float)*(MapWidth*MapWidth));
  if (pHeightmap->MapData == NULL)
    return NULL;

  // Set all elements to 0:
  for (Loop=0; Loop < (MapWidth*MapWidth); Loop++)
  {
    pHeightmap->MapData[Loop] = 0.0f;
  }

	for (Loop=0; Loop<(MapWidth-1); Loop++)
	{
		for (Loop2=0; Loop2<(MapWidth-1); Loop2++)
		{
			HeightValue = (((float)(CharData[Loop*(MapWidth-1)+Loop2]))/255.0f)*Height;
			pHeightmap->MapData[Loop*MapWidth+Loop2] = HeightValue;
		}
	}
	return pHeightmap;
}


fxeHeightmapInfoPtr fxeCreateHeightMap(
  int NumRecs, int RecWidth, int RecHeight, int MapWidth)
{
  int Loop, Loop2, Loop3;
  int RecX, RecY, RecX2, RecY2;
  fxeHeightmapInfoPtr pHeightmap;

  pHeightmap = (fxeHeightmapInfoPtr)FxMalloc(sizeof(fxeHeightmapInfo));
  if (pHeightmap == NULL)
    return NULL;

  // Create high resolution heightmap:
  pHeightmap->Width = MapWidth;

  // allocate memory for MapData:
  pHeightmap->MapData = (float *)FxMalloc(sizeof(float)*(MapWidth*MapWidth));
  if (pHeightmap->MapData == NULL)
    return NULL;

  // Set all elements to 1:
  for (Loop=0; Loop < (MapWidth*MapWidth); Loop++)
  {
    pHeightmap->MapData[Loop] = 1.0;
  }


  for (Loop=0; Loop<NumRecs; Loop++)
  {
    RecX = (rand()%(MapWidth-2));
    RecY = (rand()%(MapWidth-2));
    RecX2 = RecX+10+rand()%RecWidth;
    RecY2 = RecY+10+rand()%RecWidth;
    if (RecX2 > (MapWidth-1))
      RecX2 = MapWidth-1;
    if (RecY2 > (MapWidth-1))
      RecY2 = MapWidth-1;

    for (Loop2=RecY; Loop2<RecY2; Loop2++)
    {
      for (Loop3=RecX; Loop3<RecX2; Loop3++)
      {
        pHeightmap->MapData[Loop2*MapWidth+Loop3] = pHeightmap->MapData
          [Loop2*MapWidth+Loop3] + (float)(rand()%RecHeight);
      }
    }
  }
  return pHeightmap;
}

void CalculateNormalTriangle(fxeLDrawVertex *pVer1, fxeLDrawVertex* pVer2,
  fxeLDrawVertex *pVer3, fxeVector *pNormal)
{
  fxeVector Vec1, Vec2;

  Vec1.x = pVer3->x - pVer1->x;
  Vec1.y = pVer3->y - pVer1->y;
  Vec1.z = pVer3->z - pVer1->z;

  Vec2.x = pVer2->x - pVer1->x;
  Vec2.y = pVer2->y - pVer1->y;
  Vec2.z = pVer2->z - pVer1->z;

  CrossProduct3d(pNormal,&Vec1,&Vec2);
}

void CalculateVertexIntensity(fxeVector *pVertexNormal, 
	fxeColor *pColor)
{
  //int Finished;
  //fxeVector LightVector;
  //float LightVectorLenght;
  float NormalVectorLenght;
  float Dp3d;
  float ReflectionIntensity;
  //float DistanceAffection;
  float TotalShadeR, TotalShadeG, TotalShadeB;


  NormalVectorLenght = VectorLenght(pVertexNormal);
  if (NormalVectorLenght == 0.0)
    NormalVectorLenght = 0.0001f;

  pColor->r = 0.0;
  pColor->g = 0.0;
  pColor->b = 0.0;

			/*
      // Calculate reflection level of this vertex
      // with the current lightsource

      // Calculate lightsource vector:
      LightVector.x = Lightsources->x - pVertex->x;
      LightVector.y = Lightsources->y - pVertex->y;
      LightVector.z = Lightsources->z - pVertex->z;

      // Calculate lenght of lightsource vector:
      // add lightvectorLenght with 0.001 to avoid division by zero

			
      LightVectorLenght = VectorLenght(&LightVector);
      if (LightVectorLenght == 0.0)
        LightVectorLenght = 0.0001f;

      // Calculate the dot product with the lightsource vector
      // and the added normal vector for this vertex
			*/


      Dp3d = pLightNormal->x*pVertexNormal->x+pLightNormal->y*pVertexNormal->y+
        pLightNormal->z*pVertexNormal->z;


      ReflectionIntensity = ((Dp3d /
        (NormalVectorLenght)) + 1.0f) *0.5f;

      //DistanceAffection = (LightVectorLenght*Lightsources->DistanceAffection);

      TotalShadeR = ReflectionIntensity*GlobalLightIntensity; //(Lightsources->r*ReflectionIntensity)-DistanceAffection;
      TotalShadeG = ReflectionIntensity*GlobalLightIntensity; //(Lightsources->g*ReflectionIntensity)-DistanceAffection;
      TotalShadeB = ReflectionIntensity*GlobalLightIntensity; //(Lightsources->b*ReflectionIntensity)-DistanceAffection;


      if (TotalShadeR < 0.0)
        TotalShadeR = 0.0;

      if (TotalShadeG < 0.0)
        TotalShadeG = 0.0;

      if (TotalShadeB < 0.0)
        TotalShadeB = 0.0;

      pColor->r = TotalShadeR;
      pColor->g = TotalShadeG;
      pColor->b = TotalShadeB;


  if (pColor->r > 1.0)
    pColor->r = 1.0;

  if (pColor->g > 1.0)
    pColor->g = 1.0;

  if (pColor->b > 1.0)
    pColor->b = 1.0;

  if (pColor->r < 0.0)
    pColor->r = 0.0;

  if (pColor->g < 0.0)
    pColor->g = 0.0;

  if (pColor->b < 0.0)
    pColor->b = 0.0;

}


int SetupLandscapeLightning(int LandscapeLevel, fxeLandscapePtr pLandscape,
  int x, int z, fxeNormalBufferPtr pNormalBuffer)
{
  int NumPartsX;
  int IncHeightMap;
	int Loop, Loop2;
  fxeVector Normal;
	int CoordinateIndex;
	float StartR[8];
/*
	float StartG[8];
	float StartB[8];
*/
	float EndR[8];
/*
	float EndG[8];
	float EndB[8];
*/
	float StartX[8];
	float StartY[8];
	float StartZ[8];
	float EndX[8];
	float EndY[8];
	float EndZ[8];
	//fxeColor VertexColor[4];
	fxeColor NewVertexColors[4];
	fxeColor TempColor;
	fxeVector Normals[4];
	fxeVector StartNormals[8];
	fxeVector EndNormals[8];
	float Intensity;
	float VecLen;	

  NumPartsX = (int)(pow(4,LandscapeLevel));
  IncHeightMap = ((pNormalBuffer->Width-1))/NumPartsX;


  Normals[0].x=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*x].x;
  Normals[0].y=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*x].y;
  Normals[0].z=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*x].z;

  Normals[1].x=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*x].x;
  Normals[1].y=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*x].y;
  Normals[1].z=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*x].z;

	Normals[2].x=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*(x+1)].x;
  Normals[2].y=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*(x+1)].y;
  Normals[2].z=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*(x+1)].z;

  Normals[3].x=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*(x+1)].x;
  Normals[3].y=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*(x+1)].y;
  Normals[3].z=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*(x+1)].z;

	/*
  Normal.x=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*x].x;
  Normal.y=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*x].y;
  Normal.z=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*x].z;

  CalculateVertexIntensity(&Normal,
    &pLandscape->VerticesWorld[0],
		&pLandscape->VerticesCamera[0],&VertexColor[0]);

	

  Normal.x=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*x].x;
  Normal.y=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*x].y;
  Normal.z=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*x].z;
  CalculateVertexIntensity(&Normal,
    &pLandscape->VerticesWorld[1],
		&pLandscape->VerticesCamera[1],&VertexColor[1]);

  Normal.x=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*(x+1)].x;
  Normal.y=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*(x+1)].y;
  Normal.z=pNormalBuffer->Normals[IncHeightMap*(z+1)*pNormalBuffer->Width+IncHeightMap*(x+1)].z;
  CalculateVertexIntensity(&Normal,
    &(pLandscape->VerticesWorld[2]),
		&pLandscape->VerticesCamera[2],&VertexColor[2]);

  Normal.x=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*(x+1)].x;
  Normal.y=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*(x+1)].y;
  Normal.z=pNormalBuffer->Normals[IncHeightMap*z*pNormalBuffer->Width+IncHeightMap*(x+1)].z;
  CalculateVertexIntensity(&Normal,
    &(pLandscape->VerticesWorld[3]),
		&pLandscape->VerticesCamera[3],&VertexColor[3]);
	*/

	if (fxeLightmapInit(&pLandscape->Lightmap) == FXEFALSE)
		return FXEFALSE;

/*
	NewVertexColors[0].r = VertexColor[0].r;
	NewVertexColors[0].g = VertexColor[0].g;
	NewVertexColors[0].b = VertexColor[0].b;

	NewVertexColors[1].r = VertexColor[0].r + ((VertexColor[1].r-
		VertexColor[0].r)/8.0f)*8.0f;
	NewVertexColors[1].g = VertexColor[0].g + ((VertexColor[1].g-
		VertexColor[0].g)/8.0f)*8.0f;
	NewVertexColors[1].b = VertexColor[0].b + ((VertexColor[1].b-
		VertexColor[0].b)/8.0f)*8.0f;

	NewVertexColors[3].r = VertexColor[0].r + ((VertexColor[3].r-
		VertexColor[0].r)/8.0f)*8.0f;
	NewVertexColors[3].g = VertexColor[0].g + ((VertexColor[3].g-
		VertexColor[0].g)/8.0f)*8.0f;
	NewVertexColors[3].b = VertexColor[0].b + ((VertexColor[3].b-
		VertexColor[0].b)/8.0f)*8.0f;

	TempColor.r = VertexColor[3].r + ((VertexColor[2].r-
		VertexColor[3].r)/8.0f)*8.0f;
	TempColor.g = VertexColor[3].g + ((VertexColor[2].g-
		VertexColor[3].g)/8.0f)*8.0f;
	TempColor.b = VertexColor[3].b + ((VertexColor[2].b-
		VertexColor[3].b)/8.0f)*8.0f;

	NewVertexColors[2].r = NewVertexColors[1].r + ((TempColor.r-
		NewVertexColors[1].r)/8.0f)*8.0f;
	NewVertexColors[2].g = NewVertexColors[1].g + ((TempColor.g-
		NewVertexColors[1].g)/8.0f)*8.0f;
	NewVertexColors[2].b = NewVertexColors[1].b + ((TempColor.b-
		NewVertexColors[1].b)/8.0f)*8.0f;
*/

	for (Loop=0;Loop<LIGHTMAP_WIDTH;Loop++)
	{
		StartX[Loop] = pLandscape->VerticesWorld[0].x +((pLandscape->VerticesWorld[3].x -
			pLandscape->VerticesWorld[0].x) / 8.0f) * ((float)Loop);
		StartY[Loop] = pLandscape->VerticesWorld[0].y +((pLandscape->VerticesWorld[3].y -
			pLandscape->VerticesWorld[0].y) / 8.0f) * ((float)Loop);  
		StartZ[Loop] = pLandscape->VerticesWorld[0].z +((pLandscape->VerticesWorld[3].z -
			pLandscape->VerticesWorld[0].z) / 8.0f) * ((float)Loop);


		StartNormals[Loop].x = Normals[0].x + ((Normals[3].x -
			Normals[0].x)/7.0f)*((float)Loop);
		StartNormals[Loop].y = Normals[0].y + ((Normals[3].y -
			Normals[0].y)/7.0f)*((float)Loop);
		StartNormals[Loop].z = Normals[0].z + ((Normals[3].z -
			Normals[0].z)/7.0f)*((float)Loop);

		EndNormals[Loop].x = Normals[1].x + ((Normals[2].x -
			Normals[1].x)/7.0f)*((float)Loop);
		EndNormals[Loop].y = Normals[1].y + ((Normals[2].y -
			Normals[1].y)/7.0f)*((float)Loop);
		EndNormals[Loop].z = Normals[1].z + ((Normals[2].z -
			Normals[1].z)/7.0f)*((float)Loop);

#ifdef USE_INTENSITY_MAPS
		StartR[Loop] = (NewVertexColors[0].r + ((NewVertexColors[3].r -
			NewVertexColors[0].r) / 7.0f) * ((float)Loop)) * 255.0f;
#else
		StartR[Loop] = (VertexColor[0].r + ((VertexColor[3].r -
			VertexColor[0].r) / 8.0f) * ((float)Loop)) * 31.0f;
		StartG[Loop] = (VertexColor[0].g + ((VertexColor[3].g -
			VertexColor[0].g) / 8.0f) * ((float)Loop)) * 31.0f;
		StartB[Loop] = (VertexColor[0].b + ((VertexColor[3].b -
			VertexColor[0].b) / 8.0f) * ((float)Loop)) * 31.0f;
#endif
	}
		
	for (Loop=0;Loop<LIGHTMAP_WIDTH;Loop++)
	{
		CoordinateIndex = LIGHTMAP_WIDTH*(LIGHTMAP_WIDTH-1)+Loop;
		EndX[Loop] = pLandscape->VerticesWorld[1].x +((pLandscape->VerticesWorld[2].x -
			pLandscape->VerticesWorld[1].x) / 8.0f) * ((float)Loop);  
		EndY[Loop] = pLandscape->VerticesWorld[1].y +((pLandscape->VerticesWorld[2].y -
			pLandscape->VerticesWorld[1].y) / 8.0f) * ((float)Loop);  
		EndZ[Loop] = pLandscape->VerticesWorld[1].z +((pLandscape->VerticesWorld[2].z -
			pLandscape->VerticesWorld[1].z) / 8.0f) * ((float)Loop);  

		#ifdef USE_INTENSITY_MAPS
		EndR[Loop] = ((NewVertexColors[1].r + ((NewVertexColors[2].r -
			NewVertexColors[1].r) / 7.0f) * ((float)Loop)) * 255.0f);
		#else
		EndR[Loop] = ((VertexColor[1].r + ((VertexColor[2].r -
			VertexColor[1].r) / 8.0f) * ((float)Loop)) * 31.0f);
		EndG[Loop] = ((VertexColor[1].g + ((VertexColor[2].g -
			VertexColor[1].g) / 8.0f) * ((float)Loop)) * 31.0f);
		EndB[Loop] = ((VertexColor[1].b + ((VertexColor[2].b -
			VertexColor[1].b) / 8.0f) * ((float)Loop)) * 31.0f);
		#endif
	}

	for (Loop=0;Loop<LIGHTMAP_WIDTH;Loop++)
	{
		for (Loop2=0;Loop2<LIGHTMAP_WIDTH;Loop2++)
		{
			pLandscape->Lightmap.Heigths[Loop*LIGHTMAP_WIDTH+Loop2] = 
				StartY[Loop2] +
				((EndY[Loop2] - StartY[Loop2]) / 7.0f) * ((float)Loop);
/*
			pLandscape->Lightmap.LightmapCoors[Loop*LIGHTMAP_WIDTH+Loop2].y = 
				StartY[Loop2] +
				((EndY[Loop2] - StartY[Loop2]) / 7.0f) * ((float)Loop);

			pLandscape->Lightmap.LightmapCoors[Loop*LIGHTMAP_WIDTH+Loop2].z = 
				StartZ[Loop2] +
				((EndZ[Loop2] - StartZ[Loop2]) / 7.0f) * ((float)Loop);
*/			
#ifdef USE_INTENSITY_MAPS

			// Do some phong shading by iterating vertex normals
			// and calculate vertex intensity on the iterated normal.

			Normal.x = StartNormals[Loop2].x + ((EndNormals[Loop2].x -
				StartNormals[Loop2].x)/7.0f)*((float)Loop);

			Normal.y = StartNormals[Loop2].y + ((EndNormals[Loop2].y -
				StartNormals[Loop2].y)/7.0f)*((float)Loop);

			Normal.z = StartNormals[Loop2].z + ((EndNormals[Loop2].z -
				StartNormals[Loop2].z)/7.0f)*((float)Loop);

			VecLen = VectorLenght(&Normal);

			if (VecLen == 0.0f)
				VecLen = 0.00001f;

			Normal.x /= VecLen;
			Normal.y /= VecLen;
			Normal.z /= VecLen;

			CalculateVertexIntensity(&Normal,&TempColor);

			Intensity = TempColor.r*255.0f;

			pLandscape->Lightmap.StaticData[Loop*LIGHTMAP_WIDTH+Loop2] = (unsigned char)Intensity;
			/*
				((unsigned char)(StartR[Loop2] + ((EndR[Loop2] - StartR[Loop2]) / 7.0f) * ((float)Loop)) ) ;
				*/
#else
			pLandscape->Lightmap.StaticData[Loop*LIGHTMAP_WIDTH+Loop2] =  
				(((unsigned short)(StartR[Loop2] + ((EndR[Loop2] - StartR[Loop2]) / 7.0f) * ((float)Loop)) ) << 10) +
				(((unsigned short)(StartG[Loop2] + ((EndG[Loop2] - StartG[Loop2]) / 7.0f) * ((float)Loop)) ) <<  5) +
				 ((unsigned short)(StartB[Loop2] + ((EndB[Loop2] - StartB[Loop2]) / 7.0f) * ((float)Loop)) );
#endif
	

		}
	}

	/*
	// Download texture to tmu on 3dfx card
	
	if (NumTexelFx == 1)
	{
		LightmapTexInfo.data = pLandscape->Lightmap.StaticData;
		grTexDownloadMipMap(GR_TMU0,pLandscape->Lightmap.TmuAddress,GR_MIPMAPLEVELMASK_ODD,
			&LightmapTexInfo);
	}
	else
	{
		LightmapTexInfo.data = pLandscape->Lightmap.StaticData;
		grTexDownloadMipMap(GR_TMU1,pLandscape->Lightmap.TmuAddress,GR_MIPMAPLEVELMASK_ODD,
			&LightmapTexInfo);
	}
	*/


	return FXETRUE;	
}

void SetupLandscapeTexture(fxeLandscapePtr pLandscape)
{
  float Width, Height;

  Width = 1.0;
  Height = 1.0;

  pLandscape->VerticesCamera[0].u0 = 0.0;
  pLandscape->VerticesCamera[0].v0 = 0.0;

  pLandscape->VerticesCamera[1].u0 = 0.0;
  pLandscape->VerticesCamera[1].v0 = Height;

  pLandscape->VerticesCamera[2].u0 = Width;
  pLandscape->VerticesCamera[2].v0 = Height;

  pLandscape->VerticesCamera[3].u0 = Width;
  pLandscape->VerticesCamera[3].v0 = 0.0;

  pLandscape->VerticesCamera[0].u1 = 0.0;
  pLandscape->VerticesCamera[0].v1 = 0.0;

  pLandscape->VerticesCamera[1].u1 = 0.0;
  pLandscape->VerticesCamera[1].v1 = Height;

  pLandscape->VerticesCamera[2].u1 = Width;
  pLandscape->VerticesCamera[2].v1 = Height;

  pLandscape->VerticesCamera[3].u1 = Width;
  pLandscape->VerticesCamera[3].v1 = 0.0;
}


fxeLandscapePtr SetupLandscape(int LandscapeLevel,
  float Width, float Depth, int x, int z, fxeHeightmapInfoPtr pHeightmap,
  fxeTextureHandle TerrainTexture, int NumChildsX, fxeNormalBufferPtr
  pNormalBuffer)
{
  int NumPartsX;
  float IncWidth, IncDepth;
  int IncHeightMap;
  float BeginX, BeginZ;
  fxeLandscapePtr pLandscape;
  fxeVector Normal;


  pLandscape = (fxeLandscapePtr) FxMalloc(sizeof(fxeLandscape));
  if (pLandscape == NULL)
    return NULL;

  pLandscape->NumLandscapesX = NumChildsX;

  if (NumChildsX > 0)
  {
    pLandscape->Landscapes = (unsigned int*)
      FxMalloc(sizeof(fxeLandscapePtr)*(NumChildsX*NumChildsX));
    if (pLandscape->Landscapes == NULL)
      return NULL;
  }
  else
    pLandscape->Landscapes = NULL;

  

  NumPartsX = (int)pow(4,LandscapeLevel);
  IncHeightMap = (int)((float)((pHeightmap->Width-1))/(float)NumPartsX);
  IncWidth = Width/((float)NumPartsX);
  IncDepth = Depth/((float)NumPartsX);
  BeginX = -(Width/2.0f)+x*IncWidth;
  BeginZ = Depth/2.0f-z*IncDepth;

  

  // Setup the vertices of the landscape
  pLandscape->VerticesWorld[0].x = BeginX;
  pLandscape->VerticesWorld[0].z = BeginZ;
  pLandscape->VerticesWorld[0].y = pHeightmap->MapData[IncHeightMap*z*
    pHeightmap->Width+IncHeightMap*x];



  pLandscape->VerticesWorld[1].x = BeginX;
  pLandscape->VerticesWorld[1].z = BeginZ-IncDepth;
  pLandscape->VerticesWorld[1].y = pHeightmap->MapData[IncHeightMap*
    pHeightmap->Width*(z+1)+IncHeightMap*x];



  pLandscape->VerticesWorld[2].x = BeginX+IncWidth;
  pLandscape->VerticesWorld[2].z = BeginZ-IncDepth;
  pLandscape->VerticesWorld[2].y = pHeightmap->MapData[IncHeightMap*
    pHeightmap->Width*(z+1)+IncHeightMap*(x+1)];

  pLandscape->VerticesWorld[3].x = BeginX+IncWidth;
  pLandscape->VerticesWorld[3].z = BeginZ;
  pLandscape->VerticesWorld[3].y = pHeightmap->MapData[IncHeightMap*
    pHeightmap->Width*z+IncHeightMap*(x+1)];

  
  pLandscape->VertexWorld.x = (pLandscape->VerticesWorld[0].x +
    pLandscape->VerticesWorld[1].x + pLandscape->VerticesWorld[2].x +
    pLandscape->VerticesWorld[3].x) / 4.0f;

  pLandscape->VertexWorld.y = (pLandscape->VerticesWorld[0].y +
    pLandscape->VerticesWorld[1].y + pLandscape->VerticesWorld[2].y +
    pLandscape->VerticesWorld[3].y) / 4.0f;

  pLandscape->VertexWorld.z = (pLandscape->VerticesWorld[0].z +
    pLandscape->VerticesWorld[1].z + pLandscape->VerticesWorld[2].z +
    pLandscape->VerticesWorld[3].z) / 4.0f;

  pLandscape->Radius = (float)sqrt(
		(float)pow(pLandscape->VerticesWorld[0].x-
    pLandscape->VertexWorld.x,2)+
    (float)pow(pLandscape->VerticesWorld[0].z-
		pLandscape->VertexWorld.z,2)+
		(float)pow(pLandscape->VerticesWorld[0].y-
		pLandscape->VertexWorld.y,2));

  Normal.x = 0.0f;
  Normal.y = 0.0f;
  Normal.z = 0.0f;

  
	CalculateNormalTriangle(&(pLandscape->VerticesWorld[0]),
    &(pLandscape->VerticesWorld[1]),&(pLandscape->VerticesWorld[2]),&pLandscape->Normal1);
	CalculateNormalTriangle(&(pLandscape->VerticesWorld[2]),
		&(pLandscape->VerticesWorld[3]),&(pLandscape->VerticesWorld[0]),&pLandscape->Normal2);

	
  // Add normals to new landscape in the normal buffer if landscapelevel is 3 - 
	// used during lightning
	if (LandscapeLevel == 3)
	{
		pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*x].x += pLandscape->Normal1.x;
		pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*x].y += pLandscape->Normal1.y;
		pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*x].z += pLandscape->Normal1.z;

		pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*x].x +=
			pLandscape->Normal1.x;
		pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*x].y +=
			pLandscape->Normal1.y;
		pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*x].z +=
			pLandscape->Normal1.z;

		pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*(x+1)].x +=
			pLandscape->Normal1.x;
		pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*(x+1)].y +=
			pLandscape->Normal1.y;
		pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*(x+1)].z +=
			pLandscape->Normal1.z;


		pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*(x+1)].x +=
			pLandscape->Normal2.x;
		pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*(x+1)].y +=
			pLandscape->Normal2.y;
		pNormalBuffer->Normals[IncHeightMap*(z+1)*pHeightmap->Width+IncHeightMap*(x+1)].z +=
			pLandscape->Normal2.z;
  
		pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*(x+1)].x +=
			pLandscape->Normal2.x;
		pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*(x+1)].y +=
			pLandscape->Normal2.y;
		pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*(x+1)].z +=
			pLandscape->Normal2.z;

		pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*x].x += pLandscape->Normal2.x;
		pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*x].y += pLandscape->Normal2.y;
		pNormalBuffer->Normals[IncHeightMap*z*pHeightmap->Width+IncHeightMap*x].z += pLandscape->Normal2.z;
	}


  pLandscape->TextureMode = TEXTURE_MAPPED;
  pLandscape->TH = TerrainTexture;
  SetupLandscapeTexture(pLandscape);

	pLandscape->Displayed = FXEFALSE;

  return pLandscape;
}

fxeLandscapeHandle
  fxeLandscapeCreate(float Width, float Depth, fxeHeightmapInfoPtr
  pHeightmap, fxeNormalBufferPtr pNormalBuffer, fxeTextureHandle TerrainTexture)
{
	// Creates a landscape using two different sizes:
	// 256x256 or 64x64
  int Loop, Loop2, Loop3, Loop4, Loop5, Loop6, Loop7, Loop8;
  fxeLandscapePtr pParentLandscape;
  fxeLandscapePtr pCurrLandscape, pCurrLandscape2, pCurrLandscape3;

	if (pHeightmap->Width == 65)
	{

		pParentLandscape = SetupLandscape(0,Width,Depth,0,0,pHeightmap,
			TerrainTexture,4,pNormalBuffer);
		if (pParentLandscape == NULL)
			return NULL;



		for (Loop=0; Loop<4; Loop++)
		{
			for (Loop2=0; Loop2<4; Loop2++)
			{
				pParentLandscape->Landscapes[Loop*4+Loop2] = (unsigned int)
					SetupLandscape(1,Width,Depth,Loop2,Loop,pHeightmap,TerrainTexture,4,
					pNormalBuffer);

				pCurrLandscape = (fxeLandscapePtr)(
					pParentLandscape->Landscapes[Loop*4+Loop2]);
				if (pCurrLandscape == NULL)
					return NULL;

				for (Loop3=0; Loop3<4; Loop3++)
				{
					for (Loop4=0; Loop4<4; Loop4++)
					{
						pCurrLandscape->Landscapes[Loop3*4+Loop4] = (unsigned int)
							SetupLandscape(2,Width,Depth,Loop2*4+Loop4,Loop*4+Loop3,pHeightmap,
							TerrainTexture,4,pNormalBuffer);

						pCurrLandscape2 = (fxeLandscapePtr)(
							pCurrLandscape->Landscapes[Loop3*4+Loop4]);
						if (pCurrLandscape2 == NULL)
							return NULL;

						for (Loop5=0; Loop5<4; Loop5++)
						{
							for (Loop6=0; Loop6<4; Loop6++)
							{
								pCurrLandscape2->Landscapes[Loop5*4+Loop6] = (unsigned int)
									SetupLandscape(3,Width,Depth,Loop2*16+Loop4*4+Loop6,
									Loop*16+Loop3*4+Loop5,pHeightmap,TerrainTexture,0,
									pNormalBuffer);
								if ((fxeLandscapePtr)pCurrLandscape2->Landscapes[Loop5*4+Loop6] == NULL)
									return NULL;
							}
						}
					}
				}
			}
		}
  
		SetupLandscapeLightning(0,pParentLandscape,0,0,pNormalBuffer);

		for (Loop=0; Loop<4; Loop++)
		{
			for (Loop2=0; Loop2<4; Loop2++)
			{
				SetupLandscapeLightning(1,(fxeLandscapePtr)pParentLandscape->Landscapes[Loop*4+Loop2],
				Loop2,Loop,pNormalBuffer);

				pCurrLandscape = (fxeLandscapePtr)(
					pParentLandscape->Landscapes[Loop*4+Loop2]);

				for (Loop3=0; Loop3<4; Loop3++)
				{
					for (Loop4=0; Loop4<4; Loop4++)
					{
						SetupLandscapeLightning(2,(fxeLandscapePtr)pCurrLandscape->Landscapes[Loop3*4+Loop4],
							Loop2*4+Loop4,Loop*4+Loop3,pNormalBuffer);

						pCurrLandscape2 = (fxeLandscapePtr)(
							pCurrLandscape->Landscapes[Loop3*4+Loop4]);

						for (Loop5=0; Loop5<4; Loop5++)
						{
							for (Loop6=0; Loop6<4; Loop6++)
							{
								SetupLandscapeLightning(3,
									(fxeLandscapePtr)pCurrLandscape2->Landscapes[Loop5*4+Loop6],
									Loop2*16+Loop4*4+Loop6,Loop*16+Loop3*4+Loop5,
									pNormalBuffer);
							}
						}
					}
				}
			}
		}
	}
	else if (pHeightmap->Width == 257)
	{

		pParentLandscape = SetupLandscape(0,Width,Depth,0,0,pHeightmap,
			TerrainTexture,4,pNormalBuffer);
		if (pParentLandscape == NULL)
			return NULL;



		for (Loop=0; Loop<4; Loop++)
		{
			for (Loop2=0; Loop2<4; Loop2++)
			{
				pParentLandscape->Landscapes[Loop*4+Loop2] = (unsigned int)
					SetupLandscape(1,Width,Depth,Loop2,Loop,pHeightmap,TerrainTexture,4,
					pNormalBuffer);

				pCurrLandscape = (fxeLandscapePtr)(
					pParentLandscape->Landscapes[Loop*4+Loop2]);
				if (pCurrLandscape == NULL)
					return NULL;

				for (Loop3=0; Loop3<4; Loop3++)
				{
					for (Loop4=0; Loop4<4; Loop4++)
					{
						pCurrLandscape->Landscapes[Loop3*4+Loop4] = (unsigned int)
							SetupLandscape(2,Width,Depth,Loop2*4+Loop4,Loop*4+Loop3,pHeightmap,
							TerrainTexture,4,pNormalBuffer);

						pCurrLandscape2 = (fxeLandscapePtr)(
							pCurrLandscape->Landscapes[Loop3*4+Loop4]);
						if (pCurrLandscape2 == NULL)
							return NULL;

						for (Loop5=0; Loop5<4; Loop5++)
						{
							for (Loop6=0; Loop6<4; Loop6++)
							{
								pCurrLandscape2->Landscapes[Loop5*4+Loop6] = (unsigned int)
									SetupLandscape(3,Width,Depth,Loop2*16+Loop4*4+Loop6,
									Loop*16+Loop3*4+Loop5,pHeightmap,TerrainTexture,4,
									pNormalBuffer);
								if ((fxeLandscapePtr)pCurrLandscape2->Landscapes[Loop5*4+Loop6] == NULL)
									return NULL;

								pCurrLandscape3 = (fxeLandscapePtr)(
									pCurrLandscape2->Landscapes[Loop5*4+Loop6]);
								if (pCurrLandscape3 == NULL)
									return NULL;

								for (Loop7=0; Loop7<4; Loop7++)
								{
									for (Loop8=0; Loop8<4; Loop8++)
									{
										pCurrLandscape3->Landscapes[Loop7*4+Loop8] = (unsigned int)
											SetupLandscape(4,Width,Depth,Loop2*64+Loop4*16+Loop6*4+Loop8,
											Loop*64+Loop3*16+Loop5*4+Loop7,pHeightmap,TerrainTexture,0,
											pNormalBuffer);
										if ((fxeLandscapePtr)(pCurrLandscape3->Landscapes[Loop7*4+Loop8]) == NULL)
											return NULL;

									}
								}
							}
						}
					}
				}
			}
		}
  
		SetupLandscapeLightning(0,pParentLandscape,0,0,pNormalBuffer);

		for (Loop=0; Loop<4; Loop++)
		{
			for (Loop2=0; Loop2<4; Loop2++)
			{
				SetupLandscapeLightning(1,(fxeLandscapePtr)pParentLandscape->Landscapes[Loop*4+Loop2],
				Loop2,Loop,pNormalBuffer);

				pCurrLandscape = (fxeLandscapePtr)(
					pParentLandscape->Landscapes[Loop*4+Loop2]);

				for (Loop3=0; Loop3<4; Loop3++)
				{
					for (Loop4=0; Loop4<4; Loop4++)
					{
						SetupLandscapeLightning(2,(fxeLandscapePtr)pCurrLandscape->Landscapes[Loop3*4+Loop4],
							Loop2*4+Loop4,Loop*4+Loop3,pNormalBuffer);

						pCurrLandscape2 = (fxeLandscapePtr)(
							pCurrLandscape->Landscapes[Loop3*4+Loop4]);

						for (Loop5=0; Loop5<4; Loop5++)
						{
							for (Loop6=0; Loop6<4; Loop6++)
							{
								SetupLandscapeLightning(3,
									(fxeLandscapePtr)pCurrLandscape2->Landscapes[Loop5*4+Loop6],
									Loop2*16+Loop4*4+Loop6,Loop*16+Loop3*4+Loop5,
									pNormalBuffer);

								pCurrLandscape3 = (fxeLandscapePtr)(
									pCurrLandscape2->Landscapes[Loop5*4+Loop6]);

								for (Loop7=0; Loop7<4; Loop7++)
								{
									for (Loop8=0; Loop8<4; Loop8++)
									{
										SetupLandscapeLightning(
											4,(fxeLandscapePtr)pCurrLandscape3->Landscapes[Loop7*4+Loop8],
											Loop2*64+Loop4*16+Loop6*4+Loop8,Loop*64+Loop3*16+Loop5*4+Loop7,
											pNormalBuffer);
									}
								}
							}
						}
					}
				}
			}
		}
	}
	else
		return NULL;

  return pParentLandscape;
}

// Creates a landscape and sets it to world positon 0,0,0. It uses
// some random calculations, the landscape generated by this function
// is never the same. The function assumes that you have created some
// lightsources before this function is called. The Landscape is
// lit using gouraud shading by these lightsources. You also need to create
// heightmaps using the fxeCreateHeightMaps function before this function
// is called.


fxeLandscapeHandle
  fxeLandscapeCreateRandom(float Width, float Depth, fxeHeightmapInfoPtr
  pHeightmap, fxeNormalBufferPtr pNormalBuffer, fxeTextureHandle TerrainTexture,
  int Seed)
{
  // The landscape generator creates a landscape which consists
  // of 4*4 = 16 child landscapes. Each of these child landscapes
  // can consist of 4*4 = 16 child landscapes. Each of these child
  // landscapes can consist of 4*4 = 16 childlandscapes which makes a
  // total 16^3+16^2+16^1 = 4360 child landscapes.
  int Loop, Loop2, Loop3, Loop4, Loop5, Loop6;
  fxeLandscapePtr pParentLandscape;
  fxeLandscapePtr pCurrLandscape, pCurrLandscape2;

  srand(Seed);


  pParentLandscape = SetupLandscape(0,Width,Depth,0,0,pHeightmap,
    TerrainTexture,4,pNormalBuffer);
  if (pParentLandscape == NULL)
    return NULL;



  for (Loop=0; Loop<4; Loop++)
  {
    for (Loop2=0; Loop2<4; Loop2++)
    {
      pParentLandscape->Landscapes[Loop*4+Loop2] = (unsigned int)
        SetupLandscape(1,Width,Depth,Loop2,Loop,pHeightmap,TerrainTexture,4,
        pNormalBuffer);

      pCurrLandscape = (fxeLandscapePtr)(
        pParentLandscape->Landscapes[Loop*4+Loop2]);
      if (pCurrLandscape == NULL)
        return NULL;

      for (Loop3=0; Loop3<4; Loop3++)
      {
        for (Loop4=0; Loop4<4; Loop4++)
        {
          pCurrLandscape->Landscapes[Loop3*4+Loop4] = (unsigned int)
            SetupLandscape(2,Width,Depth,Loop2*4+Loop4,Loop*4+Loop3,pHeightmap,
            TerrainTexture,4,pNormalBuffer);

          pCurrLandscape2 = (fxeLandscapePtr)(
            pCurrLandscape->Landscapes[Loop3*4+Loop4]);
          if (pCurrLandscape2 == NULL)
            return NULL;

          for (Loop5=0; Loop5<4; Loop5++)
          {
            for (Loop6=0; Loop6<4; Loop6++)
            {
              pCurrLandscape2->Landscapes[Loop5*4+Loop6] = (unsigned int)
                SetupLandscape(3,Width,Depth,Loop2*16+Loop4*4+Loop6,
                Loop*16+Loop3*4+Loop5,pHeightmap,TerrainTexture,0,
                pNormalBuffer);
              if ((fxeLandscapePtr)pCurrLandscape2->Landscapes[Loop5*4+Loop6] == NULL)
                return NULL;
            }
          }
        }
      }
    }
  }



  SetupLandscapeLightning(0,pParentLandscape,0,0,pNormalBuffer);

  for (Loop=0; Loop<4; Loop++)
  {
    for (Loop2=0; Loop2<4; Loop2++)
    {
      SetupLandscapeLightning(1,(fxeLandscapePtr)pParentLandscape->Landscapes[Loop*4+Loop2],
      Loop2,Loop,pNormalBuffer);

      pCurrLandscape = (fxeLandscapePtr)(
        pParentLandscape->Landscapes[Loop*4+Loop2]);

      for (Loop3=0; Loop3<4; Loop3++)
      {
        for (Loop4=0; Loop4<4; Loop4++)
        {
          SetupLandscapeLightning(2,(fxeLandscapePtr)pCurrLandscape->Landscapes[Loop3*4+Loop4],
            Loop2*4+Loop4,Loop*4+Loop3,pNormalBuffer);

          pCurrLandscape2 = (fxeLandscapePtr)(
            pCurrLandscape->Landscapes[Loop3*4+Loop4]);

          for (Loop5=0; Loop5<4; Loop5++)
          {
            for (Loop6=0; Loop6<4; Loop6++)
            {
              SetupLandscapeLightning(3,
                (fxeLandscapePtr)pCurrLandscape2->Landscapes[Loop5*4+Loop6],
                Loop2*16+Loop4*4+Loop6,Loop*16+Loop3*4+Loop5,
                pNormalBuffer);
            }
          }
        }
      }
    }
  }
  return pParentLandscape;
}

void CreateLocalLightmapCoordinates(int NumLandscapesX, float Width, float Depth)
{
	int Loop,Loop2;

	Width = Width / ((float)NumLandscapesX);
	Width = Width / ((float)(LIGHTMAP_WIDTH-1));
	Depth = Depth / ((float)NumLandscapesX);
	Depth = Depth / ((float)(LIGHTMAP_WIDTH-1));


	for (Loop=0; Loop<LIGHTMAP_WIDTH;Loop++)
	{
		for (Loop2=0; Loop2<LIGHTMAP_WIDTH; Loop2++)
		{
			LocalLightmapCoors[LIGHTMAP_WIDTH*Loop+Loop2].x =
				Width*Loop2;
			LocalLightmapCoors[LIGHTMAP_WIDTH*Loop+Loop2].z =
				-Depth*Loop;
		}
	}
}




// This function loads a landscape from a 3df file on disk. The Geometry parameter
// specifies a pointer to a filename for a 3df file that will be used for
// the geometry data of the landscape. The 3df file must be in grayscale where
// bright areas of the image will be heights and dark areas will be walleys.
// The color parameter specifies the pointer to the filename for a 3df file that
// is specifying the color of the landscape. If Color parameter is NULL then
// the landscapes get its colors from lightsources previously added to the world
// with the fxeLightsourceAdd function. The TH parameter specifies a handle to a texture 
// to use as the surface material of the landscape.
// The size of the Geometry file and the Color file must be equal and must be in the
// intensity 8 (i8) format. Valid sizes is 256x256 and 64x64.
// No mipmapping is recommended because it is not needed a eats memory.
// Width specifies the width of the landscapes in world units and depth specifies
// the depth of the landscape in world units. Height is the height of the world in
// world units. 
// Function returns a handle to a Landscape if successful and NULL if failure.


fxeLandscapesParentHandle fxeLandscapeLoadFromFile(char *Geometry, char *Color, fxeTextureHandle TH,
	float Width, float Depth, float Height, fxeVector *LightNormal, float LightIntensity)
{
	fxeTextureHandle GTH;
	int Size;
	int Loop, Loop2;
	float HeightValue;
	fxeHeightmapInfo *pHeightmap;
	fxeNormalBuffer *pNormalBuffer;
	fxeLandscapesParentHandle LspH;
	fxeLandscapeHandle LsH;

	GlobalLightIntensity = LightIntensity;

	pLightNormal = LightNormal;

	// Create parent landscape:
	LspH = (fxeLandscapesParentHandle)FxMalloc(sizeof(fxeLandscapesParent));
	if (LspH == NULL)
		return NULL;

	// Change this to another function call later, just for test
	GTH = fxeTextureLoadToMemory(Geometry);
	if (GTH == NULL)
		return NULL;

	// Check the texture format - must be intensity 8 format.
	if (GTH->TexInfo.format != GR_TEXFMT_INTENSITY_8)
		return NULL;

	// Check the aspect ratio, it must be 1x1:

	if (GTH->TexInfo.aspectRatioLog2 != GR_ASPECT_LOG2_1x1)
		return NULL;

	// Check the size, must be 256x256 or 64x64.
	if (GTH->TexInfo.largeLodLog2 == GR_LOD_LOG2_256)
		Size = 257;
	else if (GTH->TexInfo.largeLodLog2 == GR_LOD_LOG2_64)
		Size = 65;
	else
		return NULL;

	// Allocate memory for buffer holding pointers to landscapes
	// that will be rendered.
	LspH->ppLandscape = (fxeLandscape**)FxMalloc(sizeof(fxeLandscapePtr)*Size*Size);

	pNormalBuffer = fxeCreateNormalBuffer(Size);
	if (pNormalBuffer == NULL)
		return NULL;

	pHeightmap = fxeCreateHeightMap2(Size,GTH->TexInfo.data,Height);
	if (pHeightmap == NULL)
		return NULL;

	LspH->HeightBuffer = (float*)FxMalloc(sizeof(float)*Size*Size);
	LspH->Width = Width;
	LspH->Depth = Depth;
	LspH->NumPolysX = Size;

	// Set all elements in HeightBuffer to 0
	for (Loop=0; Loop<(Size*Size); Loop++)
	{
		LspH->HeightBuffer[Loop] = 0.0f;
	}

	// Fill HeightBuffer with heightdata from bitmap:
	// Normalize it so that maximum height is height given in the
	// height parameter.
	for (Loop=0; Loop<(Size-1);Loop++)
	{
		for (Loop2=0; Loop2<(Size-1); Loop2++)
		{
			HeightValue = 
				(((float)(((unsigned char*)(GTH->TexInfo.data))
					[Loop*(Size-1)+Loop2]))/255.0f)*Height;
			LspH->HeightBuffer[Loop*Size+Loop2] = HeightValue;
		}
	}

  LsH = fxeLandscapeCreate(Width,Depth,pHeightmap,
    pNormalBuffer,TH);
  if (LsH == NULL)
    return NULL;

	// Create local lightmap coordinates
	CreateLocalLightmapCoordinates(Size-1,Width,Depth);

	LspH->pLandscape = LsH;

  fxeFreeHeightMap(pHeightmap);
  fxeFreeNormalBuffer(pNormalBuffer);
  return LspH;
}


fxeLandscapesParentHandle fxeLandscapeCreate64x64(float Width1, float Depth1,
  fxeTextureHandle TerrainTexture, int RandomSeed)
{
	fxeLandscapesParentHandle LspH;
  fxeHeightmapInfo *pHeightmap;
  fxeNormalBuffer *pNormalBuffer;
  fxeLandscapeHandle LsH;

	LspH = (fxeLandscapesParentHandle)FxMalloc(sizeof(fxeLandscapesParent));

  pNormalBuffer = fxeCreateNormalBuffer(65);
  if (pNormalBuffer == NULL)
    return NULL;

  pHeightmap = fxeCreateHeightMap(1000,10,10,65);
  if (pHeightmap == NULL)
    return NULL;

 

  LsH = fxeLandscapeCreateRandom(Width1,Depth1,pHeightmap,
    pNormalBuffer,TerrainTexture,RandomSeed);

  
  if (LsH == NULL)
    return NULL;

  fxeFreeHeightMap(pHeightmap);
  fxeFreeNormalBuffer(pNormalBuffer);

	LspH->pLandscape = LsH;
  return LspH;
}



void fxeLandscapeAddToWorld(fxeLandscapesParentHandle pParentLandscape)
{
  CurrLandscape = pParentLandscape->pLandscape;
	CurrLandscapeParent = pParentLandscape;
}

void fxeLandscapeRemoveFromWorld(void)
{
  CurrLandscape = NULL;
}

fxeLandscapesParent *CurrLandscapeParent = NULL;
fxeLandscape *CurrLandscape = NULL;
  




