

#include "fxparticle.h"
#include "FxGlobals.h"
#include <glide.h>
#include "FxMalloc.h"
#include "FxTexture.h"


void fxeEmitterAddToWorld(fxeEmitterHandle pEmitter)
{
	if (pEmitters == NULL)
	{
		pEmitters = pEmitter;
		pEmitter->Next = NULL;
		pEmitter->Prev = NULL;
	}
	else
	{
		pEmitters->Next = pEmitter;
		pEmitter->Prev = pEmitters;
		pEmitter->Next = NULL;
		pEmitters = pEmitter;
	}
}

void fxeEmitterRemoveFromWorld(fxeEmitterHandle pEmitter)
{
	if (pEmitter->Next == NULL && pEmitter->Prev == NULL)
	{
		pEmitters == NULL;
	}
	else if (pEmitter->Next == NULL)
	{
		((fxeEmitter*)(pEmitter->Prev))->Next = NULL;
		if (pEmitters == pEmitter)
			pEmitters = pEmitter->Prev;
	}
	else if (pEmitter->Prev == NULL)
	{
		((fxeEmitter*)(pEmitter->Next))->Prev = NULL;
	}
	else
	{
		((fxeEmitter*)(pEmitter->Prev))->Next = pEmitter->Next;
		((fxeEmitter*)(pEmitter->Next))->Prev = pEmitter->Prev;
	}
}

void fxeParticleSystemInit(void)
{
	pEmitters = NULL;
}


// Returns a handle to a emitter if successful and NULL
// if failure.
fxeEmitterHandle fxeEmitterCreate(float x, float y, float z, 
	int NumParticles, fxeColor *InitColor,
  fxeColor *FinishColor, fxeTextureHandle TexHandle,
  fxeVector *ForceVector, float ForceAcceleration, float
	ParticleRadius, float Magnitude)
{
  int Loop;
  fxeEmitter *pEmitter;

  pEmitter = (fxeEmitter*)FxMalloc(sizeof(fxeEmitter));
	pEmitter->x = x;
	pEmitter->y = y;
	pEmitter->z = z;
	pEmitter->TexHandle = TexHandle;
  pEmitter->InitColor.r = InitColor->r;
  pEmitter->InitColor.g = InitColor->g;
  pEmitter->InitColor.b = InitColor->b;
  pEmitter->FinishColor.r = FinishColor->r;
  pEmitter->FinishColor.g = FinishColor->g;
  pEmitter->FinishColor.b = FinishColor->b;
  pEmitter->TotalParticles = NumParticles;
  pEmitter->NumParticles = 0;
  pEmitter->UsedParticles = (fxeParticle*)FxMalloc(sizeof(fxeParticle)*NumParticles);
	if (pEmitter->UsedParticles == NULL)
		return NULL;
  pEmitter->UnusedParticles = pEmitter->UsedParticles;
  pEmitter->ForceVector.x = ForceVector->x;
  pEmitter->ForceVector.y = ForceVector->y;
  pEmitter->ForceVector.z = ForceVector->z;
  pEmitter->ForceAcceleration = ForceAcceleration;
	pEmitter->ParticleRadius = ParticleRadius;
	pEmitter->Magnitude = Magnitude;

  for (Loop=0;Loop<(NumParticles-1);Loop++)
  {
    pEmitter->UsedParticles[Loop].Next =
      (void*)&pEmitter->UsedParticles[Loop+1];
  }
  pEmitter->UsedParticles[NumParticles-1].Next = NULL;
  return pEmitter;
}


void fxeEmitterAddParticles(fxeEmitterHandle pEmitter, int NumParticles)
{
	int Loop;

	if (NumParticles == 0)
		return;

	for (Loop=0; Loop<NumParticles; Loop++)
	{
		fxeEmitterAddParticle(pEmitter);
	}
}

// Adds a new particle with a special direction vector
// and a special life length to a emitter specified:

void fxeEmitterAddParticle(fxeEmitterHandle pEmitter)
{
	fxeParticle *pCurrParticle;

	// only add particle if it's room for it:
	if (pEmitter->NumParticles == pEmitter->TotalParticles)
		return;

	pEmitter->UnusedParticles->x = pEmitter->x;
	pEmitter->UnusedParticles->y = pEmitter->y;
	pEmitter->UnusedParticles->z = pEmitter->z;

	pEmitter->UnusedParticles->NumFramesLeft = (rand()%50)+50;
	
	pEmitter->UnusedParticles->CurrColor.r = pEmitter->InitColor.r;
	pEmitter->UnusedParticles->CurrColor.g = pEmitter->InitColor.g;
	pEmitter->UnusedParticles->CurrColor.b = pEmitter->InitColor.b;
	
	pEmitter->UnusedParticles->IncColor.r = 
		(pEmitter->FinishColor.r - pEmitter->InitColor.r) /
		(float)pEmitter->UnusedParticles->NumFramesLeft;

	pEmitter->UnusedParticles->IncColor.g = 
		(pEmitter->FinishColor.g - pEmitter->InitColor.g) /
		(float)pEmitter->UnusedParticles->NumFramesLeft;

	pEmitter->UnusedParticles->IncColor.b = 
		(pEmitter->FinishColor.b - pEmitter->InitColor.b) /
		(float)pEmitter->UnusedParticles->NumFramesLeft;

	pEmitter->UnusedParticles->Radius = pEmitter->ParticleRadius;

	pEmitter->UnusedParticles->DirectionVector.x = (((float)(rand()%100))/50.0f-1.0f)*pEmitter->Magnitude;
	pEmitter->UnusedParticles->DirectionVector.y = (((float)(rand()%100))/50.0f-1.0f)*pEmitter->Magnitude;
	pEmitter->UnusedParticles->DirectionVector.z = (((float)(rand()%100))/50.0f-1.0f)*pEmitter->Magnitude;

	pEmitter->UnusedParticles->ForceVector.x = pEmitter->ForceVector.x;
	pEmitter->UnusedParticles->ForceVector.y = pEmitter->ForceVector.y;
	pEmitter->UnusedParticles->ForceVector.z = pEmitter->ForceVector.z;

	pCurrParticle = pEmitter->UnusedParticles;

	pEmitter->UnusedParticles = (fxeParticle*)pEmitter->UnusedParticles->Next;
	pEmitter->NumParticles++;

	pCurrParticle->Next = pEmitter->UsedParticles;
	pEmitter->UsedParticles->Prev = pCurrParticle;
	pEmitter->UsedParticles = pCurrParticle;
}


void fxeEmitterUpdateFrame(fxeEmitterHandle pEmitter)
{
	fxeEmitterTranslateFrame(pEmitter);
	fxeEmitterClipAndDraw(pEmitter);
}

void fxeEmitterTranslateFrame(fxeEmitterHandle pEmitter)
{
	int Loop;
	fxeParticle *pParticle;
	int NumParticles;
	fxeParticle *SavedParticle;

	if (pEmitter->NumParticles == 0)
		return;

	pParticle = pEmitter->UsedParticles;

	NumParticles = pEmitter->NumParticles;

	for (Loop=0;Loop<NumParticles;Loop++)
	{
		if (pParticle->NumFramesLeft == 0)
		{
			if (pParticle == pEmitter->UsedParticles)
			{
				// This particle is the first particle in the
				// emitter. We need to change the first particle to
				// this particle.
				pEmitter->UsedParticles = pParticle->Next;
				SavedParticle = pParticle->Next;
				pEmitter->NumParticles--;
				pParticle->Next = pEmitter->UnusedParticles;
				pEmitter->UnusedParticles = pParticle;
				pParticle = SavedParticle;
			}
			else if (Loop==(NumParticles-1))
			{
				// This is the last used particle in the emitter,
				// this means we just have to decrement the particle
				// counter.

				pEmitter->NumParticles--;
				pParticle->Next = pEmitter->UnusedParticles;
				pEmitter->UnusedParticles = pParticle;
			}
			else
			{
				// This is not the first particle and not the last
				// so we need to take into account previous and next
				// particle
				SavedParticle = pParticle->Next;
				pEmitter->NumParticles--;
				((fxeParticle*)(pParticle->Prev))->Next = 
					pParticle->Next;
				((fxeParticle*)(pParticle->Next))->Prev =
					pParticle->Prev;
				pParticle->Next = pEmitter->UnusedParticles;
				pEmitter->UnusedParticles = pParticle;
				pParticle = SavedParticle;
				
			}
		}

		else
		{
			// The particle have frames left to live, so translate
			// it and change its rgb color and decrement its life
			// time.

			pParticle->NumFramesLeft--;
			
			pParticle->CurrColor.r += pParticle->IncColor.r;
			pParticle->CurrColor.g += pParticle->IncColor.g;
			pParticle->CurrColor.b += pParticle->IncColor.b;
			
			pParticle->x += pParticle->DirectionVector.x 
				+ pParticle->ForceVector.x;
			pParticle->y += pParticle->DirectionVector.y 
				+ pParticle->ForceVector.y;
			pParticle->z += pParticle->DirectionVector.z 
				+ pParticle->ForceVector.z;

			pParticle->ForceVector.x *= pEmitter->ForceAcceleration;
			pParticle->ForceVector.y *= pEmitter->ForceAcceleration;
			pParticle->ForceVector.z *= pEmitter->ForceAcceleration;
			pParticle = pParticle->Next;
		}

	}
}

void fxeEmitterClipAndDraw(fxeEmitterHandle pEmitter)
{
	fxeDrawVertex v1,v2,v3,v4;
  void *VertexPointers[4];
	int Loop;
	fxeParticle *pParticle;
	fxeTextureHandle TexHandle;
	fxeRectangle2d Rec2d;
	fxeRectCoordinate *RC;
	fxeDrawVertex WorldVertex, CameraVertex;
	float CurrQ;
	


	
	if (pEmitter->NumParticles == 0)
		return;

	WorldVertex.x = pEmitter->x;
	WorldVertex.y = pEmitter->y;
	WorldVertex.z = pEmitter->z;
	MatXDrawVerToDrawVer(CameraMatrix,&WorldVertex,&CameraVertex);
	pEmitter->cz = CameraVertex.z;

	pParticle = pEmitter->UsedParticles;

	if (GlobalFogMode != NO_FOGGING)
	{
    grFogMode(GR_FOG_DISABLE);
	}
	
	grColorCombine( GR_COMBINE_FUNCTION_SCALE_OTHER,
		GR_COMBINE_FACTOR_LOCAL,GR_COMBINE_LOCAL_ITERATED,
		GR_COMBINE_OTHER_TEXTURE,FXFALSE);

	grConstantColorValue(255<<24);

	grDepthBufferFunction(GR_CMP_ALWAYS);


	grAlphaCombine( GR_COMBINE_FUNCTION_LOCAL,
    GR_COMBINE_FACTOR_ONE,GR_COMBINE_LOCAL_CONSTANT,
    GR_COMBINE_OTHER_TEXTURE,FXFALSE );

	grAlphaBlendFunction( GR_BLEND_ONE, GR_BLEND_ONE,
    GR_BLEND_ZERO, GR_BLEND_ZERO );

	TexHandle = pEmitter->TexHandle;

	SetupTexture(TexHandle);




	// Setup texture unit with particle texture:

	VertexPointers[0] = (void*)&v1;
	VertexPointers[1] = (void*)&v2;
	VertexPointers[2] = (void*)&v3;
	VertexPointers[3] = (void*)&v4;
  
	for (Loop=0; Loop<pEmitter->NumParticles; Loop++)
	{
		
		// Transform to camera space:
		WorldVertex.x = pParticle->x;
		WorldVertex.y = pParticle->y;
		WorldVertex.z = pParticle->z;

		MatXDrawVerToDrawVer(CameraMatrix,&WorldVertex,&CameraVertex);

		// Check if the particle need to be clipped:
		if ((CameraVertex.x + pParticle->Radius) < (-CameraVertex.z) ||
			(CameraVertex.x - pParticle->Radius) > CameraVertex.z ||
			(CameraVertex.y + pParticle->Radius) < (-CameraVertex.z) ||
			(CameraVertex.y - pParticle->Radius) > CameraVertex.z ||
			CameraVertex.z < NearClippingPlane || 
			CameraVertex.z > FarClippingPlane)
		{
			pParticle = pParticle->Next;
		}
		else
		{
			if ((CameraVertex.x + pParticle->Radius) < CameraVertex.z &&
				(CameraVertex.x - pParticle->Radius) > (-CameraVertex.z) &&
				(CameraVertex.y + pParticle->Radius) < CameraVertex.z &&
				(CameraVertex.y - pParticle->Radius) > (-CameraVertex.z))
			{

				v1.q = 1.0f/CameraVertex.z;      //pEmitter->cz;
				CurrQ = 1.0f/CameraVertex.z;
				v1.x = ((CameraVertex.x-pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				Snap(&v1.x);
				v1.y = ((CameraVertex.y+pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;
				Snap(&v1.y);
				v1.r = pParticle->CurrColor.r * 255.0f;
				v1.g = pParticle->CurrColor.g * 255.0f;
				v1.b = pParticle->CurrColor.b * 255.0f;
				v1.u0 = 0.0f;
				v1.v0 = 0.0f;
				v1.u1 = 0.0f;
				v1.v1 = 0.0f;

				v2.q = v1.q;
				v2.x = ((CameraVertex.x-pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				Snap(&v2.x);
				v2.y = ((CameraVertex.y-pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;
				Snap(&v2.y);
				v2.r = v1.r;
				v2.g = v1.g;
				v2.b = v1.b;
				v2.u0 = 0.0f;
				v2.v0 = 255.0f*v1.q;
				v2.u1 = 0.0f;
				v2.v1 = v2.v0;

				v3.q = v1.q;
				v3.x = ((CameraVertex.x+pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				Snap(&v3.x);
				v3.y = ((CameraVertex.y-pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;
				Snap(&v3.y);
				v3.r = v1.r;
				v3.g = v1.g;
				v3.b = v1.b;
				v3.u0 = 255.0f*v1.q;
				v3.v0 = v3.u0;
				v3.u1 = v3.u0;
				v3.v1 = v3.u0;

				v4.q = v1.q;
				v4.x = ((CameraVertex.x+pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				Snap(&v4.x);
				v4.y = ((CameraVertex.y+pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;
				Snap(&v4.y);
				v4.r = v1.r;
				v4.g = v1.g;
				v4.b = v1.b;
				v4.u0 = 255.0f*v1.q;
				v4.v0 = 0.0f;
				v4.u1 = v4.u0;
				v4.v1 = 0.0f;
			}
			else
			{
				// Project all vertices:

				v1.q = 1.0f/CameraVertex.z; //pEmitter->cz;
				CurrQ = 1.0f/CameraVertex.z;
				v1.x = ((CameraVertex.x-pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				v1.y = ((CameraVertex.y+pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;
				v1.r = pParticle->CurrColor.r * 255.0f;
				v1.g = pParticle->CurrColor.g * 255.0f;
				v1.b = pParticle->CurrColor.b * 255.0f;

				v2.q = v1.q;
				v2.x = ((CameraVertex.x-pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				v2.y = ((CameraVertex.y-pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;
				v2.r = v1.r;
				v2.g = v1.g;
				v2.b = v1.b;

				v3.q = v1.q;
				v3.x = ((CameraVertex.x+pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				v3.y = ((CameraVertex.y-pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;
				v3.r = v1.r;
				v3.g = v1.g;
				v3.b = v1.b;

				v4.q = v1.q;
				v4.x = ((CameraVertex.x+pParticle->Radius) * CurrQ +
					1.0f) * ProjectWidthDiv2 + ProjectStartX;
				v4.y = ((CameraVertex.y+pParticle->Radius) * CurrQ +
					1.0f) * ProjectHeightDiv2 + ProjectStartY;
				v4.r = v1.r;
				v4.g = v1.g;
				v4.b = v1.b;

				RC = Rec2d.RectCoordinates;
				RC[0].x = v1.x;
				RC[0].y = v1.y;
				RC[1].x = v2.x;
				RC[1].y = v2.y;
				RC[2].x = v3.x;
				RC[2].y = v3.y;
				RC[3].x = v4.x;
				RC[3].y = v4.y;

				ClipTextureRectangle2d(&Rec2d);

				v1.x = RC[0].x;
				v1.y = RC[0].y;
				Snap(&v1.x);
				Snap(&v1.y);
				v1.u0 = RC[0].u*255.0f*v1.q;
				v1.v0 = RC[0].v*255.0f*v1.q;
				v1.u1 = RC[0].u*255.0f*v1.q;
				v1.v1 = RC[0].v*255.0f*v1.q;

				v2.x = RC[1].x;
				v2.y = RC[1].y;
				Snap(&v2.x);
				Snap(&v2.y);
				v2.u0 = RC[1].u*255.0f*v1.q;
				v2.v0 = RC[1].v*255.0f*v1.q;
				v2.u1 = RC[1].u*255.0f*v1.q;
				v2.v1 = RC[1].v*255.0f*v1.q;

				v3.x = RC[2].x;
				v3.y = RC[2].y;
				Snap(&v3.x);
				Snap(&v3.y);
				v3.u0 = RC[2].u*255.0f*v1.q;
				v3.v0 = RC[2].v*255.0f*v1.q;
				v3.u1 = RC[2].u*255.0f*v1.q;
				v3.v1 = RC[2].v*255.0f*v1.q;

				v4.x = RC[3].x;
				v4.y = RC[3].y;
				Snap(&v4.x);
				Snap(&v4.y);
				v4.u0 = RC[3].u*255.0f*v1.q;
				v4.v0 = RC[3].v*255.0f*v1.q;
				v4.u1 = RC[3].u*255.0f*v1.q;
				v4.v1 = RC[3].v*255.0f*v1.q;


			}

			Performance.NumPolys++;

	
			grDrawVertexArray(GR_TRIANGLE_FAN,4,VertexPointers);
			
			pParticle = pParticle->Next;
		}
		

	}
	
  grAlphaCombine(GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_ONE,
    GR_COMBINE_LOCAL_NONE,GR_COMBINE_OTHER_CONSTANT,FXFALSE);

  grAlphaBlendFunction(GR_BLEND_SRC_ALPHA,GR_BLEND_ONE_MINUS_SRC_ALPHA,
      GR_BLEND_ZERO,GR_BLEND_ZERO);

	grDepthBufferFunction(GR_CMP_LEQUAL);

	if (GlobalFogMode != NO_FOGGING)
	{
    grFogMode(GR_FOG_WITH_TABLE_ON_Q);
	}
}


void fxeEmitterChangeProperties(fxeEmitter *pEmitter, 
	float x, float y, float z, fxeVector *ForceVector)
{
	pEmitter->x = x;
	pEmitter->y = y;
	pEmitter->z = z;
	pEmitter->ForceVector.x = ForceVector->x;
	pEmitter->ForceVector.y = ForceVector->y;
	pEmitter->ForceVector.z = ForceVector->z;
}





