
// Massive 3d Power - FxEngine ...

/*
01/04/1999 - Fixed object lightning bug, Fixed Lensflare bug, 
						 discovered new bug: when
						 ALT-TAB from app and back the clipping routines doesn't
						 work when Z < 0, very strange...
*/

/*
01/03/1999 - Fixed enviroment mapping clipping problem.
						 Fixed problem when application is loosing
						 focus and getting focus from windows.
						 Use fxeLooseFocus when your application is
						 loosing focus and fxeSetFocus when your
						 application is getting focus.
*/

/*
01/02/1999 - Implemented texture caching. Now you can load more
						 textures then texture memory of 3dfx cards allows.
						 Fixed 2MB texture memory bug.
						 Recoded texure loading so that FxEngine isn't loading
						 textures that already have been loaded, instead 
						 the handle of the old texture is returned.
						 Fixed lensflare effect, only displays lensflare when
						 lightsource is visible.
						 Recoded object clipping routines for the 3rd time...
             Implemeexitnted Skyboxes.
*/

/*
12/Something/1998 - Fixed bugs in lightmap system, fixed
										landscape clipping bug, optimized landscape
										rendering. Fixed particle system clipping bug,
										added smoke system. Dynamic lightsources now
										also affects objects. Still dynamic lightning bug
										with flat shaded objects. Fixed 2d bitmap drawing
										bug. Ported FxEngine to delphi. Made new simple fxengine
										sample program. Now the engine is using dithered
										texture mapping with bilinear filtering, looks
										very good.
*/

/*
12/01/1998 - Now lightmapping for landscapes works. Yes!! It was real hard to add this one...
						 You can have as many dynamic lightsources as you want. Still some code needed to make
						 ambient light work for landscapes and dynamic lightsources to work with objects.
						 for the moment around 11000 lines of code.
*/

/*
11/24/1998 - Wrote structures for lightmaps and added some to the
						 landscape structures to make use of lightmaps.
*/

/*
11/23/1998 - Plans to implement lighting through colored light maps allowing
             for really cool lightsource effects such as a red light
             around explosions that can vary and is very detailed.
             The lightmaps will probably be 8x8 and they are bilinear
             filtered by 3dfx so they will certainly look good.
             It will be a two pass operation on voodoos and single pass
             on voodoo2.
*/

/* 
11/22/1998 - Added lensflare effects to the engine. Looks really cool!
						 Also added collision detection for object-landscape.
						 Many thanks goes to Daniel M Jones for parts of the
						 Lensflare effect.
*/

/*
11/17/1998 - Conversion now finished except clipping of particles.
*/

/*
11/15/1998 - Began converting the engine to window coordinates. I hope it 
             will work on voodoo1, voodoo banshee and voodoo rush when
             finished. Finished converting landscapes and is almost
             finished with particle system. 
						 Added function so you can specify where the 3d rendering 
						 window will be positioned on screen and how big it will be.
						 This lets you add such things as split screens.
*/

/*
10/24/1998 - Now you can load 3ds files from the exe-file, you need
						 to make a custom "FILE_3DS" resource type and add the
						 3ds object files as "FILE_3DS" resources to the 
						 resources of the project. Still some shading errors when
						 using gouraud shading.
*/

/*
10/23/1998 - Now sorts emitters in z. Looks better when multiple
						 emitters is involved.   
*/


/*
10/22/1998 - Optimized the object structure for size. Made 
						 precalculation of polygon and vertex normals.

/*
10/21/1998 - Now 3ds loader can load ambient and diffuse
             colors from 3ds files. Also alpha values but
						 they are not currently used by FxEngine.
						 Fixed bug when double sided polygons is used. 
						 This problem is only fixed for flat shaded 
						 objects, not for gouraud shaded objects.
						 Optimized memory use too.

*/

/*

Welcome to the world of FxEngine, a 3d engine running on top of
   glide3x which uses 3dfx
   cards for incredible detail and smoothness.
   Key features supported (or will be supported when finished)

   - Trilinear filtered gouraud shaded texture mapping
   - Unlimited amount of RGB lightsources (moving lightsources only supported
     on objects)
   - Unlimited amount of objects
   - Unlimited size of Landscapes
   - Transparency (only in objects)
   - fogging
   - objects
   - Landscapes
   - Texture Mapping with mipmapping, bilinear filtering/trilinear filtering
   - Enviroment mapping with mipmapping, bilinear filtering/trilinear filt.
   - Advanced 2d bitmap/text functions (zoom in/out, translucency/transparency)
   - Bump mapping (coming)
   - Collision detection
   - Particle system
	 - 3ds object loader
	 - Used defined render windows
	 - Lens flares
	 - Terrain Following (coming)
	 - Lightmapping (coming)
	 - Skybox support

   Main source file for FxEngine, I allows anybody use this code as
   much as you want, make changes and so on but you're NOT allowed to
   call the code for FxEngine without permission from the author of
   FxEngine. If you want to contact the author you can go to
   http://welcome.to/3dfxPS
   From that site you can contact the author and can get info on
   how to contribute on FxEngine and more.
   You can of course also contact him directly by e-mail:
   andreas.ingo@swipnet.se

*/

// This version is limited to 1000 vertices per polygon, but you will
// probably NEVER need so much...


#include <glide.h>
#include "resource.h"
#include "FxMalloc.h"
#include "FxResource.h"
#include "dynlight.h"
#include "FxLog.h"
#include "FxEnginedrv.h"
#include "fxlandscapedrv.h"
#include "FxParticle.h"
#include "FxSmoke.h"
#include "FxCharacter.h"
#include "FxSkybox.h"
#include "FxTexture.h"
#ifndef _FXLIGHTMAP_
#include "Lightmap.h"
#endif

#include "3dsload.h"
#include <math.h>
#include <stdio.h>

#define MAX_VERTICES_PER_POLY 1000
#define MAX_OBJECTS_IN_WORLD 10000
#define MAX_POLYGONS_PER_OBJECT 30000
#define MAX_EMITTERS_IN_WORLD 1000
#define MAX_SMOKE_EMITTERS_IN_WORLD 1000

fxeObjectPtr WorldObjectPointers[MAX_OBJECTS_IN_WORLD];
fxePolygonPtr PolygonPointers[MAX_POLYGONS_PER_OBJECT];

fxePolygonList SortedPolygons[MAX_POLYGONS_PER_OBJECT];

// Globals:

GrTexInfo TexInfo;
FxU32 DownloadAddress;

int TextureAlign;

int LostFocus;

int BL;

fxeCharacterset DefaultCS;

unsigned char *GlideState;

HWND GlobalHwnd;
 
GrScreenResolution_t Res;

int CombineTextureOnly;

int TextureMode;

int ChromaKey;

unsigned long DrawShape;

float ViewingDistance;


float HorInc, VerInc;

float ProjectWidthDiv2;
float ProjectHeightDiv2;
float ProjectStartX;
float ProjectStartY;
float ProjectEndX;
float ProjectEndY;

int HardwareSelected;

int SwapFlag;

fxeTextureHandle CurrentTextureSource = NULL;

float CosLockup[361];
float SinLockup[361];

float NearClippingPlane;
float FarClippingPlane;



long sqrt_table[0x100];

fxePolygon **GlobalPolygonList;

fxeObject *CurrentObject;

FxI32 NumTexelFx;

FxU32 WMaxValue;


int TextureFilteringMode;

GrContext_t Context;

int GlobalFogMode;


// All objects in world
fxeObject *WorldObjects;


int TotNumLightsources;

// World camera position and angle
fxeCamera Camera;

fxePerformance Performance;


fxeCharactersetHandle DefaultCharacterset;

int AlphaBlendMode;

fxeColor BackgroundColor;

float SnapValue;





// All textures in world
fxeTextureHandle Textures;

int GlobalVertexIndices[3];
int GlobalVertexIndices2[3];


float HFOWInc;
float VFOWInc;

int FrameDisplayed;

// Ambient light level.
float AmbientR, AmbientG, AmbientB;
float AmbientR2, AmbientG2, AmbientB2;

GrFog_t *FxeFogTable;
FxI32 nFogEntries;

fxeTextureHandle NULLTexture;

unsigned long FrameCounter;

float VertexSnapper;

float DotProduct3d(fxeVector *Vector1, fxeVector *Vector2);

void SetupTmusWithTexture(fxeTextureHandle TexHandle)
{
		// check which texture filtering method which
    // is used:

		SetupTexture(TexHandle);
}

void fxeResolutionGet(float *HorisontalRes, float *VerticalRes)
{
	*HorisontalRes = (float)HorRes;
	*VerticalRes = (float)VerRes;
}

void SetupSnapping(void)
{
	int MemVar;
	
	VertexSnapper = (float)(3L << 18);
	
	// Initialize FPU to 24 bit precision
	_asm
	{
		push eax
		finit
		fwait
		fstcw [MemVar]
		fwait
		mov eax, [MemVar]
		and eax, 0xfffffcff
		mov [MemVar], eax
		fldcw [MemVar]
		fwait
		pop eax
	}
}

void Snap(float *ValueToSnap)
{
	*ValueToSnap += VertexSnapper;
	*ValueToSnap -= VertexSnapper;
}

void fxeRenderWindowSet(int x, int y, int Width, int Height)
{
	ProjectStartX = (float)x;
	ProjectStartY = (float)y;
	ProjectWidthDiv2 = ((float)Width)/2.0f;
	ProjectHeightDiv2 = ((float)Height)/2.0f;
	ProjectEndX = (float)x+(float)Width;
	ProjectEndY = (float)y+(float)Height;
}

void fxeFogSetMode(int FogMode);

// Here comes some matrix math functions by Hugues Landry which also is
// going to be used in a 3d engine called Sp@cE_iNvAdr.
// Big thanks to you Hugues for this! The functions have been modified
// some by Andreas Ingo to fit better in FxEngine.

typedef float fxeMatrix[4][4];


// Clips a textured rectangle in 2d 
// Vertex 0 in rectangle is upper left corner, next vertex
// is down left corner, next is down right corner and last
// vertex is upper right corner.
// When a vertex is y=0 and x=0 then the vertex resides
// at the down left corner of the screen.
// A rectangle MUST NOT be completely outside the clipping region
// incapsulated by ProjectStartX,ProjectStartY,ProjectEndX,
// and ProjectEndY.

void ClipTextureRectangle2d(fxeRectangle2d *pRec2d)
{
	float Width, Height;
	float IncX,IncY;
	fxeRectCoordinate *RC;

	RC = pRec2d->RectCoordinates;

	Width = RC[3].x - RC[0].x;
	Height = RC[0].y - RC[1].y;

	if (RC[0].x < ProjectStartX)
	{
		// We must clip this rectangle in the left side.
		IncX = ProjectStartX - RC[0].x;
		RC[0].u = IncX/Width;
		RC[1].u = RC[0].u;
		RC[0].x = ProjectStartX;
		RC[1].x = ProjectStartX;
	}
	else
	{
		RC[0].u = 0.0f;
		RC[1].u = 0.0f;
	}

	if (RC[3].x > ProjectEndX)
	{
		// We must clip this rectangle in the right side.
		IncX = RC[3].x - ProjectEndX;
		RC[3].u = 1.0f-IncX/Width;
		RC[2].u = RC[3].u;
		RC[3].x = ProjectEndX;
		RC[2].x = ProjectEndX;
	}
	else
	{
		RC[3].u = 1.0f;
		RC[2].u = 1.0f;
	}

	if (RC[1].y < ProjectStartY)
	{
		// We must clip this rectangle in the lower side.
		IncY = ProjectStartY - RC[1].y;
		RC[1].v = IncY/Height;
		RC[2].v = RC[1].v;
		RC[1].y = ProjectStartY;
		RC[2].y = RC[1].y;
	}
	else
	{
		RC[1].v = 0.0f;
		RC[2].v = 0.0f;
	}

	if (RC[0].y > ProjectEndY)
	{
		// We must clip this rectangle in the upper side.
		IncY = RC[0].y - ProjectEndY;
		RC[0].v = 1.0f-IncY/Height;
		RC[3].v = RC[0].v;
		RC[0].y = ProjectEndY;
		RC[3].y = RC[0].y;
	}
	else
	{
		RC[0].v = 1.0f;
		RC[3].v = 1.0f;
	}
}



void ConcatXforms(fxeMatrix SourceXform1, fxeMatrix SourceXform2,
  fxeMatrix DestXform);

float GetMaximumRadius(fxeObject *pObject)
{
  int Loop;
  float MaxRadius;
  float TempRadius;

  MaxRadius = 0.0;

  for (Loop=0; Loop<pObject->NumVertices; Loop++)
  {
    TempRadius = (float)sqrt(pObject->VertexList[Loop].x*
      pObject->VertexList[Loop].x+pObject->VertexList[Loop].y*
      pObject->VertexList[Loop].y+pObject->VertexList[Loop].z*
      pObject->VertexList[Loop].z);

    if (TempRadius > MaxRadius) MaxRadius = TempRadius;
  }

  return MaxRadius;
}


float GetPolygonRadius(fxePolygon *pPoly, fxeObject *pObject)
{
  int Loop;

  float ObjectRadius, LargestRadius;
  fxeVector Vector;

  LargestRadius = 0;



  for (Loop=1; Loop<pPoly->NumVertices; Loop++)
  {
    Vector.x = pObject->VertexList[pPoly->VertexIndices[Loop]].x -
      pObject->VertexList[pPoly->VertexIndices[0]].x;
    
    Vector.y = pObject->VertexList[pPoly->VertexIndices[Loop]].y -
      pObject->VertexList[pPoly->VertexIndices[0]].y;
    
    Vector.z = pObject->VertexList[pPoly->VertexIndices[Loop]].z -
      pObject->VertexList[pPoly->VertexIndices[0]].z;

    ObjectRadius = (float)sqrt(Vector.x*Vector.x+Vector.y*Vector.y+Vector.z*Vector.z);
    if (ObjectRadius > LargestRadius)
      LargestRadius = ObjectRadius;
  }

  return LargestRadius;
}

void SetupViewFrustrum(void);

// Here is some huge lockup tables which stores the rotation
// matrices for all rotation angles from 0 to 360.


////////////////////////////////////////////////
// Transformation Module Function declaration //
////////////////////////////////////////////////

void InitMatrix();						// identity Matrix initialization

void MatCopy(fxeMatrix Source, fxeMatrix Dest);

// Matrix math Transformation Functions

void MoveMatrix(fxeMatrix Matrix, float x, float y, float z);
void RotXMatrix(fxeMatrix Matrix, float Angle);
void RotYMatrix(fxeMatrix Matrix, float Angle);
void RotZMatrix(fxeMatrix Matrix, float Angle);
void ScaleMatrix(fxeMatrix Matrix, float SFactor);
void ProjectMatrix(fxeMatrix Matrix);

extern void _cdecl CrossProduct3dASM(fxeVector *Result, fxeVector *Vector1,
  fxeVector *Vector2);
extern void _cdecl SetPrecisionASM(void);

void CopyVertex(fxeDrawVertex *Src, fxeDrawVertex *Dst);
void CopyLVertex(fxeLDrawVertex *Src, fxeLDrawVertex *Dst);

int ClipLineToPlane(fxePlane *pPlane, fxeLine *pInLine, fxeLine *pOutLine);
int ClipLLineToPlane(fxePlane *pPlane, fxeLLine *pInLine, fxeLLine *pOutLine);

// Matrix Math multiply Functions

void Matmult(fxeMatrix MatrixA, fxeMatrix MatrixB);
void MatXVer(fxeMatrix Matrix, fxeVertex *Vertex, fxeVertex *Result);
void MatXVerToDrawVer(fxeMatrix Matrix, fxeVertex *Vertex, fxeDrawVertex *Result);
void MatXDrawVerToDrawVer(fxeMatrix Matrix, fxeDrawVertex *Vertex, fxeDrawVertex *Result);
void LMatXDrawVerToDrawVer(fxeMatrix Matrix, fxeLDrawVertex *Vertex, fxeLDrawVertex *Result);
void MatXVec(fxeMatrix Matrix, fxeVector *Vector, fxeVector *Result);

// Optimized Assembly version of MatXVer
//void extern _cdecl
//  MatXVerASM(fxeMatrix Matrix, fxeVertex *Vertex, fxeVertex *Result);


// This matrix is calculated for every call to fxeRenderWorld.
// fxeRenderWorld calculates the camera transformation matrix. Because
// the camera isn't changed under that function call we only need
// to calculate it once.

fxeMatrix CameraMatrix;
fxeMatrix CameraRotX;
fxeMatrix CameraRotY;
fxeMatrix CameraRotZ;
fxeMatrix CameraMove;

/*
int _stdcall LibMain(void *hinstDll,unsigned long dwReason,void *reserved)
{
	return(1);
}
*/

void fxeDriverGetInfo(fxeDriverInfo *DriverInfo)
{
  DriverInfo->DriverString = "FxEngine Glide3 Driver version 0.40";
  DriverInfo->MoreInfo = "FxEngineG3.txt";
  DriverInfo->TextFileExist = FXETRUE;
  DriverInfo->MainVersion = 0;
  DriverInfo->UnderVersion = 40;
}
     


// Fast square root functions:
void build_sqrt_table(void) {
    unsigned short i;
    float f;
    unsigned int *fi = (unsigned*)&f;   // To access the bits of a float in
                                        // C quickly we must misuse pointers
    for (i = 0; i <= 0x7f; i++) {
        *fi = 0;

            // Build a float with the bit pattern i as mantissa
            // and an exponent of 0, stored as 127
        *fi = (i << 16) | (127 << 23);
        f = (float)sqrt(f);

            // Take the square root then strip the first 7 bits of
            // the mantissa into the table
        sqrt_table[i] = (*fi & 0x7fffff);

            // Repeat the process, this time with an exponent of 1,
            // stored as 128
        *fi = 0;
        *fi = (i << 16) | (128 << 23);
        f = (float)sqrt(f);
        sqrt_table[i+0x80] = (*fi & 0x7fffff);
    }
}


/* original C version */
float fast_fsqrt(float n) {
    unsigned int *num = (unsigned *)&n; // to access the bits of a float in C
                                        // we must misuse pointers

    short e;        // the exponent
    if (n == 0) return (0); /* check for square root of 0 */
    e = (*num >> 23) - 127; /* get the exponent - on a SPARC the */
                            /* exponent is stored with 127 added */
    *num &= 0x7fffff;   /* leave only the mantissa */
    if (e & 0x01) *num |= 0x800000;
                /* the exponent is odd so we have to */
                /* look it up in the second half of  */
                /* the lookup table, so we set the high bit */
    e >>= 1;    /* divide the exponent by two */
                /* note that in C the shift */
                /* operators are sign preserving */
                /* for signed operands */
        // Do the table lookup, based on the quaternary mantissa,
        // then reconstruct the result back into a float
    *num = ((sqrt_table[*num >> 16])) + ((e + 127) << 23);
    return(n);
}


void fxeVSyncEnable(void)
{
  SwapFlag = 1;
}

void fxeVSyncDisable(void)
{
  SwapFlag = 0;
}

void fxeFogSetMode(int FogMode)
{
  if (FogMode == NO_FOGGING)
  {
    grFogMode(GR_FOG_DISABLE);
  }
  else if (FogMode == DISTANCE_FOGGING)
  {
    guFogGenerateLinear(FxeFogTable,NearClippingPlane,FarClippingPlane);
    grFogTable(FxeFogTable);
    grFogMode(GR_FOG_WITH_TABLE_ON_Q);
  }
  GlobalFogMode = FogMode;
}


// this function returns the number of frames that have
// been rendered since last call to fxeStartFrameCount
// if the function haven't been called the function

void fxeStartFrameCount(void)
{
  FrameCounter = 0;
}

void fxePerformanceInit(void)
{
  Performance.NumPolys = 0;
  Performance.FrameCount = 0;
}

fxePerformance * fxePerformanceGet(void)
{
  return &Performance;
}

// this function returns the number of frames that have
// been rendered since last call to fxeStartFrameCount
unsigned long fxeGetFrameCount(void)
{
  return FrameCounter;
}

void fxeFogSetColor(fxeColor *FogColor)
{
  grFogColorValue((unsigned long)(FogColor->r*255.0 * 65536.0+ FogColor->g*255.0 * 256.0
    + FogColor->b*255.0));
}


// This function converts a angle in degrees to a angle in radians
float DegToRad(float Deg)
{
  return (Deg*0.017453292f);
}

float RadToDeg(float Rad)
{
  return (Rad*57.29577951f);
}

// converts a floating point value to a integer value
int FloatToInt(float Value)
{
  return (int)Value;
}


// Text rendering associated functions:

void fxeSetDefaultCharacterset(fxeCharactersetHandle CSH)
{
  DefaultCharacterset = CSH;
}

fxeCharactersetHandle fxeCharactersetLoad(char *FilePrefix,
  int UseDefaultSettings, int Translucent,
  float AlphaLevel, int Transparent, fxeColor *TransColor, float Spacing,
  float CharWidth, float CharHeight)
{
  fxeCharactersetHandle CSH;
  int Loop;
  char Filename[256];

  CSH = (fxeCharactersetHandle)FxMalloc(sizeof(fxeCharacterset));
  if (CSH == NULL)
    return NULL;

  for (Loop=0; Loop<256; Loop++)
  {
    sprintf(Filename,"%s%c%s",FilePrefix,(char)(Loop),".3df");

    CSH->Characterset[Loop].TextureHandle = fxeTextureLoad(Filename,FXE_TEXTURE_MAP);
   
    CSH->Spacing = Spacing;
    CSH->CharWidth = CharWidth;
    CSH->CharHeight = CharHeight;
    CSH->Translucent = Translucent;
    CSH->AlphaLevel = AlphaLevel;
    CSH->Transparent = Transparent;
    if (Transparent == FXETRUE && TransColor != NULL)
    {
      CSH->TransColor.r = TransColor->r;
      CSH->TransColor.g = TransColor->g;
      CSH->TransColor.b = TransColor->b;
    }
    CSH->UseDefaults = UseDefaultSettings;

    if (CSH->Characterset[Loop].TextureHandle != NULL)
    {
      CSH->Characterset[Loop].Used = FXETRUE;
    }
    else
      CSH->Characterset[Loop].Used = FXEFALSE;
  }
  return CSH;
}


void fxeCharactersetSetDefaultDrawingMode(fxeCharactersetHandle CSH)
{
  CSH->UseDefaults = FXETRUE;
}

void fxeCharactersetSetSpecialDrawingMode(fxeCharactersetHandle CSH,
  int Translucent, float AlphaLevel, int Transparent, fxeColor *TransColor,
  float Spacing, float CharacterWidth, float CharacterHeight)
{
  
  CSH->UseDefaults = FXEFALSE;
  CSH->Translucent = Translucent;
  CSH->AlphaLevel = AlphaLevel;
  CSH->Transparent = Transparent;
  if (Transparent == FXETRUE && TransColor != NULL)
  {
    CSH->TransColor.r = TransColor->r;
    CSH->TransColor.g = TransColor->g;
    CSH->TransColor.b = TransColor->b;
  }
  CSH->Spacing = Spacing;
  CSH->CharWidth = CharacterWidth;
  CSH->CharHeight = CharacterHeight;
}
  
  
  
void fxeStringDrawXY(char *String, float X, float Y)
{
  float CurrX;
  fxeCharactersetHandle Characterset;
  CurrX = X;
  
  if (DefaultCharacterset->UseDefaults == FXETRUE)
  {
    Characterset = &DefaultCS;
  }
  else
  {
    Characterset = DefaultCharacterset;
  }



  if (Characterset->Translucent == FXETRUE)
  {
    grConstantColorValue( (((int)DefaultCharacterset->AlphaLevel)<<24) );
  }

  else if (Characterset->Translucent == FXEFALSE)
  {
    grConstantColorValue(255<<24);
  }

  if (Characterset->Transparent == FXETRUE && ChromaKey == FXEFALSE)
  {
    grChromakeyMode(GR_CHROMAKEY_ENABLE);
    ChromaKey = FXETRUE;
  }
  if (Characterset->Transparent == FXETRUE)
  {
    grChromakeyValue( (((int)(Characterset->TransColor.r*255.0)) << 16) +
      (((int)(Characterset->TransColor.g*255.0)) << 8) +
      ((int)(Characterset->TransColor.b*255.0)) );
  }

  if (Characterset->Transparent == FXEFALSE && ChromaKey == FXETRUE)
  {
    grChromakeyMode(GR_CHROMAKEY_DISABLE);
    ChromaKey = FXEFALSE;
  }



  if (String != NULL)
  {
    while ((*String) != 0)
    {
      if (DefaultCharacterset->Characterset[(*String)].Used == FXETRUE)
      {
        fxeCharacterDraw(&(DefaultCharacterset->Characterset[(*String)]),
          CurrX,Y,Characterset->CharWidth,
          Characterset->CharHeight);
      }
      CurrX += Characterset->Spacing;
      String++;
    }
  }
}
    
  
fxeCharacterHandle fxeBitmapCreate(fxeTextureHandle TextureHandle,
  int Translucent, float AlphaLevel, int Transparent,
  fxeColor* TransparentColor)
{
  fxeCharacter *pCharacter;


  pCharacter = (fxeCharacter*)FxMalloc(sizeof(fxeCharacter));
  if (pCharacter == NULL)
    return NULL;

  pCharacter->TextureHandle = TextureHandle;
  pCharacter->Translucent = Translucent;
  pCharacter->AlphaLevel = AlphaLevel;
  pCharacter->Transparent = Transparent;
	if (TransparentColor != NULL)
	{
		pCharacter->TransparentColor.r = TransparentColor->r;
		pCharacter->TransparentColor.g = TransparentColor->g;
		pCharacter->TransparentColor.b = TransparentColor->b;
	}
  return pCharacter;
}


  

void fxeSetRenderState(int RenderState)
{
  if (RenderState == FXE_RENDER_2D)
  {

    // Disable fogging if needed:
    if (GlobalFogMode != NO_FOGGING)
      grFogMode(GR_FOG_DISABLE);

  }

  else if (RenderState == FXE_RENDER_3D)
  {

    // Disable chroma keying
    grChromakeyMode(GR_CHROMAKEY_DISABLE);
    ChromaKey = FXEFALSE;

    // Enable fogging if needed:

    if (GlobalFogMode != NO_FOGGING)
    {
      fxeFogSetMode(GlobalFogMode);
    }

  }
}


void fxeBitmapDrawXY(fxeCharacterHandle CharHandle, float X, float Y,
  float Width, float Height)
{

  fxeDrawVertex Character[4];
  fxeTextureHandle TexHandle;


  if (CharHandle->Translucent == FXETRUE)
	{
    grConstantColorValue( (((int)CharHandle->AlphaLevel)<<24) );
  }
  else if (CharHandle->Translucent == FXEFALSE)
  {
    grConstantColorValue(255<<24);
  }

  if (CharHandle->Transparent == FXETRUE && ChromaKey == FXEFALSE)
  {
    grChromakeyMode(GR_CHROMAKEY_ENABLE);
    ChromaKey = FXETRUE;
  }
  if (CharHandle->Transparent == FXETRUE)
  {
    grChromakeyValue( (((int)(CharHandle->TransparentColor.r*255.0)) << 16) +
      (((int)(CharHandle->TransparentColor.g*255.0)) << 8) +
      ((int)(CharHandle->TransparentColor.b*255.0)) );
  }

  if (CharHandle->Transparent == FXEFALSE && ChromaKey == FXETRUE)
  {
    grChromakeyMode(GR_CHROMAKEY_DISABLE);
    ChromaKey = FXEFALSE;
  }



  TexHandle = CharHandle->TextureHandle;


  Character[0].x = X*HorInc;
  Character[0].y = VerRes-Y*VerInc;
  Character[0].q = 1.0;
  Character[0].r = 255.0;
  Character[0].g = 255.0;
  Character[0].b = 255.0;

  Character[1].x = X*HorInc;
  Character[1].y = VerRes-((Y+Height)*VerInc);
  Character[1].q = 1.0;
  Character[1].r = 255.0;
  Character[1].g = 255.0;
  Character[1].b = 255.0;

  Character[2].x = (X+Width)*HorInc;
  Character[2].y = Character[1].y;
  Character[2].q = 1.0;
  Character[2].r = 255.0;
  Character[2].g = 255.0;
  Character[2].b = 255.0;

  Character[3].x = Character[2].x;
  Character[3].y = Character[0].y;
  Character[3].q = 1.0;
  Character[3].r = 255.0;
  Character[3].g = 255.0;
  Character[3].b = 255.0;




    Character[0].u0 = 0.0;
    Character[0].v0 = 0.0;
    Character[1].u0 = 0.0;
    Character[1].v0 = 256.0;
    Character[2].u0 = 256.0;
    Character[2].v0 = 256.0;
    Character[3].u0 = 256.0;
    Character[3].v0 = 0.0;

    Character[0].u1 = 0.0;
    Character[0].v1 = 0.0;
    Character[1].u1 = 0.0;
    Character[1].v1 = 256.0;
    Character[2].u1 = 256.0;
    Character[2].v1 = 256.0;
    Character[3].u1 = 256.0;
    Character[3].v1 = 0.0;

	SetupTexture(TexHandle);

  // Draw the character:

  grDrawTriangle(&Character[0],&Character[1],&Character[2]);
  grDrawTriangle(&Character[2],&Character[3],&Character[0]);
}


void fxeCharacterDraw(fxeCharacterHandle CharHandle, float X, float Y,
  float Width, float Height)
{
  fxeDrawVertex Character[4];
  fxeTextureHandle TexHandle;

  TexHandle = CharHandle->TextureHandle;


  Character[0].x = X*HorInc;
  Character[0].y = VerRes-Y*VerInc;
  Character[0].q = 1.0;
  Character[0].r = 255.0;
  Character[0].g = 255.0;
  Character[0].b = 255.0;
 
  Character[1].x = X*HorInc;
  Character[1].y = VerRes-((Y+Height)*VerInc);
  Character[1].q = 1.0;
  Character[1].r = 255.0;
  Character[1].g = 255.0;
  Character[1].b = 255.0;

  Character[2].x = (X+Width)*HorInc;
  Character[2].y = Character[1].y;
  Character[2].q = 1.0;
  Character[2].r = 255.0;
  Character[2].g = 255.0;
  Character[2].b = 255.0;

  Character[3].x = Character[2].x;
  Character[3].y = Character[0].y;
  Character[3].q = 1.0;
  Character[3].r = 255.0;
  Character[3].g = 255.0;
  Character[3].b = 255.0;



    // Copy texture coordinates:

    Character[0].u0 = 0.0;
    Character[0].v0 = 0.0;
    Character[1].u0 = 0.0;
    Character[1].v0 = 256.0;
    Character[2].u0 = 256.0;
    Character[2].v0 = 256.0;
    Character[3].u0 = 256.0;
    Character[3].v0 = 0.0;

    Character[0].u1 = 0.0;
    Character[0].v1 = 0.0;
    Character[1].u1 = 0.0;
    Character[1].v1 = 256.0;
    Character[2].u1 = 256.0;
    Character[2].v1 = 256.0;
    Character[3].u1 = 256.0;
		Character[3].v1 = 0.0;

		SetupTexture(TexHandle);

  // Draw the character:

  grDrawTriangle(&Character[0],&Character[1],&Character[2]);
  grDrawTriangle(&Character[2],&Character[3],&Character[0]);
}






// Landscape rendering functions:


void ClipLandscapeToPlane(fxePlane *pPlane, fxeLandscapePtr pLandscape)
{
	int Loop;
  fxeLLine ClipLine, ClipLine2;
  int NAddedVerts;
  int NextVert;
  fxeLDrawVertex TempVertex[200];
  int Result;
	int ClipIndex;
	fxeVector TmpVertex;

		NAddedVerts = 0;

    for (Loop=0;Loop<pLandscape->NumAddedVertices1;Loop++)
    {
      NextVert = (Loop+1)%pLandscape->NumAddedVertices1;

      // We creates a line going from the current point to the
      // next point:

      CopyLVertex(&pLandscape->ClippedVertices1[Loop],&ClipLine.v0);
      CopyLVertex(&pLandscape->ClippedVertices1[NextVert],&ClipLine.v1);

      Result = ClipLLineToPlane(pPlane,&ClipLine,&ClipLine2);

      switch(Result)
      {
        case 1:
          CopyLVertex(&ClipLine2.v0,&TempVertex[NAddedVerts]);
          CopyLVertex(&ClipLine2.v1,&TempVertex[NAddedVerts+1]);
          NAddedVerts+=2;
          break;

        case 2:
          CopyLVertex(&ClipLine2.v1,&TempVertex[NAddedVerts]);
          NAddedVerts++;
          break;
      }
    }
		ClipIndex=0;
		TmpVertex.x = 100000.0f;
		TmpVertex.y = 100000.0f;
		TmpVertex.z = 100000.0f;
		if (TempVertex[NAddedVerts-1].x == TempVertex[0].x &&
			TempVertex[NAddedVerts-1].y == TempVertex[0].y &&
			TempVertex[NAddedVerts-1].z == TempVertex[0].z)
		{
				NAddedVerts--;

		}
				if (TempVertex[NAddedVerts-1].x == TempVertex[0].x &&
			TempVertex[NAddedVerts-1].y == TempVertex[0].y &&
			TempVertex[NAddedVerts-1].z == TempVertex[0].z)
		{
				NAddedVerts--;
		}


    for (Loop=0;Loop<NAddedVerts;Loop++)
    {
			if (TempVertex[Loop].x == TmpVertex.x && TempVertex[Loop].y ==
					TmpVertex.y && TempVertex[Loop].z == TmpVertex.z)
			{
				TmpVertex.x = TempVertex[Loop].x;
				TmpVertex.y = TempVertex[Loop].y;
				TmpVertex.z = TempVertex[Loop].z;
			}
			else
			{
				CopyLVertex(&TempVertex[Loop],&pLandscape->ClippedVertices1[ClipIndex]);
				ClipIndex++;
				TmpVertex.x = TempVertex[Loop].x;
				TmpVertex.y = TempVertex[Loop].y;
				TmpVertex.z = TempVertex[Loop].z;
			}
    }
    pLandscape->NumAddedVertices1 = ClipIndex;


		NAddedVerts = 0;

    for (Loop=0;Loop<pLandscape->NumAddedVertices2;Loop++)
    {
      NextVert = (Loop+1)%pLandscape->NumAddedVertices2;

      // We creates a line going from the current point to the
      // next point:

      CopyLVertex(&pLandscape->ClippedVertices2[Loop],&ClipLine.v0);
      CopyLVertex(&pLandscape->ClippedVertices2[NextVert],&ClipLine.v1);

      Result = ClipLLineToPlane(pPlane,&ClipLine,&ClipLine2);

      switch(Result)
      {
        case 1:
          CopyLVertex(&ClipLine2.v0,&TempVertex[NAddedVerts]);
          CopyLVertex(&ClipLine2.v1,&TempVertex[NAddedVerts+1]);
          NAddedVerts+=2;
          break;

        case 2:
          CopyLVertex(&ClipLine2.v1,&TempVertex[NAddedVerts]);
          NAddedVerts++;
          break;
      }
    }
		ClipIndex=0;
		TmpVertex.x = 100000.0f;
		TmpVertex.y = 100000.0f;
		TmpVertex.z = 100000.0f;
		if (TempVertex[NAddedVerts-1].x == TempVertex[0].x &&
			TempVertex[NAddedVerts-1].y == TempVertex[0].y &&
			TempVertex[NAddedVerts-1].z == TempVertex[0].z)
		{
				NAddedVerts--;

		}
				if (TempVertex[NAddedVerts-1].x == TempVertex[0].x &&
			TempVertex[NAddedVerts-1].y == TempVertex[0].y &&
			TempVertex[NAddedVerts-1].z == TempVertex[0].z)
		{
				NAddedVerts--;
		}


    for (Loop=0;Loop<NAddedVerts;Loop++)
    {
			if (TempVertex[Loop].x == TmpVertex.x && TempVertex[Loop].y ==
					TmpVertex.y && TempVertex[Loop].z == TmpVertex.z)
			{
				TmpVertex.x = TempVertex[Loop].x;
				TmpVertex.y = TempVertex[Loop].y;
				TmpVertex.z = TempVertex[Loop].z;
			}
			else
			{
				CopyLVertex(&TempVertex[Loop],&pLandscape->ClippedVertices2[ClipIndex]);
				ClipIndex++;
				TmpVertex.x = TempVertex[Loop].x;
				TmpVertex.y = TempVertex[Loop].y;
				TmpVertex.z = TempVertex[Loop].z;
			}
    }
    pLandscape->NumAddedVertices2 = ClipIndex;


}
	

void CopyVertexColor(fxeDrawVertex *v1, fxeDrawVertex *v2)
{
	v2->r = v1->r;
	v2->g = v1->g;
	v2->b = v1->b;
}

// Returns 0 if landscape not completely clipped and 1
// if completely clipped
int ClipLandscape(fxeLandscapePtr pLandscape)
{
  int Loop;
  fxeLandscapePtr pLandscapeLocal = pLandscape;
  int ClippedZ1, ClippedZ2, ClippedX1, ClippedX2, ClippedY1, ClippedY2;

  ClippedZ1 = 0;
  ClippedZ2 = 0;
  ClippedX1 = 0;
  ClippedX2 = 0;
  ClippedY1 = 0;
  ClippedY2 = 0;

  pLandscapeLocal->Clipped = FXEFALSE;
  pLandscapeLocal->PartlyClipped = FXEFALSE;
	pLandscapeLocal->NumAddedVertices1 = 0;
	pLandscapeLocal->NumAddedVertices2 = 0;

  // Transform all vertices to camera coordinates and clip those vertices:
  for (Loop=0; Loop<4; Loop++)
  {
    LMatXDrawVerToDrawVer(CameraMatrix,&(pLandscapeLocal->VerticesWorld[Loop]),
      &(pLandscapeLocal->VerticesCamera[Loop]));

    if (pLandscapeLocal->VerticesCamera[Loop].z <= NearClippingPlane)
      ClippedZ1++;
    if (pLandscapeLocal->VerticesCamera[Loop].z >= FarClippingPlane)
      ClippedZ2++;
    if (pLandscapeLocal->VerticesCamera[Loop].x <= (-(pLandscapeLocal->
      VerticesCamera[Loop].z)))
      ClippedX1++;
    if (pLandscapeLocal->VerticesCamera[Loop].x >= (pLandscapeLocal->
      VerticesCamera[Loop].z))
      ClippedX2++;
    if (pLandscapeLocal->VerticesCamera[Loop].y <=
      (-(pLandscapeLocal->VerticesCamera[Loop].z)))
      ClippedY1++;
    if (pLandscapeLocal->VerticesCamera[Loop].y >= 
			(pLandscapeLocal->VerticesCamera[Loop].z))
      ClippedY2++;
  }


  if (pLandscape->Landscapes != NULL)
	{
		if (ClippedZ1 == 4 || ClippedZ2 == 4 || ClippedX1 == 4 || 
			ClippedX2 == 4)
		{
			pLandscapeLocal->Clipped = FXETRUE;
			return 1;
		}
	}
  else
	{
		if (ClippedZ1 == 4 || ClippedZ2 == 4 || ClippedX1 == 4 || 
			ClippedX2 == 4 || ClippedY1 == 4 || ClippedY2 == 4)
		{
			pLandscapeLocal->Clipped = FXETRUE;
			return 1;
		}
	}
	
	// Only perform partial clipping if it is needed,
	// it is not needed if landscape has child landscapes.
	
	if (pLandscape->Landscapes != NULL)
		return 0;

  if (ClippedZ1 > 0 || ClippedZ2 > 0 || ClippedX1 > 0 ||
	  ClippedX2 > 0 || ClippedY1 > 0 || ClippedY2 > 0)
  {

/*
		CopyVertexColor(&pLandscapeLocal->VerticesWorld[0],
			&pLandscapeLocal->VerticesCamera[0]);
		CopyVertexColor(&pLandscapeLocal->VerticesWorld[1],
			&pLandscapeLocal->VerticesCamera[1]);
		CopyVertexColor(&pLandscapeLocal->VerticesWorld[2],
			&pLandscapeLocal->VerticesCamera[2]);
		CopyVertexColor(&pLandscapeLocal->VerticesWorld[3],
			&pLandscapeLocal->VerticesCamera[3]);
*/

		// Copy camera vertices to ClippedVertices1 and ClippedVertices2
		CopyLVertex(&pLandscapeLocal->VerticesCamera[0],
			&pLandscapeLocal->ClippedVertices1[0]);
		CopyLVertex(&pLandscapeLocal->VerticesCamera[1],
			&pLandscapeLocal->ClippedVertices1[1]);
		CopyLVertex(&pLandscapeLocal->VerticesCamera[2],
			&pLandscapeLocal->ClippedVertices1[2]);
		pLandscapeLocal->NumAddedVertices1 = 3;

		CopyLVertex(&pLandscapeLocal->VerticesCamera[2],
			&pLandscapeLocal->ClippedVertices2[0]);
		CopyLVertex(&pLandscapeLocal->VerticesCamera[3],
			&pLandscapeLocal->ClippedVertices2[1]);
		CopyLVertex(&pLandscapeLocal->VerticesCamera[0],
			&pLandscapeLocal->ClippedVertices2[2]);
		pLandscapeLocal->NumAddedVertices2 = 3;

		pLandscapeLocal->PartlyClipped = FXETRUE;

		if (ClippedZ1 > 0)
			ClipLandscapeToPlane(&NearPlane,pLandscapeLocal);
		if (ClippedZ2 > 0)
			ClipLandscapeToPlane(&FarPlane,pLandscapeLocal);
		if (ClippedX1 > 0)
			ClipLandscapeToPlane(&LeftPlane,pLandscapeLocal);
		if (ClippedX2 > 0)
			ClipLandscapeToPlane(&RightPlane,pLandscapeLocal);

		if (ClippedY1 > 0)
			ClipLandscapeToPlane(&DownPlane,pLandscapeLocal);
		if (ClippedY2 > 0)
			ClipLandscapeToPlane(&UpPlane,pLandscapeLocal);
  }

	return 0;
}

void AddLandscapeToGlobalRenderList(fxeLandscapePtr pLandscape)
{
	fxeVector LineOfSight;
	// Only add landscape if it is facing forward to the camera,
	// remove all backfaces:
		
	LineOfSight.x = ActiveCamera->x-pLandscape->VerticesWorld[1].x;
  LineOfSight.y = ActiveCamera->y-pLandscape->VerticesWorld[1].y;
  LineOfSight.z = ActiveCamera->z-pLandscape->VerticesWorld[1].z;

  // Make Dot product with the normal of the polygon
  // and the light of sight vector, if the sign
  // of the result is positive then the polygon
  // is visible, if not it's invisible:

	// This increased performance even with 50% on my voodoo2

  if (DotProduct3d(&LineOfSight,&pLandscape->Normal1) >= 0.0)
  {
		CurrLandscapeParent->ppLandscape
			[CurrLandscapeParent->NumLandscapesInList] = pLandscape;
		CurrLandscapeParent->NumLandscapesInList++;
	}
	else if (DotProduct3d(&LineOfSight,&pLandscape->Normal2) >= 0.0)
	{
		CurrLandscapeParent->ppLandscape
			[CurrLandscapeParent->NumLandscapesInList] = pLandscape;
		CurrLandscapeParent->NumLandscapesInList++;
	}

}

int fxeHaveBuiltLightmaps(void)
{
	return BL;
}

// This function builds the lightmaps for all landscapes rendered
// Quite heavy calculations...
void BuildLightmaps(void)
{
	int Loop,Loop2,Loop3,Loop4;
	fxeLandscape *pCurrLandscape;
	fxeDynLightsource *pDynLightsource;
	fxeVector LightLandscapeVector;
	float LightVecLen;
	fxeLightmap *pCurrLightmap;
	fxeVector LightmapVector;
	float LightIntensity;
	int LoopVal;
#ifdef USE_INTENSITY_MAPS
	int IIntensity;
#else
	int IRIntensity, IGIntensity, IBIntensity;
	float RIntensity, GIntensity, BIntensity;
#endif

	char Changed;
	fxeDynLightsource *DynLightsources[FXE_MAX_DYN_LIGHTSOURCES];
	int CurrentLightsource;
	int FoundLightsource;
	int SomeLightsourceMoved;

	if (CurrLandscapeParent->NumLandscapesInList == 0)
		return;

	BL = FXEFALSE;

	CurrentLightsource = 0;
	Changed = FXEFALSE;

	// Go through all landscapes in list
	for (Loop=0;Loop<CurrLandscapeParent->NumLandscapesInList;Loop++)
	{
		pCurrLandscape = CurrLandscapeParent->ppLandscape[Loop];
		pCurrLightmap = &pCurrLandscape->Lightmap;


		
		// Go through all dynamic lightsources:
		if (FirstDynLightsource != NULL)
		{
			pCurrLightmap->AffectedByDynLight = FXEFALSE;

			pDynLightsource = FirstDynLightsource;
			FoundLightsource = FXEFALSE;

			Changed = 0;
			CurrentLightsource = 0;
			SomeLightsourceMoved = 0;

			while (pDynLightsource != NULL)
			{
				if (pDynLightsource->Active == FXETRUE)
				{
					// Check if the current landscape is affected
					// by this lightsource:

						LightLandscapeVector.x = pCurrLandscape->VertexWorld.x -
							pDynLightsource->x;
						LightLandscapeVector.y = pCurrLandscape->VertexWorld.y -
							pDynLightsource->y;
						LightLandscapeVector.z = pCurrLandscape->VertexWorld.z -
							pDynLightsource->z;

						LightVecLen = VectorLenght(&LightLandscapeVector);

					// We only need to change lightsource

					if ((LightVecLen-(pCurrLandscape->Radius)-pDynLightsource->Radius)
						<= 0) 
					{
						Changed = Changed | pDynLightsource->Changed;
						if (pCurrLandscape->Displayed == FXEFALSE)
							Changed = 1;
						FoundLightsource = FXETRUE;
						SomeLightsourceMoved = SomeLightsourceMoved | 
							pDynLightsource->HaveMoved;
						DynLightsources[CurrentLightsource] = pDynLightsource;
						CurrentLightsource++;
					}
				}
				pDynLightsource = pDynLightsource->Next;
			}

			if (FoundLightsource == FXETRUE)
			{


			if (Changed
				|| pCurrLightmap->SavedNumLightsources !=
				CurrentLightsource
				)
			{
				for (Loop4=0;Loop4<CurrentLightsource;Loop4++)
				{
					pCurrLightmap->SavedNumLightsources = CurrentLightsource;
					pDynLightsource = DynLightsources[Loop4];

					// The landscape is affected by this lightsource so we need to update
					// the lightmap of the current landscape:

					// First calculate the length between the vectors of all 
					// points in the lightmap and the lightsource.
						
					for (Loop2=0;Loop2<LIGHTMAP_WIDTH;Loop2++)
					{
						for (Loop3=0;Loop3<LIGHTMAP_WIDTH;Loop3++)
						{
								LoopVal = Loop2*LIGHTMAP_WIDTH+Loop3;
								BL = FXETRUE;

									LightmapVector.x = pDynLightsource->x-
										(LocalLightmapCoors[LoopVal].x+pCurrLandscape->
										VerticesWorld[0].x);

									LightmapVector.y = pDynLightsource->y-
										pCurrLightmap->Heigths[LoopVal];

									LightmapVector.z = pDynLightsource->z-
										(LocalLightmapCoors[LoopVal].z+pCurrLandscape->
										VerticesWorld[0].z);

									LightVecLen = VectorLenght(&LightmapVector);


								LightVecLen = pDynLightsource->Radius-LightVecLen;
								if (LightVecLen < 0.0f)
									LightVecLen = 0.0f;
								LightIntensity = LightVecLen*pDynLightsource->Oor;

									// Now we got the intensity of the light from
									// the lightsource. Lets calculate the rgb
									// components.
								if (LightIntensity != 0.0f)
								{
									#ifdef USE_INTENSITY_MAPS
										IIntensity = ((int)(LightIntensity*255.0f));
											
									#else
									RIntensity = LightIntensity*pDynLightsource->r;
									GIntensity = LightIntensity*pDynLightsource->g;
									BIntensity = LightIntensity*pDynLightsource->b;
									IRIntensity = ((int)RIntensity);
									IGIntensity = ((int)GIntensity);
									IBIntensity = ((int)BIntensity);
									#endif
								}

								// Now we have the totalintensity value we can write
								// to the lightmap buffer

								if (pCurrLightmap->AffectedByDynLight == FXEFALSE)
								{
									if (LightIntensity == 0.0f)
									{
										pCurrLightmap->DownloadData[LoopVal] = pCurrLightmap->StaticData[LoopVal];
									}
									else
									{
										#ifdef USE_INTENSITY_MAPS

										IIntensity += pCurrLightmap->StaticData[LoopVal];
										if (IIntensity > 255)
											IIntensity = 255;
										pCurrLightmap->DownloadData[LoopVal] = IIntensity;

										#else
										
										IRIntensity = ((IRIntensity<<10)+
											(pCurrLightmap->StaticData[LoopVal]&0x7c00));
										if (IRIntensity > 0x7c00)
											IRIntensity = 0x7c00;
										
										IGIntensity = ((IGIntensity<<5)+
											(pCurrLightmap->StaticData[LoopVal]&0x3e0));
										if (IGIntensity > 0x3e0)
											IGIntensity = 0x3e0;

										IBIntensity = (IBIntensity+
											(pCurrLightmap->StaticData[LoopVal]&0x1f));
										if (IBIntensity > 0x1f)
											IBIntensity = 0x1f;
										
										pCurrLightmap->DownloadData[LoopVal] = IRIntensity |
											IGIntensity | IBIntensity;
										
										#endif

									}
								}
								else
								{
									if (LightIntensity != 0.0f)
									{
										#ifdef USE_INTENSITY_MAPS

										IIntensity += pCurrLightmap->DownloadData[LoopVal];
										if (IIntensity > 255)
											IIntensity = 255;
										pCurrLightmap->DownloadData[LoopVal] = IIntensity; 

										#else

										IRIntensity = ((IRIntensity<<10)+
											(pCurrLightmap->DownloadData[LoopVal]&0x7c00));
										if (IRIntensity > 0x7c00)
											IRIntensity = 0x7c00;
										
										IGIntensity = ((IGIntensity<<5)+
											(pCurrLightmap->DownloadData[LoopVal]&0x3e0));
										if (IGIntensity > 0x3e0)
											IGIntensity = 0x3e0;

										IBIntensity = (IBIntensity+
											(pCurrLightmap->DownloadData[LoopVal]&0x1f));
										if (IBIntensity > 0x1f)
											IBIntensity = 0x1f;
										
										pCurrLightmap->DownloadData[LoopVal] = IRIntensity |
											IGIntensity | IBIntensity;

										#endif

									}
								}

						}
					}

					pCurrLightmap->AffectedByDynLight = FXETRUE;
					pCurrLightmap->StaticDataCopied = FXEFALSE;
					
				}

				BL = FXETRUE;

				// Download the lightmap data to a tmu.
				// I don't know how slow this is for the moment, if this will be slow then 
				// I will arrange downloaddata on a way so I only have to call grTexDownloadMipMap once.

				pCurrLightmap->TexHandle->TexInfo.data = pCurrLightmap->DownloadData;
				pCurrLightmap->TexHandle->Changed = FXETRUE;


			}
			}
			
			else
			{
				// No lightsource affected the landscape - lets download static data if needed:

				if (pCurrLightmap->StaticDataCopied == FXEFALSE)
				{
					BL = TRUE;
					pCurrLightmap->StaticDataCopied = FXETRUE;
					pCurrLightmap->TexHandle->TexInfo.data = pCurrLightmap->StaticData;
					pCurrLightmap->TexHandle->Changed = FXETRUE;
				}


			}
			
		}
		else
		{
			// There wasn't any dynamic lightsources, just download static
			// lightmap data to a tmu:
			// Only download data if its not already downloaded. 

			if (pCurrLightmap->StaticDataCopied == FXEFALSE)
			{
				BL = TRUE;
				
				// Set flags so that texture cache routines know it has to 
				// download data again.
				pCurrLightmap->TexHandle->TexInfo.data = pCurrLightmap->StaticData;
				pCurrLightmap->TexHandle->Changed = FXETRUE;
				pCurrLightmap->StaticDataCopied = FXETRUE;
			}
			

		}

	}
}

void ProjectAndDrawLandscapeLightmap(fxeLandscapePtr pLandscape)
{
  int Loop;
  fxeLandscapePtr pLandscapeLocal = pLandscape;
  fxeLDrawVertex ProjectedVertices[50];
	fxeTextureHandle TexHandle;
	void *VertexArray[200];

	pLandscapeLocal = pLandscape;
	TexHandle = pLandscapeLocal->TH;

	pLandscape->Displayed = FXETRUE;

  if (pLandscapeLocal->PartlyClipped == FXETRUE)
  {


		if (pLandscapeLocal->NumAddedVertices2 > 2)
		{
		for (Loop=0; Loop<pLandscapeLocal->NumAddedVertices2; Loop++)
    {
 
			// As this is window coordinates we need to project and
			// snap coordinates:
			// Snapping should not be required but I do it anyway because
			// I have heard that people got snapping problems on voodoo
			// rush in glide3.

			ProjectedVertices[Loop].q = 
				1.0f/pLandscapeLocal->ClippedVertices2[Loop].z;

			ProjectedVertices[Loop].x = 
				(pLandscapeLocal->ClippedVertices2[Loop].x * 
			  ProjectedVertices[Loop].q + 1.0f) *
				ProjectWidthDiv2 + ProjectStartX;

			ProjectedVertices[Loop].y =
				(pLandscapeLocal->ClippedVertices2[Loop].y *
				ProjectedVertices[Loop].q + 1.0f) *
				ProjectHeightDiv2 + ProjectStartY;

			Snap(&ProjectedVertices[Loop].x); 
			Snap(&ProjectedVertices[Loop].y);


			ProjectedVertices[Loop].u0 = (254.0f *  
				pLandscapeLocal->ClippedVertices2[Loop].u0) *
				ProjectedVertices[Loop].q;
			ProjectedVertices[Loop].v0 = (254.0f *
				pLandscapeLocal->ClippedVertices2[Loop].v0) *
				ProjectedVertices[Loop].q;

			ProjectedVertices[Loop].u1 = ProjectedVertices[Loop].u0;
			ProjectedVertices[Loop].v1 = ProjectedVertices[Loop].v0;

 			VertexArray[Loop] = 
				(void*)&ProjectedVertices[Loop];

    }

		SetupTexture(TexHandle);

		// Here is the interresting lightmapping code:
		// Check which tmu we shall use for lightmapping,
		// if tmu0 we need to render the texture polygon first
		// and then alpha blend with the other lightmapped polygon,
		// a two pass process. If tmu1 we don't need a two pass
		// process since the first tmu can be used to blend with
		// the texture in the other tmu, a single pass process.

		if (pLandscapeLocal->Lightmap.TmuNumber == 0)
		{


			// Render the first polygon
			grDrawVertexArray(GR_TRIANGLE_FAN,pLandscapeLocal->NumAddedVertices2,VertexArray);

			Performance.NumPolys++;

			// Now enable alpha blending - make the blending so that
			// the normal texture map is multiplied by the lightmap
			grAlphaBlendFunction(GR_BLEND_DST_COLOR,GR_BLEND_ZERO,
				GR_BLEND_ZERO,GR_BLEND_ZERO);

			LightmapTexInfo.data = pLandscapeLocal->Lightmap.StaticData;

			// setup tmu 0 to use the lightmap:
			SetupTexture(pLandscapeLocal->Lightmap.TexHandle);


			// Render the 2nd polygon with the lightmap
			grDrawVertexArray(GR_TRIANGLE_FAN,pLandscapeLocal->NumAddedVertices2,VertexArray);

			// Now set back the alpha blending to normal mode for fxengine
			grAlphaBlendFunction(GR_BLEND_SRC_ALPHA,GR_BLEND_ONE_MINUS_SRC_ALPHA,
				GR_BLEND_ZERO,GR_BLEND_ZERO);

		}

		else
		{
			// We have two tmu's so lets use them!
			
			// We will not setup render modes for tmus here, it's done only
			// once per frame.

			// Setup the texture source for tmu1:
			// setup tmu 0 to use the lightmap:

			LightmapTexInfo.data = pLandscapeLocal->Lightmap.StaticData;

      SetupTexture(pLandscapeLocal->Lightmap.TexHandle);



			// Draw the textured polygon lit by the lightmap in
			// tmu1

			// Render the one and only polygon
			// As we can do this in a single pass on voodoo2 it is 
			// very fast.
			grDrawVertexArray(GR_TRIANGLE_FAN,pLandscapeLocal->NumAddedVertices2,VertexArray);
		}
		}

		
		



  
		if (pLandscapeLocal->NumAddedVertices1 > 2)
		{
		for (Loop=0; Loop<pLandscapeLocal->NumAddedVertices1; Loop++)
    {
 
			// As this is window coordinates we need to project and
			// snap coordinates:
			// Snapping should not be required but I do it anyway because
			// I have heard that people got snapping problems on voodoo
			// rush in glide3.

			ProjectedVertices[Loop].q = 
				1.0f/pLandscapeLocal->ClippedVertices1[Loop].z;

			ProjectedVertices[Loop].x = 
				(pLandscapeLocal->ClippedVertices1[Loop].x * 
			  ProjectedVertices[Loop].q + 1.0f) *
				ProjectWidthDiv2 + ProjectStartX;

			ProjectedVertices[Loop].y =
				(pLandscapeLocal->ClippedVertices1[Loop].y *
				ProjectedVertices[Loop].q + 1.0f) *
				ProjectHeightDiv2 + ProjectStartY;

			Snap(&ProjectedVertices[Loop].x); 
			Snap(&ProjectedVertices[Loop].y);


			ProjectedVertices[Loop].u0 = (254.0f *  
				pLandscapeLocal->ClippedVertices1[Loop].u0) *
				ProjectedVertices[Loop].q;
			ProjectedVertices[Loop].v0 = (254.0f *
				pLandscapeLocal->ClippedVertices1[Loop].v0) *
				ProjectedVertices[Loop].q;

			ProjectedVertices[Loop].u1 = ProjectedVertices[Loop].u0;
			ProjectedVertices[Loop].v1 = ProjectedVertices[Loop].v0;

 			VertexArray[Loop] = 
				(void*)&ProjectedVertices[Loop];

    }


		SetupTexture(TexHandle);


		// Here is the interresting lightmapping code:
		// Check which tmu we shall use for lightmapping,
		// if tmu0 we need to render the texture polygon first
		// and then alpha blend with the other lightmapped polygon,
		// a two pass process. If tmu1 we don't need a two pass
		// process since the first tmu can be used to blend with
		// the texture in the other tmu, a single pass process.

		if (pLandscapeLocal->Lightmap.TmuNumber == 0)
		{
			// Render the first polygon
			grDrawVertexArray(GR_TRIANGLE_FAN,pLandscapeLocal->NumAddedVertices1,VertexArray);

			// Now enable alpha blending - make the blending so that
			// the normal texture map is multiplied by the lightmap
			grAlphaBlendFunction(GR_BLEND_DST_COLOR,GR_BLEND_ZERO,
				GR_BLEND_ZERO,GR_BLEND_ZERO);

			LightmapTexInfo.data = pLandscapeLocal->Lightmap.StaticData;

			// setup tmu 0 to use the lightmap:
			SetupTexture(pLandscapeLocal->Lightmap.TexHandle);

			// Render the 2nd polygon with the lightmap
			grDrawVertexArray(GR_TRIANGLE_FAN,pLandscapeLocal->NumAddedVertices1,VertexArray);

			// Now set back the alpha blending to normal mode for fxengine
			grAlphaBlendFunction(GR_BLEND_SRC_ALPHA,GR_BLEND_ONE_MINUS_SRC_ALPHA,
				GR_BLEND_ZERO,GR_BLEND_ZERO);

		}
	

		else
		{
			// We have two tmu's so lets use them!
			
			// We will not setup render modes for tmus here, it's done only
			// once per frame.

			// Setup the texture source for tmu1:
			// setup tmu 0 to use the lightmap:

			LightmapTexInfo.data = pLandscapeLocal->Lightmap.StaticData;
			
			SetupTexture(pLandscapeLocal->Lightmap.TexHandle);

			// Draw the textured polygon lit by the lightmap in
			// tmu1

			// Render the one and only polygon
			// As we can do this in a single pass on voodoo2 it is 
			// very fast.
			grDrawVertexArray(GR_TRIANGLE_FAN,pLandscapeLocal->NumAddedVertices1,VertexArray);
		}
		}
	}
	
	


  if (pLandscapeLocal->PartlyClipped == FXEFALSE)
  {
    for (Loop=0; Loop<4; Loop++)
    {
      // Transform camera vertices to projected vertices:


			// As this is window coordinates we need to project and
			// snap coordinates:
			// Snapping should not be required but I do it anyway because
			// I have heard that people got snapping problems on voodoo
			// rush in glide3.

			ProjectedVertices[Loop].q = 
				1.0f/pLandscapeLocal->VerticesCamera[Loop].z;

			ProjectedVertices[Loop].x = 
				(pLandscapeLocal->VerticesCamera[Loop].x * 
			  ProjectedVertices[Loop].q + 1.0f) *
				ProjectWidthDiv2 + ProjectStartX;

			ProjectedVertices[Loop].y = 
				(pLandscapeLocal->VerticesCamera[Loop].y *
				ProjectedVertices[Loop].q + 1.0f) *
				ProjectHeightDiv2 + ProjectStartY;

			Snap(&ProjectedVertices[Loop].x);
			Snap(&ProjectedVertices[Loop].y);


			ProjectedVertices[Loop].u0 = (254.0f *  
				pLandscapeLocal->VerticesCamera[Loop].u0) *
				ProjectedVertices[Loop].q;
			ProjectedVertices[Loop].v0 = (254.0f *
				pLandscapeLocal->VerticesCamera[Loop].v0) *
				ProjectedVertices[Loop].q;

			ProjectedVertices[Loop].u1 = ProjectedVertices[Loop].u0;
			ProjectedVertices[Loop].v1 = ProjectedVertices[Loop].v0;

 			VertexArray[Loop] = 
				(void*)&ProjectedVertices[Loop];
      

    }


    TexHandle = pLandscapeLocal->TH;

		SetupTexture(TexHandle);

		// Here is the interresting lightmapping code:
		// Check which tmu we shall use for lightmapping,
		// if tmu0 we need to render the texture polygon first
		// and then alpha blend with the other lightmapped polygon,
		// a two pass process. If tmu1 we don't need a two pass
		// process since the first tmu can be used to blend with
		// the texture in the other tmu, a single pass process.

		if (pLandscapeLocal->Lightmap.TmuNumber == 0)
		{

			// Draw triangles using triangle fan:
		
			grDrawVertexArray(GR_TRIANGLE_FAN,4,VertexArray);

			Performance.NumPolys++;

			// Now enable alpha blending - make the blending so that
			// the normal texture map is multiplied by the lightmap
			grAlphaBlendFunction(GR_BLEND_DST_COLOR,GR_BLEND_ZERO,
				GR_BLEND_ZERO,GR_BLEND_ZERO);

			LightmapTexInfo.data = pLandscapeLocal->Lightmap.StaticData;

			SetupTexture(pLandscapeLocal->Lightmap.TexHandle);

			// Draw triangles using triangle fan:
		
			grDrawVertexArray(GR_TRIANGLE_FAN,4,VertexArray);



			// Now set back the alpha blending to normal mode for fxengine
			grAlphaBlendFunction(GR_BLEND_SRC_ALPHA,GR_BLEND_ONE_MINUS_SRC_ALPHA,
				GR_BLEND_ZERO,GR_BLEND_ZERO);

		}

		else
		{
			// We have two tmu's so lets use them!
			
			// We will not setup render modes for tmus here, it's done only
			// once per frame.

			// Setup the texture source for tmu1:
			// setup tmu 0 to use the lightmap:

			LightmapTexInfo.data = pLandscapeLocal->Lightmap.StaticData;

			SetupTexture(pLandscapeLocal->Lightmap.TexHandle);

			// Draw the textured polygon lit by the lightmap in
			// tmu1

			// Render the one and only polygon
			// As we can do this in a single pass on voodoo2 it is 
			// very fast.
			// Draw triangles using triangle fan:
		
			grDrawVertexArray(GR_TRIANGLE_FAN,4,VertexArray);


			Performance.NumPolys++;
		}
  }
}


void ProjectAndDrawLandscapesUseLightmapping(void)
{

	int Loop;

	if (CurrLandscapeParent->NumLandscapesInList == 0)
		return;

	for (Loop=0;Loop<CurrLandscapeParent->NumLandscapesInList;Loop++)
	{
		ProjectAndDrawLandscapeLightmap(CurrLandscapeParent->ppLandscape[Loop]);	
	}
}

/*
void ProjectAndDrawLandscape(fxeLandscapePtr pLandscape)
{
  int Loop;
  fxeLandscapePtr pLandscapeLocal = pLandscape;
  fxeDrawVertex ProjectedVertices[50];
	fxeTextureHandle TexHandle;
	void *VertexArray[200];
	int Loop3;

  
  if (pLandscapeLocal->Clipped == FXEFALSE)
  {	  
  if (pLandscapeLocal->PartlyClipped == FXETRUE)
  {
		for (Loop=0; Loop<pLandscapeLocal->NumAddedVertices2; Loop++)
    {
 
      // Transfer vertex colors:


      pLandscapeLocal->ClippedVertices2[Loop].r += AmbientR;
      pLandscapeLocal->ClippedVertices2[Loop].g += AmbientG;
      pLandscapeLocal->ClippedVertices2[Loop].b += AmbientB;

      //pLandscapeLocal->ClippedVertices2[Loop].q =
      //  pLandscapeLocal->ClippedVertices2[Loop].z;

      if (pLandscapeLocal->ClippedVertices2[Loop].r > 1.0)
        pLandscapeLocal->ClippedVertices2[Loop].r = 1.0;
      if (pLandscapeLocal->ClippedVertices2[Loop].g > 1.0)
        pLandscapeLocal->ClippedVertices2[Loop].g = 1.0;
      if (pLandscapeLocal->ClippedVertices2[Loop].b > 1.0)
        pLandscapeLocal->ClippedVertices2[Loop].b = 1.0;


			// As this is window coordinates we need to project and
			// snap coordinates:
			// Snapping should not be required but I do it anyway because
			// I have heard that people got snapping problems on voodoo
			// rush in glide3.

			ProjectedVertices[Loop].q = 
				1.0f/pLandscapeLocal->ClippedVertices2[Loop].z;

			ProjectedVertices[Loop].x = 
				(pLandscapeLocal->ClippedVertices2[Loop].x * 
			  ProjectedVertices[Loop].q + 1.0f) *
				ProjectWidthDiv2 + ProjectStartX;

			ProjectedVertices[Loop].y =
				(pLandscapeLocal->ClippedVertices2[Loop].y *
				ProjectedVertices[Loop].q + 1.0f) *
				ProjectHeightDiv2 + ProjectStartY;

			Snap(&ProjectedVertices[Loop].x); 
			Snap(&ProjectedVertices[Loop].y);

			ProjectedVertices[Loop].r = 
				pLandscapeLocal->ClippedVertices2[Loop].r * 255.0f;
			ProjectedVertices[Loop].g = 
				pLandscapeLocal->ClippedVertices2[Loop].g * 255.0f;
			ProjectedVertices[Loop].b = 
				pLandscapeLocal->ClippedVertices2[Loop].b * 255.0f;

			ProjectedVertices[Loop].u0 = (255.0f *  
				pLandscapeLocal->ClippedVertices2[Loop].u0) *
				ProjectedVertices[Loop].q;
			ProjectedVertices[Loop].v0 = (255.0f *
				pLandscapeLocal->ClippedVertices2[Loop].v0) *
				ProjectedVertices[Loop].q;

			ProjectedVertices[Loop].u1 = ProjectedVertices[Loop].u0;
			ProjectedVertices[Loop].v1 = ProjectedVertices[Loop].v0;

 			VertexArray[Loop] = 
				(void*)&ProjectedVertices[Loop];

    }

    if (pLandscapeLocal->TextureMode != NOT_TEXTURED)
    {

      TexHandle = pLandscapeLocal->TH;

      if (TexHandle != CurrentTextureSource)
      {

        if (TexHandle->FilteringMethod == BILINEAR_FILTERED)
        {
          // Setup texture unit 0's texture source
          grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
            GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
          CurrentTextureSource = TexHandle;
        }

        else if (TexHandle->FilteringMethod == TRILINEAR_FILTERED)
        {

          // Setup texture unit 0's texture source
          grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
            GR_MIPMAPLEVELMASK_EVEN,&TexHandle->TexInfo);

          // Setup texture unit 1's texture source
          grTexSource(GR_TMU1,TexHandle->VideoMemoryAddressTMU1,
            GR_MIPMAPLEVELMASK_ODD,&TexHandle->TexInfo);
          CurrentTextureSource = TexHandle;
        }
      }
    }

    grDrawVertexArray(GR_TRIANGLE_FAN,3,VertexArray);

		if (pLandscapeLocal->NumAddedVertices2 > 3)
		{
			for (Loop3=0;Loop3<pLandscapeLocal->NumAddedVertices2-3;Loop3++)
				grDrawVertexArray(GR_TRIANGLE_FAN_CONTINUE,1,&VertexArray[Loop3+3]);
		}
    Performance.NumPolys++;
  

		for (Loop=0; Loop<pLandscapeLocal->NumAddedVertices1; Loop++)
    {
 
      // Transfer vertex colors:

      pLandscapeLocal->ClippedVertices1[Loop].r += AmbientR;
      pLandscapeLocal->ClippedVertices1[Loop].g += AmbientG;
      pLandscapeLocal->ClippedVertices1[Loop].b += AmbientB;

      //pLandscapeLocal->ClippedVertices2[Loop].q =
      //  pLandscapeLocal->ClippedVertices2[Loop].z;

      if (pLandscapeLocal->ClippedVertices1[Loop].r > 1.0)
        pLandscapeLocal->ClippedVertices1[Loop].r = 1.0;
      if (pLandscapeLocal->ClippedVertices1[Loop].g > 1.0)
        pLandscapeLocal->ClippedVertices1[Loop].g = 1.0;
      if (pLandscapeLocal->ClippedVertices1[Loop].b > 1.0)
        pLandscapeLocal->ClippedVertices1[Loop].b = 1.0;

      
			// As this is window coordinates we need to project and
			// snap coordinates:
			// Snapping should not be required but I do it anyway because
			// I have heard that people got snapping problems on voodoo
			// rush in glide3.

			ProjectedVertices[Loop].q = 
				1.0f/pLandscapeLocal->ClippedVertices1[Loop].z;

			ProjectedVertices[Loop].x = 
				(pLandscapeLocal->ClippedVertices1[Loop].x * 
			  ProjectedVertices[Loop].q + 1.0f) *
				ProjectWidthDiv2 + ProjectStartX;

			ProjectedVertices[Loop].y = 
				(pLandscapeLocal->ClippedVertices1[Loop].y *
				ProjectedVertices[Loop].q + 1.0f) *
				ProjectHeightDiv2 + ProjectStartY;

			Snap(&ProjectedVertices[Loop].x);
			Snap(&ProjectedVertices[Loop].y);

			ProjectedVertices[Loop].r = 
				pLandscapeLocal->ClippedVertices1[Loop].r * 255.0f;
			ProjectedVertices[Loop].g = 
				pLandscapeLocal->ClippedVertices1[Loop].g * 255.0f;
			ProjectedVertices[Loop].b = 
				pLandscapeLocal->ClippedVertices1[Loop].b * 255.0f;

			ProjectedVertices[Loop].u0 = (255.0f *  
				pLandscapeLocal->ClippedVertices1[Loop].u0) *
				ProjectedVertices[Loop].q;
			ProjectedVertices[Loop].v0 = (255.0f *
				pLandscapeLocal->ClippedVertices1[Loop].v0) *
				ProjectedVertices[Loop].q;

			ProjectedVertices[Loop].u1 = ProjectedVertices[Loop].u0;
			ProjectedVertices[Loop].v1 = ProjectedVertices[Loop].v0;

 			VertexArray[Loop] = 
				(void*)&ProjectedVertices[Loop];

    }

    if (pLandscapeLocal->TextureMode != NOT_TEXTURED)
    {

      TexHandle = pLandscapeLocal->TH;

      if (TexHandle != CurrentTextureSource)
      {

        if (TexHandle->FilteringMethod == BILINEAR_FILTERED)
        {
          // Setup texture unit 0's texture source
          grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
            GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
          CurrentTextureSource = TexHandle;
        }

        else if (TexHandle->FilteringMethod == TRILINEAR_FILTERED)
        {

          // Setup texture unit 0's texture source
          grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
            GR_MIPMAPLEVELMASK_EVEN,&TexHandle->TexInfo);

          // Setup texture unit 1's texture source
          grTexSource(GR_TMU1,TexHandle->VideoMemoryAddressTMU1,
            GR_MIPMAPLEVELMASK_ODD,&TexHandle->TexInfo);
          CurrentTextureSource = TexHandle;
        }
      }
    }

    grDrawVertexArray(GR_TRIANGLE_FAN,3,VertexArray);

		if (pLandscapeLocal->NumAddedVertices1 > 3)
		{
			for (Loop3=0;Loop3<pLandscapeLocal->NumAddedVertices1-3;Loop3++)
				grDrawVertexArray(GR_TRIANGLE_FAN_CONTINUE,1,&VertexArray[Loop3+3]);
		}
    Performance.NumPolys++;
	}


  if (pLandscapeLocal->PartlyClipped == FXEFALSE)
  {
    for (Loop=0; Loop<4; Loop++)
    {
      // Transform camera vertices to projected vertices:


      // Transfer vertex colors and add ambient light to the values,
      // vertex color values is clipped to 255.0

      pLandscapeLocal->VerticesCamera[Loop].r = 
				pLandscapeLocal->VerticesWorld[Loop].r + AmbientR;
      pLandscapeLocal->VerticesCamera[Loop].g = 
				pLandscapeLocal->VerticesWorld[Loop].g + AmbientG;
      pLandscapeLocal->VerticesCamera[Loop].b = 
				pLandscapeLocal->VerticesWorld[Loop].b + AmbientB;

      if (pLandscapeLocal->VerticesCamera[Loop].r > 1.0)
        pLandscapeLocal->VerticesCamera[Loop].r = 1.0;
      if (pLandscapeLocal->VerticesCamera[Loop].g > 1.0)
        pLandscapeLocal->VerticesCamera[Loop].g = 1.0;
      if (pLandscapeLocal->VerticesCamera[Loop].b > 1.0)
        pLandscapeLocal->VerticesCamera[Loop].b = 1.0;

			// As this is window coordinates we need to project and
			// snap coordinates:
			// Snapping should not be required but I do it anyway because
			// I have heard that people got snapping problems on voodoo
			// rush in glide3.

			ProjectedVertices[Loop].q = 
				1.0f/pLandscapeLocal->VerticesCamera[Loop].z;

			ProjectedVertices[Loop].x = 
				(pLandscapeLocal->VerticesCamera[Loop].x * 
			  ProjectedVertices[Loop].q + 1.0f) *
				ProjectWidthDiv2 + ProjectStartX;

			ProjectedVertices[Loop].y = 
				(pLandscapeLocal->VerticesCamera[Loop].y *
				ProjectedVertices[Loop].q + 1.0f) *
				ProjectHeightDiv2 + ProjectStartY;

			Snap(&ProjectedVertices[Loop].x);
			Snap(&ProjectedVertices[Loop].y);

			ProjectedVertices[Loop].r = 
				pLandscapeLocal->VerticesCamera[Loop].r * 255.0f;
			ProjectedVertices[Loop].g = 
				pLandscapeLocal->VerticesCamera[Loop].g * 255.0f;
			ProjectedVertices[Loop].b = 
				pLandscapeLocal->VerticesCamera[Loop].b * 255.0f;

			ProjectedVertices[Loop].u0 = (255.0f *  
				pLandscapeLocal->VerticesCamera[Loop].u0) *
				ProjectedVertices[Loop].q;
			ProjectedVertices[Loop].v0 = (255.0f *
				pLandscapeLocal->VerticesCamera[Loop].v0) *
				ProjectedVertices[Loop].q;

			ProjectedVertices[Loop].u1 = ProjectedVertices[Loop].u0;
			ProjectedVertices[Loop].v1 = ProjectedVertices[Loop].v0;

 			VertexArray[Loop] = 
				(void*)&ProjectedVertices[Loop];
      

    }


    if (pLandscapeLocal->TextureMode != NOT_TEXTURED)
    {

      TexHandle = pLandscapeLocal->TH;

      if (TexHandle != CurrentTextureSource)
      {

        if (TexHandle->FilteringMethod == BILINEAR_FILTERED)
        {
          // Setup texture unit 0's texture source
          grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
            GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
          CurrentTextureSource = TexHandle;
        }

        else if (TexHandle->FilteringMethod == TRILINEAR_FILTERED)
        {

          // Setup texture unit 0's texture source
          grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
            GR_MIPMAPLEVELMASK_EVEN,&TexHandle->TexInfo);

          // Setup texture unit 1's texture source
          grTexSource(GR_TMU1,TexHandle->VideoMemoryAddressTMU1,
            GR_MIPMAPLEVELMASK_ODD,&TexHandle->TexInfo);
          CurrentTextureSource = TexHandle;
        }
      }
    }

    // Draw triangles using triangle fan:
		
    grDrawVertexArray(GR_TRIANGLE_FAN,3,VertexArray);

		grDrawVertexArray(GR_TRIANGLE_FAN_CONTINUE,1,&VertexArray[3]);

    Performance.NumPolys++;

  }
  }
}
*/

void RenderLandscape(fxeLandscapePtr pLandscape)
{
  int Loop,Loop2;

  // Transform world pos of landscape to camera coordinates
  /*
	VertexCamera = &(pLandscape->VertexCamera);
  Radius = pLandscape->Radius;


  MatXVer(CameraMatrix,&(pLandscape->VertexWorld),VertexCamera);

  if (VertexCamera->z + Radius < NearClippingPlane ||
      VertexCamera->z - Radius > FarClippingPlane)
    return;
  else if (VertexCamera->x + Radius < (-VertexCamera->z) ||
           VertexCamera->x - Radius > VertexCamera->z)
    return;
  else if (VertexCamera->y + Radius < (-VertexCamera->z) ||
           VertexCamera->y - Radius > (VertexCamera->z))
    return;
	*/

  if ((ClipLandscape(pLandscape)) == 1)
		return;

  if (pLandscape->Landscapes != NULL)
  {
    for (Loop=0 ; Loop<pLandscape->NumLandscapesX; Loop++)
    {
      for (Loop2=0 ; Loop2<pLandscape->NumLandscapesX; Loop2++)
      {
        RenderLandscape((fxeLandscapePtr)
          (pLandscape->Landscapes
          [(Loop*(pLandscape->NumLandscapesX)+Loop2)]));
      }
    }
  }
  else
  {

		// Add to global render list
		AddLandscapeToGlobalRenderList(pLandscape);

    // Draw Polygon directly since this Landscape doesn't have child
    // landscapes:
    //ProjectAndDrawLandscape(pLandscape);
  }
}


void fxeLandscapeRender(fxeLandscapeHandle pLandscape)
{
  fxeLandscapePtr pLandscapeLocal = pLandscape;

  
  RenderLandscape(pLandscapeLocal);
}

  
void fxeGetVersion(int *MainVersion, int *UnderVersion)
{
  *MainVersion = 0;
  *UnderVersion = 37;
}

void fxeFarClippingPlaneSet(float Distance)
{
  FarClippingPlane = Distance;
	// Setup the view frustrum planes
  SetupViewFrustrum();
}

void fxeNearClippingPlaneSet(float Distance)
{
  NearClippingPlane = Distance;
	// Setup the view frustrum planes
  SetupViewFrustrum();
}


// call this function first - initializes FxEngine
// Returns FXETRUE if successful and FXEFALSE if failure
int 
fxeInitialize(HWND hwnd, int ResToUse)
{
  int Loop;
  GrResolution Query;
	GrScreenResolution_t FxRes;
  GrResolution *ResList;
  FxI32 NumBoards;
  int ResListSize;
  FxU32 WMaxValuePtr[2];
  fxeCharacterHandle CH;
  fxeTextureHandle TH;
  fxeColor TransColor;
  unsigned long BufferSize;

	// Open fxengine logfile:

	fxeLogDeviceOpen();

	NumBytesAllocated = 0;

  /*
  if (LoadProcAddresses() == 0)
    return FXEFALSE;
  */


  grGet(GR_NUM_BOARDS,sizeof(NumBoards),&NumBoards);


  if (NumBoards == 0)
	{
		fxeLogDeviceWriteString("No glide3 compatible hardware found!\n\n");
    return FXEFALSE;
	}

	grGlideInit();

	// Setup snapping so it will work on voodoo rush
	SetupSnapping();

	fxeLogDeviceWriteString("Glide version:\n");
	fxeLogDeviceWriteString((char*)grGetString(GR_VERSION));
	fxeLogDeviceWriteString("\n\n");
	fxeLogDeviceWriteString("3dfx Hardware detected:\n");
	fxeLogDeviceWriteString((char*)grGetString(GR_HARDWARE));
	fxeLogDeviceWriteString("\n\n");

  // Select first 3dfx hardware found.
  HardwareSelected = 0;
  grSstSelect(HardwareSelected);

  // query for resolution specified:

	if (ResToUse == FXE_RES_640x480)
		Query.resolution = GR_RESOLUTION_640x480;
	else if (ResToUse == FXE_RES_800x600 )
		Query.resolution = GR_RESOLUTION_800x600;
	else if (ResToUse == FXE_RES_1024x768)
		Query.resolution = GR_RESOLUTION_1024x768;
	else if (ResToUse == FXE_RES_1280x1024)
		Query.resolution = GR_RESOLUTION_1280x1024;
	else if (ResToUse == FXE_RES_1600x1200)
		Query.resolution = GR_RESOLUTION_1600x1200;
	else
		Query.resolution = GR_RESOLUTION_640x480;

	FxRes = Query.resolution;

  Query.refresh = GR_REFRESH_60Hz;
  Query.numColorBuffers = 2;
  Query.numAuxBuffers = 1;

  ResListSize = grQueryResolutions(&Query,NULL);
  ResList = (GrResolution*) FxMalloc(ResListSize);
  if (grQueryResolutions(&Query,ResList) == 0)
    return 0;

  Res = FxRes;

  // After the resolution and detaillevel varibles have been
  // we swithes to 3dfx videomode using a refresh rate of 60HZ,
  // ARGB colorformat, double buffering and a auxiliary buffer
  // which is going to be used for z buffering
	
	GlobalHwnd = hwnd;

	
	if ((Context = grSstWinOpen((FxU32)hwnd,Res,GR_REFRESH_60Hz,GR_COLORFORMAT_ARGB,
		GR_ORIGIN_LOWER_LEFT,2,1)) == 0)
	{
		return FXEFALSE;
	}


  // Retrieve the number of TexelFx chips:

  grGet(GR_NUM_TMU,sizeof(NumTexelFx),&NumTexelFx);

	grGet(GR_TEXTURE_ALIGN,4,&TextureAlign);


  switch(Res)
  {
    case GR_RESOLUTION_640x480:
      HorRes = 640;
      VerRes = 480;
      break;

    case GR_RESOLUTION_800x600:
      HorRes = 800;
      VerRes = 600;
      break;

    case GR_RESOLUTION_1024x768:
      HorRes = 1024;
      VerRes = 768;
      break;

    case GR_RESOLUTION_1280x1024:
      HorRes = 1280;
      VerRes = 1024;
      break;

    case GR_RESOLUTION_1600x1200:
      HorRes = 1600;
      VerRes = 1200;
      break;
  }


	TextureFilteringMode = BILINEAR_FILTERED;


    // Set up texture unit 0 for bilinear filtering
    grTexFilterMode(GR_TMU0,GR_TEXTUREFILTER_BILINEAR,
      GR_TEXTUREFILTER_BILINEAR);

    // Setup TMU0 for simple texture mapping
    grTexCombine(GR_TMU0,GR_COMBINE_FUNCTION_LOCAL,
      GR_COMBINE_FACTOR_NONE,
      GR_COMBINE_FUNCTION_LOCAL,
      GR_COMBINE_FACTOR_NONE,
      FXFALSE,FXFALSE);

		grEnable(GR_ALLOW_MIPMAP_DITHER);

    // enable mipmapping:
    grTexMipMapMode(GR_TMU0,GR_MIPMAP_NEAREST_DITHER,FXFALSE);

    grTexClampMode(GR_TMU0,GR_TEXTURECLAMP_CLAMP,GR_TEXTURECLAMP_CLAMP);



  if (NumTexelFx > 1)
  {

    // Set up texture unit 1 for bilinear filtering
    grTexFilterMode(GR_TMU1,GR_TEXTUREFILTER_BILINEAR,
      GR_TEXTUREFILTER_BILINEAR);
		grTexMipMapMode(GR_TMU1,GR_MIPMAP_NEAREST_DITHER,FXFALSE);
    grTexClampMode(GR_TMU1,GR_TEXTURECLAMP_CLAMP,GR_TEXTURECLAMP_CLAMP);

  }



  grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
    GR_COMBINE_FACTOR_LOCAL,GR_COMBINE_LOCAL_ITERATED,
    GR_COMBINE_OTHER_TEXTURE,FXFALSE);
  TextureMode = TEXTURE_MAPPED;

	// Initialise Lightmapping system:

	InitLightmapSystem();
    

  // Now Inititialise some structures and variabled used by
  // FxEngine
	
	CurrentTextureSource = NULL;

  // Build sqrt table:
  build_sqrt_table();



  // retrieve the size of the fog table:

  grGet(GR_FOG_TABLE_ENTRIES,4,&nFogEntries);


  FxeFogTable = (GrFog_t*)FxMalloc(sizeof(GrFog_t)*nFogEntries);


  // Setup the fogtable:

  NearClippingPlane = 2.0;
  FarClippingPlane = 3000.0;

  guFogGenerateLinear(FxeFogTable,FarClippingPlane*0.8f,FarClippingPlane);
  grFogTable(FxeFogTable);

  // Reset camera position to 0,0,0 with angles 0,0,0
  Camera.x = 0;
  Camera.y = 0;
  Camera.z = 0;
  Camera.ax = 0;
  Camera.ay = 0;
  Camera.az = 0;
  Camera.NumFramesLeft = 0;

	Camera.FollowPoint = FXEFALSE;

  // Reset world objects:

  WorldObjects = NULL;

  // Reset Lightsources:

  Lightsources = NULL;

	fxeTextureCacheInit();

  // Reset Textures:
  Textures = NULL;


  // Calculate viewing distance such as the horisontal field of view
  // is 90 degrees independent of resolution

  ViewingDistance = (float)((HorRes/2.0)) / (float)(tan(DegToRad(45.0)));

  HFOWInc = (float)(HorRes/2.0) / ViewingDistance;
  VFOWInc = (float)(VerRes/2.0) / ViewingDistance;

  // Precalculate cos and sin tables:
  for (Loop=0; Loop<361; Loop++)
  {
    CosLockup[Loop] = (float)cos(DegToRad((float)Loop));
    SinLockup[Loop] = (float)sin(DegToRad((float)Loop));
  }

	// Initialize hardware for w buffering: (same as z buffering but
	// uses w instead of z)

  grDepthBufferMode(GR_DEPTHBUFFER_WBUFFER);
  grDepthMask(FXTRUE);
  grDepthBufferFunction(GR_CMP_LEQUAL);

	grAlphaTestFunction(GR_CMP_ALWAYS);
  grAlphaTestReferenceValue(0x1);

	TotNumLightsources = 0;

  HalfHorRes = (float)(HorRes/2.0f);
  HalfVerRes = (float)(VerRes/2.0f);


  // set Textures to NULL which indicates that no texture is loaded
  // Textures = NULL;

	TexInfo.data = (void*)FxMalloc(128);
	for (Loop=0;Loop<64;Loop++)
	{
		((unsigned short*)(TexInfo.data))[Loop] = (unsigned short)31<<5;
	}
	TexInfo.format = GR_TEXFMT_ARGB_1555;
	TexInfo.smallLodLog2 = GR_LOD_LOG2_8;
	TexInfo.largeLodLog2 = GR_LOD_LOG2_8;
	TexInfo.aspectRatioLog2 = GR_ASPECT_LOG2_1x1;

	DownloadAddress = fxeTMUAllocateMemory(GR_TMU0,128);

	grTexDownloadMipMap(GR_TMU0,DownloadAddress,GR_MIPMAPLEVELMASK_ODD,
		&TexInfo);


  // Set ambient light to 0,0,0
  AmbientR = 0.0;
  AmbientG = 0.0;
  AmbientB = 0.0;
  AmbientR2 = 0.0;
  AmbientG2 = 0.0;
  AmbientB2 = 0.0;

  grAlphaCombine(GR_COMBINE_FUNCTION_BLEND_OTHER,GR_COMBINE_FACTOR_ONE,
    GR_COMBINE_LOCAL_NONE,GR_COMBINE_OTHER_CONSTANT,FXFALSE);

  grAlphaBlendFunction(GR_BLEND_SRC_ALPHA,GR_BLEND_ONE_MINUS_SRC_ALPHA,
      GR_BLEND_ZERO,GR_BLEND_ZERO);
  AlphaBlendMode = FXEFALSE;


  BackgroundColor.r = 0;
  BackgroundColor.g = 0;
  BackgroundColor.b = 0;

  CurrLandscape = NULL;

  DefaultCharacterset = NULL;

  HorInc = (float)(HorRes/640.0f);
  VerInc = (float)(VerRes/480.0f);

  ChromaKey = FXEFALSE;
   CombineTextureOnly = FXEFALSE;


  // Set coordinate space to window coordinates
  grCoordinateSpace(GR_WINDOW_COORDS);

	// Thes code not required when using window coordinates
	/*
  grViewport(0,0,
    (FxI32)HorRes,(FxI32)VerRes);

  grDepthRange(0,1);
	*/
 
  // Setup the vertex layout:
  grVertexLayout(GR_PARAM_XY,0,GR_PARAM_ENABLE);
  grVertexLayout(GR_PARAM_W,8,GR_PARAM_ENABLE);
  //grVertexLayout(GR_PARAM_Z,12,GR_PARAM_ENABLE);
  
	// Need to specify this when using window coordinates:
	grVertexLayout(GR_PARAM_Q,16,GR_PARAM_ENABLE);
  grVertexLayout(GR_PARAM_RGB,20,GR_PARAM_ENABLE);
  grVertexLayout(GR_PARAM_ST0,32,GR_PARAM_ENABLE);
  grVertexLayout(GR_PARAM_ST1,40,GR_PARAM_ENABLE);

	// Setup projection values:
	ProjectWidthDiv2 = HalfHorRes;
	ProjectHeightDiv2 = HalfVerRes;
	ProjectStartX = 0.0f;
	ProjectStartY = 0.0f;
	ProjectEndX = (float)HorRes;
	ProjectEndY = (float)VerRes;

  grGet(GR_WDEPTH_MIN_MAX,8,(FxI32*)WMaxValuePtr);
  WMaxValue = WMaxValuePtr[1];

  // set parameters for default characterset.
  DefaultCS.Translucent = FXEFALSE;
  DefaultCS.AlphaLevel = 128;
  DefaultCS.Transparent = FXETRUE;
  DefaultCS.TransColor.r = 0;
  DefaultCS.TransColor.g = 0;
  DefaultCS.TransColor.b = 0;
  DefaultCS.Spacing = 20;
  DefaultCS.CharWidth = 20;
  DefaultCS.CharHeight = 20;

  SwapFlag = 1;

  TextureMode = TEXTURE_MAPPED;

  DrawShape = GR_TRIANGLE_FAN;

  // Setup the view frustrum planes
  SetupViewFrustrum();


  // Load the white texture - use as "NULL" texture

  NULLTexture = fxeTextureLoad("white.3df",FXE_TEXTURE_MAP);
  if (NULLTexture == NULL)
    return FXEFALSE;


  grGet(GR_GLIDE_STATE_SIZE,4,(void*)&BufferSize);

  GlideState = (unsigned char*)
	  FxMalloc(sizeof(unsigned char)*BufferSize);

  LostFocus = FXEFALSE;

  GlobalFogMode = NO_FOGGING;

  ActiveCamera = &Camera;

	fxeParticleSystemInit();

	fxeSmokeSystemInit();

	// View the initialization picture
	fxeSetRenderState(FXE_RENDER_2D);

  TH = fxeTextureLoad("Init.3df",FXE_TEXTURE_MAP);
  CH = fxeBitmapCreate(TH,FXEFALSE,0,FXEFALSE,&TransColor);

  fxeBitmapDrawXY(CH,0,0,640,480);
  grBufferSwap(0);

  fxeSetRenderState(FXE_RENDER_3D);

	DefaultStrTexture = fxeTextureLoad("DefaultCS.3df",FXE_TEXTURE_MAP);

	fxeStrDefaultProperties();

	GlobalSkyboxHandle = NULL;

  // Return success:
  return FXETRUE;
}

void fxeLooseFocus(void)
{
  if (LostFocus == FXEFALSE)
  {
    // Save glide state and release window
    grGlideGetState((void*)&GlideState);
    //grSstWinClose(Context);
    LostFocus = FXETRUE;
		grSstWinClose(Context);
  }
}

void fxeSetFocus(void)
{
	fxeTexture *pTexture;

  if (LostFocus == FXETRUE)
  {
    // Open 3dfx and restore glide state.
		grGlideInit();
    Context =
			grSstWinOpen((FxU32)GlobalHwnd,Res,GR_REFRESH_60Hz,
			GR_COLORFORMAT_ARGB,GR_ORIGIN_LOWER_LEFT,2,1);

		// We must restore glide state, this is important because
		// the 3dfx hardware is shared among other programs
		// this is using it and the glide state can have changed.

    grGlideSetState((void*)&GlideState);

		// Set all textures to changed so the texture cache routines
		// must download all textures. This is required
		// because the 3dfx hardware can be shared among
		// other programs which is using texture memory.
		// Therefore we need to do this. 


		if (pFirstTexture != NULL)
		{
			pTexture = pFirstTexture;

			while (pTexture != NULL)
			{
				pTexture->Changed = FXETRUE;
				pTexture = pTexture->Next;
			}
		}


		SetupSnapping();
		

    LostFocus = FXEFALSE;
  }
}

// This function should be called before your application terminates
// For the moment it just shut down glide but in the future new things
// could be destroyed here.
void fxeShutdown(void)
{

  // We shall free all objects, lightsources, textures, landscapes
  // and the like here too.

  // Shut down glide:
  grSstWinClose(Context);
  grGlideShutdown();
	fxeLogDeviceClose();
}	

// Camera related functions:

void fxeCameraMove(float x, float y, float z)
{
  ActiveCamera->x = x;
  ActiveCamera->y = y;
  ActiveCamera->z = z;
}


void fxeCameraRotate(float ax, float ay, float az)
{
  ActiveCamera->ax = ax;
  ActiveCamera->ay = ay;
  ActiveCamera->az = az;
}

// Ambient light related functions:

void fxeAmbientlightLight(float red, float green, float blue)
{
  AmbientR = red;
  AmbientG = green;
  AmbientB = blue;
  AmbientR2 = red*255.0f;
  AmbientG2 = green*255.0f;
  AmbientG2 = blue*255.0f;
}

// Lightsource related functions:

// Draws Lensflare of a lightsource.
void fxeLightsourceDrawLensflare(fxeLightsource *pLightsource)
{
	fxeVertex VertexCamera;
	fxeVertex VertexWorld;
	fxeDrawVertex ProjectedVertex;
	float VecLen;
	fxeVector LightVector;
	fxeVector CameraVector;
	float Dp3d;
	int Loop;
	float WindowCenterX, WindowCenterY;
	float FlareRelX, FlareRelY;
	float FlareX, FlareY;
	fxeDrawVertex Lensflare[4];
	float Alpha;
	fxeRectangle2d Rec2d;
	fxeRectCoordinate *RC;
	void *VertexPointers[4];
	unsigned short Pixel;


	if (pLightsource->LFH == NULL) 
		return;

	VertexPointers[0] = (void*)&Lensflare[0];
	VertexPointers[1] = (void*)&Lensflare[1];
	VertexPointers[2] = (void*)&Lensflare[2];
	VertexPointers[3] = (void*)&Lensflare[3];

	// We must transform lightsource to camera coordinates:
	VertexWorld.x = pLightsource->x;
	VertexWorld.y = pLightsource->y;
	VertexWorld.z = pLightsource->z;

	MatXVer(CameraMatrix,&VertexWorld,&VertexCamera);


	// We don't want divide by zero...
	if (VertexCamera.z <= NearClippingPlane)
		return;

	// Project the camera vertex so we can clip the lightsource in
	// 2D.

	ProjectedVertex.q = 1.0f/VertexCamera.z;

	ProjectedVertex.x = (VertexCamera.x*ProjectedVertex.q+1.0f)*
		ProjectWidthDiv2+ProjectStartX;
	Snap(&ProjectedVertex.x);

	ProjectedVertex.y = (VertexCamera.y*ProjectedVertex.q+1.0f)*		
		ProjectHeightDiv2+ProjectStartY;
	Snap(&ProjectedVertex.y);


	// Don't use this flare if its outside the render window:
	if (ProjectedVertex.x < ProjectStartX || ProjectedVertex.x >
		ProjectEndX || ProjectedVertex.y < ProjectStartY ||
		ProjectedVertex.y > ProjectEndY)
		return;

	// Check if the lightsource pixel we rendered is visible,
	// if visible render lensflare, else exit.

	if (grLfbReadRegion(GR_BUFFER_BACKBUFFER,
		(FxU32)ProjectedVertex.x,
		(FxU32)VerRes-(FxU32)ProjectedVertex.y,
		1,1,
		2,&Pixel) == FXFALSE)
	{
		fxeLogDeviceWriteString("Unable to read data from Linear Frame Buffer!");
		return;
	}

	if (Pixel != 0xffff)
		return;
			

	// As this is only made once per frame for a limited amount of
	// lightsources we can perform some advanced code here.
	
	// We need to create a normalized vector from the origin of 
	// camera space to the lightsource position in camera space.
	// Aahhh, this is simple because we already have the vector,
	// we just need to make its lenght 1.0.


	LightVector.x = VertexCamera.x;
	LightVector.y = VertexCamera.y;
	LightVector.z = VertexCamera.z;

	VecLen = VectorLenght(&LightVector);

	LightVector.x /= VecLen;
	LightVector.y /= VecLen;
	LightVector.z /= VecLen;

	CameraVector.x = 0.0f;
	CameraVector.y = 0.0f;
	CameraVector.z = 1.0f;
	
	Dp3d = CameraVector.x*LightVector.x+CameraVector.y*
		LightVector.y+CameraVector.z*LightVector.z;

	// Now Dp3d should be a value from 0..1. 1 means the lightsource
	// is just in front of the camera and should be drawn with maximum
	// intensity. This works perfect for lightsources with very long
	// distance such as the sun but not very well lightsources with
	// a limited distance.

	Alpha = 255.0f*Dp3d;

	WindowCenterX = ProjectWidthDiv2+ProjectStartX;
	WindowCenterY = ProjectHeightDiv2+ProjectStartY;
	FlareRelX = ProjectedVertex.x-WindowCenterX;
	FlareRelY = ProjectedVertex.y-WindowCenterY;

	
	grDepthBufferMode(GR_DEPTHBUFFER_DISABLE);
	grTexFilterMode(GR_TMU0,GR_TEXTUREFILTER_POINT_SAMPLED,
		GR_TEXTUREFILTER_POINT_SAMPLED);
	if (NumTexelFx > 1) 
		grTexFilterMode(GR_TMU1,GR_TEXTUREFILTER_POINT_SAMPLED,
			GR_TEXTUREFILTER_POINT_SAMPLED);

	//alpha blending stuff...
  grColorCombine( GR_COMBINE_FUNCTION_SCALE_OTHER,
                       GR_COMBINE_FACTOR_LOCAL_ALPHA,
                       GR_COMBINE_LOCAL_NONE,
                       GR_COMBINE_OTHER_TEXTURE,
                       FXFALSE );
  grAlphaCombine( GR_COMBINE_FUNCTION_LOCAL_ALPHA,
                        GR_COMBINE_FACTOR_NONE,
                        GR_COMBINE_LOCAL_CONSTANT,
                        GR_COMBINE_OTHER_NONE,
                        FXFALSE );
	grConstantColorValue(((int)Alpha)<<24);
	grAlphaBlendFunction(GR_BLEND_ONE,GR_BLEND_ONE,
		GR_BLEND_ZERO,GR_BLEND_ZERO);
  grAlphaTestFunction( GR_CMP_ALWAYS );

	//this is where the magic happens...
	

	for (Loop = 0; Loop<pLightsource->LFH->Numflares; Loop++)
	{
		FlareX = pLightsource->LFH->Lensfls[Loop].MulFactor *
			FlareRelX + WindowCenterX;
		FlareY = pLightsource->LFH->Lensfls[Loop].MulFactor *
			FlareRelY + WindowCenterY;

		
		RC = Rec2d.RectCoordinates;
		// Upper left corner:


		Lensflare[0].q = 1.0f;
		Lensflare[0].x = FlareX - pLightsource->LFH->Lensfls[Loop].Radius;
		Lensflare[0].y = FlareY + pLightsource->LFH->Lensfls[Loop].Radius;

		RC[0].x = Lensflare[0].x;
		RC[0].y = Lensflare[0].y;
	

		// Down left corner:

		Lensflare[1].q = 1.0f;
		Lensflare[1].x = FlareX - pLightsource->LFH->Lensfls[Loop].Radius;
		Lensflare[1].y = FlareY - pLightsource->LFH->Lensfls[Loop].Radius;
		RC[1].x = Lensflare[1].x;
		RC[1].y = Lensflare[1].y;

		// Down Right corner:
	
		Lensflare[2].q = 1.0f;
		Lensflare[2].x = FlareX + pLightsource->LFH->Lensfls[Loop].Radius;
		Lensflare[2].y = FlareY - pLightsource->LFH->Lensfls[Loop].Radius;
		RC[2].x = Lensflare[2].x;
		RC[2].y = Lensflare[2].y;

		// Upper right corner:

		Lensflare[3].q = 1.0f;
		Lensflare[3].x = FlareX + pLightsource->LFH->Lensfls[Loop].Radius;
		Lensflare[3].y = FlareY + pLightsource->LFH->Lensfls[Loop].Radius;
		RC[3].x = Lensflare[3].x;
		RC[3].y = Lensflare[3].y;

		ClipTextureRectangle2d(&Rec2d);

		Lensflare[0].x = RC[0].x;
		Lensflare[0].y = RC[0].y;
		Snap(&Lensflare[0].x);
		Snap(&Lensflare[0].y);
		Lensflare[0].u0 = RC[0].u*255.0f*Lensflare[0].q;
		Lensflare[0].v0 = RC[0].v*255.0f*Lensflare[0].q;
		Lensflare[0].u1 = Lensflare[0].u0;
		Lensflare[0].v1 = Lensflare[0].v0;

		Lensflare[1].x = RC[1].x;
		Lensflare[1].y = RC[1].y;
		Snap(&Lensflare[1].x);
		Snap(&Lensflare[1].y);
		Lensflare[1].u0 = RC[1].u*255.0f*Lensflare[1].q;
		Lensflare[1].v0 = RC[1].v*255.0f*Lensflare[1].q;
		Lensflare[1].u1 = Lensflare[1].u0;
		Lensflare[1].v1 = Lensflare[1].v0;

		Lensflare[2].x = RC[2].x;
		Lensflare[2].y = RC[2].y;
		Snap(&Lensflare[2].x);
		Snap(&Lensflare[2].y);
		Lensflare[2].u0 = RC[2].u*255.0f*Lensflare[2].q;
		Lensflare[2].v0 = RC[2].v*255.0f*Lensflare[2].q;
		Lensflare[2].u1 = Lensflare[2].u0;
		Lensflare[2].v1 = Lensflare[2].v0;

		Lensflare[3].x = RC[3].x;
		Lensflare[3].y = RC[3].y;
		Snap(&Lensflare[3].x);
		Snap(&Lensflare[3].y);
		Lensflare[3].u0 = RC[3].u*255.0f*Lensflare[3].q;
		Lensflare[3].v0 = RC[3].v*255.0f*Lensflare[3].q;
		Lensflare[3].u1 = Lensflare[3].u0;
		Lensflare[3].v1 = Lensflare[3].v0;

		// Setup tmu[s] with texture of lensflare:
		SetupTmusWithTexture
			(pLightsource->LFH->Lensfls[Loop].LensTexture);

		grDrawVertexArray(GR_TRIANGLE_FAN,4,VertexPointers);
	}

	// Restore default glide state for FxEngine:

	grTexFilterMode(GR_TMU0,GR_TEXTUREFILTER_BILINEAR,
		GR_TEXTUREFILTER_BILINEAR);
	if (NumTexelFx > 1) 
		grTexFilterMode(GR_TMU1,GR_TEXTUREFILTER_BILINEAR,
			GR_TEXTUREFILTER_BILINEAR);

  grColorCombine( GR_COMBINE_FUNCTION_SCALE_OTHER,
                       GR_COMBINE_FACTOR_LOCAL,
                       GR_COMBINE_LOCAL_ITERATED,
                       GR_COMBINE_OTHER_TEXTURE,
                       FXFALSE );
  grAlphaCombine( GR_COMBINE_FUNCTION_LOCAL_ALPHA,
                        GR_COMBINE_FACTOR_NONE,
                        GR_COMBINE_LOCAL_CONSTANT,
                        GR_COMBINE_OTHER_NONE,
                        FXFALSE );
	grConstantColorValue(255<<24);
	grAlphaBlendFunction(GR_BLEND_SRC_ALPHA,GR_BLEND_ONE_MINUS_SRC_ALPHA,
		GR_BLEND_ZERO,GR_BLEND_ZERO);
	
	grDepthBufferMode(GR_DEPTHBUFFER_WBUFFER);
}


// Creates a lensflare which you can bind to one or more lightsources.
fxeLensflareHandle fxeLensflareCreate(int NumFlares, fxeTextureHandle *TextureHandles,
	float *Radiuses, float *MulFactors)
{
	int Loop;
	fxeLensflareHandle LensFlares;

	LensFlares = (fxeLensflares*)FxMalloc(sizeof(fxeLensflares));
	if (LensFlares == NULL)
		return NULL;
	LensFlares->Numflares = NumFlares;
	LensFlares->Lensfls	= (fxeLensflare*)FxMalloc(sizeof(fxeLensflare)*NumFlares);
	if (LensFlares->Lensfls == NULL)
		return NULL;
	for (Loop=0;Loop<NumFlares;Loop++)
	{
		LensFlares->Lensfls[Loop].LensTexture = TextureHandles[Loop];
		LensFlares->Lensfls[Loop].MulFactor = MulFactors[Loop];
		LensFlares->Lensfls[Loop].Radius = Radiuses[Loop];
	}
	return LensFlares;
}
																					

// Allocates memory for a lightsource and inserts it into the
// linked list of lightsources

fxeLightsourceHandle fxeLightsourceAdd(float r, float g, float b,
  float x, float y, float z, float DistanceAffection, fxeLensflareHandle LFH,
	fxeVector *pNormal)
{
	fxeLightsource *Lightsource;

  Lightsource = (fxeLightsource *)FxMalloc(sizeof(fxeLightsource));
	Lightsource->r = r;
	Lightsource->g = g;
	Lightsource->b = b;
  Lightsource->x = x;
  Lightsource->y = y;
  Lightsource->z = z;
  Lightsource->DistanceAffection = DistanceAffection;
	Lightsource->Next = NULL;
	Lightsource->LFH = LFH;
	Lightsource->Enabled = FXETRUE;
	Lightsource->Normal.x = pNormal->x;
	Lightsource->Normal.y = pNormal->y;
	Lightsource->Normal.z = pNormal->z;

	// go to last lightsource in linked list of lightsources
	if (Lightsources == NULL)
  {
		// This is the first lightsource
		Lightsource->Prev = NULL;
	  Lightsources = Lightsource;
	}
	else
	{
		while (Lightsources->Next != NULL)
		{
			Lightsources = Lightsources->Next;
	  }
	  Lightsource->Prev = Lightsources;
    Lightsources->Next = Lightsource;
	  Lightsources = Lightsource;
	}
	return Lightsources;
}

void fxeLightsourceDrawPoint(fxeLightsource *pLightsource)
{
	fxeVertex VertexWorld, VertexCamera;
	fxeDrawVertex ProjectedVertex;
	unsigned short Pixel;

	// Only draw point if lightsource have a lensflare
	// associated with it.

	if (pLightsource->LFH == NULL)
		return;

	// We must transform lightsource to camera coordinates:
	VertexWorld.x = pLightsource->x;
	VertexWorld.y = pLightsource->y;
	VertexWorld.z = pLightsource->z;

	MatXVer(CameraMatrix,&VertexWorld,&VertexCamera);


	// We don't want divide by zero...
	if (VertexCamera.z <= NearClippingPlane)
		return;

	// Project the camera vertex so we can clip the lightsource in
	// 2D.

	ProjectedVertex.q = 1.0f/VertexCamera.z;

	ProjectedVertex.x = (VertexCamera.x*ProjectedVertex.q+1.0f)*
		ProjectWidthDiv2+ProjectStartX;
	Snap(&ProjectedVertex.x);

	ProjectedVertex.y = (VertexCamera.y*ProjectedVertex.q+1.0f)*		
		ProjectHeightDiv2+ProjectStartY;
	Snap(&ProjectedVertex.y);


	// Don't use this lightsource if its outside the render window:
	if (ProjectedVertex.x < ProjectStartX || ProjectedVertex.x >
		ProjectEndX || ProjectedVertex.y < ProjectStartY ||
		ProjectedVertex.y > ProjectEndY)
		return;

	// The lightsource is inside clipping frustrum, now we are free
	// to draw the white point in the frame buffer.

	Pixel = 0xffff;

	if (grLfbWriteRegion(GR_BUFFER_BACKBUFFER,
		(FxU32)ProjectedVertex.x,
		(FxU32)VerRes-(FxU32)ProjectedVertex.y,
		GR_LFB_SRC_FMT_565,
		1,1,
		FXFALSE,
		2,(void*)&Pixel) == FXFALSE)
	{
		fxeLogDeviceWriteString("Unable to write to linear frame buffer!");
		return;
	}

}


void fxeLightsourcesDrawPoints(void)
{
	fxeLightsource *pLightsource;

	pLightsource = Lightsources;

	if (Lightsources == NULL)
		return;
	
	// Go to first lightsource:
	while (Lightsources->Prev != NULL)
		Lightsources = Lightsources->Prev;
	
	// Now we are at the first lightsource
	// Go through all lightsources and draw point
	// lightsource is enabled

	while (Lightsources != NULL)
	{
		if (Lightsources->Enabled == FXETRUE) 
			fxeLightsourceDrawPoint(Lightsources);
		Lightsources = Lightsources->Next;
	}

	Lightsources = pLightsource;
}

void fxeLightsourcesDraw(void)
{
	fxeLightsource *pLightsource;

	pLightsource = Lightsources;

	if (Lightsources == NULL)
		return;
	// Go to first lightsource:
	while (Lightsources->Prev != NULL)
		Lightsources = Lightsources->Prev;
	
	// Now we are at the first lightsource
	// Go through all lightsources and draw lensflares
	// if enabled

	while (Lightsources != NULL)
	{
		if (Lightsources->Enabled == FXETRUE) 
			fxeLightsourceDrawLensflare(Lightsources);
		Lightsources = Lightsources->Next;
	}

	Lightsources = pLightsource;
}

void fxeLightsourceMove(fxeLightsourceHandle LightsourceHandle,
	float x, float y, float z)
{
  LightsourceHandle->x = x;
  LightsourceHandle->y = y;
  LightsourceHandle->z = z;
}


void fxeLightsourceLight(fxeLightsourceHandle LightsourceHandle,
	float r, float g, float b)
{
	LightsourceHandle->r = r;
	LightsourceHandle->g = g;
	LightsourceHandle->b = b;
}

// Note that if LightsourceHandle isn't a valid lightsource undefined
// results may occure such as a halting system crash!
void fxeLightsourceDelete(fxeLightsourceHandle LightsourceHandle)
{
	if (LightsourceHandle->Prev == NULL && LightsourceHandle->Next == NULL)
	{
		Lightsources = NULL;
	}
	else if (LightsourceHandle->Prev == NULL)
	{
		Lightsources = LightsourceHandle->Next;
		Lightsources->Prev = NULL;
  }
	else if (LightsourceHandle->Next == NULL)
	{
		Lightsources = LightsourceHandle->Prev;
		Lightsources->Next = NULL;
	}
	else
	{
		Lightsources = LightsourceHandle->Prev;
		Lightsources->Next = LightsourceHandle->Next;
		Lightsources = LightsourceHandle->Next;
		Lightsources->Prev = LightsourceHandle->Prev;
	}
	free(LightsourceHandle);
}

// Texture related functions:

// This function loads only a specific type of mipmap, only mipmaps
// with the argb4444 format, loading a mipmap with another format will
// cause the function to return NULL.

/*

fxeTextureHandle fxeBumpMapLoad(char *filename)
{
  fxeTextureHandle TexHandle;
  int Loop;

  TexHandle = fxeTextureLoad(filename);
  if (TexHandle == NULL)
    return NULL;

  if (TexHandle->TexInfo.format != GR_TEXFMT_ARGB_4444)
  {
    fxeTextureDelete(TexHandle);
    return NULL;
  }

  for (Loop=0; Loop<TexHandle->TexInfo.smallLod
  return TexHandle;
}
*/


// Loads a 3df texture file from disk, returns
// NULL if failure and handle to texture if
// successful.

fxeTextureHandle fxeTextureLoadToMemory(char *filename)
{
	Gu3dfInfo TexInfo;
  fxeTexture *Texture;



	// get the texture info:
  if (gu3dfGetInfo(filename,&TexInfo) == FXFALSE)
    return NULL;

  // allocate memory for the mipmap:
  TexInfo.data = (void*)FxMalloc(TexInfo.mem_required);
  if (TexInfo.data == NULL)
    return NULL;

  // Load the mipmap from disk:
  gu3dfLoad(filename,&TexInfo);


	Texture = (fxeTexture*)FxMalloc(sizeof(fxeTexture));
	if (Texture == NULL)
	{
    free(TexInfo.data);
		return NULL;
	}

  Texture->TexInfo.data = TexInfo.data;
	Texture->TexInfo.format = TexInfo.header.format;
  Texture->TexInfo.aspectRatioLog2 = TexInfo.header.aspect_ratio;
  Texture->TexInfo.smallLodLog2 = TexInfo.header.small_lod;
  Texture->TexInfo.largeLodLog2 = TexInfo.header.large_lod;

	return Texture;
}




/*
// Deletes a texture - TextureHandle must be a valid
// texturehandle or else strange things may happen!
void fxeTextureDelete(fxeTextureHandle TextureHandle)
{
	if (TextureHandle->Next == NULL && TextureHandle->Prev
		== NULL)
		Textures = NULL;
	else if (TextureHandle->Next == NULL)
	{
		Textures = TextureHandle->Prev;
		Textures->Next = NULL;
	}
	else if (TextureHandle->Prev == NULL)
	{
		Textures = TextureHandle->Next;
		Textures->Prev = NULL;
	}
	else
	{
		Textures = TextureHandle->Prev;
		Textures->Next = TextureHandle->Next;
		Textures = TextureHandle->Next;
		Textures->Prev = TextureHandle->Prev;
	}
	free(TextureHandle->TexInfo.data);
	free(TextureHandle);
}
*/

// Centers a object at 0,0,0:
void CenterObject(fxeObject *pObject)
{
  int Loop;
  float ax,ay,az;
  float minx,miny,minz,maxx,maxy,maxz;
  
  // Calculate average x,y,z;

  minx = 65535;
  miny = 65535;
  minz = 65535;
  maxx = -65535;
  maxy = -65535;
  maxz = -65535;

  for (Loop=0;Loop<pObject->NumVertices;Loop++)
  {
    if (pObject->VertexList[Loop].x < minx)
      minx = pObject->VertexList[Loop].x;
    if (pObject->VertexList[Loop].x > maxx)
      maxx = pObject->VertexList[Loop].x;

    if (pObject->VertexList[Loop].y < miny)
      miny = pObject->VertexList[Loop].y;
    if (pObject->VertexList[Loop].y > maxy)
      maxy = pObject->VertexList[Loop].y;

    if (pObject->VertexList[Loop].z < minz)
      minz = pObject->VertexList[Loop].z;
    if (pObject->VertexList[Loop].z > maxz)
      maxz = pObject->VertexList[Loop].z;
  }

  ax = (minx+maxx)/2;
  ay = (miny+maxy)/2;
  az = (minz+maxz)/2;

  // Subtract the average x from all vertices x component
  // Subtract the average y from all vertices y component
  // Subtract the average z from all vertices z component

  for (Loop=0;Loop<pObject->NumVertices;Loop++)
  {
    pObject->VertexList[Loop].x -= ax;
    pObject->VertexList[Loop].y -= ay;
    pObject->VertexList[Loop].z -= az;
  }
}

// Scales a object so that it's maximum radius is ScaleFactor

void ScaleObject(fxeObject *pObject, float ScaleFactor)
{
  int Loop;

  // Divide every vertex by current object radius so that
  // we get all vertices x,y,z values between 0..1

  for (Loop=0;Loop<pObject->NumVertices;Loop++)
  {
    pObject->VertexList[Loop].x /= pObject->ObjectRadius;
    pObject->VertexList[Loop].y /= pObject->ObjectRadius;
    pObject->VertexList[Loop].z /= pObject->ObjectRadius;
  }

  pObject->ObjectRadius = ScaleFactor;


  // Now we can multiply every vertex's x,y,z value with ScaleFactor:
  
  for (Loop=0;Loop<pObject->NumVertices;Loop++)
  {
    pObject->VertexList[Loop].x *= ScaleFactor;
    pObject->VertexList[Loop].y *= ScaleFactor;
    pObject->VertexList[Loop].z *= ScaleFactor;
  }

  // We must setup all polygons radius too:
  for (Loop=0;Loop<pObject->NumPolys;Loop++)
  {

    pObject->PolygonList[Loop].PolygonRadius =
      GetPolygonRadius(&(pObject->PolygonList[Loop]),pObject);
  }


}


// Object related functions:

int fxeObjectDetectCollisionLandscape(fxeObjectHandle OH,
	fxeLandscapesParentHandle LspH)
{
	fxeVector DistanceVector;
	float Distance;
	float Height;

	Height = fxeLandscapeGetHeightXZ(LspH,OH->x,OH->z);
	DistanceVector.x = 0.0f;
	DistanceVector.y = Height - OH->y;
	DistanceVector.z = 0.0f;

	Distance =	VectorLenght(&DistanceVector);
	Distance = Distance - OH->ObjectRadius;
	if (Distance <= 0)
		return FXETRUE;
	return FXEFALSE;
}

// Detects if two objects have collided. If the objects have collided
// this function returns FXETRUE and if the objects haven't collided
// this function returns FXEFALSE
int fxeObjectDetectCollision(fxeObjectHandle OH1, fxeObjectHandle OH2)
{
  fxeVector DistanceVector;
  float Distance;

  DistanceVector.x = OH1->x - OH2->x;
  DistanceVector.y = OH1->y - OH2->y;
  DistanceVector.z = OH1->z - OH2->z;

  Distance = VectorLenght(&DistanceVector);

  if ((Distance - OH1->ObjectRadius - OH2->ObjectRadius) <= 0)
    return FXETRUE;

  return FXEFALSE;
}

void fxeObjectAddVertex(fxeObjectHandle ObjectHandle, float x, float y,
  float z)
{
  ObjectHandle->NumVertices++;

  if (ObjectHandle->VertexList == NULL)
  {
    ObjectHandle->VertexList = (fxeVertex*)FxMalloc(sizeof(fxeVertex));
    ObjectHandle->VerticesWorld = (fxeVertex*)FxMalloc(sizeof(fxeVertex));
    ObjectHandle->VerticesCamera = (fxeDrawVertex*)FxMalloc(sizeof(fxeDrawVertex));
    ObjectHandle->EnviromentVerts = (fxeTextureCoordinate*)FxMalloc(
      sizeof(fxeTextureCoordinate));
		ObjectHandle->VertexNormals = 
				(fxeVector*)FxMalloc(sizeof(fxeVector));
		ObjectHandle->DynColors =
				(fxeColor*)FxMalloc(sizeof(fxeColor));

  }
  else
  {
    ObjectHandle->VertexList = (fxeVertex*)realloc(ObjectHandle->VertexList,
      sizeof(fxeVertex)*ObjectHandle->NumVertices);
    ObjectHandle->VerticesWorld = (fxeVertex*)realloc(ObjectHandle->VerticesWorld,
      sizeof(fxeVertex)*ObjectHandle->NumVertices);
    ObjectHandle->VerticesCamera = (fxeDrawVertex*)realloc(ObjectHandle->VerticesCamera,
      sizeof(fxeDrawVertex)*ObjectHandle->NumVertices);
    ObjectHandle->EnviromentVerts = (fxeTextureCoordinate*)
      realloc(ObjectHandle->EnviromentVerts,sizeof(fxeTextureCoordinate)*ObjectHandle->NumVertices);
		ObjectHandle->VertexNormals = 
				(fxeVector*)realloc(ObjectHandle->VertexNormals,
				sizeof(fxeVector)*ObjectHandle->NumVertices);
		ObjectHandle->DynColors =
				(fxeColor*)realloc(ObjectHandle->DynColors,
				sizeof(fxeColor)*ObjectHandle->NumVertices);
  }

  ObjectHandle->VertexList[ObjectHandle->NumVertices-1].x = x;
  ObjectHandle->VertexList[ObjectHandle->NumVertices-1].y = y;
  ObjectHandle->VertexList[ObjectHandle->NumVertices-1].z = z;

  ObjectHandle->VerticesCamera[ObjectHandle->NumVertices-1].NumPolys = 0;
}

void fxeObjectAddPolygon(fxeObjectHandle ObjectHandle, int VertexIndex1,
  int VertexIndex2, int VertexIndex3, int DoubleSidedState,
  fxeTextureHandle TH, float u0, float v0, float u1, float v1, float
  u2, float v2)
{
  ObjectHandle->NumPolys++;

  if (ObjectHandle->PolygonList == NULL)
  {
    ObjectHandle->PolygonList = (fxePolygon*)FxMalloc(sizeof(fxePolygon));
  }
  else
  {
    ObjectHandle->PolygonList = (fxePolygon*)realloc(ObjectHandle->PolygonList,
      sizeof(fxePolygon)*ObjectHandle->NumPolys);
  }

  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].NumVertices = 3;
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].VertexIndices = (fxeVertexIndex*)
    FxMalloc(sizeof(fxeVertexIndex)*3);
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].VertexIndices[0] = VertexIndex1;
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].VertexIndices[1] = VertexIndex2;
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].VertexIndices[2] = VertexIndex3;

  //ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].ClippedVertices = (fxeDrawVertex*)
  //  FxMalloc(sizeof(fxeDrawVertex)*10);
  //ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].ClippedEnvVer = (fxeTextureCoordinate*)
  //  FxMalloc(sizeof(fxeTextureCoordinate)*10);

  // These two variables isn't used for the moment:
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].TexMode = TEXTURE_MAPPED;
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].TextureHandle = TH;
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].Transparent = FXEFALSE;

  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].DoubleSided = DoubleSidedState;
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].DiffuseColor.r = 1.0f;
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].DiffuseColor.g = 1.0f;
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].DiffuseColor.b = 1.0f;
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].AmbientColor.r = 0.0f;
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].AmbientColor.g = 0.0f;
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].AmbientColor.b = 0.0f;

  // Get the polygons radius:
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].PolygonRadius =
    GetPolygonRadius(&ObjectHandle->PolygonList[ObjectHandle->NumPolys-1],ObjectHandle);

  // Set up texture coordinates:
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].TextureCoordinates =
    (fxeTextureCoordinate*)FxMalloc(sizeof(fxeTextureCoordinate)*3);

  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].TextureCoordinates[0].u
    = u0;
    
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].TextureCoordinates[0].v
    = v0;

  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].TextureCoordinates[1].u
    = u1;
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].TextureCoordinates[1].v
    = v1;

  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].TextureCoordinates[2].u
    = u2;
  ObjectHandle->PolygonList[ObjectHandle->NumPolys-1].TextureCoordinates[2].v
    = v2;

  // Add pointer to this polygon in vertices that uses it.
  if (ObjectHandle->VerticesCamera[VertexIndex1].NumPolys == 0)
  {
    ObjectHandle->VerticesCamera[VertexIndex1].PolygonIndices = (fxePolygonIndex*)
      FxMalloc(sizeof(fxePolygonIndex));

    ObjectHandle->VerticesCamera[VertexIndex1].PolygonIndices[0] =
      ObjectHandle->NumPolys-1;

    ObjectHandle->VerticesCamera[VertexIndex1].NumPolys++;
  }
  else
  {
    ObjectHandle->VerticesCamera[VertexIndex1].NumPolys++;

    ObjectHandle->VerticesCamera[VertexIndex1].PolygonIndices = (fxePolygonIndex*)
      realloc(ObjectHandle->VerticesCamera[VertexIndex1].PolygonIndices,
        sizeof(fxePolygonIndex)*ObjectHandle->VerticesCamera[VertexIndex1].NumPolys);

    ObjectHandle->VerticesCamera[VertexIndex1].PolygonIndices[
      ObjectHandle->VerticesCamera[VertexIndex1].NumPolys-1] =
        ObjectHandle->NumPolys-1;
  }  

  // Add pointer to this polygon in vertices that uses it.
  if (ObjectHandle->VerticesCamera[VertexIndex2].NumPolys == 0)
  {
    ObjectHandle->VerticesCamera[VertexIndex2].PolygonIndices = (fxePolygonIndex*)
      FxMalloc(sizeof(fxePolygonIndex));

    ObjectHandle->VerticesCamera[VertexIndex2].PolygonIndices[0] =
      ObjectHandle->NumPolys-1;

    ObjectHandle->VerticesCamera[VertexIndex2].NumPolys++;
  }
  else
  {
    ObjectHandle->VerticesCamera[VertexIndex2].NumPolys++;

    ObjectHandle->VerticesCamera[VertexIndex2].PolygonIndices = (fxePolygonIndex*)
      realloc(ObjectHandle->VerticesCamera[VertexIndex2].PolygonIndices,
        sizeof(fxePolygonIndex)*ObjectHandle->VerticesCamera[VertexIndex2].NumPolys);

    ObjectHandle->VerticesCamera[VertexIndex2].PolygonIndices[
      ObjectHandle->VerticesCamera[VertexIndex2].NumPolys-1] =
        ObjectHandle->NumPolys-1;
  }  

  // Add pointer to this polygon in vertices that uses it.
  if (ObjectHandle->VerticesCamera[VertexIndex3].NumPolys == 0)
  {
    ObjectHandle->VerticesCamera[VertexIndex3].PolygonIndices = (fxePolygonIndex*)
      FxMalloc(sizeof(fxePolygonIndex));

    ObjectHandle->VerticesCamera[VertexIndex3].PolygonIndices[0] =
      ObjectHandle->NumPolys-1;

    ObjectHandle->VerticesCamera[VertexIndex3].NumPolys++;
  }
  else
  {
    ObjectHandle->VerticesCamera[VertexIndex3].NumPolys++;

    ObjectHandle->VerticesCamera[VertexIndex3].PolygonIndices = (fxePolygonIndex*)
      realloc(ObjectHandle->VerticesCamera[VertexIndex3].PolygonIndices,
        sizeof(fxePolygonIndex)*ObjectHandle->VerticesCamera[VertexIndex3].NumPolys);

    ObjectHandle->VerticesCamera[VertexIndex3].PolygonIndices[
      ObjectHandle->VerticesCamera[VertexIndex3].NumPolys-1] =
        ObjectHandle->NumPolys-1;
  }
}

void PrecalculateVertexNormals(fxeObject *pObject);
void CalculateNormalsLocal(fxeObjectHandle ObjectHandle);
void NormalizeNormals(fxeObject *pObject);



void fxeObjectPerformPrecalculations(fxeObjectHandle ObjectHandle, float
  ObjectSize)
{
  CenterObject(ObjectHandle);
  ObjectHandle->ObjectRadius = GetMaximumRadius(ObjectHandle);
  ScaleObject(ObjectHandle,ObjectSize);
			
	// Precalculate normals for all polygons in object:
	CalculateNormalsLocal(ObjectHandle);
	NormalizeNormals(ObjectHandle);

	// Precalculate normals for all vertices in object:
	PrecalculateVertexNormals(ObjectHandle);
}
  

// This function returns a pointer to a object which is initialized
// with default values. You can't use this object just as it is.
// You have to add vertices, texture coordinates and polygons to it
// using the fxeObjectAddVertex, fxeObjectAddPoly functions.

fxeObject * fxeObjectInitDefaultState(void)
{
  fxeObject *pObject;

  // Allocate memory for the object structure.
  pObject = (fxeObject*)FxMalloc(sizeof(fxeObject));

  // Set the position of the object to 0,0,0
  pObject->x = 0.0;
  pObject->y = 0.0;
  pObject->z = 0.0;

  // Set the default drawing shape - polygons
  pObject->DrawShape = GR_TRIANGLE_FAN;

  // Set the default shading type - gouraud shading
  pObject->ShadingType = GOURAUD_SHADING;

  // Set the default transparent type - not transparent
  pObject->Transparent = FXEFALSE;

  // Set the default enviroment mapping type - not enviroment mapped
  pObject->EnviromentMapped = FXEFALSE;

  // Set the default texturingtype - texture mapped

  pObject->ObjectType = OBJECT_COMPLETELY_TEXTURED;

  pObject->LowAccuracyLightning = FXEFALSE;

  // 0 polys and 0 vertices to begin with - nothing added to the object yet.
  pObject->NumPolys = 0;
  pObject->NumVertices = 0;

  // All lists of the object is set to NULL
  pObject->VertexList = NULL;
  pObject->VerticesWorld = NULL;
  pObject->VerticesCamera = NULL;
  pObject->EnviromentVerts = NULL;
  pObject->PolygonList = NULL;

  return pObject;
}



void PrecalculateVertexNormals(fxeObject *pObject)
{
	int Loop, Loop2;
	fxeDrawVertex *CamVertex = pObject->VerticesCamera;
	fxePolygon *PolyList = pObject->PolygonList;
	fxeVector AddedNormal;
	float NormalLength;

	for (Loop=0;Loop<pObject->NumVertices;Loop++)
	{
		AddedNormal.x = 0.0f;
		AddedNormal.y = 0.0f;
		AddedNormal.z = 0.0f;

		for (Loop2=0;Loop2<CamVertex[Loop].NumPolys;Loop2++)
		{
			AddedNormal.x = PolyList[CamVertex[Loop].PolygonIndices[Loop2]].Normal.x;
			AddedNormal.y = PolyList[CamVertex[Loop].PolygonIndices[Loop2]].Normal.y;
			AddedNormal.z = PolyList[CamVertex[Loop].PolygonIndices[Loop2]].Normal.z;
		}

		NormalLength = VectorLenght(&AddedNormal);

		if (NormalLength == 0)
		{
			NormalLength = 0.0001f;
		}

		AddedNormal.x /= NormalLength;
		AddedNormal.y /= NormalLength;
		AddedNormal.z /= NormalLength;

		pObject->VertexNormals[Loop].x = AddedNormal.x;
		pObject->VertexNormals[Loop].y = AddedNormal.y;
		pObject->VertexNormals[Loop].z = AddedNormal.z;
	}
}

// Calculates normals for all polygons in object
void CalculateNormalsLocal(fxeObjectHandle ObjectHandle)
{
	int Loop;
	fxePolygon *Poly;
	fxeVector Vector1, Vector2;
	fxeVertexIndex *VIndex;
	fxeVertex *Vertex;
  fxeVertex *CurrVertex, *CurrVertex2, *CurrVertex3;

	Poly = ObjectHandle->PolygonList;
	Vertex = ObjectHandle->VertexList;

  for (Loop=0; Loop<ObjectHandle->NumPolys; Loop++)
  {
			VIndex = Poly[Loop].VertexIndices;
      CurrVertex = &(Vertex[VIndex[2]]);
      CurrVertex2 = &(Vertex[VIndex[1]]);
      CurrVertex3 = &(Vertex[VIndex[0]]);
      Vector1.x = CurrVertex->x-CurrVertex3->x;
      Vector1.y = CurrVertex->y-CurrVertex3->y;
      Vector1.z = CurrVertex->z-CurrVertex3->z;
      Vector2.x = CurrVertex2->x-CurrVertex3->x;
      Vector2.y = CurrVertex2->y-CurrVertex3->y;
      Vector2.z = CurrVertex2->z-CurrVertex3->z;

      CrossProduct3d(&(Poly[Loop].Normal),&Vector2,&Vector1);
	}
}



void NormalizeNormals(fxeObject *pObject)
{
	fxePolygon *PolyList = pObject->PolygonList;
	int Loop;
	float NormalLength;

	for (Loop=0;Loop<pObject->NumPolys;Loop++)
	{
		NormalLength = VectorLenght(&PolyList[Loop].Normal);

		if (NormalLength == 0)
		{
			NormalLength = 0.0001f;
		}

		PolyList[Loop].Normal.x /= NormalLength;
		PolyList[Loop].Normal.y /= NormalLength;
		PolyList[Loop].Normal.z /= NormalLength;
	}
}
		


// Loads a object from disk and positions it at 0,0,0
// the loader can only load 3ds files for the moment.
// The 3ds loader loads a entire scene with multiple objects
// and makes these objects to a big object.
fxeObject * fxeObjectLoad(char *Filename, int FileFormat,
  float ScaleFactor, int UseMaterials)
{
  FILE *FilePointer;
  unsigned long size;
  H3dsScene *pScene;
  fxeObject  *pObject;
  int Loop, Loop2;
  int TotalPolys, TotalVertices;
  int CurrVertex, CurrPolygon;
  int VertexIndex;
  H3dsScene *pNewScene;
  int NVB;
  H3dsMaterial *pMaterial;
  char TexFilename[80];
  int n;
  fxeTextureHandle TH;
  int NPB;
  int PIndex;
  int FoundMap;
  unsigned short Swap;


  switch(FileFormat)
  {
    case OBJECT_3DS:

      FilePointer = fopen(Filename,"r+b");
      if (FilePointer == NULL)
        return NULL;

      fseek(FilePointer,0,SEEK_END);
      size = ftell(FilePointer);
      rewind(FilePointer);

      pScene = HRead3dsScene(FilePointer,0,size);
      if (pScene == 0)
        return NULL;

      fclose(FilePointer);

			fxeLogDeviceWriteString("Loading object: ");
			fxeLogDeviceWriteString(Filename);
			fxeLogDeviceWriteString("\n\n");



      // Now we should have a pointer to the scene in pScene

      pObject = (fxeObject*)FxMalloc(sizeof(fxeObject));

      TotalPolys = 0;
      TotalVertices= 0;

      

      for (Loop=0; Loop<pScene->meshobjs; Loop++)
      {
        TotalPolys += (int)pScene->meshobjlist[Loop].faces;
        TotalVertices += (int)pScene->meshobjlist[Loop].verts;
      }



      // Now we creates a 3ds scene with only 1 object - we takes all
      // objects in the meshobjlist of the loaded scene
      // and sets them in one object only in the new scene - this is just
      // to make the code easier.

      pNewScene = (H3dsScene*)FxMalloc(sizeof(H3dsScene*));
      pNewScene->meshobjs = 1;
      pNewScene->meshobjlist = (H3dsMeshObj*)FxMalloc(sizeof(H3dsMeshObj));
      pNewScene->meshobjlist[0].verts = TotalVertices;
      pNewScene->meshobjlist[0].faces = TotalPolys;
      pNewScene->meshobjlist[0].maps = TotalVertices;
      pNewScene->meshobjlist[0].facelist =
        (H3dsFace*)FxMalloc(sizeof(H3dsFace)*TotalPolys);
      pNewScene->meshobjlist[0].vertlist =
        (H3dsVert*)FxMalloc(sizeof(H3dsVert)*TotalVertices);
      pNewScene->meshobjlist[0].maplist =
        (H3dsMap*)FxMalloc(sizeof(H3dsMap)*TotalVertices);

      // Copy all vertices and polys from old scene to new scene:

      CurrVertex = 0;
      CurrPolygon = 0;
      NVB = 0;

      for (Loop=0; Loop<pScene->meshobjs; Loop++)
      {
        for (Loop2=0; Loop2<pScene->meshobjlist[Loop].faces; Loop2++)
        {
          pNewScene->meshobjlist[0].facelist[CurrPolygon].p0 =
            pScene->meshobjlist[Loop].facelist[Loop2].p0+NVB;
          pNewScene->meshobjlist[0].facelist[CurrPolygon].p1 =
            pScene->meshobjlist[Loop].facelist[Loop2].p1+NVB;
          pNewScene->meshobjlist[0].facelist[CurrPolygon].p2 =
            pScene->meshobjlist[Loop].facelist[Loop2].p2+NVB;
          pNewScene->meshobjlist[0].facelist[CurrPolygon].flags =
            pScene->meshobjlist[Loop].facelist[Loop2].flags;
          CurrPolygon++;
        }

        for (Loop2=0; Loop2<pScene->meshobjlist[Loop].verts; Loop2++)
        {
          pNewScene->meshobjlist[0].vertlist[CurrVertex].x =
            pScene->meshobjlist[Loop].vertlist[Loop2].x;
          pNewScene->meshobjlist[0].vertlist[CurrVertex].y =
            pScene->meshobjlist[Loop].vertlist[Loop2].y;
          pNewScene->meshobjlist[0].vertlist[CurrVertex].z =
            pScene->meshobjlist[Loop].vertlist[Loop2].z;

          if (pScene->meshobjlist[Loop].maps != 0)
          {
            pNewScene->meshobjlist[0].maplist[CurrVertex].u =
              pScene->meshobjlist[Loop].maplist[Loop2].u;
            pNewScene->meshobjlist[0].maplist[CurrVertex].v =
              pScene->meshobjlist[Loop].maplist[Loop2].v;
          }
          CurrVertex++;
        }



        NVB += pScene->meshobjlist[Loop].verts;
      }

      pObject->x = 0.0;
      pObject->y = 0.0;
      pObject->z = 0.0;

      pObject->DrawShape = GR_TRIANGLE_FAN;

      pObject->ShadingType = FLAT_SHADING;
      pObject->Transparent = FXEFALSE;
      pObject->EnviromentMapped = FXEFALSE;
      pObject->pPolygonSortList = NULL;


      if (UseMaterials == FXETRUE)
        pObject->ObjectType = OBJECT_COMPLETELY_TEXTURED;
      else
        pObject->ObjectType = OBJECT_NOT_TEXTURED;

      pObject->LowAccuracyLightning = FXEFALSE;

      pObject->NumPolys = TotalPolys;
      pObject->NumVertices = TotalVertices;

      pObject->VertexList = (fxeVertex*)FxMalloc(sizeof(fxeVertex)*TotalVertices);
      pObject->VerticesWorld = (fxeVertex*)FxMalloc(sizeof(fxeVertex)*TotalVertices);
      pObject->VerticesCamera = (fxeDrawVertex*)FxMalloc(sizeof(fxeDrawVertex)*TotalVertices);
      pObject->EnviromentVerts = (fxeTextureCoordinate*)
        FxMalloc(sizeof(fxeTextureCoordinate)*TotalVertices);
			pObject->VertexNormals = 
				(fxeVector*)FxMalloc(sizeof(fxeVector)*TotalVertices);
      pObject->PolygonList = (fxePolygon*)FxMalloc(sizeof(fxePolygon)*TotalPolys);
      if (pObject->PolygonList == NULL)
				printf("Not enough memory\n");
			pObject->DynColors = (fxeColor*)FxMalloc(sizeof(fxeColor)*TotalVertices);



        for (Loop2=0; Loop2<pNewScene->meshobjlist[0].verts; Loop2++)
        {
          pObject->VertexList[Loop2].x =
            pNewScene->meshobjlist[0].vertlist[Loop2].x;
          pObject->VertexList[Loop2].y =
            pNewScene->meshobjlist[0].vertlist[Loop2].y;
          pObject->VertexList[Loop2].z =
            pNewScene->meshobjlist[0].vertlist[Loop2].z;
         
          pObject->VerticesCamera[Loop2].NumPolys = 0;
          pObject->VerticesCamera[Loop2].PolygonIndices = NULL;
        }


        for (Loop2=0; Loop2<pNewScene->meshobjlist[0].faces; Loop2++)
        {
          pObject->VerticesCamera[pNewScene->meshobjlist[0].facelist[Loop2].p0].
            NumPolys++;

          pObject->VerticesCamera[pNewScene->meshobjlist[0].facelist[Loop2].p1].
            NumPolys++;

          pObject->VerticesCamera[pNewScene->meshobjlist[0].facelist[Loop2].p2].
            NumPolys++;
        }



        // allocate memory for all vertices polygonindices:
        for (Loop2=0; Loop2<pNewScene->meshobjlist[0].verts; Loop2++)
        {
          pObject->VerticesCamera[Loop2].PolygonIndices =
            (fxePolygonIndex*) FxMalloc(sizeof(fxePolygonIndex)*
            pObject->VerticesCamera[Loop2].NumPolys);
          pObject->VerticesCamera[Loop2].NumPolys = 0;
        }





        for (Loop2=0; Loop2<pNewScene->meshobjlist[0].faces; Loop2++)
        {
          pObject->PolygonList[Loop2].NumVertices = 3;

          pObject->PolygonList[Loop2].VertexIndices =
            (fxeVertexIndex*)FxMalloc(sizeof(fxeVertexIndex)*3);

          pObject->PolygonList[Loop2].TextureCoordinates =
            (fxeTextureCoordinate*)FxMalloc(sizeof(fxeTextureCoordinate)*3);


          if (pNewScene->meshobjlist[0].facelist[Loop2].flags == 0x07)
          {
            // Vertices are given in the opposite order - swap
            // vertices. Saw this in the source to a 3ds loader.

            Swap = pNewScene->meshobjlist[0].facelist[Loop2].p2;

            pNewScene->meshobjlist[0].facelist[Loop2].p2 =
              pNewScene->meshobjlist[0].facelist[Loop2].p1;

            pNewScene->meshobjlist[0].facelist[Loop2].p1 = Swap;
          }

              
          VertexIndex = pNewScene->meshobjlist[0].facelist[Loop2].p0;

          pObject->PolygonList[Loop2].VertexIndices[0] = VertexIndex;

          pObject->VerticesCamera[VertexIndex].PolygonIndices[
            (pObject->VerticesCamera[VertexIndex].NumPolys)] = Loop2;
          pObject->VerticesCamera[VertexIndex].NumPolys++;

          pObject->PolygonList[Loop2].TextureCoordinates[0].u =
            pNewScene->meshobjlist[0].maplist[VertexIndex].u;
          pObject->PolygonList[Loop2].TextureCoordinates[0].v =
            pNewScene->meshobjlist[0].maplist[VertexIndex].v;
          


          VertexIndex = pNewScene->meshobjlist[0].facelist[Loop2].p1;

          pObject->PolygonList[Loop2].VertexIndices[1] = VertexIndex;

          pObject->VerticesCamera[VertexIndex].PolygonIndices[
            (pObject->VerticesCamera[VertexIndex].NumPolys)] = Loop2;
          pObject->VerticesCamera[VertexIndex].NumPolys++;

          pObject->PolygonList[Loop2].TextureCoordinates[1].u =
            pNewScene->meshobjlist[0].maplist[VertexIndex].u;
          pObject->PolygonList[Loop2].TextureCoordinates[1].v =
            pNewScene->meshobjlist[0].maplist[VertexIndex].v;


          VertexIndex = pNewScene->meshobjlist[0].facelist[Loop2].p2;

          pObject->PolygonList[Loop2].VertexIndices[2] = VertexIndex;

          pObject->VerticesCamera[VertexIndex].PolygonIndices[
            (pObject->VerticesCamera[VertexIndex].NumPolys)] = Loop2;
          pObject->VerticesCamera[VertexIndex].NumPolys++;

          pObject->PolygonList[Loop2].TextureCoordinates[2].u =
            pNewScene->meshobjlist[0].maplist[VertexIndex].u;
          pObject->PolygonList[Loop2].TextureCoordinates[2].v =
            pNewScene->meshobjlist[0].maplist[VertexIndex].v;

          // Allocate memory for clipped vertices: (very many for
          // the moment, when you clip a triangle to six planes
          // I don't know how many new vertices can be added so it's just
          // to be sure and we got a HUGE amount of memory in win32
          // anyway...)

          //pObject->PolygonList[Loop2].ClippedVertices =
          //  (fxeDrawVertex*)FxMalloc(sizeof(fxeDrawVertex)*10);
          //pObject->PolygonList[Loop2].ClippedEnvVer =
          //  (fxeTextureCoordinate*)FxMalloc(sizeof(fxeTextureCoordinate)*10);

          pObject->PolygonList[Loop2].TexMode = NOT_TEXTURED;
          pObject->PolygonList[Loop2].Transparent = FXEFALSE;

          pObject->PolygonList[Loop2].DoubleSided = FXEFALSE;

          pObject->PolygonList[Loop2].PolygonRadius =
            GetPolygonRadius(&(pObject->PolygonList[Loop2]),pObject);

          pObject->PolygonList[Loop2].DiffuseColor.r = 1.0f;
          pObject->PolygonList[Loop2].DiffuseColor.g = 1.0f;
          pObject->PolygonList[Loop2].DiffuseColor.b = 1.0f;
          
					pObject->PolygonList[Loop2].AmbientColor.r = 0.0f;
          pObject->PolygonList[Loop2].AmbientColor.g = 0.0f;
          pObject->PolygonList[Loop2].AmbientColor.b = 0.0f;
				

        }



        if (UseMaterials == FXETRUE)
        {

        FoundMap = FXEFALSE;

        // go through all objects, load textures needed

        // Load all textures:

        pMaterial = &pScene->Materials;

        while (pMaterial != NULL)
        {
          if (pMaterial->TexFilename != NULL)
          {
            n = 0;
            while(pMaterial->TexFilename[n] != '.')
            {
              TexFilename[n] = pMaterial->TexFilename[n];
              n++;
            }
            TexFilename[n] = '.';
            TexFilename[n+1] = '3';
            TexFilename[n+2] = 'd';
            TexFilename[n+3] = 'f';
            TexFilename[n+4] = 0;

						//fxeLogDeviceWriteString(TexFilename);
						//fxeLogDeviceWriteString("\n");

            pMaterial->TH = (void*)fxeTextureLoad(TexFilename,FXE_TEXTURE_MAP);
            if (pMaterial->TH != NULL)
              FoundMap = FXETRUE;
          }
          pMaterial = pMaterial->Next;
        }

        if (FoundMap == FXEFALSE)
          pObject->ObjectType = OBJECT_NOT_TEXTURED;

        NPB = 0;

        for (Loop=0;Loop<pScene->meshobjs;Loop++)
        {
          pScene->meshobjlist[Loop].CurrFaceMaterial =
            &pScene->meshobjlist[Loop].FaceMaterials;

          while (pScene->meshobjlist[Loop].CurrFaceMaterial != NULL)
          {
            if (pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial != NULL)
            {
              TH = pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->TH;

                if (pScene->meshobjlist[Loop].CurrFaceMaterial->nPolygons != 0)
                {

                  for (Loop2=0;Loop2<
                    pScene->meshobjlist[Loop].CurrFaceMaterial->nPolygons;
                    Loop2++)
                  {
                  
                    PIndex =
                      pScene->meshobjlist[Loop].
                      CurrFaceMaterial->PolygonIndices[Loop2];

                    pObject->PolygonList[PIndex+NPB].TextureHandle = TH;

                    pObject->PolygonList[PIndex+NPB].DoubleSided =
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->
											DoubleSided;

										pObject->PolygonList[PIndex+NPB].AmbientColor.r =
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->
											ambient.red;
/*
										sprintf(ColorStr,"Ambient Red Color %f",
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->ambient.red);
										fxeLogDeviceWriteString(ColorStr);
										sprintf(ColorStr,"Ambient Green Color %f",
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->ambient.green);
										fxeLogDeviceWriteString(ColorStr);
										sprintf(ColorStr,"Ambient Blue Color %f",
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->ambient.blue);
										fxeLogDeviceWriteString(ColorStr);
										
										sprintf(ColorStr,"Diffuse Red Color %f",
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->diffuse.red);
										fxeLogDeviceWriteString(ColorStr);
										sprintf(ColorStr,"Diffuse Green Color %f",
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->diffuse.green);
										fxeLogDeviceWriteString(ColorStr);
										sprintf(ColorStr,"Diffuse Blue Color %f",
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->diffuse.blue);
										fxeLogDeviceWriteString(ColorStr);

										pObject->PolygonList[PIndex+NPB].AmbientColor.b =
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->
											ambient.blue;
	*/									
										pObject->PolygonList[PIndex+NPB].DiffuseColor.r =
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->
											diffuse.red;

										pObject->PolygonList[PIndex+NPB].DiffuseColor.g =
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->
											diffuse.green;

										pObject->PolygonList[PIndex+NPB].DiffuseColor.b =
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->
											diffuse.blue;

										pObject->PolygonList[PIndex+NPB].AlphaLevel =
											(unsigned char)(pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->
											alpha * 255.0f);

                    if (TH != NULL)
                      pObject->PolygonList[PIndex+NPB].TexMode = TEXTURE_MAPPED;
                  }
                }
            }
            pScene->meshobjlist[Loop].CurrFaceMaterial =
              pScene->meshobjlist[Loop].CurrFaceMaterial->Next;
          }
          NPB += pScene->meshobjlist[Loop].faces;
        }
        }

        else
          pObject->ObjectType = OBJECT_NOT_TEXTURED;



        // Now destroy the new created 3ds scene:
        free(pNewScene->meshobjlist[0].vertlist);
        free(pNewScene->meshobjlist[0].facelist);
        free(pNewScene->meshobjlist[0].maplist);
        free(pNewScene->meshobjlist);
        free(pNewScene);

 
      // Center the object at 0,0,0 so rotation works properly:

      CenterObject(pObject);



      pObject->ObjectRadius = GetMaximumRadius(pObject);

      if (ScaleFactor != 0)
      {
        ScaleObject(pObject,ScaleFactor);
      }

			// Precalculate normals for all polygons in object:

			CalculateNormalsLocal(pObject);
			NormalizeNormals(pObject);

			// Precalculate normals for all vertices in object:

			PrecalculateVertexNormals(pObject);
       

    HFree3dsScene(pScene);
    return pObject;
  }
  return NULL;

}


// Loads a object from a 'FILE_3DS' resource from the executable 
// and positions it at 0,0,0
// the loader can only load 3ds files for the moment
// The 3ds loader loads a entire scene with multiple objects
// and makes these objects to a big object.
fxeObject * fxeObjectLoadFromExe(int ResID, int FileFormat,
  float ScaleFactor, int UseTextures)
{

  unsigned long size;
  H3dsScene *pScene;
  fxeObject  *pObject;
  int Loop, Loop2;
  int TotalPolys, TotalVertices;
  int CurrVertex, CurrPolygon;
  int VertexIndex;
  H3dsScene *pNewScene;
  int NVB;
  H3dsMaterial *pMaterial;
  char TexFilename[80];
  int n;
  fxeTextureHandle TH;
  int NPB;
  int PIndex;
  int FoundMap;
  unsigned short Swap;
	void* ObjMem;

  switch(FileFormat)
  {
    case OBJECT_3DS:

      ObjMem = fxeLoadResource(ResID,"FILE_3DS",&size);
			if (ObjMem == NULL)
				return NULL;

      pScene = HRead3dsScene(ObjMem,1,size);
      if (pScene == 0)
        return NULL;

      // Now we should have a pointer to the scene in pScene

      pObject = (fxeObject*)FxMalloc(sizeof(fxeObject));

      TotalPolys = 0;
      TotalVertices= 0;

      

      for (Loop=0; Loop<pScene->meshobjs; Loop++)
      {
        TotalPolys += (int)pScene->meshobjlist[Loop].faces;
        TotalVertices += (int)pScene->meshobjlist[Loop].verts;
      }



      // Now we creates a 3ds scene with only 1 object - we takes all
      // objects in the meshobjlist of the loaded scene
      // and sets them in one object only in the new scene - this is just
      // to make the code easier.

      pNewScene = (H3dsScene*)FxMalloc(sizeof(H3dsScene*));
      pNewScene->meshobjs = 1;
      pNewScene->meshobjlist = (H3dsMeshObj*)FxMalloc(sizeof(H3dsMeshObj));
      pNewScene->meshobjlist[0].verts = TotalVertices;
      pNewScene->meshobjlist[0].faces = TotalPolys;
      pNewScene->meshobjlist[0].maps = TotalVertices;
      pNewScene->meshobjlist[0].facelist =
        (H3dsFace*)FxMalloc(sizeof(H3dsFace)*TotalPolys);
      pNewScene->meshobjlist[0].vertlist =
        (H3dsVert*)FxMalloc(sizeof(H3dsVert)*TotalVertices);
      pNewScene->meshobjlist[0].maplist =
        (H3dsMap*)FxMalloc(sizeof(H3dsMap)*TotalVertices);

      // Copy all vertices and polys from old scene to new scene:

      CurrVertex = 0;
      CurrPolygon = 0;
      NVB = 0;

      for (Loop=0; Loop<pScene->meshobjs; Loop++)
      {
        for (Loop2=0; Loop2<pScene->meshobjlist[Loop].faces; Loop2++)
        {
          pNewScene->meshobjlist[0].facelist[CurrPolygon].p0 =
            pScene->meshobjlist[Loop].facelist[Loop2].p0+NVB;
          pNewScene->meshobjlist[0].facelist[CurrPolygon].p1 =
            pScene->meshobjlist[Loop].facelist[Loop2].p1+NVB;
          pNewScene->meshobjlist[0].facelist[CurrPolygon].p2 =
            pScene->meshobjlist[Loop].facelist[Loop2].p2+NVB;
          pNewScene->meshobjlist[0].facelist[CurrPolygon].flags =
            pScene->meshobjlist[Loop].facelist[Loop2].flags;
          CurrPolygon++;
        }

        for (Loop2=0; Loop2<pScene->meshobjlist[Loop].verts; Loop2++)
        {
          pNewScene->meshobjlist[0].vertlist[CurrVertex].x =
            pScene->meshobjlist[Loop].vertlist[Loop2].x;
          pNewScene->meshobjlist[0].vertlist[CurrVertex].y =
            pScene->meshobjlist[Loop].vertlist[Loop2].y;
          pNewScene->meshobjlist[0].vertlist[CurrVertex].z =
            pScene->meshobjlist[Loop].vertlist[Loop2].z;

          if (pScene->meshobjlist[Loop].maps != 0)
          {
            pNewScene->meshobjlist[0].maplist[CurrVertex].u =
              pScene->meshobjlist[Loop].maplist[Loop2].u;
            pNewScene->meshobjlist[0].maplist[CurrVertex].v =
              pScene->meshobjlist[Loop].maplist[Loop2].v;
          }
          CurrVertex++;
        }



        NVB += pScene->meshobjlist[Loop].verts;
      }

      pObject->x = 0.0;
      pObject->y = 0.0;
      pObject->z = 0.0;

      pObject->DrawShape = GR_TRIANGLE_FAN;

      pObject->ShadingType = FLAT_SHADING;
      pObject->Transparent = FXEFALSE;
      pObject->EnviromentMapped = FXEFALSE;
      pObject->pPolygonSortList = NULL;


      if (UseTextures == FXETRUE)
        pObject->ObjectType = OBJECT_COMPLETELY_TEXTURED;
      else
        pObject->ObjectType = OBJECT_NOT_TEXTURED;

      pObject->LowAccuracyLightning = FXEFALSE;

      pObject->NumPolys = TotalPolys;
      pObject->NumVertices = TotalVertices;

      pObject->VertexList = (fxeVertex*)FxMalloc(sizeof(fxeVertex)*TotalVertices);
      pObject->VerticesWorld = (fxeVertex*)FxMalloc(sizeof(fxeVertex)*TotalVertices);
      pObject->VerticesCamera = (fxeDrawVertex*)FxMalloc(sizeof(fxeDrawVertex)*TotalVertices);
      pObject->EnviromentVerts = (fxeTextureCoordinate*)
        FxMalloc(sizeof(fxeTextureCoordinate)*TotalVertices);
			pObject->VertexNormals = 
				(fxeVector*)FxMalloc(sizeof(fxeVector)*TotalVertices);
      pObject->PolygonList = (fxePolygon*)FxMalloc(sizeof(fxePolygon)*TotalPolys);
      if (pObject->PolygonList == NULL)
      printf("Not enough memory\n");


        for (Loop2=0; Loop2<pNewScene->meshobjlist[0].verts; Loop2++)
        {
          pObject->VertexList[Loop2].x =
            pNewScene->meshobjlist[0].vertlist[Loop2].x;
          pObject->VertexList[Loop2].y =
            pNewScene->meshobjlist[0].vertlist[Loop2].y;
          pObject->VertexList[Loop2].z =
            pNewScene->meshobjlist[0].vertlist[Loop2].z;
         
          pObject->VerticesCamera[Loop2].NumPolys = 0;
          pObject->VerticesCamera[Loop2].PolygonIndices = NULL;
        }


        for (Loop2=0; Loop2<pNewScene->meshobjlist[0].faces; Loop2++)
        {
          pObject->VerticesCamera[pNewScene->meshobjlist[0].facelist[Loop2].p0].
            NumPolys++;

          pObject->VerticesCamera[pNewScene->meshobjlist[0].facelist[Loop2].p1].
            NumPolys++;

          pObject->VerticesCamera[pNewScene->meshobjlist[0].facelist[Loop2].p2].
            NumPolys++;
        }



        // allocate memory for all vertices polygonindices:
        for (Loop2=0; Loop2<pNewScene->meshobjlist[0].verts; Loop2++)
        {
          pObject->VerticesCamera[Loop2].PolygonIndices =
            (fxePolygonIndex*) FxMalloc(sizeof(fxePolygonIndex)*
            pObject->VerticesCamera[Loop2].NumPolys);
          pObject->VerticesCamera[Loop2].NumPolys = 0;
        }





        for (Loop2=0; Loop2<pNewScene->meshobjlist[0].faces; Loop2++)
        {
          pObject->PolygonList[Loop2].NumVertices = 3;

          pObject->PolygonList[Loop2].VertexIndices =
            (fxeVertexIndex*)FxMalloc(sizeof(fxeVertexIndex)*3);

          pObject->PolygonList[Loop2].TextureCoordinates =
            (fxeTextureCoordinate*)FxMalloc(sizeof(fxeTextureCoordinate)*3);


          if (pNewScene->meshobjlist[0].facelist[Loop2].flags == 0x07)
          {
            // Vertices are given in the opposite order - swap
            // vertices. Saw this in the source to a 3ds loader.

            Swap = pNewScene->meshobjlist[0].facelist[Loop2].p2;

            pNewScene->meshobjlist[0].facelist[Loop2].p2 =
              pNewScene->meshobjlist[0].facelist[Loop2].p1;

            pNewScene->meshobjlist[0].facelist[Loop2].p1 = Swap;
          }

              
          VertexIndex = pNewScene->meshobjlist[0].facelist[Loop2].p0;

          pObject->PolygonList[Loop2].VertexIndices[0] = VertexIndex;

          pObject->VerticesCamera[VertexIndex].PolygonIndices[
            (pObject->VerticesCamera[VertexIndex].NumPolys)] = Loop2;
          pObject->VerticesCamera[VertexIndex].NumPolys++;

          pObject->PolygonList[Loop2].TextureCoordinates[0].u =
            pNewScene->meshobjlist[0].maplist[VertexIndex].u;
          pObject->PolygonList[Loop2].TextureCoordinates[0].v =
            pNewScene->meshobjlist[0].maplist[VertexIndex].v;
          


          VertexIndex = pNewScene->meshobjlist[0].facelist[Loop2].p1;

          pObject->PolygonList[Loop2].VertexIndices[1] = VertexIndex;

          pObject->VerticesCamera[VertexIndex].PolygonIndices[
            (pObject->VerticesCamera[VertexIndex].NumPolys)] = Loop2;
          pObject->VerticesCamera[VertexIndex].NumPolys++;

          pObject->PolygonList[Loop2].TextureCoordinates[1].u =
            pNewScene->meshobjlist[0].maplist[VertexIndex].u;
          pObject->PolygonList[Loop2].TextureCoordinates[1].v =
            pNewScene->meshobjlist[0].maplist[VertexIndex].v;


          VertexIndex = pNewScene->meshobjlist[0].facelist[Loop2].p2;

          pObject->PolygonList[Loop2].VertexIndices[2] = VertexIndex;

          pObject->VerticesCamera[VertexIndex].PolygonIndices[
            (pObject->VerticesCamera[VertexIndex].NumPolys)] = Loop2;
          pObject->VerticesCamera[VertexIndex].NumPolys++;

          pObject->PolygonList[Loop2].TextureCoordinates[2].u =
            pNewScene->meshobjlist[0].maplist[VertexIndex].u;
          pObject->PolygonList[Loop2].TextureCoordinates[2].v =
            pNewScene->meshobjlist[0].maplist[VertexIndex].v;

          // Allocate memory for clipped vertices: (very many for
          // the moment, when you clip a triangle to six planes
          // I don't know how many new vertices can be added so it's just
          // to be sure and we got a HUGE amount of memory in win32
          // anyway...)

          //pObject->PolygonList[Loop2].ClippedVertices =
          //  (fxeDrawVertex*)FxMalloc(sizeof(fxeDrawVertex)*10);
          //pObject->PolygonList[Loop2].ClippedEnvVer =
          //  (fxeTextureCoordinate*)FxMalloc(sizeof(fxeTextureCoordinate)*10);

          pObject->PolygonList[Loop2].TexMode = NOT_TEXTURED;
          pObject->PolygonList[Loop2].Transparent = FXEFALSE;

          pObject->PolygonList[Loop2].DoubleSided = FXEFALSE;

          pObject->PolygonList[Loop2].PolygonRadius =
            GetPolygonRadius(&(pObject->PolygonList[Loop2]),pObject);

          pObject->PolygonList[Loop2].DiffuseColor.r = 1.0f;
          pObject->PolygonList[Loop2].DiffuseColor.g = 1.0f;
          pObject->PolygonList[Loop2].DiffuseColor.b = 1.0f;
          
					pObject->PolygonList[Loop2].AmbientColor.r = 0.0f;
          pObject->PolygonList[Loop2].AmbientColor.g = 0.0f;
          pObject->PolygonList[Loop2].AmbientColor.b = 0.0f;
				

        }



        if (UseTextures == FXETRUE)
        {

        FoundMap = FXEFALSE;

        // go through all objects, load textures needed

        // Load all textures:

        pMaterial = &pScene->Materials;

        while (pMaterial != NULL)
        {
          if (pMaterial->TexFilename != NULL)
          {
            n = 0;
            while(pMaterial->TexFilename[n] != '.')
            {
              TexFilename[n] = pMaterial->TexFilename[n];
              n++;
            }
            TexFilename[n] = '.';
            TexFilename[n+1] = '3';
            TexFilename[n+2] = 'd';
            TexFilename[n+3] = 'f';
            TexFilename[n+4] = 0;
            pMaterial->TH = (void*)fxeTextureLoad(TexFilename,FXE_TEXTURE_MAP);
            if (pMaterial->TH != NULL)
              FoundMap = FXETRUE;
          }
          pMaterial = pMaterial->Next;
        }

        if (FoundMap == FXEFALSE)
          pObject->ObjectType = OBJECT_NOT_TEXTURED;

        NPB = 0;

        for (Loop=0;Loop<pScene->meshobjs;Loop++)
        {
          pScene->meshobjlist[Loop].CurrFaceMaterial =
            &pScene->meshobjlist[Loop].FaceMaterials;

          while (pScene->meshobjlist[Loop].CurrFaceMaterial != NULL)
          {
            if (pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial != NULL)
            {
              TH = pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->TH;

                if (pScene->meshobjlist[Loop].CurrFaceMaterial->nPolygons != 0)
                {

                  for (Loop2=0;Loop2<
                    pScene->meshobjlist[Loop].CurrFaceMaterial->nPolygons;
                    Loop2++)
                  {
                  
                    PIndex =
                      pScene->meshobjlist[Loop].
                      CurrFaceMaterial->PolygonIndices[Loop2];

                    pObject->PolygonList[PIndex+NPB].TextureHandle = TH;

                    pObject->PolygonList[PIndex+NPB].DoubleSided =
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->
											DoubleSided;

										pObject->PolygonList[PIndex+NPB].AmbientColor.r =
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->
											ambient.red;

										pObject->PolygonList[PIndex+NPB].AmbientColor.g =
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->
											ambient.green;

										pObject->PolygonList[PIndex+NPB].AmbientColor.b =
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->
											ambient.blue;
										
										pObject->PolygonList[PIndex+NPB].DiffuseColor.r =
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->
											diffuse.red;

										pObject->PolygonList[PIndex+NPB].DiffuseColor.g =
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->
											diffuse.green;

										pObject->PolygonList[PIndex+NPB].DiffuseColor.b =
											pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->
											diffuse.blue;

										pObject->PolygonList[PIndex+NPB].AlphaLevel =
											(unsigned char)(pScene->meshobjlist[Loop].CurrFaceMaterial->pMaterial->
											alpha * 255.0f);

                    if (TH != NULL)
                      pObject->PolygonList[PIndex+NPB].TexMode = TEXTURE_MAPPED;
                  }
                }
            }
            pScene->meshobjlist[Loop].CurrFaceMaterial =
              pScene->meshobjlist[Loop].CurrFaceMaterial->Next;
          }
          NPB += pScene->meshobjlist[Loop].faces;
        }
        }

        else
          pObject->ObjectType = OBJECT_NOT_TEXTURED;



        // Now destroy the new created 3ds scene:
        free(pNewScene->meshobjlist[0].vertlist);
        free(pNewScene->meshobjlist[0].facelist);
        free(pNewScene->meshobjlist[0].maplist);
        free(pNewScene->meshobjlist);
        free(pNewScene);

 
      // Center the object at 0,0,0 so rotation works properly:

      CenterObject(pObject);



      pObject->ObjectRadius = GetMaximumRadius(pObject);

      if (ScaleFactor != 0)
      {
        ScaleObject(pObject,ScaleFactor);
      }

			// Precalculate normals for all polygons in object:

			CalculateNormalsLocal(pObject);
			NormalizeNormals(pObject);

			// Precalculate normals for all vertices in object:

			PrecalculateVertexNormals(pObject);
       

    HFree3dsScene(pScene);
    return pObject;
  }
  return NULL;

}


void fxeObjectMakeTranslucent(fxeObjectHandle ObjectHandle,
  float AlphaLevel)
{
  ObjectHandle->Transparent = FXETRUE;
  ObjectHandle->AlphaLevel = AlphaLevel;
}

void fxeObjectDisableTranslucency(fxeObjectHandle ObjectHandle)
{
  ObjectHandle->Transparent = FXEFALSE;
}



void fxeObjectSetShadingType(fxeObjectHandle ObjectHandle, int ShadingType)
{
 
  ObjectHandle->ShadingType = ShadingType;
}

void fxeObjectMakeEnviromentMapped(fxeObjectHandle OH,
  fxeTextureHandle TH, float AlphaLevel)
{
  OH->EnviromentMapped = FXETRUE;
  OH->ETH = TH;
  OH->EnvTexFactor = AlphaLevel;
}

void fxeObjectSetEnviromentAlpha(fxeObjectHandle ObjectHandle,
  float EnviromentAlphaLevel)
{
  ObjectHandle->EnvTexFactor = EnviromentAlphaLevel;
}

void fxeObjectSetShape(fxeObjectHandle ObjectHandle, int Shape)
{
  if (Shape == FXE_SHAPE_POLYGONS)
    ObjectHandle->DrawShape = GR_TRIANGLE_FAN;
  else if (Shape == FXE_SHAPE_LINES)
    ObjectHandle->DrawShape = GR_LINE_STRIP;
  else if (Shape == FXE_SHAPE_POINTS)
    ObjectHandle->DrawShape = GR_POINTS;
}

// Adds an already created object to the worlds objects:
fxeObjectHandle fxeObjectAdd(fxeObject *Object)
{
	if (WorldObjects == NULL)
	{
		Object->Prev = NULL;
		Object->Next = NULL;
		WorldObjects = Object;
		return WorldObjects;
	}
	else
	{
		// Go to last object in world:
		if (WorldObjects->Next != NULL)
			WorldObjects = WorldObjects->Next;

		WorldObjects->Next = Object;
		Object->Prev = WorldObjects;
		Object->Next = NULL;
		WorldObjects = WorldObjects->Next;
		return WorldObjects;
	}
}

// Scales a object with SFactor.
void fxeObjectScale(fxeObjectHandle ObjectHandle, float SFactor)
{
  int Loop;
  fxeVertex *Vertex;

  Vertex = ObjectHandle->VertexList;

  for (Loop=0 ; Loop<ObjectHandle->NumVertices ; Loop++)
  {
    Vertex[Loop].x *= SFactor;
    Vertex[Loop].y *= SFactor;
    Vertex[Loop].z *= SFactor;
  }
}

// the angles is given in degress and must be in the range of
// 0..359
void fxeObjectRotate(fxeObjectHandle ObjectHandle,
  float xa, float ya, float za)
{
	int Loop;
	fxeVertex *Vertex;
  fxeVertex Vertex2;
  fxeMatrix XRot,YRot,ZRot,ConcatMatrix;
	fxeVector *VertexNormals = ObjectHandle->VertexNormals;
	fxeVector VertexNormal;
	fxePolygon *PolygonList = ObjectHandle->PolygonList;

	if (xa == 0.0 && ya == 0.0 && za == 0.0) return;

	// we performs special logic depending on which angles we must rotate
	// around, when we only needs to rotate around one axis we dont need
	// than doing just that!

	// Check if only the x rotation matrix is required
	if (xa != 0.0 && ya == 0.0 && za == 0.0)
	{
		// Now we only need to multiply the x rotation matrix with every
		// vertex in the object:

    RotXMatrix(ConcatMatrix,360.0f-xa);
	}

	// Check if only the y rotation matrix is required
	else if (xa == 0.0 && ya != 0.0 && za == 0.0)
	{
		// Now we only need to multiply the y rotation matrix with every
		// vertex in the object:

    RotYMatrix(ConcatMatrix,360.0f-ya);
	}

	// Check if only the z rotation matrix is required
	else if (xa == 0.0 && ya == 0.0 && za != 0.0)
	{
		// Now we only need to multiply the z rotation matrix with every
		// vertex in the object:

    RotZMatrix(ConcatMatrix,360.0f-za);
	}

	else if (xa != 0.0 && ya != 0.0 && za == 0.0)
	{
		// We must concatenate the x rotation matrix by the y rotation matrix:

    RotXMatrix(XRot,360.0f-xa);
    RotYMatrix(YRot,360.0f-ya);
    Matmult(XRot,YRot);
    MatCopy(YRot,ConcatMatrix);
	}

	else if (xa != 0.0 && ya == 0.0 && za != 0.0)
	{
		// We must concatenate the x rotation matrix by the z rotation matrix:

    RotXMatrix(XRot,360.0f-xa);
    RotZMatrix(ZRot,360.0f-za);
    Matmult(XRot,ZRot);
    MatCopy(ZRot,ConcatMatrix);
	}

	else if (xa == 0.0 && ya != 0.0 && za != 0.0)
	{
		// We must concatenate the y rotation matrix by the z rotation matrix:
    RotYMatrix(YRot,360.0f-ya);
    RotZMatrix(ZRot,360.0f-za);
    Matmult(YRot,ZRot);
    MatCopy(ZRot,ConcatMatrix);
	}

  else if (xa != 0.0 && ya != 0.0 && za != 0.0)
	{
		// We must concatenate the x rotation matrix by the y rotation matrix:

    RotXMatrix(XRot,360.0f-xa);
    RotYMatrix(YRot,360.0f-ya);
    RotZMatrix(ZRot,360.0f-za);
    Matmult(YRot,ZRot);
    Matmult(ZRot,XRot);
    MatCopy(XRot,ConcatMatrix);
	}

	Vertex = ObjectHandle->VertexList;

	if (Vertex != NULL)
  {
		if (ObjectHandle->ShadingType == GOURAUD_SHADING ||
			ObjectHandle->EnviromentMapped == FXETRUE)
		{
			for (Loop=0 ; Loop<ObjectHandle->NumVertices; Loop++)
			{
				MatXVer(ConcatMatrix,&(Vertex[Loop]),&Vertex2);
				Vertex[Loop].x = Vertex2.x;
				Vertex[Loop].y = Vertex2.y;
				Vertex[Loop].z = Vertex2.z;
				// We need to rotate the precalculated normals vectors for
				// all vertices too:
				MatXVec(ConcatMatrix,
					&VertexNormals[Loop],&VertexNormal);
				VertexNormals[Loop].x = VertexNormal.x;
				VertexNormals[Loop].y = VertexNormal.y;
				VertexNormals[Loop].z = VertexNormal.z;
			}
		}
		else
		{
			for (Loop=0 ; Loop<ObjectHandle->NumVertices; Loop++)
			{
				MatXVer(ConcatMatrix,&(Vertex[Loop]),&Vertex2);
				Vertex[Loop].x = Vertex2.x;
				Vertex[Loop].y = Vertex2.y;
				Vertex[Loop].z = Vertex2.z;
			}
		}
	}

	for (Loop=0 ; Loop<ObjectHandle->NumPolys; Loop++)
	{
		// We need to rotate the precalculated normals of the polygons too:

		MatXVec(ConcatMatrix,
			&PolygonList[Loop].Normal,&VertexNormal);
		PolygonList[Loop].Normal.x = VertexNormal.x;
		PolygonList[Loop].Normal.y = VertexNormal.y;
		PolygonList[Loop].Normal.z = VertexNormal.z;
	}
}


void fxeObjectMove(fxeObjectHandle ObjectHandle,
  float x, float y, float z)
{
  ObjectHandle->x = x;
  ObjectHandle->y = y;
  ObjectHandle->z = z;
}

// This function is not for use by a application, just internally.
void FreeObjectMemory(fxeObjectHandle ObjectHandle)
{
	fxePolygon *Poly;
	int Loop;

	// go through all polygons and destroy respective polygons vertex
  // indices and texture coordinates if any
	Poly = ObjectHandle->PolygonList;
	if (Poly != NULL)
  {
		for (Loop=0 ; Loop<ObjectHandle->NumPolys; Loop++)
		{
			free(Poly[Loop].VertexIndices);
      if (Poly[Loop].TexMode != NOT_TEXTURED)
        free(Poly[Loop].TextureCoordinates);
	  }

		// free polygon list
		free(ObjectHandle->PolygonList);
	}

  // Free all vertices:
	if (ObjectHandle->VertexList != NULL)
	{
		free(ObjectHandle->VertexList);
		free(ObjectHandle->VerticesWorld);
		free(ObjectHandle->VerticesCamera);
	}
}

// Note that if ObjectHandle isn't a valid Object handle undefined
// results may occure such as a halting system crash!
void fxeObjectDelete(fxeObjectHandle ObjectHandle)
{
	if (ObjectHandle->Prev == NULL && ObjectHandle->Next == NULL)
	{
		WorldObjects = NULL;
	}
	else if (ObjectHandle->Prev == NULL)
	{
		WorldObjects = ObjectHandle->Next;
		WorldObjects->Prev = NULL;
  }
	else if (ObjectHandle->Next == NULL)
	{
		WorldObjects = ObjectHandle->Prev;
		WorldObjects->Next = NULL;
	}
	else
	{
		WorldObjects = ObjectHandle->Prev;
		WorldObjects->Next = ObjectHandle->Next;
		WorldObjects = ObjectHandle->Next;
		WorldObjects->Prev = ObjectHandle->Prev;
	}
	// destroy polygons and vertices associated with object
	FreeObjectMemory(ObjectHandle);
	free(ObjectHandle);
}

// Rendering associated functions:

// takes a objects local coordinates and transforms them
// to world coordinates:
void TransformToWorld(fxeObjectHandle ObjectHandle)
{
	fxeVertex *LocVer;
	fxeVertex *WorVer;
	int Loop;

	LocVer = ObjectHandle->VertexList;
	WorVer = ObjectHandle->VerticesWorld;


	if (LocVer != NULL)
	{
		for (Loop=0 ; Loop<ObjectHandle->NumVertices; Loop++)
		{
			WorVer[Loop].x = LocVer[Loop].x + ObjectHandle->x;
			WorVer[Loop].y = LocVer[Loop].y + ObjectHandle->y;
			WorVer[Loop].z = LocVer[Loop].z + ObjectHandle->z;
    }
	}
}

void CalculateCameraTransformationMatrix(void)
{
  fxeMatrix Move,XRot,YRot,ZRot;

  // Now lets compute the transformation matrix:

  RotXMatrix(XRot,360.0f-ActiveCamera->ax);
  RotYMatrix(YRot,360.0f-ActiveCamera->ay);
  RotZMatrix(ZRot,360.0f-ActiveCamera->az);
  MoveMatrix(Move,-ActiveCamera->x,-ActiveCamera->y,-ActiveCamera->z);

  Matmult(Move,YRot);
  Matmult(YRot,ZRot);
  Matmult(ZRot,XRot);
  MatCopy(XRot,CameraMatrix);

}


// This function clips invisible objects from the
// rendering pipeline - speeds up rendering a HUGE amount
void ClipObject(fxeObjectHandle ObjectHandle)
{
	float ObjectRadius;
	fxeVertex VerRelCamera;
	fxeVertex TempVertex;

	// When we shall clip objects we need to convert the objects
	// position in world to relative to camera

	// We need to invert the camera angles because when we
	// rotates the object to relative the camera position we must do
	// this in the inverse direction of the camera angles.

	ObjectHandle->Clipped = FXEFALSE;
  ObjectHandle->CompletelyVisible = FXETRUE;


  TempVertex.x = ObjectHandle->x;
  TempVertex.y = ObjectHandle->y;
  TempVertex.z = ObjectHandle->z;

  MatXVer(CameraMatrix,&TempVertex,&VerRelCamera);


	// Rotate the position of the object in the inverse angle of
	// the camera:


	// now we have the position of the object in camera coordinates -
	// lets test the objects bounding box against the viewing volume:

  // Get the radius of the object:
  ObjectRadius = ObjectHandle->ObjectRadius;

  if ((VerRelCamera.z + ObjectRadius) < NearClippingPlane) 
	{
    ObjectHandle->Clipped = FXETRUE;
	}
  else if ((VerRelCamera.z - ObjectRadius) > FarClippingPlane)
	{
		ObjectHandle->Clipped = FXETRUE;
	}
	
  else if ((VerRelCamera.x + ObjectRadius) < (-VerRelCamera.z))
	{
		ObjectHandle->Clipped = FXETRUE;
	}

  else if ((VerRelCamera.x - ObjectRadius) > VerRelCamera.z)
	{
		ObjectHandle->Clipped = FXETRUE;
	}

  else if ((VerRelCamera.y + ObjectRadius) < (-VerRelCamera.z))
	{
		ObjectHandle->Clipped = FXETRUE;
	}

  else if ((VerRelCamera.y - ObjectRadius) > (VerRelCamera.z))
	{
		ObjectHandle->Clipped = FXETRUE;
	}

	
  if ((VerRelCamera.z - ObjectRadius*2) < NearClippingPlane) 
	{
    ObjectHandle->CompletelyVisible = FXEFALSE;
	}
  else if ((VerRelCamera.z + ObjectRadius*2) > FarClippingPlane)
	{
		ObjectHandle->CompletelyVisible = FXEFALSE;
	}
	
  else if ((VerRelCamera.x - ObjectRadius*2) < (-VerRelCamera.z))
	{
		ObjectHandle->CompletelyVisible = FXEFALSE;
	}

  else if ((VerRelCamera.x + ObjectRadius*2) > VerRelCamera.z)
	{
		ObjectHandle->CompletelyVisible = FXEFALSE;
	}

  else if ((VerRelCamera.y - ObjectRadius*2) < (-VerRelCamera.z))
	{
		ObjectHandle->CompletelyVisible = FXEFALSE;
	}

  else if ((VerRelCamera.y + ObjectRadius*2) > (VerRelCamera.z))
	{
		ObjectHandle->CompletelyVisible = FXEFALSE;
	}

}


void CrossProduct3d(fxeVector *Result, fxeVector *Vector1,
	fxeVector *Vector2)
{
	Result->x = (Vector1->y)*(Vector2->z)-(Vector1->z)*(Vector2->y);
	Result->y = -((Vector1->x)*(Vector2->z)-(Vector1->z)*(Vector2->x));
	Result->z = (Vector1->x)*(Vector2->y)-(Vector1->y)*(Vector2->x);
}

float DotProduct3d(fxeVector *Vector1, fxeVector *Vector2)
{
	return Vector1->x*Vector2->x + Vector1->y*Vector2->y +
		Vector1->z*Vector2->z;
}

float VectorLenght(fxeVector *Vector)
{
	return
    fast_fsqrt(Vector->x*Vector->x+Vector->y*Vector->y+Vector->z*Vector->z);
}


// Calculates normals for all polygons in object
void CalculateNormals(fxeObjectHandle ObjectHandle)
{
	int Loop;
	fxePolygon *Poly;
	fxeVector Vector1, Vector2;
	fxeVertexIndex *VIndex;
	fxeVertex *Vertex;
  fxeVertex *CurrVertex, *CurrVertex2, *CurrVertex3;

	Poly = ObjectHandle->PolygonList;
	Vertex = ObjectHandle->VerticesWorld;

  for (Loop=0; Loop<ObjectHandle->NumPolys; Loop++)
  {
			VIndex = Poly[Loop].VertexIndices;
      CurrVertex = &(Vertex[VIndex[2]]);
      CurrVertex2 = &(Vertex[VIndex[1]]);
      CurrVertex3 = &(Vertex[VIndex[0]]);
      Vector1.x = CurrVertex->x-CurrVertex3->x;
      Vector1.y = CurrVertex->y-CurrVertex3->y;
      Vector1.z = CurrVertex->z-CurrVertex3->z;
      Vector2.x = CurrVertex2->x-CurrVertex3->x;
      Vector2.y = CurrVertex2->y-CurrVertex3->y;
      Vector2.z = CurrVertex2->z-CurrVertex3->z;

      CrossProduct3d(&(Poly[Loop].Normal),&Vector2,&Vector1);
	}
}


void RemoveBackfaces(fxeObjectHandle ObjectHandle)
{
	int Loop;
	int Loop2;
	fxePolygon *Poly;
	fxeVector LineOfSight;
	fxeVertexIndex *VIndex;
	fxeVertex *Vertex;
	fxeDrawVertex *CamVertex;
	int Flag;

	Flag = 0;
	Poly = ObjectHandle->PolygonList;
  Vertex = ObjectHandle->VerticesWorld;
	CamVertex = ObjectHandle->VerticesCamera;

	if (ObjectHandle->ShadingType == GOURAUD_SHADING ||
		ObjectHandle->EnviromentMapped == FXETRUE)
	{
		Flag = 1;
		for (Loop=0; Loop<ObjectHandle->NumVertices; Loop++)
		{
			CamVertex[Loop].Used = FXEFALSE;
		}
	}

	for (Loop=0; Loop<ObjectHandle->NumPolys; Loop++)
  {
    // Check if the polygon is doublesided:

    Poly[Loop].Culled = FXEFALSE;

		

    if (Poly[Loop].Clipped == FXEFALSE)
    {

      if (Poly[Loop].DoubleSided == FXEFALSE)
      {

        // Calculate the line of sight vector, the vector
        // going from one point of the polygon to
        // the camera position:

        VIndex = Poly[Loop].VertexIndices;

        LineOfSight.x = ActiveCamera->x-Vertex[VIndex[0]].x;
        LineOfSight.y = ActiveCamera->y-Vertex[VIndex[0]].y;
        LineOfSight.z = ActiveCamera->z-Vertex[VIndex[0]].z;

        // Make Dot product with the normal of the polygon
        // and the light of sight vector, if the sign
        // of the result is positive then the polygon
        // is visible, if not it's invisible:

        if (DotProduct3d(&LineOfSight,&(Poly[Loop].Normal))
          < 0.0)
        {
          Poly[Loop].Culled = FXETRUE;
        }
				else
				{
					if (Flag)
					{
  					for (Loop2=0;Loop2<Poly[Loop].NumVertices;Loop2++)
						{
							CamVertex[Poly[Loop].VertexIndices[Loop2]].Used = FXETRUE;
						}
					}
				}

      }
			else
			{
				if (Flag)
				{
					for (Loop2=0;Loop2<Poly[Loop].NumVertices;Loop2++)
					{
						CamVertex[Poly[Loop].VertexIndices[Loop2]].Used = FXETRUE;
					}
				}
			}

		}
  }
}


//fxeVector *GetPolygonNormal(fxeVertex *Vertex);


void ShadeObject(fxeObjectHandle ObjectHandle)
{
	// This is propably one of the most complex functions in the 3d engine
 	// since it must take into considereration an unlimited amount of
	// lightsources and perform a lot of calculations.

	fxePolygon *Poly;
	fxeVertex *Vertex;
	fxeDrawVertex *CamVertex;
  fxeVertex *VertexList;
	fxeVector LightVector;
	fxeVector AddedNormal;
	fxeVertexIndex *VIndex;
  float LightVectorLenght;
  float Dp3d;
	float ReflectionIntensity;
  fxeColor ConstantColor;
  fxeColor *FlatShadedColor;
  float TotalShadeR, TotalShadeG, TotalShadeB;
  float DistanceAffection;
  fxePolygon *CurrPoly;
  int Loop, Finished;
	float VertexIntensity;
	fxeDynLightsource *pCurrLightsource;
	int VertexIndex;

	Finished = FXEFALSE;

	Vertex = ObjectHandle->VerticesWorld;
  VertexList = ObjectHandle->VertexList;
	CamVertex = ObjectHandle->VerticesCamera;
	Poly = ObjectHandle->PolygonList;

  if (ObjectHandle->ShadingType == FLAT_SHADING)
  {
    // Go through all polygons:
    for (Loop=0; Loop<ObjectHandle->NumPolys; Loop++)
	  {
				// Check if this polygon is clipped or culled:
				if (Poly[Loop].Clipped == FXEFALSE && 
					Poly[Loop].Culled == FXEFALSE)
				{
			
					VIndex = Poly[Loop].VertexIndices;
					ConstantColor.r = Poly[Loop].DiffuseColor.r;
					ConstantColor.g = Poly[Loop].DiffuseColor.g; 
					ConstantColor.b = Poly[Loop].DiffuseColor.b;

					FlatShadedColor = &(Poly[Loop].FlatShadedColor);
					FlatShadedColor->r = 0.0f;
					FlatShadedColor->g = 0.0f;
					FlatShadedColor->b = 0.0f;

					//NormalVectorLenght = VectorLenght(&Poly[Loop].Normal);



					//if (NormalVectorLenght == 0.0f)
					//	NormalVectorLenght = 0.0001f;

					CurrPoly = &Poly[Loop];
					if (Lightsources != NULL)
					{
						// go to first lightsource in world:
						while (Lightsources->Prev != NULL)
							Lightsources = Lightsources->Prev;


						Finished = FXEFALSE;
						// Loop through all lightsources
						while (Finished == FXEFALSE)
						{
							// Calculate reflection level of this Polygon
							// with the current lightsource

							// Calculate lightsource vector:
							LightVector.x = Lightsources->x - Vertex[VIndex[0]].x;
							LightVector.y = Lightsources->y - Vertex[VIndex[0]].y;
							LightVector.z = Lightsources->z - Vertex[VIndex[0]].z;

							// Calculate lenght of lightsource vector:
							// add lightvectorLenght with 0.001 to avoid division by zero
							LightVectorLenght = VectorLenght(&LightVector)+0.001f;

							// Calculate the dot product with the lightsource vector
							// and the normal vector for this polygon

							Dp3d = LightVector.x*Poly[Loop].Normal.x+LightVector.y*Poly[Loop].
								Normal.y+LightVector.z*Poly[Loop].Normal.z;

							ReflectionIntensity = Dp3d / LightVectorLenght;

							if (Poly[Loop].DoubleSided == FXETRUE)
							{
								// check if the Reflectionintensity is negative:
								if (ReflectionIntensity < 0)
									ReflectionIntensity = -ReflectionIntensity;
							}

							ReflectionIntensity = (ReflectionIntensity+1.0f)*0.5f;
						
							// we adds the reflection level the polygon gives
							// from the current lightsource to the total reflection level
							// which is stored in the vertex structure of the object
							// As FxEngine uses RGB lightning the RGB intensity components
							// from the lightsource is used. In the formula also the
							// lenght of the light vector is used so that lightsources
							// near polygon does lit the polygon more and lightsources far
							// away from the polygon lits the polygon less.

							DistanceAffection =
								(Lightsources->DistanceAffection*LightVectorLenght);

							TotalShadeR =
								(ConstantColor.r*(ReflectionIntensity*Lightsources->r))-
								DistanceAffection;

							TotalShadeG =
								(ConstantColor.g*(ReflectionIntensity*Lightsources->g))-
								DistanceAffection;

							TotalShadeB =
								(ConstantColor.b*(ReflectionIntensity*Lightsources->b))-
								DistanceAffection;

							if (TotalShadeR < 0.0f)
								TotalShadeR = 0.0f;

							if (TotalShadeG < 0.0f)
								TotalShadeG = 0.0f;

							if (TotalShadeB < 0.0f)
								TotalShadeB = 0.0f;

							FlatShadedColor->r += TotalShadeR;
							FlatShadedColor->g += TotalShadeG;
							FlatShadedColor->b += TotalShadeB;

							if (Lightsources->Next == NULL)
								Finished = FXETRUE;
							else
								Lightsources = Lightsources->Next;
						}
					}

					if (FlatShadedColor->r < 0.0f)
						FlatShadedColor->r = 0.0f;
					if (FlatShadedColor->g < 0.0f)
						FlatShadedColor->g = 0.0f;
					if (FlatShadedColor->b < 0.0f)
						FlatShadedColor->b = 0.0f;

				

					FlatShadedColor->r += AmbientR + Poly[Loop].AmbientColor.r;
					FlatShadedColor->g += AmbientG + Poly[Loop].AmbientColor.g;
					FlatShadedColor->b += AmbientB + Poly[Loop].AmbientColor.b;
					/*
					FlatShadedColor->r += Poly[Loop].AmbientColor.r;
					FlatShadedColor->g += Poly[Loop].AmbientColor.g;
					FlatShadedColor->b += Poly[Loop].AmbientColor.b;
					*/

					// Make sure light of the Vertex is always smaller or
					// equal to 1.0
					if (FlatShadedColor->r > 1.0f)
						FlatShadedColor->r = 1.0f;

					if (FlatShadedColor->g > 1.0f)
						FlatShadedColor->g = 1.0f;

					if (FlatShadedColor->b > 1.0f)
						FlatShadedColor->b = 1.0f;
				}


    }
  }

  // I have put in the enviroment mapping code here too because we don't
  // need to calculate the addednormal twice for each vertex.

  if (ObjectHandle->ShadingType == GOURAUD_SHADING || ObjectHandle->
    EnviromentMapped == FXETRUE)
  {
		// Loop through the vertex list of the object -
		// check which polygons who shares the vertex, calculate
		// normals for polygon1,polygon2,..,polygonN who shares the vertex
		// and sum these normals together. A dot product between this
		// normalsum and every light source is performed to calculate
		// the RGB intensity of every vertex in the object.


    for (Loop=0; Loop<ObjectHandle->NumVertices; Loop++)
		{
		if (ObjectHandle->VerticesCamera[Loop].Used == FXETRUE)
		{
			/*
			AddedNormal.x = 0.0f;
      AddedNormal.y = 0.0f;
      AddedNormal.z = 0.0f;

      CurrVertex = &(CamVertex[Loop]);


      for (Loop2=0; Loop2<CurrVertex->NumPolys; Loop2++)
      {
        CurrVector =
          &(Poly[CurrVertex->PolygonIndices[Loop2]].Normal);

        AddedNormal.x += CurrVector->x;
        AddedNormal.y += CurrVector->y;
        AddedNormal.z += CurrVector->z;
      }

      NormalVectorLenght = VectorLenght(&AddedNormal);

      if (NormalVectorLenght == 0.0f)
        NormalVectorLenght = 0.0001f;
			*/

			AddedNormal.x = ObjectHandle->VertexNormals[Loop].x;
			AddedNormal.y = ObjectHandle->VertexNormals[Loop].y;
			AddedNormal.z = ObjectHandle->VertexNormals[Loop].z;

      if (ObjectHandle->EnviromentMapped == FXETRUE)
      {
        //AddedNormal2.x = AddedNormal.x/NormalVectorLenght;
        //AddedNormal2.y = AddedNormal.y/NormalVectorLenght;
        ObjectHandle->EnviromentVerts[Loop].u = AddedNormal.x*0.5f+0.5f;
        ObjectHandle->EnviromentVerts[Loop].v = AddedNormal.y*0.5f+0.5f;
      }


      if (ObjectHandle->ShadingType == GOURAUD_SHADING)
      {
      // Do the dot product between this total normal vector and
      // all lightsources in the 3d world

      CamVertex[Loop].r = 0.0f;
      CamVertex[Loop].g = 0.0f;
      CamVertex[Loop].b = 0.0f;

      if (Lightsources != NULL)
      {
        // go to first lightsource in world:
        while (Lightsources->Prev != NULL)
          Lightsources = Lightsources->Prev;

        Finished = FXEFALSE;
        // Loop through all lightsources
        while (Finished == FXEFALSE)
        {
          // Calculate reflection level of this vertex
          // with the current lightsource

          // Calculate lightsource vector:
          LightVector.x = Lightsources->x - Vertex[Loop].x;
          LightVector.y = Lightsources->y - Vertex[Loop].y;
          LightVector.z = Lightsources->z - Vertex[Loop].z;

          // Calculate lenght of lightsource vector:
          // add lightvectorLenght with 0.001 to avoid division by zero

          LightVectorLenght = VectorLenght(&LightVector)+0.001f;

          // Calculate the dot product with the lightsource vector
          // and the added normal vector for this vertex

          Dp3d = LightVector.x*AddedNormal.x+LightVector.y*AddedNormal.y+
            LightVector.z*AddedNormal.z;


          // calculate the level of light that this vertex reflects
          // from the current lightsource: (scale it to 0..1)


            ReflectionIntensity = ((Dp3d /
              (LightVectorLenght)) + 1.0f) *0.5f;


          // we adds the reflection level the vertex gives
          // from the current lightsource to the total reflection level
          // which is stored in the vertex structure of the object
          // As FxEngine uses RGB lightning the RGB intensity components
          // from the lightsource is used. In the formula also the
          // lenght of the light vector is used so that lightsources
          // near vertex does lit the vertex more and lightsources far
          // away from the verted lits the vertex less.

          DistanceAffection = (LightVectorLenght*Lightsources->DistanceAffection);

          TotalShadeR = (Lightsources->r*ReflectionIntensity)
            -DistanceAffection;

          TotalShadeG = (Lightsources->g*ReflectionIntensity)
            -DistanceAffection;

          TotalShadeB = (Lightsources->b*ReflectionIntensity)
            -DistanceAffection;

          if (TotalShadeR < 0.0f)
            TotalShadeR = 0.0f;

          if (TotalShadeG < 0.0f)
            TotalShadeG = 0.0f;

          if (TotalShadeB < 0.0f)
            TotalShadeB = 0.0f;


          CamVertex[Loop].r += TotalShadeR;
          CamVertex[Loop].g += TotalShadeG;
          CamVertex[Loop].b += TotalShadeB;

          if (Lightsources->Next == NULL)
            Finished = FXETRUE;
          else
            Lightsources = Lightsources->Next;
        }


      }

      }
    }
		}
    
  }
	
	if (ObjectHandle->ShadingType == GOURAUD_SHADING)
	{
		memset(ObjectHandle->DynColors,0,
			sizeof(fxeColor)*ObjectHandle->NumVertices);

	// Go through all dynamic lightsources 
	if (FirstDynLightsource != NULL)
	{
		pCurrLightsource = FirstDynLightsource;
		while (pCurrLightsource != NULL)
		{
			// Check if this object is affected by the current lightsource.

			LightVector.x = ObjectHandle->x - pCurrLightsource->x;
			LightVector.y = ObjectHandle->y - pCurrLightsource->y;
			LightVector.z = ObjectHandle->z - pCurrLightsource->z;
			LightVectorLenght = VectorLenght(&LightVector);

			if ((LightVectorLenght-ObjectHandle->ObjectRadius-pCurrLightsource->Radius) <=
				0)
			{
				for (Loop=0;Loop<ObjectHandle->NumVertices;Loop++)
				{
					// Only perform calculations on vertices that will be visible
					LightVector.x = ObjectHandle->VerticesWorld[Loop].x - pCurrLightsource->x;
					LightVector.y = ObjectHandle->VerticesWorld[Loop].y - pCurrLightsource->y;
					LightVector.z = ObjectHandle->VerticesWorld[Loop].z - pCurrLightsource->z;
					LightVectorLenght = VectorLenght(&LightVector);
					VertexIntensity = pCurrLightsource->Radius-LightVectorLenght;
					if (VertexIntensity < 0.0f)
						VertexIntensity = 0.0f;
					else
						VertexIntensity = VertexIntensity * pCurrLightsource->Oor;

					ObjectHandle->DynColors[Loop].r += VertexIntensity*pCurrLightsource->RDiv31;
					ObjectHandle->DynColors[Loop].g += VertexIntensity*pCurrLightsource->GDiv31;
					ObjectHandle->DynColors[Loop].b += VertexIntensity*pCurrLightsource->BDiv31;


				}
			}
			pCurrLightsource = pCurrLightsource->Next;
		}
	}
	}
	else if (ObjectHandle->ShadingType == FLAT_SHADING)
	{
	// Go through all dynamic lightsources 
	if (FirstDynLightsource != NULL)
	{
		pCurrLightsource = FirstDynLightsource;
		while (pCurrLightsource != NULL)
		{
			// Check if this object is affected by the current lightsource.

			LightVector.x = ObjectHandle->x - pCurrLightsource->x;
			LightVector.y = ObjectHandle->y - pCurrLightsource->y;
			LightVector.z = ObjectHandle->z - pCurrLightsource->z;
			LightVectorLenght = VectorLenght(&LightVector);

			if ((LightVectorLenght-ObjectHandle->ObjectRadius-pCurrLightsource->Radius) <=
				0)
			{
				for (Loop=0;Loop<ObjectHandle->NumPolys;Loop++)
				{
					if (ObjectHandle->PolygonList[Loop].Clipped == FXEFALSE &&
						ObjectHandle->PolygonList[Loop].Culled == FXEFALSE)
					{
					// Only perform calculations on vertices that will be visible
					VertexIndex = ObjectHandle->PolygonList[Loop].VertexIndices[0];
					LightVector.x = ObjectHandle->VerticesWorld[VertexIndex].x - pCurrLightsource->x;
					LightVector.y = ObjectHandle->VerticesWorld[VertexIndex].y - pCurrLightsource->y;
					LightVector.z = ObjectHandle->VerticesWorld[VertexIndex].z - pCurrLightsource->z;
					LightVectorLenght = VectorLenght(&LightVector);
					VertexIntensity = pCurrLightsource->Radius-LightVectorLenght;
					if (VertexIntensity < 0.0f)
						VertexIntensity = 0.0f;
					else
						VertexIntensity = VertexIntensity * pCurrLightsource->Oor;

					ObjectHandle->PolygonList[Loop].FlatShadedColor.r += VertexIntensity*pCurrLightsource->RDiv31;
					ObjectHandle->PolygonList[Loop].FlatShadedColor.g += VertexIntensity*pCurrLightsource->GDiv31;
					ObjectHandle->PolygonList[Loop].FlatShadedColor.b += VertexIntensity*pCurrLightsource->BDiv31;

					if (ObjectHandle->PolygonList[Loop].FlatShadedColor.r > 1.0f)
						ObjectHandle->PolygonList[Loop].FlatShadedColor.r = 1.0f;
					if (ObjectHandle->PolygonList[Loop].FlatShadedColor.g > 1.0f)
						ObjectHandle->PolygonList[Loop].FlatShadedColor.g = 1.0f;
					if (ObjectHandle->PolygonList[Loop].FlatShadedColor.b > 1.0f)
						ObjectHandle->PolygonList[Loop].FlatShadedColor.b = 1.0f;
					}

				}
			}
			pCurrLightsource = pCurrLightsource->Next;
		}
	}
		
	}


}




void TransformToCameraMatrix(fxeObjectHandle ObjectHandle)
{
	fxeVertex *VerticesWorld;
	fxeDrawVertex *VerticesCamera;
	int Loop;

	VerticesWorld = ObjectHandle->VerticesWorld;
	VerticesCamera = ObjectHandle->VerticesCamera;

	for(Loop=0 ; Loop<ObjectHandle->NumVertices; Loop++)
	{
    MatXVerToDrawVer(CameraMatrix,&VerticesWorld[Loop],&VerticesCamera[Loop]);
	}
}

void SetupViewFrustrum(void)
{
  // calculate normals and distances of all 6 planes in
  // the view frustrum
 

  NearPlane.Distance = NearClippingPlane;
  NearPlane.Normal.x = 0;
  NearPlane.Normal.y = 0;
  NearPlane.Normal.z = -1;

  FarPlane.Distance = FarClippingPlane;
  FarPlane.Normal.x = 0;
  FarPlane.Normal.y = 0;
  FarPlane.Normal.z = 1;

  LeftPlane.Distance = 0.0;
  RightPlane.Distance = 0.0;
  UpPlane.Distance = 0.0;
  DownPlane.Distance = 0.0;

  LeftPlane.Normal.x = -0.707106781f;
  LeftPlane.Normal.z = -0.707106781f;
  LeftPlane.Normal.y = 0.0;

  RightPlane.Normal.x = 0.707106781f;
  RightPlane.Normal.z = -0.707106781f;
  RightPlane.Normal.y = 0.0;

  UpPlane.Normal.x = 0.0;
  UpPlane.Normal.y = 0.707106781f;
  UpPlane.Normal.z = -0.707106781f;

  DownPlane.Normal.x = 0.0;
  DownPlane.Normal.y = -0.707106781f;
  DownPlane.Normal.z = -0.707106781f;
}

// Copies a vertex from Src to Dst
void CopyVertex(fxeDrawVertex *Src, fxeDrawVertex *Dst)
{
  // both points is inside:
  Dst->x = Src->x;
  Dst->y = Src->y;
  Dst->z = Src->z;
  Dst->r = Src->r;
  Dst->g = Src->g;
  Dst->b = Src->b;
  Dst->u0 = Src->u0;
  Dst->v0 = Src->v0;
  Dst->u1 = Src->u1;
  Dst->v1 = Src->v1;
}

// Copies a vertex from Src to Dst
void CopyLVertex(fxeLDrawVertex *Src, fxeLDrawVertex *Dst)
{
  // both points is inside:
  Dst->x = Src->x;
  Dst->y = Src->y;
  Dst->z = Src->z;
  Dst->u0 = Src->u0;
  Dst->v0 = Src->v0;
  Dst->u1 = Src->u1;
  Dst->v1 = Src->v1;
}


void CopyEnvVer(fxeTextureCoordinate *Src, fxeTextureCoordinate *Dst)
{
  Dst->u = Src->u;
  Dst->v = Src->v;
}

// returns 0 if completely outside plane and 1 if some point or both
// points is inside.
// the clipped line is returned in pOutLine:
int ClipLLineToPlane(fxePlane *pPlane, fxeLLine *pInLine, fxeLLine *pOutLine)
{
  float dp1,dp2;
  float MulValue;

  // take the dot product with the unit normal of the plane and the
  // two points of the line.

  dp1 = pPlane->Normal.x*pInLine->v0.x +
        pPlane->Normal.y*pInLine->v0.y +
        pPlane->Normal.z*pInLine->v0.z;

  dp2 = pPlane->Normal.x*pInLine->v1.x +
        pPlane->Normal.y*pInLine->v1.y +
        pPlane->Normal.z*pInLine->v1.z;

  if (dp1 >= pPlane->Distance && dp2 >= pPlane->Distance)
    return 0;

  else if (dp1 < pPlane->Distance && dp2 < pPlane->Distance)
  {
    // both points is inside:

    CopyLVertex(&pInLine->v0,&pOutLine->v0);
    CopyLVertex(&pInLine->v1,&pOutLine->v1);
    CopyEnvVer(&pInLine->e0,&pOutLine->e0);
    CopyEnvVer(&pInLine->e1,&pOutLine->e1);

    return 2;
  }

  else
  {
    // Check which point in the line is outside/inside

    if (dp1 >= pPlane->Distance)
    {
      // We must find a new point instead of v0 in pInLine

      MulValue = (pPlane->Distance-dp2) / (dp1-dp2);

      // Get new geometry data:
      pOutLine->v0.x = pInLine->v1.x +
        (pInLine->v0.x - pInLine->v1.x) * MulValue;
      pOutLine->v0.y = pInLine->v1.y +
        (pInLine->v0.y - pInLine->v1.y) * MulValue;
      pOutLine->v0.z = pInLine->v1.z +
        (pInLine->v0.z - pInLine->v1.z) * MulValue;
/*
      pOutLine->v0.r = pInLine->v1.r +
        (pInLine->v0.r - pInLine->v1.r) * MulValue;
      pOutLine->v0.g = pInLine->v1.g +
        (pInLine->v0.g - pInLine->v1.g) * MulValue;
      pOutLine->v0.b = pInLine->v1.b +
        (pInLine->v0.b - pInLine->v1.b) * MulValue;
*/
      // Get new texture coordinates:
      pOutLine->v0.u0 = pInLine->v1.u0 +
        (pInLine->v0.u0 - pInLine->v1.u0) * MulValue;
      pOutLine->v0.v0 = pInLine->v1.v0 +
        (pInLine->v0.v0 - pInLine->v1.v0) * MulValue;
      pOutLine->v0.u1 = pOutLine->v0.u0;
      pOutLine->v0.v1 = pOutLine->v0.v0;

      pOutLine->e0.u = pInLine->e1.u +
        (pInLine->e0.u - pInLine->e1.u) * MulValue;
      pOutLine->e0.v = pInLine->e1.v +
        (pInLine->e0.v - pInLine->e1.v) * MulValue;

        

      // Just copy the other vertex since it is inside the plane
      // and don't need to be clipped:
      CopyEnvVer(&pInLine->e1,&pOutLine->e1);
      CopyLVertex(&pInLine->v1,&pOutLine->v1);
    }            
    else
    {
      // We must find a new point instead of v1 in pInLine

      MulValue = (pPlane->Distance-dp1) / (dp2-dp1);

      // Get new geometry data:
      pOutLine->v1.x = pInLine->v0.x +
        (pInLine->v1.x - pInLine->v0.x) * MulValue;
      pOutLine->v1.y = pInLine->v0.y +
        (pInLine->v1.y - pInLine->v0.y) * MulValue;
      pOutLine->v1.z = pInLine->v0.z +
        (pInLine->v1.z - pInLine->v0.z) * MulValue;
/*
      pOutLine->v1.r = pInLine->v0.r +
        (pInLine->v1.r - pInLine->v0.r) * MulValue;
      pOutLine->v1.g = pInLine->v0.g +
        (pInLine->v1.g - pInLine->v0.g) * MulValue;
      pOutLine->v1.b = pInLine->v0.b +
        (pInLine->v1.b - pInLine->v0.b) * MulValue;
*/
      // Get new texture coordinates:
      pOutLine->v1.u0 = pInLine->v0.u0 +
        (pInLine->v1.u0 - pInLine->v0.u0) * MulValue;
      pOutLine->v1.v0 = pInLine->v0.v0 +
        (pInLine->v1.v0 - pInLine->v0.v0) * MulValue;
      pOutLine->v1.u1 = pOutLine->v1.u0;
      pOutLine->v1.v1 = pOutLine->v1.v0;


      pOutLine->e1.u = pInLine->e0.u +
        (pInLine->e1.u - pInLine->e0.u) * MulValue;
      pOutLine->e1.v = pInLine->e0.v +
        (pInLine->e1.v - pInLine->e0.v) * MulValue;

      // Just copy the other vertex since it is inside the plane
      // and don't need to be clipped:

      CopyEnvVer(&pInLine->e0,&pOutLine->e0);
      CopyLVertex(&pInLine->v0,&pOutLine->v0);
    }
  return 1;
  }
}


// returns 0 if completely outside plane and 1 if some point or both
// points is inside.
// the clipped line is returned in pOutLine:
int ClipLineToPlane(fxePlane *pPlane, fxeLine *pInLine, fxeLine *pOutLine)
{
  float dp1,dp2;
  float MulValue;

  // take the dot product with the unit normal of the plane and the
  // two points of the line.

  dp1 = pPlane->Normal.x*pInLine->v0.x +
        pPlane->Normal.y*pInLine->v0.y +
        pPlane->Normal.z*pInLine->v0.z;

  dp2 = pPlane->Normal.x*pInLine->v1.x +
        pPlane->Normal.y*pInLine->v1.y +
        pPlane->Normal.z*pInLine->v1.z;

  if (dp1 >= pPlane->Distance && dp2 >= pPlane->Distance)
    return 0;

  else if (dp1 < pPlane->Distance && dp2 < pPlane->Distance)
  {
    // both points is inside:

    CopyVertex(&pInLine->v0,&pOutLine->v0);
    CopyVertex(&pInLine->v1,&pOutLine->v1);
    CopyEnvVer(&pInLine->e0,&pOutLine->e0);
    CopyEnvVer(&pInLine->e1,&pOutLine->e1);

    return 2;
  }

  else
  {
    // Check which point in the line is outside/inside

    if (dp1 >= pPlane->Distance)
    {
      // We must find a new point instead of v0 in pInLine

      MulValue = (pPlane->Distance-dp2) / (dp1-dp2);

      // Get new geometry data:
      pOutLine->v0.x = pInLine->v1.x +
        (pInLine->v0.x - pInLine->v1.x) * MulValue;
      pOutLine->v0.y = pInLine->v1.y +
        (pInLine->v0.y - pInLine->v1.y) * MulValue;
      pOutLine->v0.z = pInLine->v1.z +
        (pInLine->v0.z - pInLine->v1.z) * MulValue;

      pOutLine->v0.r = pInLine->v1.r +
        (pInLine->v0.r - pInLine->v1.r) * MulValue;
      pOutLine->v0.g = pInLine->v1.g +
        (pInLine->v0.g - pInLine->v1.g) * MulValue;
      pOutLine->v0.b = pInLine->v1.b +
        (pInLine->v0.b - pInLine->v1.b) * MulValue;

      // Get new texture coordinates:
      pOutLine->v0.u0 = pInLine->v1.u0 +
        (pInLine->v0.u0 - pInLine->v1.u0) * MulValue;
      pOutLine->v0.v0 = pInLine->v1.v0 +
        (pInLine->v0.v0 - pInLine->v1.v0) * MulValue;
      pOutLine->v0.u1 = pOutLine->v0.u0;
      pOutLine->v0.v1 = pOutLine->v0.v0;

      pOutLine->e0.u = pInLine->e1.u +
        (pInLine->e0.u - pInLine->e1.u) * MulValue;
      pOutLine->e0.v = pInLine->e1.v +
        (pInLine->e0.v - pInLine->e1.v) * MulValue;

        

      // Just copy the other vertex since it is inside the plane
      // and don't need to be clipped:
      CopyEnvVer(&pInLine->e1,&pOutLine->e1);
      CopyVertex(&pInLine->v1,&pOutLine->v1);
    }            
    else
    {
      // We must find a new point instead of v1 in pInLine

      MulValue = (pPlane->Distance-dp1) / (dp2-dp1);

      // Get new geometry data:
      pOutLine->v1.x = pInLine->v0.x +
        (pInLine->v1.x - pInLine->v0.x) * MulValue;
      pOutLine->v1.y = pInLine->v0.y +
        (pInLine->v1.y - pInLine->v0.y) * MulValue;
      pOutLine->v1.z = pInLine->v0.z +
        (pInLine->v1.z - pInLine->v0.z) * MulValue;

      pOutLine->v1.r = pInLine->v0.r +
        (pInLine->v1.r - pInLine->v0.r) * MulValue;
      pOutLine->v1.g = pInLine->v0.g +
        (pInLine->v1.g - pInLine->v0.g) * MulValue;
      pOutLine->v1.b = pInLine->v0.b +
        (pInLine->v1.b - pInLine->v0.b) * MulValue;

      // Get new texture coordinates:
      pOutLine->v1.u0 = pInLine->v0.u0 +
        (pInLine->v1.u0 - pInLine->v0.u0) * MulValue;
      pOutLine->v1.v0 = pInLine->v0.v0 +
        (pInLine->v1.v0 - pInLine->v0.v0) * MulValue;
      pOutLine->v1.u1 = pOutLine->v1.u0;
      pOutLine->v1.v1 = pOutLine->v1.v0;


      pOutLine->e1.u = pInLine->e0.u +
        (pInLine->e1.u - pInLine->e0.u) * MulValue;
      pOutLine->e1.v = pInLine->e0.v +
        (pInLine->e1.v - pInLine->e0.v) * MulValue;

      // Just copy the other vertex since it is inside the plane
      // and don't need to be clipped:

      CopyEnvVer(&pInLine->e0,&pOutLine->e0);
      CopyVertex(&pInLine->v0,&pOutLine->v0);
    }
  return 1;
  }
}

/*
// This routine shall only be called when a polygon is intersecting with
// the plane, this is for speed reasons.
void ClipPolygonToPlane(fxePlane *pPlane, fxePolygon *pPolygon,
  fxeObject *pObject)
{
  int Loop;
  fxeLine ClipLine, ClipLine2;
  int NAddedVerts;
  int NextVert;
  fxeDrawVertex TempVertex[200];
  fxeTextureCoordinate TempEnvVer[200];
  int Result;
  fxeVector CmpVertex;
	int ClipCount;

  NAddedVerts = 0;

  if (pPolygon->PartlyClipped == FXETRUE)
  {
    for (Loop=0;Loop<pPolygon->NumAddedVertices;Loop++)
    {
      NextVert = (Loop+1)%pPolygon->NumAddedVertices;

      // We creates a line going from the current point to the
      // next point:

      CopyVertex(&pPolygon->ClippedVertices[Loop],&ClipLine.v0);
      CopyVertex(&pPolygon->ClippedVertices[NextVert],&ClipLine.v1);
      CopyEnvVer(&pPolygon->ClippedEnvVer[Loop],&ClipLine.e0);
      CopyEnvVer(&pPolygon->ClippedEnvVer[NextVert],&ClipLine.e1);

      Result = ClipLineToPlane(pPlane,&ClipLine,&ClipLine2);

      switch(Result)
      {
        case 1:
          CopyVertex(&ClipLine2.v0,&TempVertex[NAddedVerts]);
          CopyVertex(&ClipLine2.v1,&TempVertex[NAddedVerts+1]);
          CopyEnvVer(&ClipLine2.e0,&TempEnvVer[NAddedVerts]);
          CopyEnvVer(&ClipLine2.e1,&TempEnvVer[NAddedVerts+1]);
          NAddedVerts+=2;
          break;

        case 2:
          CopyVertex(&ClipLine2.v1,&TempVertex[NAddedVerts]);
          CopyEnvVer(&ClipLine2.e1,&TempEnvVer[NAddedVerts]);
          NAddedVerts++;
          break;
      }
    }
    // Copy all temporary vertices that not is duplicated 
		// to ClippedVertices
		CmpVertex.x = 100000.0f;
		CmpVertex.y = 100000.0f;
		CmpVertex.z = 100000.0f;
		ClipCount = 0;

    for (Loop=0;Loop<NAddedVerts;Loop++)
    {
			if (TempVertex[Loop].x == CmpVertex.x && TempVertex[Loop].y ==
				CmpVertex.y && TempVertex[Loop].z == CmpVertex.z)
			{
				CmpVertex.x = TempVertex[Loop].x;
				CmpVertex.y = TempVertex[Loop].y;
				CmpVertex.z = TempVertex[Loop].z;
			}
			else
			{
				CopyVertex(&TempVertex[Loop],&pPolygon->ClippedVertices[ClipCount]);
				CopyEnvVer(&TempEnvVer[Loop],&pPolygon->ClippedEnvVer[ClipCount]);
				ClipCount++;
				CmpVertex.x = TempVertex[Loop].x;
				CmpVertex.y = TempVertex[Loop].y;
				CmpVertex.z = TempVertex[Loop].z;
			}
    }
    pPolygon->NumAddedVertices = ClipCount;
  }
  else
  {
    // We have our vertices in the objects vertices camera member
    for (Loop=0;Loop<pPolygon->NumVertices;Loop++)
    {
      NextVert = (Loop+1)%pPolygon->NumVertices;

      // We creates a line going from the current point to the
      // next point:

      CopyVertex(&pObject->VerticesCamera[pPolygon->VertexIndices[Loop]],
        &ClipLine.v0);
      CopyEnvVer(&pObject->EnviromentVerts[pPolygon->VertexIndices[Loop]],
        &ClipLine.e0);

      // Copy texture u,v coordinates from polygons texture coordinates:
      ClipLine.v0.u0 = pPolygon->TextureCoordinates[Loop].u;
      ClipLine.v0.v0 = pPolygon->TextureCoordinates[Loop].v;
      ClipLine.v0.u1 = ClipLine.v0.u0;
      ClipLine.v0.v1 = ClipLine.v0.v0;

      CopyVertex(&pObject->VerticesCamera[pPolygon->VertexIndices[NextVert]],
        &ClipLine.v1);
      CopyEnvVer(&pObject->EnviromentVerts[pPolygon->VertexIndices[NextVert]],
        &ClipLine.e1);

      ClipLine.v1.u0 = pPolygon->TextureCoordinates[NextVert].u;
      ClipLine.v1.v0 = pPolygon->TextureCoordinates[NextVert].v;
      ClipLine.v1.u1 = ClipLine.v1.u0;
      ClipLine.v1.v1 = ClipLine.v1.v0;

      Result = ClipLineToPlane(pPlane,&ClipLine,&ClipLine2);

      switch(Result)
      {
        case 1:
          CopyVertex(&ClipLine2.v0,&TempVertex[NAddedVerts]);
          CopyVertex(&ClipLine2.v1,&TempVertex[NAddedVerts+1]);
          CopyEnvVer(&ClipLine2.e0,&TempEnvVer[NAddedVerts]);
          CopyEnvVer(&ClipLine2.e1,&TempEnvVer[NAddedVerts+1]);
          NAddedVerts+=2;
          break;

        case 2:
          CopyVertex(&ClipLine2.v1,&TempVertex[NAddedVerts]);
          CopyEnvVer(&ClipLine2.e1,&TempEnvVer[NAddedVerts]);
          NAddedVerts++;
          break;

      }

    }

    // Copy all temporary vertices to ClippedVertices
    if (NAddedVerts!=0)
    {
			    // Copy all temporary vertices that not is duplicated 
		// to ClippedVertices
		
		CmpVertex.x = 100000.0f;
		CmpVertex.y = 100000.0f;
		CmpVertex.z = 100000.0f;
		ClipCount = 0;

    for (Loop=0;Loop<NAddedVerts;Loop++)
    {
			if (TempVertex[Loop].x == CmpVertex.x && TempVertex[Loop].y ==
				CmpVertex.y && TempVertex[Loop].z == CmpVertex.z)
			{
				CmpVertex.x = TempVertex[Loop].x;
				CmpVertex.y = TempVertex[Loop].y;
				CmpVertex.z = TempVertex[Loop].z;
			}
			else
			{
				CopyVertex(&TempVertex[Loop],&pPolygon->ClippedVertices[ClipCount]);
				CopyEnvVer(&TempEnvVer[Loop],&pPolygon->ClippedEnvVer[ClipCount]);
				ClipCount++;
				CmpVertex.x = TempVertex[Loop].x;
				CmpVertex.y = TempVertex[Loop].y;
				CmpVertex.z = TempVertex[Loop].z;
			}
    }

      pPolygon->NumAddedVertices = ClipCount;
      pPolygon->PartlyClipped = FXETRUE;
    }
    else
      pPolygon->Clipped = FXETRUE;
  }

}
*/

/*
void ClipIntersectedPolygons(fxeObjectHandle ObjectHandle)
{
  int Loop;

  for (Loop=0; Loop<ObjectHandle->NumPolys; Loop++)
  {
    if (ObjectHandle->PolygonList[Loop].PartlyClipped == FXETRUE)
    {

      ObjectHandle->PolygonList[Loop].PartlyClipped = FXEFALSE;
      // Clip polygon to all planes:


      // Check if we need to clip to Near Plane:
      if ((ObjectHandle->PolygonList[Loop].PartlyClipStatus & 133152) != 0)
      {
        ClipPolygonToPlane(&NearPlane ,&ObjectHandle->PolygonList[Loop],
          ObjectHandle);
      }

      if ((ObjectHandle->PolygonList[Loop].PartlyClipStatus & 66576) != 0)
      {
        ClipPolygonToPlane(&FarPlane  ,&ObjectHandle->PolygonList[Loop],
          ObjectHandle);
      }

      if ((ObjectHandle->PolygonList[Loop].PartlyClipStatus & 33288) != 0)
      {
        ClipPolygonToPlane(&UpPlane   ,&ObjectHandle->PolygonList[Loop],
          ObjectHandle);
      }

      if ((ObjectHandle->PolygonList[Loop].PartlyClipStatus & 16644) != 0)
      {
        ClipPolygonToPlane(&DownPlane ,&ObjectHandle->PolygonList[Loop],
          ObjectHandle);
      }

      if ((ObjectHandle->PolygonList[Loop].PartlyClipStatus & 4161) != 0)
      {
        ClipPolygonToPlane(&LeftPlane ,&ObjectHandle->PolygonList[Loop],
          ObjectHandle);
      }

      if ((ObjectHandle->PolygonList[Loop].PartlyClipStatus & 8322) != 0)
      {
        ClipPolygonToPlane(&RightPlane,&ObjectHandle->PolygonList[Loop],
          ObjectHandle);
      }
    }
  }
}
*/


void ClipPolygon(fxeObjectHandle ObjectHandle, fxePolygon *pPolygon)
{
  int Loop;
	float x,y,z;
	int value;

  pPolygon->Clipped = FXEFALSE;
  pPolygon->PartlyClipped = FXEFALSE;
  pPolygon->PartlyClipStatus = 0;

	if (ObjectHandle->CompletelyVisible == FXETRUE)
		return;

  for (Loop=0;Loop<3;Loop++)
  {
		x = ObjectHandle->VerticesCamera[pPolygon->VertexIndices[Loop]].x;
		y = ObjectHandle->VerticesCamera[pPolygon->VertexIndices[Loop]].y;
		z = ObjectHandle->VerticesCamera[pPolygon->VertexIndices[Loop]].z;
		value = Loop*6;

		// 2^0 + 2^6 + 2^12 = 4161
    pPolygon->PartlyClipStatus = pPolygon->PartlyClipStatus | 
      ((x < (-z)) << (value));

		// 2^1 + 2^7 + 2^13 = 8322
    pPolygon->PartlyClipStatus = pPolygon->PartlyClipStatus | 
			((x > z) << (1+value));

		// 2^2 + 2^8 + 2^14 = 16644
    pPolygon->PartlyClipStatus = pPolygon->PartlyClipStatus | 
			((y < (-z)) << (2+value));

		// 2^3 + 2^9 + 2^15 = 33288
    pPolygon->PartlyClipStatus = pPolygon->PartlyClipStatus | 
			((y > z) << (3+value));

		// 2^4 + 2^10+ 2^16 = 66576
    pPolygon->PartlyClipStatus = pPolygon->PartlyClipStatus | 
			((z > FarClippingPlane) << (4+value));

		// 2^5 + 2^11+ 2^17 = 133152
    pPolygon->PartlyClipStatus = pPolygon->PartlyClipStatus |
			((z < NearClippingPlane) << (5+value));
  }

  if (pPolygon->PartlyClipStatus == 0)
    return;

 if ((pPolygon->PartlyClipStatus & 133152) == 133152 ||
    (pPolygon->PartlyClipStatus & 66576) == 66576  ||
    (pPolygon->PartlyClipStatus & 33288) == 33288  ||
    (pPolygon->PartlyClipStatus & 16644) == 16644  ||
    (pPolygon->PartlyClipStatus & 8322)  == 8322   ||
    (pPolygon->PartlyClipStatus & 4161)  == 4161)
  {
    pPolygon->Clipped = FXETRUE;
  }

  else if (pPolygon->PartlyClipStatus != 0)
  {
    pPolygon->PartlyClipped = FXETRUE;
  }
}

// Clips polygons in camera space:
void ClipPolygons(fxeObjectHandle ObjectHandle)
{
  int Loop;

	
  for (Loop=0; Loop<ObjectHandle->NumPolys; Loop++)
  {
			ClipPolygon(ObjectHandle,&ObjectHandle->PolygonList[Loop]);
  }
}


/*
void ClipPolygonsToNearPlane(fxeObjectHandle ObjectHandle)
{
  int NumPolygons;
  int Loop, Loop2;
  fxePolygon *PolygonList;
  fxeVertex *VerticesCamera;
  fxeVertexIndex *VertexIndices;
  fxeVertex *CurrVertex;
  float MultValue;
  int NumVertices;
  int DestVertex;
  int Nextvert;
  int Curin, Nextin;
  int CurZ, NextZ;
  fxeVertex *ClippedVertices;
  fxeTextureCoordinate *ClippedTexCoors;
  fxeTextureCoordinate *TextureCoors;
  fxeTextureCoordinate *CurrTexCoor;

  PolygonList = ObjectHandle->PolygonList;
  NumPolygons = ObjectHandle->NumPolys;
  VerticesCamera = ObjectHandle->VerticesCamera;

  for (Loop=0; Loop<NumPolygons ; Loop++)
  {
    if (PolygonList[Loop].PartlyClippedZ == FXETRUE &&
      PolygonList[Loop].Culled == FXEFALSE)
    {
        // Go through all edges of the polygon

        NumVertices = PolygonList[Loop].NumVertices;

        PolygonList[Loop].NumAddedVertices = 0;

        DestVertex = 0;

        ClippedVertices = PolygonList[Loop].ClippedVertices;

        ClippedTexCoors = PolygonList[Loop].ClippedTextureCoordinates;

        TextureCoors = PolygonList[Loop].TextureCoordinates;

        VertexIndices = PolygonList[Loop].VertexIndices;

        CurZ = VerticesCamera[VertexIndices[0]].z;
        Curin = (CurZ >= NearClippingPlane);


        for (Loop2=0; Loop2<NumVertices; Loop2++)
        {
          Nextvert = (Loop2 + 1) % NumVertices;

          if (Curin)
          {
            CurrVertex = &(VerticesCamera[VertexIndices[Loop2]]);

            ClippedVertices[DestVertex].x = CurrVertex->x;
            ClippedVertices[DestVertex].y = CurrVertex->y;
            ClippedVertices[DestVertex].z = CurrVertex->z;
            
            if (PolygonList[Loop].TexMode != NOT_TEXTURED)
            {
              CurrTexCoor = &(TextureCoors[Loop2]);
              ClippedTexCoors[DestVertex].u = CurrTexCoor->u;
              ClippedTexCoors[DestVertex].v = CurrTexCoor->v;
            }

            ClippedVertices[DestVertex].r = CurrVertex->r;
            ClippedVertices[DestVertex].g = CurrVertex->g;
            ClippedVertices[DestVertex].b = CurrVertex->b;

            PolygonList[Loop].NumAddedVertices++;

            DestVertex++;
          }

          NextZ = VerticesCamera[VertexIndices[Nextvert]].z;
          Nextin = (NextZ >= NearClippingPlane);

          if (Curin != Nextin)
          {
            MultValue = (NearClippingPlane - CurZ) /
              (NextZ - CurZ);

            CurrVertex = &(VerticesCamera[VertexIndices[Loop2]]);

            ClippedVertices[DestVertex].x = CurrVertex->x +
              ((VerticesCamera[VertexIndices[Nextvert]].x - CurrVertex->x)
              * MultValue);

            ClippedVertices[DestVertex].y = CurrVertex->y +
              ((VerticesCamera[VertexIndices[Nextvert]].y - CurrVertex->y)
              * MultValue);
            
            ClippedVertices[DestVertex].z = CurrVertex->z +
              ((VerticesCamera[VertexIndices[Nextvert]].z - CurrVertex->z)
              * MultValue);

            if (PolygonList[Loop].TexMode != NOT_TEXTURED)
            {
              CurrTexCoor = &(TextureCoors[Loop2]);

              ClippedTexCoors[DestVertex].u = CurrTexCoor->u +
                ((TextureCoors[Nextvert].u - CurrTexCoor->u)
                * MultValue);

              ClippedTexCoors[DestVertex].v = CurrTexCoor->v +
                ((TextureCoors[Nextvert].v - CurrTexCoor->v)
                * MultValue);
            }

            ClippedVertices[DestVertex].r = CurrVertex->r;

            ClippedVertices[DestVertex].g = CurrVertex->g;

            ClippedVertices[DestVertex].b = CurrVertex->b;

            PolygonList[Loop].NumAddedVertices++;

            DestVertex++;
          }

          CurZ = NextZ;
          Curin = Nextin;
        }
      }
    }
}
*/

/*
int SortPolygonsCallback(const void **ppPolygon1,
  const void **ppPolygon2)
{
  int Loop;
  fxePolygonPtr pPolygon1 = (fxePolygonPtr)*ppPolygon1;
  fxePolygonPtr pPolygon2 = (fxePolygonPtr)*ppPolygon2;
  fxeVertex *VerticesCamera;
  float Z1,Z2;

  Z1 = 0.0f;
  Z1 = 0.0f;

  VerticesCamera = CurrentObject->VerticesCamera;

  // Calculate average z for polygon:

  for (Loop=0; Loop<pPolygon1->NumVertices; Loop++)
  {
    Z1 += VerticesCamera[pPolygon1->VertexIndices[Loop]].z;
  }

  for (Loop=0; Loop<pPolygon2->NumVertices; Loop++)
  {
    Z2 += VerticesCamera[pPolygon2->VertexIndices[Loop]].z;
  }

  Z1 /= pPolygon1->NumVertices;
  Z2 /= pPolygon2->NumVertices;

  if (Z1 > Z2)
    return -1;
  else if (Z1 < Z2)
    return 1;
}
*/

/*
// Calculates minimum z value of polygon. Checks if polygon is clipped or
// not. Takes z from clippedvertices if clipped and from VerticesCamera
// if not clipped.
float PolyCalculateMinimumZ(fxePolygon *pPolygon, fxeDrawVertex *VerticesCamera)
{
  float MinZ;
  int Loop;

  MinZ = 65535.0f;

  if (pPolygon->PartlyClipped == FXETRUE)
  {
    for (Loop=0;Loop<pPolygon->NumAddedVertices;Loop++)
    {
      if (pPolygon->ClippedVertices[Loop].z < MinZ)
        MinZ = pPolygon->ClippedVertices[Loop].z;
    }
  }
  else
  {
    for (Loop=0;Loop<pPolygon->NumVertices;Loop++)
    {
      if (VerticesCamera[pPolygon->VertexIndices[Loop]].z < MinZ)
        MinZ = VerticesCamera[pPolygon->VertexIndices[Loop]].z;
    }
  }
  return MinZ;
}
*/

/*
void SortTranslucentPolygons(fxeObjectHandle ObjectHandle)
{
  int Loop;
  fxePolygon *pPolygon;
  fxeDrawVertex *VerticesCamera;
  fxePolygonList *pCurrSortNode;
  int IsNull;
  int FoundNull;

	if (ObjectHandle->Transparent == FXETRUE)
	{

		pPolygon = ObjectHandle->PolygonList;
		VerticesCamera = ObjectHandle->VerticesCamera;

		// Add first polygon to sorted list:
		ObjectHandle->pPolygonSortList = &SortedPolygons[0];

		ObjectHandle->pPolygonSortList->Next = NULL;
		ObjectHandle->pPolygonSortList->Prev = NULL;
		ObjectHandle->pPolygonSortList->pPolygon = &ObjectHandle->PolygonList[0];

		// Calculate minimum z for this sorted polygon:
		ObjectHandle->pPolygonSortList->MinZ =
		PolyCalculateMinimumZ(&ObjectHandle->PolygonList[0],VerticesCamera);

		pCurrSortNode = ObjectHandle->pPolygonSortList;


	  for (Loop=1;Loop<ObjectHandle->NumPolys;Loop++)
		{
			// Calculate minimumZ for this polygon:

			SortedPolygons[Loop].MinZ =
				PolyCalculateMinimumZ(&ObjectHandle->PolygonList[Loop],VerticesCamera);

			if (SortedPolygons[Loop].MinZ > pCurrSortNode->MinZ)
			{
				IsNull = FXETRUE;
				FoundNull = FXEFALSE;
				while(FoundNull == FXEFALSE && IsNull == FXETRUE)
				{
					if (SortedPolygons[Loop].MinZ <= pCurrSortNode->MinZ)
					{
						SortedPolygons[Loop].Prev = pCurrSortNode;
						SortedPolygons[Loop].Next = pCurrSortNode->Next;
						if (SortedPolygons[Loop].Next != NULL)
							((fxePolygonList*)SortedPolygons[Loop].Next)->Prev = &SortedPolygons[Loop];
						pCurrSortNode->Next = &SortedPolygons[Loop];
						SortedPolygons[Loop].pPolygon = &ObjectHandle->PolygonList[Loop];
						pCurrSortNode = &SortedPolygons[Loop];
						IsNull = FXEFALSE;
					}
					else
					{
						if (pCurrSortNode->Prev != NULL)
							pCurrSortNode = pCurrSortNode->Prev;
						else
							FoundNull = FXETRUE;
					}
				}

				if (IsNull == FXETRUE)
				{
					SortedPolygons[Loop].Prev = NULL;
					SortedPolygons[Loop].Next = pCurrSortNode;
					pCurrSortNode->Prev = &SortedPolygons[Loop];
					SortedPolygons[Loop].pPolygon = &ObjectHandle->PolygonList[Loop];
					pCurrSortNode = &SortedPolygons[Loop];
				}

			}
			else if (SortedPolygons[Loop].MinZ <= pCurrSortNode->MinZ)
			{
				IsNull = FXETRUE;
				FoundNull = FXEFALSE;

				while(FoundNull == FXEFALSE && IsNull == FXETRUE)
				{
					if (SortedPolygons[Loop].MinZ > pCurrSortNode->MinZ)
					{
						SortedPolygons[Loop].Next = pCurrSortNode;
						SortedPolygons[Loop].Prev = pCurrSortNode->Prev;
						if (pCurrSortNode->Prev != NULL)
							((fxePolygonList*)pCurrSortNode->Prev)->Next = &SortedPolygons[Loop];
						pCurrSortNode->Prev = &SortedPolygons[Loop];
						SortedPolygons[Loop].pPolygon = &ObjectHandle->PolygonList[Loop];
						pCurrSortNode = &SortedPolygons[Loop];
						IsNull = FXEFALSE;
					}
					else
					{
						if (pCurrSortNode->Next != NULL)
							pCurrSortNode = pCurrSortNode->Next;
						else
							FoundNull = FXETRUE;
					}

				}

				if (IsNull == FXETRUE)
				{
					SortedPolygons[Loop].Next = NULL;
					SortedPolygons[Loop].Prev = pCurrSortNode;
					pCurrSortNode->Next = &SortedPolygons[Loop];
					SortedPolygons[Loop].pPolygon = &ObjectHandle->PolygonList[Loop];
					pCurrSortNode = &SortedPolygons[Loop];
				}
          
			}
		}
	}
}
*/


fxeDrawVertex ClippedVertices[50];
fxeDrawVertex ClippedVertices2[50];
fxeTextureCoordinate ClippedEnvVertices[50];
fxeTextureCoordinate ClippedEnvVertices2[50];
int NumClippedVertices;
char ClipEnv;


void ClipPolygonLeft(int *NumVertices, 
	fxeDrawVertex *pInVertex, fxeDrawVertex *pOutVertex)
{
	int CurrVertex;
	int NextVertex;
	int VertexCount;
	int FirstVertexResult, NextVertexResult;
	float Length;
	float Difference;

	VertexCount = 0;

	if (NumVertices[0] == 0)
		return;

	for (CurrVertex=0;CurrVertex<NumVertices[0];CurrVertex++)
	{
		NextVertex = (CurrVertex+1) % NumVertices[0];

		FirstVertexResult = pInVertex[CurrVertex].x >= ProjectStartX;
		NextVertexResult = pInVertex[NextVertex].x >= ProjectStartX;

		if ((!FirstVertexResult) && NextVertexResult)
		{
			Length = pInVertex[NextVertex].x - pInVertex[CurrVertex].x;
			Difference = ProjectStartX-pInVertex[CurrVertex].x;
			Difference = 1.0f/Length*Difference;

			// Setup vertices
			pOutVertex[VertexCount].x = ProjectStartX;

			pOutVertex[VertexCount].y = pInVertex[CurrVertex].y + 
				(pInVertex[NextVertex].y - pInVertex[CurrVertex].y)*Difference;

			pOutVertex[VertexCount].q = pInVertex[CurrVertex].q +
				(pInVertex[NextVertex].q - pInVertex[CurrVertex].q)*Difference;

			// Setup colors
			pOutVertex[VertexCount].r = pInVertex[CurrVertex].r +
				(pInVertex[NextVertex].r - pInVertex[CurrVertex].r)*Difference;

			pOutVertex[VertexCount].g = pInVertex[CurrVertex].g +
				(pInVertex[NextVertex].g - pInVertex[CurrVertex].g)*Difference;

			pOutVertex[VertexCount].b = pInVertex[CurrVertex].b +
				(pInVertex[NextVertex].b - pInVertex[CurrVertex].b)*Difference;

			// Setup texture coordinates
			pOutVertex[VertexCount].u0 = pInVertex[CurrVertex].u0 +
				(pInVertex[NextVertex].u0 - pInVertex[CurrVertex].u0)*Difference;

			pOutVertex[VertexCount].v0 = pInVertex[CurrVertex].v0 +
				(pInVertex[NextVertex].v0 - pInVertex[CurrVertex].v0)*Difference;

			if (ClipEnv)
			{
				// Clip enviroment mapping vertices if needed.
				ClippedEnvVertices2[VertexCount].u = ClippedEnvVertices[CurrVertex].u+
					(ClippedEnvVertices[NextVertex].u - ClippedEnvVertices[CurrVertex].u)*Difference;

				ClippedEnvVertices2[VertexCount].v = ClippedEnvVertices[CurrVertex].v+
					(ClippedEnvVertices[NextVertex].v - ClippedEnvVertices[CurrVertex].v)*Difference;
			}


			VertexCount++;

		}
		else if (FirstVertexResult && (!NextVertexResult))
		{
			// Copy first vertex as it is inside clipping frustrum
			memcpy(&pOutVertex[VertexCount],
				&pInVertex[CurrVertex],sizeof(fxeDrawVertex));

			if (ClipEnv)
			{
				memcpy(&ClippedEnvVertices2[VertexCount],
					&ClippedEnvVertices[CurrVertex],sizeof(fxeTextureCoordinate));
			}

			VertexCount++;

			Length = pInVertex[NextVertex].x - pInVertex[CurrVertex].x;
			Difference = ProjectStartX-pInVertex[CurrVertex].x;
			Difference = 1.0f/Length*Difference;

			// Setup vertices
			pOutVertex[VertexCount].x = ProjectStartX;

			pOutVertex[VertexCount].y = pInVertex[CurrVertex].y + 
				(pInVertex[NextVertex].y - pInVertex[CurrVertex].y)*Difference;

			pOutVertex[VertexCount].q = pInVertex[CurrVertex].q +
				(pInVertex[NextVertex].q - pInVertex[CurrVertex].q)*Difference;

			// Setup colors
			pOutVertex[VertexCount].r = pInVertex[CurrVertex].r +
				(pInVertex[NextVertex].r - pInVertex[CurrVertex].r)*Difference;

			pOutVertex[VertexCount].g = pInVertex[CurrVertex].g +
				(pInVertex[NextVertex].g - pInVertex[CurrVertex].g)*Difference;

			pOutVertex[VertexCount].b = pInVertex[CurrVertex].b +
				(pInVertex[NextVertex].b - pInVertex[CurrVertex].b)*Difference;

			// Setup texture coordinates
			pOutVertex[VertexCount].u0 = pInVertex[CurrVertex].u0 +
				(pInVertex[NextVertex].u0 - pInVertex[CurrVertex].u0)*Difference;

			pOutVertex[VertexCount].v0 = pInVertex[CurrVertex].v0 +
				(pInVertex[NextVertex].v0 - pInVertex[CurrVertex].v0)*Difference;

			if (ClipEnv)
			{
				// Clip enviroment mapping vertices if needed.
				ClippedEnvVertices2[VertexCount].u = ClippedEnvVertices[CurrVertex].u+
					(ClippedEnvVertices[NextVertex].u - ClippedEnvVertices[CurrVertex].u)*Difference;

				ClippedEnvVertices2[VertexCount].v = ClippedEnvVertices[CurrVertex].v+
					(ClippedEnvVertices[NextVertex].v - ClippedEnvVertices[CurrVertex].v)*Difference;
			}

			VertexCount++;

		}
		else if (FirstVertexResult && NextVertexResult)
		{
			// Copy vertex as it is inside clipping frustrum
			memcpy(&pOutVertex[VertexCount],
				&pInVertex[CurrVertex],sizeof(fxeDrawVertex));

			if (ClipEnv)
			{
				memcpy(&ClippedEnvVertices2[VertexCount],
					&ClippedEnvVertices[CurrVertex],sizeof(fxeTextureCoordinate));
			}

			VertexCount++;
		}
	}
	NumVertices[0] = VertexCount;

	// Copy all vertices from pOutVertex to pInVertex
	memcpy(pInVertex,pOutVertex,sizeof(fxeDrawVertex)*VertexCount);

	if (ClipEnv)
	{
		memcpy(ClippedEnvVertices,ClippedEnvVertices2,
			sizeof(fxeTextureCoordinate)*VertexCount);
	}
}

		
void ClipPolygonRight(int *NumVertices, 
	fxeDrawVertex *pInVertex, fxeDrawVertex *pOutVertex)
{
	int CurrVertex;
	int NextVertex;
	int VertexCount;
	int FirstVertexResult, NextVertexResult;
	float Length;
	float Difference;

	VertexCount = 0;

	if (NumVertices[0] == 0)
		return;

	
	for (CurrVertex=0;CurrVertex<NumVertices[0];CurrVertex++)
	{
		NextVertex = (CurrVertex+1) % NumVertices[0];

		FirstVertexResult = pInVertex[CurrVertex].x < ProjectEndX;
		NextVertexResult = pInVertex[NextVertex].x < ProjectEndX;

		if ((!FirstVertexResult) && NextVertexResult)
		{
			Length = pInVertex[NextVertex].x - pInVertex[CurrVertex].x;
			Difference = ProjectEndX-pInVertex[CurrVertex].x;
			Difference = 1.0f/Length*Difference;

			// Setup vertices
			pOutVertex[VertexCount].x = ProjectEndX;

			pOutVertex[VertexCount].y = pInVertex[CurrVertex].y + 
				(pInVertex[NextVertex].y - pInVertex[CurrVertex].y)*Difference;

			pOutVertex[VertexCount].q = pInVertex[CurrVertex].q +
				(pInVertex[NextVertex].q - pInVertex[CurrVertex].q)*Difference;

			// Setup colors
			pOutVertex[VertexCount].r = pInVertex[CurrVertex].r +
				(pInVertex[NextVertex].r - pInVertex[CurrVertex].r)*Difference;

			pOutVertex[VertexCount].g = pInVertex[CurrVertex].g +
				(pInVertex[NextVertex].g - pInVertex[CurrVertex].g)*Difference;

			pOutVertex[VertexCount].b = pInVertex[CurrVertex].b +
				(pInVertex[NextVertex].b - pInVertex[CurrVertex].b)*Difference;

			// Setup texture coordinates
			pOutVertex[VertexCount].u0 = pInVertex[CurrVertex].u0 +
				(pInVertex[NextVertex].u0 - pInVertex[CurrVertex].u0)*Difference;

			pOutVertex[VertexCount].v0 = pInVertex[CurrVertex].v0 +
				(pInVertex[NextVertex].v0 - pInVertex[CurrVertex].v0)*Difference;


			if (ClipEnv)
			{
				// Clip enviroment mapping vertices if needed.
				ClippedEnvVertices2[VertexCount].u = ClippedEnvVertices[CurrVertex].u+
					(ClippedEnvVertices[NextVertex].u - ClippedEnvVertices[CurrVertex].u)*Difference;

				ClippedEnvVertices2[VertexCount].v = ClippedEnvVertices[CurrVertex].v+
					(ClippedEnvVertices[NextVertex].v - ClippedEnvVertices[CurrVertex].v)*Difference;
			}

			VertexCount++;

		}
		else if (FirstVertexResult && (!NextVertexResult))
		{
			// Copy first vertex as it is inside clipping frustrum
			memcpy(&pOutVertex[VertexCount],
				&pInVertex[CurrVertex],sizeof(fxeDrawVertex));

			if (ClipEnv)
			{
				memcpy(&ClippedEnvVertices2[VertexCount],
					&ClippedEnvVertices[CurrVertex],sizeof(fxeDrawVertex));
			}

			VertexCount++;

			Length = pInVertex[NextVertex].x - pInVertex[CurrVertex].x;
			Difference = ProjectEndX-pInVertex[CurrVertex].x;
			Difference = 1.0f/Length*Difference;

			// Setup vertices
			pOutVertex[VertexCount].x = ProjectEndX;

			pOutVertex[VertexCount].y = pInVertex[CurrVertex].y + 
				(pInVertex[NextVertex].y - pInVertex[CurrVertex].y)*Difference;

			pOutVertex[VertexCount].q = pInVertex[CurrVertex].q +
				(pInVertex[NextVertex].q - pInVertex[CurrVertex].q)*Difference;

			// Setup colors
			pOutVertex[VertexCount].r = pInVertex[CurrVertex].r +
				(pInVertex[NextVertex].r - pInVertex[CurrVertex].r)*Difference;

			pOutVertex[VertexCount].g = pInVertex[CurrVertex].g +
				(pInVertex[NextVertex].g - pInVertex[CurrVertex].g)*Difference;

			pOutVertex[VertexCount].b = pInVertex[CurrVertex].b +
				(pInVertex[NextVertex].b - pInVertex[CurrVertex].b)*Difference;

			// Setup texture coordinates
			pOutVertex[VertexCount].u0 = pInVertex[CurrVertex].u0 +
				(pInVertex[NextVertex].u0 - pInVertex[CurrVertex].u0)*Difference;

			pOutVertex[VertexCount].v0 = pInVertex[CurrVertex].v0 +
				(pInVertex[NextVertex].v0 - pInVertex[CurrVertex].v0)*Difference;

			if (ClipEnv)
			{
				// Clip enviroment mapping vertices if needed.
				ClippedEnvVertices2[VertexCount].u = ClippedEnvVertices[CurrVertex].u+
					(ClippedEnvVertices[NextVertex].u - ClippedEnvVertices[CurrVertex].u)*Difference;

				ClippedEnvVertices2[VertexCount].v = ClippedEnvVertices[CurrVertex].v+
					(ClippedEnvVertices[NextVertex].v - ClippedEnvVertices[CurrVertex].v)*Difference;
			}

			VertexCount++;

		}
		else if (FirstVertexResult && NextVertexResult)
		{
			// Copy vertex as it is inside clipping frustrum
			memcpy(&pOutVertex[VertexCount],
				&pInVertex[CurrVertex],sizeof(fxeDrawVertex));

			if (ClipEnv)
			{
				memcpy(&ClippedEnvVertices2[VertexCount],
					&ClippedEnvVertices[CurrVertex],sizeof(fxeDrawVertex));
			}

			VertexCount++;
		}
	}
	NumVertices[0] = VertexCount;
	// Copy all vertices from pOutVertex to pInVertex
	memcpy(pInVertex,pOutVertex,sizeof(fxeDrawVertex)*VertexCount);

	if (ClipEnv)
	{
		memcpy(ClippedEnvVertices,ClippedEnvVertices2,
			sizeof(fxeTextureCoordinate)*VertexCount);
	}
}
		

void ClipPolygonTop(int *NumVertices, 
	fxeDrawVertex *pInVertex, fxeDrawVertex *pOutVertex)
{
	int CurrVertex;
	int NextVertex;
	int VertexCount;
	int FirstVertexResult, NextVertexResult;
	float Length;
	float Difference;

	VertexCount = 0;

	if (NumVertices[0] == 0)
		return;

	
	for (CurrVertex=0;CurrVertex<NumVertices[0];CurrVertex++)
	{
		NextVertex = (CurrVertex+1) % NumVertices[0];

		FirstVertexResult = pInVertex[CurrVertex].y >= ProjectStartY;
		NextVertexResult = pInVertex[NextVertex].y >= ProjectStartY;

		if ((!FirstVertexResult) && NextVertexResult)
		{
			Length = pInVertex[NextVertex].y - pInVertex[CurrVertex].y;
			Difference = ProjectStartY-pInVertex[CurrVertex].y;
			Difference = 1.0f/Length*Difference;

			// Setup vertices
			pOutVertex[VertexCount].y = ProjectStartY;

			pOutVertex[VertexCount].x = pInVertex[CurrVertex].x + 
				(pInVertex[NextVertex].x - pInVertex[CurrVertex].x)*Difference;

			pOutVertex[VertexCount].q = pInVertex[CurrVertex].q +
				(pInVertex[NextVertex].q - pInVertex[CurrVertex].q)*Difference;

			// Setup colors
			pOutVertex[VertexCount].r = pInVertex[CurrVertex].r +
				(pInVertex[NextVertex].r - pInVertex[CurrVertex].r)*Difference;

			pOutVertex[VertexCount].g = pInVertex[CurrVertex].g +
				(pInVertex[NextVertex].g - pInVertex[CurrVertex].g)*Difference;

			pOutVertex[VertexCount].b = pInVertex[CurrVertex].b +
				(pInVertex[NextVertex].b - pInVertex[CurrVertex].b)*Difference;

			// Setup texture coordinates
			pOutVertex[VertexCount].u0 = pInVertex[CurrVertex].u0 +
				(pInVertex[NextVertex].u0 - pInVertex[CurrVertex].u0)*Difference;

			pOutVertex[VertexCount].v0 = pInVertex[CurrVertex].v0 +
				(pInVertex[NextVertex].v0 - pInVertex[CurrVertex].v0)*Difference;


			if (ClipEnv)
			{
				// Clip enviroment mapping vertices if needed.
				ClippedEnvVertices2[VertexCount].u = ClippedEnvVertices[CurrVertex].u+
					(ClippedEnvVertices[NextVertex].u - ClippedEnvVertices[CurrVertex].u)*Difference;

				ClippedEnvVertices2[VertexCount].v = ClippedEnvVertices[CurrVertex].v+
					(ClippedEnvVertices[NextVertex].v - ClippedEnvVertices[CurrVertex].v)*Difference;
			}

			VertexCount++;

		}
		else if (FirstVertexResult && (!NextVertexResult))
		{
			// Copy first vertex as it is inside clipping frustrum
			memcpy(&pOutVertex[VertexCount],
				&pInVertex[CurrVertex],sizeof(fxeDrawVertex));

			if (ClipEnv)
			{
				memcpy(&ClippedEnvVertices2[VertexCount],
					&ClippedEnvVertices[CurrVertex],sizeof(fxeDrawVertex));
			}

			VertexCount++;

			Length = pInVertex[NextVertex].y - pInVertex[CurrVertex].y;
			Difference = ProjectStartY-pInVertex[CurrVertex].y;
			Difference = 1.0f/Length*Difference;

			// Setup vertices
			pOutVertex[VertexCount].y = ProjectStartY;

			pOutVertex[VertexCount].x = pInVertex[CurrVertex].x + 
				(pInVertex[NextVertex].x - pInVertex[CurrVertex].x)*Difference;

			pOutVertex[VertexCount].q = pInVertex[CurrVertex].q +
				(pInVertex[NextVertex].q - pInVertex[CurrVertex].q)*Difference;

			// Setup colors
			pOutVertex[VertexCount].r = pInVertex[CurrVertex].r +
				(pInVertex[NextVertex].r - pInVertex[CurrVertex].r)*Difference;

			pOutVertex[VertexCount].g = pInVertex[CurrVertex].g +
				(pInVertex[NextVertex].g - pInVertex[CurrVertex].g)*Difference;

			pOutVertex[VertexCount].b = pInVertex[CurrVertex].b +
				(pInVertex[NextVertex].b - pInVertex[CurrVertex].b)*Difference;

			// Setup texture coordinates
			pOutVertex[VertexCount].u0 = pInVertex[CurrVertex].u0 +
				(pInVertex[NextVertex].u0 - pInVertex[CurrVertex].u0)*Difference;

			pOutVertex[VertexCount].v0 = pInVertex[CurrVertex].v0 +
				(pInVertex[NextVertex].v0 - pInVertex[CurrVertex].v0)*Difference;


			if (ClipEnv)
			{
				// Clip enviroment mapping vertices if needed.
				ClippedEnvVertices2[VertexCount].u = ClippedEnvVertices[CurrVertex].u+
					(ClippedEnvVertices[NextVertex].u - ClippedEnvVertices[CurrVertex].u)*Difference;

				ClippedEnvVertices2[VertexCount].v = ClippedEnvVertices[CurrVertex].v+
					(ClippedEnvVertices[NextVertex].v - ClippedEnvVertices[CurrVertex].v)*Difference;
			}

			VertexCount++;

		}
		else if (FirstVertexResult && NextVertexResult)
		{
			// Copy vertex as it is inside clipping frustrum
			memcpy(&pOutVertex[VertexCount],
				&pInVertex[CurrVertex],sizeof(fxeDrawVertex));

			if (ClipEnv)
			{
				memcpy(&ClippedEnvVertices2[VertexCount],
					&ClippedEnvVertices[CurrVertex],sizeof(fxeDrawVertex));
			}

			VertexCount++;
		}
	}
	NumVertices[0] = VertexCount;
	// Copy all vertices from pOutVertex to pInVertex
	memcpy(pInVertex,pOutVertex,sizeof(fxeDrawVertex)*VertexCount);

	if (ClipEnv)
	{
		memcpy(ClippedEnvVertices,ClippedEnvVertices2,
			sizeof(fxeTextureCoordinate)*VertexCount);
	}
}


void ClipPolygonBottom(int *NumVertices, 
	fxeDrawVertex *pInVertex, fxeDrawVertex *pOutVertex)
{
	int CurrVertex;
	int NextVertex;
	int VertexCount;
	int FirstVertexResult, NextVertexResult;
	float Length;
	float Difference;

	VertexCount = 0;

	if (NumVertices[0] == 0)
		return;

	
	for (CurrVertex=0;CurrVertex<NumVertices[0];CurrVertex++)
	{
		NextVertex = (CurrVertex+1) % NumVertices[0];

		FirstVertexResult = pInVertex[CurrVertex].y < ProjectEndY;
		NextVertexResult = pInVertex[NextVertex].y < ProjectEndY;

		if ((!FirstVertexResult) && NextVertexResult)
		{
			Length = pInVertex[NextVertex].y - pInVertex[CurrVertex].y;
			Difference = ProjectEndY-pInVertex[CurrVertex].y;
			Difference = 1.0f/Length*Difference;

			// Setup vertices
			pOutVertex[VertexCount].y = ProjectEndY;

			pOutVertex[VertexCount].x = pInVertex[CurrVertex].x + 
				(pInVertex[NextVertex].x - pInVertex[CurrVertex].x)*Difference;

			pOutVertex[VertexCount].q = pInVertex[CurrVertex].q +
				(pInVertex[NextVertex].q - pInVertex[CurrVertex].q)*Difference;

			// Setup colors
			pOutVertex[VertexCount].r = pInVertex[CurrVertex].r +
				(pInVertex[NextVertex].r - pInVertex[CurrVertex].r)*Difference;

			pOutVertex[VertexCount].g = pInVertex[CurrVertex].g +
				(pInVertex[NextVertex].g - pInVertex[CurrVertex].g)*Difference;

			pOutVertex[VertexCount].b = pInVertex[CurrVertex].b +
				(pInVertex[NextVertex].b - pInVertex[CurrVertex].b)*Difference;

			// Setup texture coordinates
			pOutVertex[VertexCount].u0 = pInVertex[CurrVertex].u0 +
				(pInVertex[NextVertex].u0 - pInVertex[CurrVertex].u0)*Difference;

			pOutVertex[VertexCount].v0 = pInVertex[CurrVertex].v0 +
				(pInVertex[NextVertex].v0 - pInVertex[CurrVertex].v0)*Difference;


			if (ClipEnv)
			{
				// Clip enviroment mapping vertices if needed.
				ClippedEnvVertices2[VertexCount].u = ClippedEnvVertices[CurrVertex].u+
					(ClippedEnvVertices[NextVertex].u - ClippedEnvVertices[CurrVertex].u)*Difference;

				ClippedEnvVertices2[VertexCount].v = ClippedEnvVertices[CurrVertex].v+
					(ClippedEnvVertices[NextVertex].v - ClippedEnvVertices[CurrVertex].v)*Difference;
			}

			VertexCount++;

		}
		else if (FirstVertexResult && (!NextVertexResult))
		{
			// Copy first vertex as it is inside clipping frustrum
			memcpy(&pOutVertex[VertexCount],
				&pInVertex[CurrVertex],sizeof(fxeDrawVertex));

			if (ClipEnv)
			{
				memcpy(&ClippedEnvVertices2[VertexCount],
					&ClippedEnvVertices[CurrVertex],sizeof(fxeDrawVertex));
			}

			VertexCount++;

			Length = pInVertex[NextVertex].y - pInVertex[CurrVertex].y;
			Difference = ProjectEndY-pInVertex[CurrVertex].y;
			Difference = 1.0f/Length*Difference;

			// Setup vertices
			pOutVertex[VertexCount].y = ProjectEndY;

			pOutVertex[VertexCount].x = pInVertex[CurrVertex].x + 
				(pInVertex[NextVertex].x - pInVertex[CurrVertex].x)*Difference;

			pOutVertex[VertexCount].q = pInVertex[CurrVertex].q +
				(pInVertex[NextVertex].q - pInVertex[CurrVertex].q)*Difference;

			// Setup colors
			pOutVertex[VertexCount].r = pInVertex[CurrVertex].r +
				(pInVertex[NextVertex].r - pInVertex[CurrVertex].r)*Difference;

			pOutVertex[VertexCount].g = pInVertex[CurrVertex].g +
				(pInVertex[NextVertex].g - pInVertex[CurrVertex].g)*Difference;

			pOutVertex[VertexCount].b = pInVertex[CurrVertex].b +
				(pInVertex[NextVertex].b - pInVertex[CurrVertex].b)*Difference;

			// Setup texture coordinates
			pOutVertex[VertexCount].u0 = pInVertex[CurrVertex].u0 +
				(pInVertex[NextVertex].u0 - pInVertex[CurrVertex].u0)*Difference;

			pOutVertex[VertexCount].v0 = pInVertex[CurrVertex].v0 +
				(pInVertex[NextVertex].v0 - pInVertex[CurrVertex].v0)*Difference;


			if (ClipEnv)
			{
				// Clip enviroment mapping vertices if needed.
				ClippedEnvVertices2[VertexCount].u = ClippedEnvVertices[CurrVertex].u+
					(ClippedEnvVertices[NextVertex].u - ClippedEnvVertices[CurrVertex].u)*Difference;

				ClippedEnvVertices2[VertexCount].v = ClippedEnvVertices[CurrVertex].v+
					(ClippedEnvVertices[NextVertex].v - ClippedEnvVertices[CurrVertex].v)*Difference;
			}

			VertexCount++;

		}
		else if (FirstVertexResult && NextVertexResult)
		{
			// Copy vertex as it is inside clipping frustrum
			memcpy(&pOutVertex[VertexCount],
				&pInVertex[CurrVertex],sizeof(fxeDrawVertex));

			if (ClipEnv)
			{
				memcpy(&ClippedEnvVertices2[VertexCount],
					&ClippedEnvVertices[CurrVertex],sizeof(fxeDrawVertex));
			}

			VertexCount++;
		}
	}
	NumVertices[0] = VertexCount;
	// Copy all vertices from pOutVertex to pInVertex
	memcpy(pInVertex,pOutVertex,sizeof(fxeDrawVertex)*VertexCount);

	if (ClipEnv)
	{
		memcpy(ClippedEnvVertices,ClippedEnvVertices2,
			sizeof(fxeTextureCoordinate)*VertexCount);
	}
}

void ClipPolygonNear(int *NumVertices, 
	fxeDrawVertex *pInVertex, fxeDrawVertex *pOutVertex)
{
	int CurrVertex;
	int NextVertex;
	int VertexCount;
	int FirstVertexResult, NextVertexResult;
	float Length;
	float Difference;

	VertexCount = 0;

	if (NumVertices[0] == 0)
		return;
	
	for (CurrVertex=0;CurrVertex<NumVertices[0];CurrVertex++)
	{
		NextVertex = (CurrVertex+1) % NumVertices[0];

		FirstVertexResult = pInVertex[CurrVertex].z >= NearClippingPlane;
		NextVertexResult = pInVertex[NextVertex].z >= NearClippingPlane;

		if ((!FirstVertexResult) && NextVertexResult)
		{
			Length = pInVertex[NextVertex].z - pInVertex[CurrVertex].z;
			Difference = NearClippingPlane-pInVertex[CurrVertex].z;
			Difference = 1.0f/Length*Difference;

			// Setup vertices
			pOutVertex[VertexCount].z = NearClippingPlane;

			pOutVertex[VertexCount].x = pInVertex[CurrVertex].x + 
				(pInVertex[NextVertex].x - pInVertex[CurrVertex].x)*Difference;

			pOutVertex[VertexCount].y = pInVertex[CurrVertex].y +
				(pInVertex[NextVertex].y - pInVertex[CurrVertex].y)*Difference;

			// Setup colors
			pOutVertex[VertexCount].r = pInVertex[CurrVertex].r +
				(pInVertex[NextVertex].r - pInVertex[CurrVertex].r)*Difference;

			pOutVertex[VertexCount].g = pInVertex[CurrVertex].g +
				(pInVertex[NextVertex].g - pInVertex[CurrVertex].g)*Difference;

			pOutVertex[VertexCount].b = pInVertex[CurrVertex].b +
				(pInVertex[NextVertex].b - pInVertex[CurrVertex].b)*Difference;

			// Setup texture coordinates
			pOutVertex[VertexCount].u0 = pInVertex[CurrVertex].u0 +
				(pInVertex[NextVertex].u0 - pInVertex[CurrVertex].u0)*Difference;

			pOutVertex[VertexCount].v0 = pInVertex[CurrVertex].v0 +
				(pInVertex[NextVertex].v0 - pInVertex[CurrVertex].v0)*Difference;


			if (ClipEnv)
			{
				// Clip enviroment mapping vertices if needed.
				ClippedEnvVertices2[VertexCount].u = ClippedEnvVertices[CurrVertex].u+
					(ClippedEnvVertices[NextVertex].u - ClippedEnvVertices[CurrVertex].u)*Difference;

				ClippedEnvVertices2[VertexCount].v = ClippedEnvVertices[CurrVertex].v+
					(ClippedEnvVertices[NextVertex].v - ClippedEnvVertices[CurrVertex].v)*Difference;
			}

			VertexCount++;

		}
		else if (FirstVertexResult && (!NextVertexResult))
		{
			// Copy first vertex as it is inside clipping frustrum
			memcpy(&pOutVertex[VertexCount],
				&pInVertex[CurrVertex],sizeof(fxeDrawVertex));

			if (ClipEnv)
			{
				memcpy(&ClippedEnvVertices2[VertexCount],
					&ClippedEnvVertices[CurrVertex],sizeof(fxeDrawVertex));
			}

			VertexCount++;

			Length = pInVertex[NextVertex].z - pInVertex[CurrVertex].z;
			Difference = NearClippingPlane-pInVertex[CurrVertex].z;
			Difference = 1.0f/Length*Difference;

			// Setup vertices
			pOutVertex[VertexCount].z = NearClippingPlane;

			pOutVertex[VertexCount].x = pInVertex[CurrVertex].x + 
				(pInVertex[NextVertex].x - pInVertex[CurrVertex].x)*Difference;

			pOutVertex[VertexCount].y = pInVertex[CurrVertex].y +
				(pInVertex[NextVertex].y - pInVertex[CurrVertex].y)*Difference;

			// Setup colors
			pOutVertex[VertexCount].r = pInVertex[CurrVertex].r +
				(pInVertex[NextVertex].r - pInVertex[CurrVertex].r)*Difference;

			pOutVertex[VertexCount].g = pInVertex[CurrVertex].g +
				(pInVertex[NextVertex].g - pInVertex[CurrVertex].g)*Difference;

			pOutVertex[VertexCount].b = pInVertex[CurrVertex].b +
				(pInVertex[NextVertex].b - pInVertex[CurrVertex].b)*Difference;

			// Setup texture coordinates
			pOutVertex[VertexCount].u0 = pInVertex[CurrVertex].u0 +
				(pInVertex[NextVertex].u0 - pInVertex[CurrVertex].u0)*Difference;

			pOutVertex[VertexCount].v0 = pInVertex[CurrVertex].v0 +
				(pInVertex[NextVertex].v0 - pInVertex[CurrVertex].v0)*Difference;


			if (ClipEnv)
			{
				// Clip enviroment mapping vertices if needed.
				ClippedEnvVertices2[VertexCount].u = ClippedEnvVertices[CurrVertex].u+
					(ClippedEnvVertices[NextVertex].u - ClippedEnvVertices[CurrVertex].u)*Difference;

				ClippedEnvVertices2[VertexCount].v = ClippedEnvVertices[CurrVertex].v+
					(ClippedEnvVertices[NextVertex].v - ClippedEnvVertices[CurrVertex].v)*Difference;
			}

			VertexCount++;

		}
		else if (FirstVertexResult && NextVertexResult)
		{
			// Copy vertex as it is inside clipping frustrum
			memcpy(&pOutVertex[VertexCount],
				&pInVertex[CurrVertex],sizeof(fxeDrawVertex));

			if (ClipEnv)
			{
				memcpy(&ClippedEnvVertices2[VertexCount],
					&ClippedEnvVertices[CurrVertex],sizeof(fxeDrawVertex));
			}

			VertexCount++;
		}
	}
	NumVertices[0] = VertexCount;
	// Copy all vertices from pOutVertex to pInVertex
	memcpy(pInVertex,pOutVertex,sizeof(fxeDrawVertex)*VertexCount);

	if (ClipEnv)
	{
		memcpy(ClippedEnvVertices,ClippedEnvVertices2,
			sizeof(fxeTextureCoordinate)*VertexCount);
	}
}


void ClipPolygonFar(int *NumVertices, 
	fxeDrawVertex *pInVertex, fxeDrawVertex *pOutVertex)
{
	int CurrVertex;
	int NextVertex;
	int VertexCount;
	int FirstVertexResult, NextVertexResult;
	float Length;
	float Difference;

	VertexCount = 0;


	if (NumVertices[0] == 0) 
		return;

	for (CurrVertex=0;CurrVertex<NumVertices[0];CurrVertex++)
	{
		NextVertex = (CurrVertex+1) % NumVertices[0];

		FirstVertexResult = pInVertex[CurrVertex].z < FarClippingPlane;
		NextVertexResult = pInVertex[NextVertex].z < FarClippingPlane;

		if ((!FirstVertexResult) && NextVertexResult)
		{
			Length = pInVertex[NextVertex].z - pInVertex[CurrVertex].z;
			Difference = FarClippingPlane-pInVertex[CurrVertex].z;
			Difference = 1.0f/Length*Difference;

			// Setup vertices
			pOutVertex[VertexCount].z = FarClippingPlane;

			pOutVertex[VertexCount].x = pInVertex[CurrVertex].x + 
				(pInVertex[NextVertex].x - pInVertex[CurrVertex].x)*Difference;

			pOutVertex[VertexCount].y = pInVertex[CurrVertex].y +
				(pInVertex[NextVertex].y - pInVertex[CurrVertex].y)*Difference;

			// Setup colors
			pOutVertex[VertexCount].r = pInVertex[CurrVertex].r +
				(pInVertex[NextVertex].r - pInVertex[CurrVertex].r)*Difference;

			pOutVertex[VertexCount].g = pInVertex[CurrVertex].g +
				(pInVertex[NextVertex].g - pInVertex[CurrVertex].g)*Difference;

			pOutVertex[VertexCount].b = pInVertex[CurrVertex].b +
				(pInVertex[NextVertex].b - pInVertex[CurrVertex].b)*Difference;

			// Setup texture coordinates
			pOutVertex[VertexCount].u0 = pInVertex[CurrVertex].u0 +
				(pInVertex[NextVertex].u0 - pInVertex[CurrVertex].u0)*Difference;

			pOutVertex[VertexCount].v0 = pInVertex[CurrVertex].v0 +
				(pInVertex[NextVertex].v0 - pInVertex[CurrVertex].v0)*Difference;


			if (ClipEnv)
			{
				// Clip enviroment mapping vertices if needed.
				ClippedEnvVertices2[VertexCount].u = ClippedEnvVertices[CurrVertex].u+
					(ClippedEnvVertices[NextVertex].u - ClippedEnvVertices[CurrVertex].u)*Difference;

				ClippedEnvVertices2[VertexCount].v = ClippedEnvVertices[CurrVertex].v+
					(ClippedEnvVertices[NextVertex].v - ClippedEnvVertices[CurrVertex].v)*Difference;
			}

			VertexCount++;

		}
		else if (FirstVertexResult && (!NextVertexResult))
		{
			// Copy first vertex as it is inside clipping frustrum
			memcpy(&pOutVertex[VertexCount],
				&pInVertex[CurrVertex],sizeof(fxeDrawVertex));

			if (ClipEnv)
			{
				memcpy(&ClippedEnvVertices2[VertexCount],
					&ClippedEnvVertices[CurrVertex],sizeof(fxeDrawVertex));
			}

			VertexCount++;

			Length = pInVertex[NextVertex].z - pInVertex[CurrVertex].z;
			Difference = FarClippingPlane-pInVertex[CurrVertex].z;
			Difference = 1.0f/Length*Difference;

			// Setup vertices
			pOutVertex[VertexCount].z = FarClippingPlane;

			pOutVertex[VertexCount].x = pInVertex[CurrVertex].x + 
				(pInVertex[NextVertex].x - pInVertex[CurrVertex].x)*Difference;

			pOutVertex[VertexCount].y = pInVertex[CurrVertex].y +
				(pInVertex[NextVertex].y - pInVertex[CurrVertex].y)*Difference;

			// Setup colors
			pOutVertex[VertexCount].r = pInVertex[CurrVertex].r +
				(pInVertex[NextVertex].r - pInVertex[CurrVertex].r)*Difference;

			pOutVertex[VertexCount].g = pInVertex[CurrVertex].g +
				(pInVertex[NextVertex].g - pInVertex[CurrVertex].g)*Difference;

			pOutVertex[VertexCount].b = pInVertex[CurrVertex].b +
				(pInVertex[NextVertex].b - pInVertex[CurrVertex].b)*Difference;

			// Setup texture coordinates
			pOutVertex[VertexCount].u0 = pInVertex[CurrVertex].u0 +
				(pInVertex[NextVertex].u0 - pInVertex[CurrVertex].u0)*Difference;

			pOutVertex[VertexCount].v0 = pInVertex[CurrVertex].v0 +
				(pInVertex[NextVertex].v0 - pInVertex[CurrVertex].v0)*Difference;


			if (ClipEnv)
			{
				// Clip enviroment mapping vertices if needed.
				ClippedEnvVertices2[VertexCount].u = ClippedEnvVertices[CurrVertex].u+
					(ClippedEnvVertices[NextVertex].u - ClippedEnvVertices[CurrVertex].u)*Difference;

				ClippedEnvVertices2[VertexCount].v = ClippedEnvVertices[CurrVertex].v+
					(ClippedEnvVertices[NextVertex].v - ClippedEnvVertices[CurrVertex].v)*Difference;
			}

			VertexCount++;

		}
		else if (FirstVertexResult && NextVertexResult)
		{
			// Copy vertex as it is inside clipping frustrum
			memcpy(&pOutVertex[VertexCount],
				&pInVertex[CurrVertex],sizeof(fxeDrawVertex));

			if (ClipEnv)
			{
				memcpy(&ClippedEnvVertices2[VertexCount],
					&ClippedEnvVertices[CurrVertex],sizeof(fxeDrawVertex));
			}

			VertexCount++;
		}
	}
	NumVertices[0] = VertexCount;
	// Copy all vertices from pOutVertex to pInVertex
	memcpy(pInVertex,pOutVertex,sizeof(fxeDrawVertex)*VertexCount);

	if (ClipEnv)
	{
		memcpy(ClippedEnvVertices,ClippedEnvVertices2,
			sizeof(fxeTextureCoordinate)*VertexCount);
	}
}



/*
int ClipAndDrawPolygon2d(void)
{
	int ClipFlags;
	
	ClipFlags = 0;

	// Check if poly is completely clipped.

	// Clip against left side of viewing frustrum
	ClipFlags += (v1->x < ProjectStartX);
	ClipFlags += ((v2->x < ProjectStartX)<<1);
	ClipFlags += ((v3->x < ProjectStartX)<<2);

	if (ClipFlags == 7)
		return 1;

	// Clip against right side of viewing frustrum
	
	ClipFlags += ((v1->x > ProjectEndX)<<3);
	ClipFlags += ((v2->x > ProjectEndX)<<4);
	ClipFlags += ((v3->x > ProjectEndX)<<5);

	if ((ClipFlags & 56) == 56)
		return 1;

	// Clip against down side of viewing frustrum

	ClipFlags += ((v1->y < ProjectStartY)<<6);
	ClipFlags += ((v2->y < ProjectStartY)<<7);
	ClipFlags += ((v3->y < ProjectStartY)<<8);

	if ((ClipFlags & 448) == 448)
		return 1;

	// Clip against upper side of viewing frustrum

	ClipFlags += ((v1->y > ProjectEndY)<<9);
	ClipFlags += ((v2->y > ProjectEndY)<<10);
	ClipFlags += ((v3->y > ProjectEndY)<<11);

	if ((ClipFlags & 7168) == 7168)
		return 1;

	// Clip againt near side of viewing frustrum

	ClipFlags += ((v1->z < NearClippingPlane) << 12);
	ClipFlags += ((v2->z < NearClippingPlane) << 13);
	ClipFlags += ((v3->z < NearClippingPlane) << 14);

	if ((ClipFlags & 57344) == 57344)
		return 1;

	// Clip against far side of viewing frustrum

	ClipFlags += ((v1->z > FarClippingPlane) << 15);
	ClipFlags += ((v2->z > FarClippingPlane) << 16);
	ClipFlags += ((v3->z > FarClippingPlane) << 17);

	if ((ClipFlags & 458752) == 458752)
		return 1;

	// Check if we don't need to clip at all:
	if (ClipFlags == 0)
		return 2;

	NumClippedVertices = 0;

	// Check which coordinates have been clipped to the left
	// side of the viewing frustrum.
	switch (ClipFlags & 7)
	{
		case 0: 
			// Easy just copy vertices:
			ClippedVertices[
		case 1: 
		case 2:
		case 3:
		case 4:
		case 5:
		case 6:
	}

}
*/

void ProjectAndDrawObject(fxeObjectHandle ObjectHandle)
{
	// Note!
  // DrawVertex is a array of 200 vertices, if polygon has
  // more than 200 vertices this function will bail out!

  void *VertexPointers[200];
	int Loop, Loop2;
	fxePolygon *Poly;
  fxeDrawVertex *CamVertex;
	fxeVertexIndex *VIndex;
	fxeTextureHandle TexHandle;
  int NumVertices;
  int PartlyClipped;
  fxeDrawVertex *CurrVertex;
  fxePolygon *CurrPoly;
  fxeDrawVertex EnvVer[200];
  fxePolygonList *pSortList;
	int Loop3;
	fxeColor SavedRGBColor[3];
	fxeDrawVertex ProjectedVertices[20];

	Poly = ObjectHandle->PolygonList;
	CamVertex = ObjectHandle->VerticesCamera;

  DrawShape = ObjectHandle->DrawShape;

  if (ObjectHandle->Transparent == FXEFALSE)
  {


  //Setup the constant color value:
  grConstantColorValue(255<<24);

	for (Loop=0 ; Loop<ObjectHandle->NumPolys; Loop++)
	{

    // only project and draw polygon if visible
    CurrPoly = &(Poly[Loop]);

    if (CurrPoly->Transparent == FXEFALSE)
    {
      // We draws the transparent last as they need the rgb values
      // from the other polygons rendered in the linear frame buffer

 
    if (CurrPoly->Clipped == FXEFALSE && (CurrPoly->Culled == FXEFALSE ||
      DrawShape == GR_POINTS || DrawShape == GR_LINE_STRIP ))
		{
      VIndex = CurrPoly->VertexIndices;

      if (CurrPoly->PartlyClipped == FXETRUE)
      {
        NumVertices = 0;
        PartlyClipped = FXETRUE;
      }
      else
      {
        NumVertices = CurrPoly->NumVertices;
        PartlyClipped = FXEFALSE;
      }

        if (PartlyClipped == FXETRUE)
        {
					ClipEnv = ObjectHandle->EnviromentMapped;
					NumClippedVertices = 3;
					
					// First we must clip our polygons to near plane to
					// avoid divide by zero when z = 0
					for (Loop2=0;Loop2<3;Loop2++)
					{
						ClippedVertices[Loop2].x = CamVertex[VIndex[Loop2]].x;
						ClippedVertices[Loop2].y = CamVertex[VIndex[Loop2]].y;
						ClippedVertices[Loop2].z = CamVertex[VIndex[Loop2]].z;

						if (ObjectHandle->ShadingType == FLAT_SHADING)
						{
							ClippedVertices[Loop2].r = CurrPoly->FlatShadedColor.r*255.0f;
							ClippedVertices[Loop2].g = CurrPoly->FlatShadedColor.g*255.0f;
							ClippedVertices[Loop2].b = CurrPoly->FlatShadedColor.b*255.0f;
						}						
						else 
						{
							ClippedVertices[Loop2].r = (CamVertex[VIndex[Loop2]].r*
								(CurrPoly->DiffuseColor.r)+CurrPoly->AmbientColor.r+
								ObjectHandle->DynColors[VIndex[Loop2]].r+AmbientR);
							ClippedVertices[Loop2].g = (CamVertex[VIndex[Loop2]].g*
								(CurrPoly->DiffuseColor.g)+CurrPoly->AmbientColor.g+
								ObjectHandle->DynColors[VIndex[Loop2]].g+AmbientG);
							ClippedVertices[Loop2].b = (CamVertex[VIndex[Loop2]].b*
								(CurrPoly->DiffuseColor.b)+CurrPoly->AmbientColor.b+
								ObjectHandle->DynColors[VIndex[Loop2]].b+AmbientB);

							if (ClippedVertices[Loop2].r > 1.0f)
								ClippedVertices[Loop2].r = 1.0f;
							if (ClippedVertices[Loop2].g > 1.0f)
								ClippedVertices[Loop2].g = 1.0f;
							if (ClippedVertices[Loop2].b > 1.0f)
								ClippedVertices[Loop2].b = 1.0f;

							ClippedVertices[Loop2].r *= 255.0f;
							ClippedVertices[Loop2].g *= 255.0f;
							ClippedVertices[Loop2].b *= 255.0f;


						}

						ClippedVertices[Loop2].u0 = CurrPoly->TextureCoordinates[Loop2].u * 255.0f;
						ClippedVertices[Loop2].v0 = CurrPoly->TextureCoordinates[Loop2].v * 255.0f;

						if (ClipEnv)
						{
							ClippedEnvVertices[Loop2].u = 
								ObjectHandle->EnviromentVerts[VIndex[Loop2]].u*255.0f;
							ClippedEnvVertices[Loop2].v =
								ObjectHandle->EnviromentVerts[VIndex[Loop2]].v*255.0f;
						}
					}

					// Check if we need to clip to Near Plane:
					if ((ObjectHandle->PolygonList[Loop].PartlyClipStatus & 133152) != 0)
					{
						ClipPolygonNear(&NumClippedVertices,ClippedVertices,
							ClippedVertices2);
					}

					if ((ObjectHandle->PolygonList[Loop].PartlyClipStatus & 66576) != 0)
					{
						// Must clip to far plane to using the z value
						ClipPolygonFar(&NumClippedVertices,ClippedVertices,
							ClippedVertices2);
					}


					if (NumClippedVertices == 0)
						continue;

					// Project vertices so that 2d clipping will work
					for (Loop2=0;Loop2<NumClippedVertices;Loop2++)
					{
						// Project every vertex:
						ClippedVertices[Loop2].q = 1.0f/ClippedVertices[Loop2].z;
						ClippedVertices[Loop2].x = (ClippedVertices[Loop2].x * ClippedVertices[Loop2].q+ 1.0f)*
							ProjectWidthDiv2+ProjectStartX;
						ClippedVertices[Loop2].y = (ClippedVertices[Loop2].y * ClippedVertices[Loop2].q+ 1.0f)*
							ProjectHeightDiv2+ProjectStartY;
						ClippedVertices[Loop2].u0 *= ClippedVertices[Loop2].q;
						ClippedVertices[Loop2].v0 *= ClippedVertices[Loop2].q;

						if (ClipEnv)
						{
							ClippedEnvVertices[Loop2].u *= ClippedVertices[Loop2].q;
							ClippedEnvVertices[Loop2].v *= ClippedVertices[Loop2].q;
						}
					}

					if ((ObjectHandle->PolygonList[Loop].PartlyClipStatus & 33288) != 0)
					{
						ClipPolygonBottom(&NumClippedVertices,ClippedVertices,
							ClippedVertices2);
					}

					if ((ObjectHandle->PolygonList[Loop].PartlyClipStatus & 16644) != 0)
					{
						ClipPolygonTop(&NumClippedVertices,ClippedVertices,
							ClippedVertices2);
					}

					if ((ObjectHandle->PolygonList[Loop].PartlyClipStatus & 4161) != 0)
					{
						ClipPolygonLeft(&NumClippedVertices,ClippedVertices,
							ClippedVertices2);
					}

					if ((ObjectHandle->PolygonList[Loop].PartlyClipStatus & 8322) != 0)
					{
						ClipPolygonRight(&NumClippedVertices,ClippedVertices,
							ClippedVertices2);
					}


					// Now we have the clipped polygon in clippedvertices and
					// the number of vertices in NumClippedVertices.
					NumVertices = NumClippedVertices;

					if (NumVertices == 0)
					{
						// if NumVertices is 0 we don't need to display
						// this polygon at all
						continue;
					}
						

					for (Loop2=0;Loop2<NumClippedVertices;Loop2++)
					{
						ClippedVertices[Loop2].u1 = ClippedVertices[Loop2].u0;
						ClippedVertices[Loop2].v1 = ClippedVertices[Loop2].v0;
						Snap(&ClippedVertices[Loop2].x);
						Snap(&ClippedVertices[Loop2].y);
						VertexPointers[Loop2] = (void*)&ClippedVertices[Loop2];
					}



        }

        else if (PartlyClipped == FXEFALSE)
        {
          CurrVertex = &CamVertex[VIndex[0]];

          VertexPointers[0] = (void*)&ProjectedVertices[0];

          if (CurrPoly->TexMode != NOT_TEXTURED)
          {
            CurrVertex->u0 = CurrPoly->TextureCoordinates[0].u;
            CurrVertex->v0 = CurrPoly->TextureCoordinates[0].v;
            CurrVertex->u1 = CurrPoly->TextureCoordinates[0].u;
            CurrVertex->v1 = CurrPoly->TextureCoordinates[0].v;
          }

          if (ObjectHandle->ShadingType == FLAT_SHADING)
          {
            CurrVertex->r = CurrPoly->FlatShadedColor.r;
            CurrVertex->g = CurrPoly->FlatShadedColor.g;
            CurrVertex->b = CurrPoly->FlatShadedColor.b;
          }
					else 
					{
						SavedRGBColor[0].r = CurrVertex->r;
						SavedRGBColor[0].g = CurrVertex->g;
						SavedRGBColor[0].b = CurrVertex->b;

						CurrVertex->r = CurrVertex->r*(CurrPoly->DiffuseColor.r)+CurrPoly->AmbientColor.r+
							ObjectHandle->DynColors[VIndex[0]].r+AmbientR;
						CurrVertex->g = CurrVertex->g*(CurrPoly->DiffuseColor.g)+CurrPoly->AmbientColor.g+
							ObjectHandle->DynColors[VIndex[0]].g+AmbientG;
						CurrVertex->b = CurrVertex->b*(CurrPoly->DiffuseColor.b)+CurrPoly->AmbientColor.b+
							ObjectHandle->DynColors[VIndex[0]].b+AmbientB;
/*
						CurrVertex->r += CurrPoly->AmbientColor.r;
						CurrVertex->g += CurrPoly->AmbientColor.g;
						CurrVertex->b += CurrPoly->AmbientColor.b;
*/
						if (CurrVertex->r > 1.0f)
							CurrVertex->r = 1.0f;
						if (CurrVertex->g > 1.0f)
							CurrVertex->g = 1.0f;
						if (CurrVertex->b > 1.0f)
							CurrVertex->b = 1.0f;
					}

					ProjectedVertices[0].q = 1.0f/CurrVertex->z;
					
					ProjectedVertices[0].x = (CurrVertex->x*ProjectedVertices[0].q+1.0f)*
						ProjectWidthDiv2+ProjectStartX;
					Snap(&ProjectedVertices[0].x);

					ProjectedVertices[0].y = (CurrVertex->y*ProjectedVertices[0].q+1.0f)*
						ProjectHeightDiv2+ProjectStartY;
					Snap(&ProjectedVertices[0].y);

					ProjectedVertices[0].r = CurrVertex->r * 255.0f;
					ProjectedVertices[0].g = CurrVertex->g * 255.0f;
					ProjectedVertices[0].b = CurrVertex->b * 255.0f;

					ProjectedVertices[0].u0 = CurrVertex->u0 * 255.0f*
						ProjectedVertices[0].q;
					ProjectedVertices[0].v0 = CurrVertex->v0 * 255.0f*
						ProjectedVertices[0].q;
					ProjectedVertices[0].u1 = ProjectedVertices[0].u0;
					ProjectedVertices[0].v1 = ProjectedVertices[0].v0;



          CurrVertex = &CamVertex[VIndex[1]];
          VertexPointers[1] = (void*)&ProjectedVertices[1];

/*						if (CurrVertex->x <= -CurrVertex->z) 
							CurrVertex->x = -CurrVertex->z+1.0f;
						if (CurrVertex->x >= CurrVertex->z)
							CurrVertex->x = CurrVertex->z-1.0f;
						if (CurrVertex->y <= -CurrVertex->z) 
							CurrVertex->y = -CurrVertex->z+1.0f;
						if (CurrVertex->y >= CurrVertex->z)
							CurrVertex->y = CurrVertex->z-1.0f;
						if (CurrVertex->z <= NearClippingPlane)
							CurrVertex->z = NearClippingPlane+1.0f;
*/

          if (CurrPoly->TexMode != NOT_TEXTURED)
          {
            CurrVertex->u0 = CurrPoly->TextureCoordinates[1].u;
            CurrVertex->v0 = CurrPoly->TextureCoordinates[1].v;
            CurrVertex->u1 = CurrPoly->TextureCoordinates[1].u;
            CurrVertex->v1 = CurrPoly->TextureCoordinates[1].v;
          }

          if (ObjectHandle->ShadingType == FLAT_SHADING)
          {
            CurrVertex->r = CurrPoly->FlatShadedColor.r;
            CurrVertex->g = CurrPoly->FlatShadedColor.g;
            CurrVertex->b = CurrPoly->FlatShadedColor.b;
          }
					else
					{
						SavedRGBColor[1].r = CurrVertex->r;
						SavedRGBColor[1].g = CurrVertex->g;
						SavedRGBColor[1].b = CurrVertex->b;

						CurrVertex->r = CurrVertex->r*(CurrPoly->DiffuseColor.r)+CurrPoly->AmbientColor.r+
							ObjectHandle->DynColors[VIndex[1]].r+AmbientR;
						CurrVertex->g = CurrVertex->g*(CurrPoly->DiffuseColor.g)+CurrPoly->AmbientColor.g+
							ObjectHandle->DynColors[VIndex[1]].g+AmbientG;
						CurrVertex->b = CurrVertex->b*(CurrPoly->DiffuseColor.b)+CurrPoly->AmbientColor.b+
							ObjectHandle->DynColors[VIndex[1]].b+AmbientB;
/*
						CurrVertex->r += CurrPoly->AmbientColor.r;
						CurrVertex->g += CurrPoly->AmbientColor.g;
						CurrVertex->b += CurrPoly->AmbientColor.b;
*/
						if (CurrVertex->r > 1.0f)
							CurrVertex->r = 1.0f;
						if (CurrVertex->g > 1.0f)
							CurrVertex->g = 1.0f;
						if (CurrVertex->b > 1.0f)
							CurrVertex->b = 1.0f;
					}

					ProjectedVertices[1].q = 1.0f/CurrVertex->z;
					
					ProjectedVertices[1].x = (CurrVertex->x*ProjectedVertices[1].q+1.0f)*
						ProjectWidthDiv2+ProjectStartX;
					Snap(&ProjectedVertices[1].x);

					ProjectedVertices[1].y = (CurrVertex->y*ProjectedVertices[1].q+1.0f)*
						ProjectHeightDiv2+ProjectStartY;
					Snap(&ProjectedVertices[1].y);

					ProjectedVertices[1].r = CurrVertex->r * 255.0f;
					ProjectedVertices[1].g = CurrVertex->g * 255.0f;
					ProjectedVertices[1].b = CurrVertex->b * 255.0f;

					ProjectedVertices[1].u0 = CurrVertex->u0 * 255.0f*
						ProjectedVertices[1].q;
					ProjectedVertices[1].v0 = CurrVertex->v0 * 255.0f*
						ProjectedVertices[1].q;
					ProjectedVertices[1].u1 = ProjectedVertices[1].u0;
					ProjectedVertices[1].v1 = ProjectedVertices[1].v0;



          CurrVertex = &CamVertex[VIndex[2]];
          VertexPointers[2] = (void*)&ProjectedVertices[2];


          if (CurrPoly->TexMode != NOT_TEXTURED)
          {
            CurrVertex->u0 = CurrPoly->TextureCoordinates[2].u;
            CurrVertex->v0 = CurrPoly->TextureCoordinates[2].v;
            CurrVertex->u1 = CurrPoly->TextureCoordinates[2].u;
            CurrVertex->v1 = CurrPoly->TextureCoordinates[2].v;
          }

          if (ObjectHandle->ShadingType == FLAT_SHADING)
          {
            CurrVertex->r = CurrPoly->FlatShadedColor.r;
            CurrVertex->g = CurrPoly->FlatShadedColor.g;
            CurrVertex->b = CurrPoly->FlatShadedColor.b;
          }
					else
					{
						SavedRGBColor[2].r = CurrVertex->r;
						SavedRGBColor[2].g = CurrVertex->g;
						SavedRGBColor[2].b = CurrVertex->b;

						CurrVertex->r = CurrVertex->r*(CurrPoly->DiffuseColor.r)+CurrPoly->AmbientColor.r+
							ObjectHandle->DynColors[VIndex[2]].r+AmbientR;
						CurrVertex->g = CurrVertex->g*(CurrPoly->DiffuseColor.g)+CurrPoly->AmbientColor.g+
							ObjectHandle->DynColors[VIndex[2]].g+AmbientG;
						CurrVertex->b = CurrVertex->b*(CurrPoly->DiffuseColor.b)+CurrPoly->AmbientColor.b+
							ObjectHandle->DynColors[VIndex[2]].b+AmbientB;
/*
						CurrVertex->r += CurrPoly->AmbientColor.r;
						CurrVertex->g += CurrPoly->AmbientColor.g;
						CurrVertex->b += CurrPoly->AmbientColor.b;
*/
						if (CurrVertex->r > 1.0f)
							CurrVertex->r = 1.0f;
						if (CurrVertex->g > 1.0f)
							CurrVertex->g = 1.0f;
						if (CurrVertex->b > 1.0f)
							CurrVertex->b = 1.0f;
					}
					
					ProjectedVertices[2].q = 1.0f/CurrVertex->z;
					
					ProjectedVertices[2].x = (CurrVertex->x*ProjectedVertices[2].q+1.0f)*
						ProjectWidthDiv2+ProjectStartX;
					Snap(&ProjectedVertices[2].x);

					ProjectedVertices[2].y = (CurrVertex->y*ProjectedVertices[2].q+1.0f)*
						ProjectHeightDiv2+ProjectStartY;
					Snap(&ProjectedVertices[2].y);

					ProjectedVertices[2].r = CurrVertex->r * 255.0f;
					ProjectedVertices[2].g = CurrVertex->g * 255.0f;
					ProjectedVertices[2].b = CurrVertex->b * 255.0f;

					ProjectedVertices[2].u0 = CurrVertex->u0 * 255.0f * 
						ProjectedVertices[2].q;
					ProjectedVertices[2].v0 = CurrVertex->v0 * 255.0f *
						ProjectedVertices[2].q;
					ProjectedVertices[2].u1 = ProjectedVertices[2].u0;
					ProjectedVertices[2].v1 = ProjectedVertices[2].v0;


        }

          if (CurrPoly->TexMode == NOT_TEXTURED)
          {
						grColorCombine(GR_COMBINE_FUNCTION_LOCAL,
							GR_COMBINE_FACTOR_NONE,GR_COMBINE_LOCAL_ITERATED,
							GR_COMBINE_OTHER_NONE,FXFALSE);

            /*
						TexHandle = NULLTexture;
            if (CurrPoly->PartlyClipped != FXEFALSE)
            {
            if (TextureFilteringMode == TRILINEAR_FILTERED)
            {
              ((fxeDrawVertex*)VertexPointers[0])->u0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->v0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->u1 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->v1 = 0.0f;
              
							// Mste multi med q:
              ((fxeDrawVertex*)VertexPointers[1])->u0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[1])->v0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[1])->q;
              ((fxeDrawVertex*)VertexPointers[1])->u1 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[1])->v1 = 255.0f*
								((fxeDrawVertex*)VertexPointers[1])->q;

              ((fxeDrawVertex*)VertexPointers[2])->u0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[2])->q;
              ((fxeDrawVertex*)VertexPointers[2])->v0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[2])->q;
              ((fxeDrawVertex*)VertexPointers[2])->u1 = 255.0f*
								((fxeDrawVertex*)VertexPointers[2])->q;
              ((fxeDrawVertex*)VertexPointers[2])->v1 = 255.0f*
								((fxeDrawVertex*)VertexPointers[2])->q;
            }
            else
            {
              ((fxeDrawVertex*)VertexPointers[0])->u0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->v0 = 0.0f;
              
              ((fxeDrawVertex*)VertexPointers[1])->u0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[1])->v0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[1])->q;

              ((fxeDrawVertex*)VertexPointers[2])->u0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[2])->q;
              ((fxeDrawVertex*)VertexPointers[2])->v0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[2])->q;
            }
            }

            else
            {
            if (TextureFilteringMode == TRILINEAR_FILTERED)
            {
              ((fxeDrawVertex*)VertexPointers[0])->u0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->v0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->u1 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->v1 = 0.0f;
              for(Loop2=1;Loop2<NumVertices;Loop2++)
              {
                ((fxeDrawVertex*)VertexPointers[Loop2])->u0 = 255.0f*
									((fxeDrawVertex*)VertexPointers[Loop2])->q;
                ((fxeDrawVertex*)VertexPointers[Loop2])->v0 = 255.0f*
									((fxeDrawVertex*)VertexPointers[Loop2])->q;
                ((fxeDrawVertex*)VertexPointers[Loop2])->u1 = 255.0f*
									((fxeDrawVertex*)VertexPointers[Loop2])->q;
                ((fxeDrawVertex*)VertexPointers[Loop2])->v1 = 255.0f*
									((fxeDrawVertex*)VertexPointers[Loop2])->q;
              }
              
            }
            else
            {
              ((fxeDrawVertex*)VertexPointers[0])->u0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->v0 = 0.0f;
              for(Loop2=1;Loop2<NumVertices;Loop2++)
              {
                ((fxeDrawVertex*)VertexPointers[Loop2])->u0 = 255.0f*
									((fxeDrawVertex*)VertexPointers[Loop2])->q;

                ((fxeDrawVertex*)VertexPointers[Loop2])->v0 = 255.0f*
									((fxeDrawVertex*)VertexPointers[Loop2])->q;
              }
            }
            }
						*/
          }
          else
					{
            TexHandle = CurrPoly->TextureHandle;
						SetupTexture(TexHandle);
					}


        // so finally send polygon to 3dfx hardware:
        // Draw polygon using triangle fans. This way any convex polygon
        // can be drawn.

        if (DrawShape == GR_LINE_STRIP)
        {
          VertexPointers[NumVertices] = VertexPointers[0];
          grDrawVertexArray(DrawShape,NumVertices+1,VertexPointers);
        }

				
        else
				{
					if (DrawShape == GR_TRIANGLE_FAN)
					{
						grDrawVertexArray(GR_TRIANGLE_FAN,3,VertexPointers);
						if (NumVertices>3)
						{
							for (Loop3=0;Loop3<NumVertices-3;Loop3++)
							{
								grDrawVertexArray(GR_TRIANGLE_FAN_CONTINUE,1,&VertexPointers[Loop3+3]);
							}
						}
					}
					else
						grDrawVertexArray(DrawShape,NumVertices,VertexPointers);

				}

        Performance.NumPolys++;

				if (CurrPoly->PartlyClipped == FXEFALSE)
				{
					// must restore old vertex colors:
					CamVertex[VIndex[0]].r = SavedRGBColor[0].r;
					CamVertex[VIndex[0]].g = SavedRGBColor[0].g;
					CamVertex[VIndex[0]].b = SavedRGBColor[0].b;
					CamVertex[VIndex[1]].r = SavedRGBColor[1].r;
					CamVertex[VIndex[1]].g = SavedRGBColor[1].g;
					CamVertex[VIndex[1]].b = SavedRGBColor[1].b;
					CamVertex[VIndex[2]].r = SavedRGBColor[2].r;
					CamVertex[VIndex[2]].g = SavedRGBColor[2].g;
					CamVertex[VIndex[2]].b = SavedRGBColor[2].b;
				}

        if (CurrPoly->TexMode == NOT_TEXTURED)
        {
					grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
						GR_COMBINE_FACTOR_LOCAL,GR_COMBINE_LOCAL_ITERATED,
						GR_COMBINE_OTHER_TEXTURE,FXFALSE);
				}
				

				// Check if object is using enviroment mapping:
        if (ObjectHandle->EnviromentMapped == FXETRUE)
        {
          // we must perform a two pass process when enviroment mapping.
          // We first draws the original polygon and then draws the
          // enviroment mapped polygon on top of it using alpha blending
          // Using this tecnique we can use enviroment mapping together
          // with texture mapping easily.

            if (CurrPoly->PartlyClipped == FXEFALSE)
            {
              VertexPointers[0] = (void*)&EnvVer[0];
              VertexPointers[1] = (void*)&EnvVer[1];
              VertexPointers[2] = (void*)&EnvVer[2];

              EnvVer[0].x = ProjectedVertices[0].x;
              EnvVer[0].y = ProjectedVertices[0].y;
              EnvVer[0].q = ProjectedVertices[0].q;
              EnvVer[0].r = 255.0f;
              EnvVer[0].g = 255.0f;
              EnvVer[0].b = 255.0f;
              EnvVer[0].u0 = ObjectHandle->EnviromentVerts[VIndex[0]].u*255.0f*EnvVer[0].q;
              EnvVer[0].v0 = ObjectHandle->EnviromentVerts[VIndex[0]].v*255.0f*EnvVer[0].q;
              EnvVer[0].u1 = EnvVer[0].u0;
              EnvVer[0].v1 = EnvVer[0].v0;

              EnvVer[1].x = ProjectedVertices[1].x;
              EnvVer[1].y = ProjectedVertices[1].y;
              EnvVer[1].q = ProjectedVertices[1].q;
              EnvVer[1].r = 255.0f;
              EnvVer[1].g = 255.0f;
              EnvVer[1].b = 255.0f;
              EnvVer[1].u0 = ObjectHandle->EnviromentVerts[VIndex[1]].u*255.0f*EnvVer[1].q;
              EnvVer[1].v0 = ObjectHandle->EnviromentVerts[VIndex[1]].v*255.0f*EnvVer[1].q;
              EnvVer[1].u1 = EnvVer[1].u0;
              EnvVer[1].v1 = EnvVer[1].v0;

              EnvVer[2].x = ProjectedVertices[2].x;
              EnvVer[2].y = ProjectedVertices[2].y;
              EnvVer[2].q = ProjectedVertices[2].q;
              EnvVer[2].r = 255.0f;
              EnvVer[2].g = 255.0f;
              EnvVer[2].b = 255.0f;
              EnvVer[2].u0 = ObjectHandle->EnviromentVerts[VIndex[2]].u*255.0f*EnvVer[2].q;
              EnvVer[2].v0 = ObjectHandle->EnviromentVerts[VIndex[2]].v*255.0f*EnvVer[2].q;
              EnvVer[2].u1 = EnvVer[2].u0;
              EnvVer[2].v1 = EnvVer[2].v0;
            }
            else
            {
              for (Loop2=0;Loop2<NumClippedVertices;Loop2++)
              {
                VertexPointers[Loop2] = (void*)&EnvVer[Loop2];
                EnvVer[Loop2].x = ClippedVertices[Loop2].x;
                EnvVer[Loop2].y = ClippedVertices[Loop2].y;
                EnvVer[Loop2].q = ClippedVertices[Loop2].q;
                EnvVer[Loop2].r = 255.0f;
                EnvVer[Loop2].g = 255.0f;
                EnvVer[Loop2].b = 255.0f;
                EnvVer[Loop2].u0 = ClippedEnvVertices[Loop2].u;
                EnvVer[Loop2].v0 = ClippedEnvVertices[Loop2].v;
                EnvVer[Loop2].u1 = EnvVer[Loop2].u0;
                EnvVer[Loop2].v1 = EnvVer[Loop2].v0;
              }
            }

					SetupTexture(ObjectHandle->ETH);

          //Setup the constant color value:
          grConstantColorValue(((int)ObjectHandle->EnvTexFactor)<<24);

          //Draw the polygon:

          if (DrawShape == GR_LINE_STRIP)
          {
            VertexPointers[NumVertices] = (void*)&EnvVer[0];
            grDrawVertexArray(DrawShape,NumVertices+1,VertexPointers);
          }
	
          else
					{
						if (DrawShape == GR_TRIANGLE_FAN)
						{
							grDrawVertexArray(GR_TRIANGLE_FAN,3,VertexPointers);
							if (NumVertices>3)
							{
								for (Loop3=0;Loop3<NumVertices-3;Loop3++)
								{
									grDrawVertexArray(GR_TRIANGLE_FAN_CONTINUE,1,&VertexPointers[Loop3+3]);
								}
							}
						}
						else
							grDrawVertexArray(DrawShape,NumVertices,VertexPointers);
					}
		  
          // Reset color combine modes and alpha blending modes back to
          // normal so new settings doesn't affect next polygon:

          //Setup the constant color value:
          grConstantColorValue(255<<24);
        }
			}    
    }
  }
  }

/*
  else if (ObjectHandle->Transparent == FXETRUE)
  {
    grConstantColorValue(((int)ObjectHandle->AlphaLevel)<<24);

    // We must sort the polygons in Z since transparent polygons needs
    // this:

    SortTranslucentPolygons(ObjectHandle);

    // Go to first polygon in sorted polygon list:

    pSortList = ObjectHandle->pPolygonSortList;

    while (pSortList->Prev != NULL)
      pSortList = pSortList->Prev;


    for (Loop=0 ; Loop<ObjectHandle->NumPolys; Loop++)
    {
    
      CurrPoly = pSortList->pPolygon;
      pSortList = pSortList->Next;


      // only project and draw polygon if visible
      // Now we project and draws all transparent polygons:

        if (CurrPoly->Clipped == FXEFALSE)
        {
          VIndex = CurrPoly->VertexIndices;

          if (CurrPoly->PartlyClipped == FXETRUE)
          {
            NumVertices = CurrPoly->NumAddedVertices;
            PartlyClipped = FXETRUE;
          }
          else
          {
            NumVertices = CurrPoly->NumVertices;
						PartlyClipped = FXEFALSE;
          }

          for (Loop2=0; Loop2<NumVertices; Loop2++)
          {

            if (PartlyClipped == FXETRUE)
            {

              VertexPointers[Loop2] = (void*)&ProjectedVertices[Loop2];
              if (ObjectHandle->ShadingType == FLAT_SHADING)
              {
                ClippedVertices[Loop2].r = CurrPoly->FlatShadedColor.r;
                ClippedVertices[Loop2].g = CurrPoly->FlatShadedColor.g;
                ClippedVertices[Loop2].b = CurrPoly->FlatShadedColor.b;
              }

							ProjectedVertices[Loop2].q = 1.0f/ClippedVertices[Loop2].z;
							
							ProjectedVertices[Loop2].x = (ClippedVertices[Loop2].x*
								ProjectedVertices[Loop2].q+1.0f)*ProjectWidthDiv2+
								ProjectStartX;
							Snap(&ProjectedVertices[Loop2].x);

							ProjectedVertices[Loop2].y = (ClippedVertices[Loop2].y*
								ProjectedVertices[Loop2].q+1.0f)*ProjectHeightDiv2+
								ProjectStartY;
							Snap(&ProjectedVertices[Loop2].y);

							ProjectedVertices[Loop2].r = ClippedVertices[Loop2].r * 255.0f;
							ProjectedVertices[Loop2].g = ClippedVertices[Loop2].g * 255.0f;
							ProjectedVertices[Loop2].b = ClippedVertices[Loop2].b * 255.0f;

							ProjectedVertices[Loop2].u0 = ClippedVertices[Loop2].u0 * 255.0f*
								ProjectedVertices[Loop2].q;
							ProjectedVertices[Loop2].v0 = ClippedVertices[Loop2].v0 * 255.0f*
								ProjectedVertices[Loop2].q;
							ProjectedVertices[Loop2].u1 = ProjectedVertices[Loop2].u0;
							ProjectedVertices[Loop2].v1 = ProjectedVertices[Loop2].v0;

							
            }

            else
            {
              CurrVertex = &(CamVertex[VIndex[Loop2]]);
            

              if (CurrPoly->TexMode != NOT_TEXTURED)
              {
                CurrVertex->u0 = CurrPoly->TextureCoordinates[Loop2].u;
                CurrVertex->v0 = CurrPoly->TextureCoordinates[Loop2].v;
                CurrVertex->u1 = CurrPoly->TextureCoordinates[Loop2].u;
                CurrVertex->v1 = CurrPoly->TextureCoordinates[Loop2].v;

              }

              if (ObjectHandle->ShadingType == FLAT_SHADING)
              {
                CurrVertex->r = CurrPoly->FlatShadedColor.r;
                CurrVertex->g = CurrPoly->FlatShadedColor.g;
                CurrVertex->b = CurrPoly->FlatShadedColor.b;
              }

							ProjectedVertices[Loop2].q = 1.0f/CurrVertex->z;
							
							ProjectedVertices[Loop2].x = (CurrVertex->x*
								ProjectedVertices[Loop2].q+1.0f)*ProjectWidthDiv2+
								ProjectStartX;
							Snap(&ProjectedVertices[Loop2].x);

							ProjectedVertices[Loop2].y = (CurrVertex->y*
								ProjectedVertices[Loop2].q+1.0f)*ProjectHeightDiv2+
								ProjectStartY;
							Snap(&ProjectedVertices[Loop2].y);

							ProjectedVertices[Loop2].r = CurrVertex->r * 255.0f;
							ProjectedVertices[Loop2].g = CurrVertex->g * 255.0f;
							ProjectedVertices[Loop2].b = CurrVertex->b * 255.0f;

							ProjectedVertices[Loop2].u0 = CurrVertex->u0 * 255.0f*
								ProjectedVertices[Loop2].q;
							ProjectedVertices[Loop2].v0 = CurrVertex->v0 * 255.0f*
								ProjectedVertices[Loop2].q;
							ProjectedVertices[Loop2].u1 = ProjectedVertices[Loop2].u0;
							ProjectedVertices[Loop2].v1 = ProjectedVertices[Loop2].v0;

              VertexPointers[Loop2] = (void*)&ProjectedVertices[Loop2];
            }
          }


          if (CurrPoly->TexMode == NOT_TEXTURED)
          {
            TexHandle = NULLTexture;
            if (CurrPoly->PartlyClipped != FXEFALSE)
            {
            if (TextureFilteringMode == TRILINEAR_FILTERED)
            {
              ((fxeDrawVertex*)VertexPointers[0])->u0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->v0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->u1 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->v1 = 0.0f;
              
              ((fxeDrawVertex*)VertexPointers[1])->u0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[1])->v0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[1])->q;
              ((fxeDrawVertex*)VertexPointers[1])->u1 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[1])->v1 = 255.0f*
								((fxeDrawVertex*)VertexPointers[1])->q;

              ((fxeDrawVertex*)VertexPointers[2])->u0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[2])->q;
              ((fxeDrawVertex*)VertexPointers[2])->v0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[2])->q;
              ((fxeDrawVertex*)VertexPointers[2])->u1 = 255.0f*
								((fxeDrawVertex*)VertexPointers[2])->q;
              ((fxeDrawVertex*)VertexPointers[2])->v1 = 255.0f*
								((fxeDrawVertex*)VertexPointers[2])->q;
            }
            else
            {
              ((fxeDrawVertex*)VertexPointers[0])->u0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->v0 = 0.0f;
              
              ((fxeDrawVertex*)VertexPointers[1])->u0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[1])->v0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[1])->q;

              ((fxeDrawVertex*)VertexPointers[2])->u0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[2])->q;
              ((fxeDrawVertex*)VertexPointers[2])->v0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[2])->q;
            }
            }

            else
            {
            if (TextureFilteringMode == TRILINEAR_FILTERED)
            {
              ((fxeDrawVertex*)VertexPointers[0])->u0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->v0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->u1 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->v1 = 0.0f;
              for(Loop2=1;Loop2<NumVertices;Loop2++)
              {
                ((fxeDrawVertex*)VertexPointers[Loop2])->u0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[Loop2])->q;
                ((fxeDrawVertex*)VertexPointers[Loop2])->v0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[Loop2])->q;
                ((fxeDrawVertex*)VertexPointers[Loop2])->u1 = 255.0f*
								((fxeDrawVertex*)VertexPointers[Loop2])->q;
                ((fxeDrawVertex*)VertexPointers[Loop2])->v1 = 255.0f*
								((fxeDrawVertex*)VertexPointers[Loop2])->q;
              }
              
            }
            else
            {
              ((fxeDrawVertex*)VertexPointers[0])->u0 = 0.0f;
              ((fxeDrawVertex*)VertexPointers[0])->v0 = 0.0f;
              for(Loop2=1;Loop2<NumVertices;Loop2++)
              {
                ((fxeDrawVertex*)VertexPointers[Loop2])->u0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[Loop2])->q;
                ((fxeDrawVertex*)VertexPointers[Loop2])->v0 = 255.0f*
								((fxeDrawVertex*)VertexPointers[Loop2])->q;
              }
            }
            }
          }
          else
            TexHandle = CurrPoly->TextureHandle;

         

            if (TexHandle != CurrentTextureSource)
            {

              // check which texture filtering method which
              // is used:


              if (TexHandle->FilteringMethod == BILINEAR_FILTERED)
              {
                // Setup texture unit 0's texture source
                grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
                  GR_MIPMAPLEVELMASK_BOTH,&TexHandle->TexInfo);
                CurrentTextureSource = TexHandle;
              }

              else if (TexHandle->FilteringMethod == TRILINEAR_FILTERED)
              {
                // Combine trilinear texture mapping with
                // bilinear filtering to produce trilinear
                // filtering

                // Setup texture unit 0's texture source
                grTexSource(GR_TMU0,TexHandle->VideoMemoryAddressTMU0,
                  GR_MIPMAPLEVELMASK_EVEN,&TexHandle->TexInfo);

                // Setup texture unit 1's texture source
                grTexSource(GR_TMU1,TexHandle->VideoMemoryAddressTMU1,
                  GR_MIPMAPLEVELMASK_ODD,&TexHandle->TexInfo);
                CurrentTextureSource = TexHandle;
              }
            }


          // so finally send polygon to 3dfx hardware:
          // Draw polygon using triangle fans. This way any convex polygon
          // can be drawn.

        if (DrawShape == GR_LINE_STRIP)
        {
          if (PartlyClipped == FXETRUE)
            VertexPointers[NumVertices] = (void*)&ProjectedVertices[0];
          else
            VertexPointers[NumVertices] = (void*)&ProjectedVertices[0];
          grDrawVertexArray(DrawShape,NumVertices+1,VertexPointers);
        }

        else
				{
					if (DrawShape == GR_TRIANGLE_FAN)
					{
						grDrawVertexArray(GR_TRIANGLE_FAN,3,VertexPointers);
						if (NumVertices>3)
						{
							for (Loop3=0;Loop3<NumVertices-3;Loop3++)
							{
								grDrawVertexArray(GR_TRIANGLE_FAN_CONTINUE,1,&VertexPointers[Loop3+3]);
							}
						}
					}
					else
						grDrawVertexArray(DrawShape,NumVertices,VertexPointers);

				}

        Performance.NumPolys++;
        }
      }
    }
		*/
}


int SortSmokeEmittersCallback(const void **ppEmitter1,
	const void **ppEmitter2)
{

	fxeSmokeEmitter *pEmitter1;
	fxeSmokeEmitter *pEmitter2;
	fxeVertex VertexWorld1;
	fxeVertex VertexWorld2;
	fxeVertex VertexCamera1;
	fxeVertex VertexCamera2;

	pEmitter1 = (fxeSmokeEmitter*)ppEmitter1[0];
	pEmitter2 = (fxeSmokeEmitter*)ppEmitter2[0];


	VertexWorld1.x = pEmitter1->x;
	VertexWorld1.y = pEmitter1->y;
	VertexWorld1.z = pEmitter1->z;

	VertexWorld2.x = pEmitter2->x;
	VertexWorld2.y = pEmitter2->y;
	VertexWorld2.z = pEmitter2->z;



	MatXVer(CameraMatrix,&VertexWorld1,&VertexCamera1);
	MatXVer(CameraMatrix,&VertexWorld2,&VertexCamera2);

  if (VertexCamera1.z > VertexCamera2.z)
    return -1;
  else if (VertexCamera1.z < VertexCamera2.z)
    return 1;
  else
    return 0;
}

int SortEmittersCallback(const void **ppEmitter1,
	const void **ppEmitter2)
{

	fxeEmitter *pEmitter1;
	fxeEmitter *pEmitter2;
	fxeVertex VertexWorld1;
	fxeVertex VertexWorld2;
	fxeVertex VertexCamera1;
	fxeVertex VertexCamera2;

	pEmitter1 = (fxeEmitter*)ppEmitter1[0];
	pEmitter2 = (fxeEmitter*)ppEmitter2[0];


	VertexWorld1.x = pEmitter1->x;
	VertexWorld1.y = pEmitter1->y;
	VertexWorld1.z = pEmitter1->z;

	VertexWorld2.x = pEmitter2->x;
	VertexWorld2.y = pEmitter2->y;
	VertexWorld2.z = pEmitter2->z;



	MatXVer(CameraMatrix,&VertexWorld1,&VertexCamera1);
	MatXVer(CameraMatrix,&VertexWorld2,&VertexCamera2);

  if (VertexCamera1.z > VertexCamera2.z)
    return -1;
  else if (VertexCamera1.z < VertexCamera2.z)
    return 1;
  else
    return 0;
}

int Sort_Objects_Callback_Function(const void **ppObject1,
  const void **ppObject2)
{
  // this function is called by qsort and sorts all objects using simple
  // z value.
  fxeVertex TempVertex;
  fxeVertex VerRelCamera;
  fxeVertex TempVertex2;
  fxeVertex VerRelCamera2;
  fxeObjectPtr pObject1 = (fxeObjectPtr)*ppObject1;
  fxeObjectPtr pObject2 = (fxeObjectPtr)*ppObject2;

  TempVertex.x = pObject1->x;
  TempVertex.y = pObject1->y;
  TempVertex.z = pObject1->z;

  TempVertex2.x = pObject2->x;
  TempVertex2.y = pObject2->y;
  TempVertex2.z = pObject2->z;

  MatXVer(CameraMatrix,&TempVertex,&VerRelCamera);
  MatXVer(CameraMatrix,&TempVertex2,&VerRelCamera2);


  if (VerRelCamera.z > VerRelCamera2.z)
    return -1;
  else if (VerRelCamera.z < VerRelCamera2.z)
    return 1;
  else
    return 0;
}

void fxeBackgroundSetColor(fxeColor *pColor)
{
  BackgroundColor.r = pColor->r;
  BackgroundColor.g = pColor->g;
  BackgroundColor.b = pColor->b;
}


void UpdateActiveCamera(void)
{
  fxeVector Vector1;
  float VectorLen;

  if (ActiveCamera->FollowPoint == FXETRUE )
  {
    Vector1.x  = ActiveCamera->PointToFollow.x - ActiveCamera->x;
    Vector1.y  = ActiveCamera->PointToFollow.y - ActiveCamera->y;
    Vector1.z  = ActiveCamera->PointToFollow.z - ActiveCamera->z;

    VectorLen  = (float)sqrt(Vector1.x*Vector1.x+Vector1.z*Vector1.z);

    if (VectorLen == 0.0f)
      VectorLen = 0.0001f;

    Vector1.x /= VectorLen;
    Vector1.y /= VectorLen;
    Vector1.z /= VectorLen;


    ActiveCamera->ay = RadToDeg((float)acos(Vector1.z));
    if (Vector1.x < 0.0f)
      ActiveCamera->ay = 360.0f-ActiveCamera->ay;

    Vector1.x  = ActiveCamera->x - ActiveCamera->PointToFollow.x;
    Vector1.y  = ActiveCamera->y - ActiveCamera->PointToFollow.y;
    Vector1.z  = ActiveCamera->z - ActiveCamera->PointToFollow.z;

    VectorLen  = (float)sqrt(Vector1.y*Vector1.y+Vector1.z*Vector1.z+
      Vector1.x*Vector1.x);

    if (VectorLen == 0.0f)
      VectorLen = 0.0001f;

    if (Vector1.y < 0)
    {
      Vector1.y = -Vector1.y;
      ActiveCamera->ax = 360.0f-RadToDeg((float)asin(Vector1.y/VectorLen));
    }
    else
      ActiveCamera->ax = RadToDeg((float)asin(Vector1.y/VectorLen));

    ActiveCamera->az = 0.0f;

    if (ActiveCamera->NumFramesLeft != 0)
    {
      ActiveCamera->x += ActiveCamera->IncX;
      ActiveCamera->y += ActiveCamera->IncY;
      ActiveCamera->z += ActiveCamera->IncZ;
      ActiveCamera->NumFramesLeft--;
    }
  }
  else
  {
    if (ActiveCamera->NumFramesLeft != 0)
    {
      ActiveCamera->x += ActiveCamera->IncX;
      ActiveCamera->y += ActiveCamera->IncY;
      ActiveCamera->z += ActiveCamera->IncZ;
      ActiveCamera->ax += ActiveCamera->RotX;
      ActiveCamera->ax = (float)((int)ActiveCamera->ax % 360);
      ActiveCamera->ay += ActiveCamera->RotY;
      ActiveCamera->ay = (float)((int)ActiveCamera->ay % 360);
      ActiveCamera->az += ActiveCamera->RotZ;
      ActiveCamera->az = (float)((int)ActiveCamera->az % 360);
      ActiveCamera->NumFramesLeft--;
    }
  }
}

// This is the big one. It transforms all objects to world space and
// then to view space. It calculates light reflection of all objects.
// It performs object clipping and polygon clipping and backface removal.
// Perspective projection is performed by Glide3x.
// Glide sorts the polygons with W buffering.


void fxeWorldRender(int DisplayWorld)
{
	int Finished;
  int NumObjects;
  int Loop;
	fxeEmitter *CurrEmitter;
	fxeSmokeEmitter *CurrSmokeEmitter;
	int NumEmitters;
	fxeEmitter *pEmitter[MAX_EMITTERS_IN_WORLD];
	fxeSmokeEmitter *pSmokeEmitter[MAX_SMOKE_EMITTERS_IN_WORLD];
	float SavedFarPlane;
	fxeVertex OldCamera;

  FrameDisplayed = FXEFALSE;

	PrepareTexturesForFrame();

	// Make no changes to the alpha and color buffers - just
	// clear the depth buffer.

	grColorMask(FXFALSE,FXFALSE);
  grBufferClear( ((int)(BackgroundColor.r*255.0)<<16)+
    ((int)(BackgroundColor.g*255.0)<<8)+((int)(BackgroundColor.b*255.0)),
    0,WMaxValue);
	grColorMask(FXTRUE,FXFALSE);

  
	UpdateActiveCamera();


  grChromakeyMode(GR_CHROMAKEY_DISABLE);
  ChromaKey = FXEFALSE;
  CombineTextureOnly = FXEFALSE;

  CalculateCameraTransformationMatrix();

	// Render skybox if skybox is defined:
	if (GlobalSkyboxHandle != NULL)
	{
			// Set far clipping plane		
			SavedFarPlane = FarClippingPlane;
			FarClippingPlane = 65535.0f;
			
			SetupViewFrustrum();

			OldCamera.x = ActiveCamera->x;
			OldCamera.y = ActiveCamera->y;
			OldCamera.z = ActiveCamera->z;

			ActiveCamera->x = 0.0f;
			ActiveCamera->y = 0.0f;
			ActiveCamera->z = 0.0f;

			CalculateCameraTransformationMatrix();

			//OldAmbient.r = AmbientR;
			//OldAmbient.g = AmbientG;
			//OldAmbient.b = AmbientB;

			//fxeAmbientlightLight(1.0f,1.0f,1.0f);

			GlobalSkyboxHandle->CompletelyVisible = FXEFALSE;
			GlobalSkyboxHandle->Clipped = FXEFALSE;
			
      TransformToCameraMatrix(GlobalSkyboxHandle);
      ClipPolygons(GlobalSkyboxHandle);
      RemoveBackfaces(GlobalSkyboxHandle);
      ShadeObject(GlobalSkyboxHandle);
      //SortTranslucentPolygons(GlobalSkyboxHandle);
      ProjectAndDrawObject(GlobalSkyboxHandle);

			// Restore far clipping plane:
			FarClippingPlane = SavedFarPlane;
			SetupViewFrustrum();
			//fxeAmbientlightLight(OldAmbient.r,OldAmbient.g,OldAmbient.b);

			
			ActiveCamera->x = OldCamera.x;
			ActiveCamera->y = OldCamera.y;
			ActiveCamera->z = OldCamera.z;
			CalculateCameraTransformationMatrix();

	}

	// Draw points of lightsources so we can see if they
	// are visible after the world have been drawn.
	fxeLightsourcesDrawPoints();

	// Render landscape if defined
	if (CurrLandscapeParent != NULL)
		CurrLandscapeParent->NumLandscapesInList = 0;
  if (CurrLandscape != NULL)
	{
		// setup color combine unit to texture only and to add a constant color, in this
		// case it is just ambient color
	
		// Setup special vertex layout for landscape
		// rendering
		grVertexLayout(GR_PARAM_XY,0,GR_PARAM_ENABLE);
		grVertexLayout(GR_PARAM_W,8,GR_PARAM_DISABLE);
		//grVertexLayout(GR_PARAM_Z,12,GR_PARAM_ENABLE);
  
		// Need to specify this when using window coordinates:
		grVertexLayout(GR_PARAM_Q,12,GR_PARAM_ENABLE);
		grVertexLayout(GR_PARAM_RGB,20,GR_PARAM_DISABLE);
		grVertexLayout(GR_PARAM_ST0,16,GR_PARAM_ENABLE);
		grVertexLayout(GR_PARAM_ST1,24,GR_PARAM_ENABLE);

		grConstantColorValue((255<<24)+(((unsigned int)(AmbientR*255.0f))<<16)+
			(((unsigned int)(AmbientG*255.0f))<<8)+
			((unsigned int)(AmbientB*255.0f)));

		grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
			GR_COMBINE_FACTOR_ONE,GR_COMBINE_LOCAL_CONSTANT,
			GR_COMBINE_OTHER_TEXTURE,FXFALSE);
	

		if (NumTexelFx > 1)
		{
			grAlphaBlendFunction(GR_BLEND_SRC_ALPHA,GR_BLEND_ONE_MINUS_SRC_ALPHA,
				GR_BLEND_ZERO,GR_BLEND_ZERO);
			grTexCombine(GR_TMU0,GR_COMBINE_FUNCTION_SCALE_OTHER,
				GR_COMBINE_FACTOR_LOCAL,GR_COMBINE_FUNCTION_LOCAL_ALPHA,
				GR_COMBINE_FACTOR_NONE,FXFALSE,FXFALSE);
			grTexCombine(GR_TMU1,GR_COMBINE_FUNCTION_LOCAL,
				GR_COMBINE_FACTOR_NONE,GR_COMBINE_FUNCTION_LOCAL_ALPHA,
				GR_COMBINE_FACTOR_NONE,FXFALSE,FXFALSE);

			fxeLandscapeRender(CurrLandscape);
			BuildLightmaps();
			ProjectAndDrawLandscapesUseLightmapping();

			grTexCombine(GR_TMU0,GR_COMBINE_FUNCTION_LOCAL,
				GR_COMBINE_FACTOR_NONE,GR_COMBINE_FUNCTION_LOCAL_ALPHA,
				GR_COMBINE_FACTOR_NONE,FXFALSE,FXFALSE);
		}
		else
		{
			fxeLandscapeRender(CurrLandscape);
			BuildLightmaps();
			ProjectAndDrawLandscapesUseLightmapping();
		}

		// Restore vertex layout to normal for fxengine
		grVertexLayout(GR_PARAM_XY,0,GR_PARAM_ENABLE);
		grVertexLayout(GR_PARAM_W,8,GR_PARAM_ENABLE);
		//grVertexLayout(GR_PARAM_Z,12,GR_PARAM_ENABLE);
  
		// Need to specify this when using window coordinates:
		grVertexLayout(GR_PARAM_Q,16,GR_PARAM_ENABLE);
		grVertexLayout(GR_PARAM_RGB,20,GR_PARAM_ENABLE);
		grVertexLayout(GR_PARAM_ST0,32,GR_PARAM_ENABLE);
		grVertexLayout(GR_PARAM_ST1,40,GR_PARAM_ENABLE);


		// restore color combining to normal for FxEngine
		grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
			GR_COMBINE_FACTOR_LOCAL,GR_COMBINE_LOCAL_ITERATED,
			GR_COMBINE_OTHER_TEXTURE,FXFALSE);

	}


	if (WorldObjects != NULL)
	{
		// go to first object in world:
		while (WorldObjects->Prev != NULL)
	 		WorldObjects = WorldObjects->Prev;

		Finished = FXEFALSE;


		// go through all objects and process them and render them

    NumObjects = 0;

  
    while(Finished == FXEFALSE)
		{


      ClipObject(WorldObjects);

      if (WorldObjects->Clipped == FXEFALSE)
      {
        WorldObjectPointers[NumObjects] = WorldObjects;
        NumObjects++;
      }

			if (WorldObjects->Next != NULL)
				WorldObjects = WorldObjects->Next;
			else
				Finished = FXETRUE;
    }

    // Sort the object list:

    qsort((char*)WorldObjectPointers,NumObjects,sizeof(fxeObjectPtr),
      Sort_Objects_Callback_Function);

    for (Loop=0; Loop<NumObjects; Loop++)
    {
			
      TransformToWorld(WorldObjectPointers[Loop]);

      TransformToCameraMatrix(WorldObjectPointers[Loop]);
      ClipPolygons(WorldObjectPointers[Loop]);
      //CalculateNormals(WorldObjectPointers[Loop]);
      RemoveBackfaces(WorldObjectPointers[Loop]);
      ShadeObject(WorldObjectPointers[Loop]);

      //ClipIntersectedPolygons(WorldObjectPointers[Loop]);
      //SortTranslucentPolygons(WorldObjectPointers[Loop]);
      ProjectAndDrawObject(WorldObjectPointers[Loop]);

    }	

	}




	// Render all added emitters:
	// First sort them in Z 
	CurrEmitter = pEmitters;
	NumEmitters = 0;

	if (CurrEmitter != NULL)
	{
		do
		{
			pEmitter[NumEmitters] = CurrEmitter;
			CurrEmitter = CurrEmitter->Prev;
			NumEmitters++;
		}
		while (CurrEmitter != NULL);
		
		// Sort Emitters by Z value;

		qsort(pEmitter,NumEmitters,sizeof(fxeEmitter*),
			SortEmittersCallback);

		for (Loop=0;Loop<NumEmitters;Loop++)
		{
			fxeEmitterUpdateFrame(pEmitter[Loop]);
		}
	}

	// Render all smoke emitters:
	// First sort them in Z 
	CurrSmokeEmitter = pSmokeEmitters;
	NumEmitters = 0;

	if (CurrSmokeEmitter != NULL)
	{
		do
		{
			pSmokeEmitter[NumEmitters] = CurrSmokeEmitter;
			CurrSmokeEmitter = CurrSmokeEmitter->Prev;
			NumEmitters++;
		}
		while (CurrSmokeEmitter != NULL);
		
		// Sort Smoke Emitters by Z value;

		qsort(pSmokeEmitter,NumEmitters,sizeof(fxeSmokeEmitter*),
			SortSmokeEmittersCallback);

		for (Loop=0;Loop<NumEmitters;Loop++)
		{
			fxeSmokeEmitterAddParticles(pSmokeEmitter[Loop]);
			fxeSmokeEmitterUpdateFrame(pSmokeEmitter[Loop]);
		}
	}

	// Draw lensflares if lightsource(s) is visible

	fxeLightsourcesDraw();
	
	fxeDynLightsourcesMakeUnchanged();
	fxeDynLightsourcesMakeUnMoved();


	MakeUnusedTexturesUnused();

  if (DisplayWorld == FXETRUE)
  {
    fxeWorldDisplay();
  }

  FrameCounter++;
  Performance.FrameCount++;
}

void fxeBackgroundClear(void)
{
	// Don't clear Depth buffer:
	grDepthMask(FXFALSE);
  grBufferClear( ((int)(BackgroundColor.r*255.0)<<16)+
    ((int)(BackgroundColor.g*255.0)<<8)+((int)(BackgroundColor.b*255.0)),
    0,WMaxValue);
	grDepthMask(FXTRUE);
}

void fxeWorldDisplay(void)
{
  if (FrameDisplayed == FXEFALSE)
  {
    // Sync to 60FPS
    grBufferSwap(SwapFlag);
    FrameDisplayed = FXETRUE;
  }
}


// Here comes some matrix math functions by Hugues Landry which also is
// going to be used in a 3d engine called Sp@cE_iNvAdr.
// Big thanks to you Hugues for this!
// Original Matrix functions by Hugues Landry,
// modification for FxEngine by Andreas Ingo

//////////////////////////////////////////////
// Transformation Module Function Definiton //
//////////////////////////////////////////////

// Initialization Function Definition

void InitMatrix(fxeMatrix Matrix)
{
  memset(Matrix,0,sizeof(fxeMatrix));

  Matrix[0][0]=1.0f;
  Matrix[1][1]=1.0f;
  Matrix[2][2]=1.0f;
  Matrix[3][3]=1.0f;
}


// Matrix Transformation Function Definitions

void MoveMatrix(fxeMatrix Matrix, float x, float y, float z)
{
	InitMatrix(Matrix);
  Matrix[3][0] = x;
  Matrix[3][1] = y;
  Matrix[3][2] = z;
}

// Note! The rotation matrix functions requires that CosLockup and
// SinLockup have been setup. Angle is the angle in degrees.
// These functions builds a inverted rotation matrices. To rotate in
// non inverted direction specify Angle argument as
// 360.0 - (Real Angle)


void RotXMatrix(fxeMatrix Matrix, float Angle)
{
	InitMatrix(Matrix);

  Matrix[1][1] = CosLockup[(int)(Angle)];
  Matrix[1][2] = SinLockup[(int)(Angle)];
  Matrix[2][1] = -SinLockup[(int)(Angle)];
  Matrix[2][2] = CosLockup[(int)(Angle)];
}


void RotYMatrix(fxeMatrix Matrix, float Angle)
{
	InitMatrix(Matrix);

  Matrix[0][0] = CosLockup[(int)(Angle)];
  Matrix[0][2] = -SinLockup[(int)(Angle)];
  Matrix[2][0] = SinLockup[(int)(Angle)];
  Matrix[2][2] = CosLockup[(int)(Angle)];
}

void RotZMatrix(fxeMatrix Matrix, float Angle)
{
	InitMatrix(Matrix);

  Matrix[0][0] = CosLockup[(int)(Angle)];
  Matrix[0][1] = SinLockup[(int)(Angle)];
  Matrix[1][0] = -SinLockup[(int)(Angle)];
  Matrix[1][1] = CosLockup[(int)(Angle)];
}

void ScaleMatrix(fxeMatrix Matrix, float SFactor)
{
	InitMatrix(Matrix);

	Matrix[0][0] = SFactor;
	Matrix[1][1] = SFactor;
	Matrix[2][2] = SFactor;
}

void ProjectMatrix(fxeMatrix Matrix)
{
	InitMatrix(Matrix);

	Matrix[3][2] = 1;
	Matrix[3][3] = 0;
}

// Matrix multiplication Function Definitions

void Matmult(fxeMatrix MatrixA, fxeMatrix MatrixB)
{
  int i,j,k;
  float Sum;
  fxeMatrix Result;

  for (i=0 ; i<4 ; i++)
  {
    for (j=0 ; j<4 ; j++)
    {
      Sum = 0;

      for (k=0 ; k < 4 ; k++)
        Sum += MatrixA[i][k] * MatrixB[k][j];

      Result[i][j] = Sum;
    }
  }
  memcpy(MatrixB,Result,sizeof(fxeMatrix));
}

// This code was taken from Michael Abrash XSharp v22
// and was modified by Andreas Ingo to fit FxEngine.

void ConcatXforms(fxeMatrix SourceXform1, fxeMatrix SourceXform2,
  fxeMatrix DestXform)
{
  int i, j;

  for (i=0; i<3; i++)
  {
    for (j=0; j<3; j++)
      DestXform[i][j] =
        SourceXform1[i][0] * SourceXform2[0][j] +
        SourceXform1[i][1] * SourceXform2[1][j] +
        SourceXform1[i][2] * SourceXform2[2][j];
      DestXform[i][3] =
        SourceXform1[i][0] * SourceXform2[0][3] +
        SourceXform1[i][1] * SourceXform2[1][3] +
        SourceXform1[i][2] * SourceXform2[2][3] +
        SourceXform1[i][3];
  }
}

/* Concatenate a rotation by Angle around the X axis to the
   transformation in XformToChange, placing the result back into
   XformToChange. */

void AppendRotationX(fxeMatrix Matrix, float Angle)
{
   float CosTemp, SinTemp;
   float Temp10, Temp11, Temp12, Temp20, Temp21, Temp22;

   /* Get the sine and cosine of the rotation angle */
   CosTemp = CosLockup[(int)Angle];
   SinTemp = SinLockup[(int)Angle];

   /* Calculate the new values of the six affected matrix entries */
   Temp10 = CosTemp * Matrix[1][0] +
         -SinTemp * Matrix[2][0];
   Temp11 = CosTemp * Matrix[1][1] +
         -SinTemp * Matrix[2][1];
   Temp12 = CosTemp * Matrix[1][2] +
         -SinTemp * Matrix[2][2];
   Temp20 = SinTemp * Matrix[1][0] +
         CosTemp * Matrix[2][0];
   Temp21 = SinTemp * Matrix[1][1] +
         CosTemp * Matrix[2][1];
   Temp22 = SinTemp * Matrix[1][2] +
         CosTemp * Matrix[2][2];
   /* Put the results back into XformToChange */
   Matrix[1][0] = Temp10; Matrix[1][1] = Temp11;
   Matrix[1][2] = Temp12; Matrix[2][0] = Temp20; 
   Matrix[2][1] = Temp21; Matrix[2][2] = Temp22;
}

/* Concatenate a rotation by Angle around the Y axis to the
   transformation in XformToChange, placing the result back into
   XformToChange. */
void AppendRotationY(fxeMatrix Matrix, float Angle)
{
   float CosTemp, SinTemp;
   float Temp00, Temp01, Temp02, Temp20, Temp21, Temp22;

   CosTemp = CosLockup[(int)Angle];
   SinTemp = SinLockup[(int)Angle];

   /* Calculate the new values of the six affected matrix entries */
   Temp00 = CosTemp * Matrix[0][0] +
         SinTemp * Matrix[2][0];
   Temp01 = CosTemp * Matrix[0][1] +
         SinTemp * Matrix[2][1];
   Temp02 = CosTemp * Matrix[0][2] +
         SinTemp * Matrix[2][2];
   Temp20 = -SinTemp * Matrix[0][0] +
         CosTemp * Matrix[2][0];
   Temp21 = -SinTemp * Matrix[0][1] +
         CosTemp * Matrix[2][1];
   Temp22 = -SinTemp * Matrix[0][2] +
         CosTemp * Matrix[2][2];
   /* Put the results back into XformToChange */
   Matrix[0][0] = Temp00; Matrix[0][1] = Temp01;
   Matrix[0][2] = Temp02; Matrix[2][0] = Temp20;
   Matrix[2][1] = Temp21; Matrix[2][2] = Temp22;
}

/* Concatenate a rotation by Angle around the Z axis to the
   transformation in XformToChange, placing the result back into
   XformToChange. */
void AppendRotationZ(fxeMatrix Matrix, float Angle)
{
   float CosTemp, SinTemp;
   float Temp00, Temp01, Temp02, Temp10, Temp11, Temp12;

   CosTemp = CosLockup[(int)Angle];
   SinTemp = SinLockup[(int)Angle];

   /* Calculate the new values of the six affected matrix entries */
   Temp00 = CosTemp * Matrix[0][0] +
         -SinTemp * Matrix[1][0];
   Temp01 = CosTemp * Matrix[0][1] +
         -SinTemp * Matrix[1][1];
   Temp02 = CosTemp * Matrix[0][2] +
         -SinTemp * Matrix[1][2];
   Temp10 = SinTemp * Matrix[0][0] +
         CosTemp * Matrix[1][0];
   Temp11 = SinTemp * Matrix[0][1] +
         CosTemp * Matrix[1][1];
   Temp12 = SinTemp * Matrix[0][2] +
         CosTemp * Matrix[1][2];
   /* Put the results back into XformToChange */
   Matrix[0][0] = Temp00; Matrix[0][1] = Temp01;
   Matrix[0][2] = Temp02; Matrix[1][0] = Temp10;
   Matrix[1][1] = Temp11; Matrix[1][2] = Temp12;
}


void MatCopy(fxeMatrix Source, fxeMatrix Dest)
{
  memcpy(Dest, Source, sizeof(fxeMatrix));
}

/*
void Matmult(fxeMatrix MatrixA, fxeMatrix MatrixB)
{
  int i,j;
  fxeMatrix res;

	for (i = 0; i < 4; i++) {
		for (j = 0; j < 4; j++) {
      res[i][j]  = MatrixA[0][j] * MatrixB[i][0];
      res[i][j] += MatrixA[1][j] * MatrixB[i][1];
      res[i][j] += MatrixA[2][j] * MatrixB[i][2];
      res[i][j] += MatrixA[3][j] * MatrixB[i][3];
		}
	}
  memcpy(MatrixB, res, sizeof(res));
}
*/

void MatXVer(fxeMatrix Matrix, fxeVertex *Vertex, fxeVertex *Result)
{
  float sum;

    sum = Vertex->x*Matrix[0][0];
    sum += Vertex->y*Matrix[1][0];
    sum += Vertex->z*Matrix[2][0];
    sum += Matrix[3][0];
    Result->x = sum;

    sum = Vertex->x*Matrix[0][1];
    sum += Vertex->y*Matrix[1][1];
    sum += Vertex->z*Matrix[2][1];
    sum += Matrix[3][1];
    Result->y = sum;

    sum = Vertex->x*Matrix[0][2];
    sum += Vertex->y*Matrix[1][2];
    sum += Vertex->z*Matrix[2][2];
    sum += Matrix[3][2];
    Result->z = sum;
}

void MatXVec(fxeMatrix Matrix, fxeVector *Vector, fxeVector *Result)
{
  float sum;

    sum = Vector->x*Matrix[0][0];
    sum += Vector->y*Matrix[1][0];
    sum += Vector->z*Matrix[2][0];
    sum += Matrix[3][0];
    Result->x = sum;

    sum = Vector->x*Matrix[0][1];
    sum += Vector->y*Matrix[1][1];
    sum += Vector->z*Matrix[2][1];
    sum += Matrix[3][1];
    Result->y = sum;

    sum = Vector->x*Matrix[0][2];
    sum += Vector->y*Matrix[1][2];
    sum += Vector->z*Matrix[2][2];
    sum += Matrix[3][2];
    Result->z = sum;
}


void MatXVerToDrawVer(fxeMatrix Matrix, fxeVertex *Vertex, fxeDrawVertex *Result)
{
  float sum;

    sum = Vertex->x*Matrix[0][0];
    sum += Vertex->y*Matrix[1][0];
    sum += Vertex->z*Matrix[2][0];
    sum += Matrix[3][0];
    Result->x = sum;

    sum = Vertex->x*Matrix[0][1];
    sum += Vertex->y*Matrix[1][1];
    sum += Vertex->z*Matrix[2][1];
    sum += Matrix[3][1];
    Result->y = sum;

    sum = Vertex->x*Matrix[0][2];
    sum += Vertex->y*Matrix[1][2];
    sum += Vertex->z*Matrix[2][2];
    sum += Matrix[3][2];
    Result->z = sum;
}

void LMatXDrawVerToDrawVer(fxeMatrix Matrix, fxeLDrawVertex *Vertex, fxeLDrawVertex *Result)
{
  float sum;

    sum = Vertex->x*Matrix[0][0];
    sum += Vertex->y*Matrix[1][0];
    sum += Vertex->z*Matrix[2][0];
    sum += Matrix[3][0];
    Result->x = sum;

    sum = Vertex->x*Matrix[0][1];
    sum += Vertex->y*Matrix[1][1];
    sum += Vertex->z*Matrix[2][1];
    sum += Matrix[3][1];
    Result->y = sum;

    sum = Vertex->x*Matrix[0][2];
    sum += Vertex->y*Matrix[1][2];
    sum += Vertex->z*Matrix[2][2];
    sum += Matrix[3][2];
    Result->z = sum;
}

void MatXDrawVerToDrawVer(fxeMatrix Matrix, fxeDrawVertex *Vertex, fxeDrawVertex *Result)
{
  float sum;

    sum = Vertex->x*Matrix[0][0];
    sum += Vertex->y*Matrix[1][0];
    sum += Vertex->z*Matrix[2][0];
    sum += Matrix[3][0];
    Result->x = sum;

    sum = Vertex->x*Matrix[0][1];
    sum += Vertex->y*Matrix[1][1];
    sum += Vertex->z*Matrix[2][1];
    sum += Matrix[3][1];
    Result->y = sum;

    sum = Vertex->x*Matrix[0][2];
    sum += Vertex->y*Matrix[1][2];
    sum += Vertex->z*Matrix[2][2];
    sum += Matrix[3][2];
    Result->z = sum;
}
